
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import ProveSubof.
Require Import Regularity.
Require Import Singleton.
Require Import Formation.
Require Import Preservation.
Require Import Determinism.
Require Import Congruence.
Require Import Contextual.
Require Import Equivalence.
Require Import Reductive.
Require Import Translation.
Require Import Split.
Require Import Inverse.
Require Import Correctness.
Require Inversion.



(* This messy bit is needed to show that we can substitute into contextual equivalence
   with substitutions (specifically trsc) that almost satisfy aof, but not quite.
*)

(* substitutends for almost all *)
Inductive aaof : context -> term -> term -> Prop :=
| aaof_cn {G c k}
    : ofc G c k
      -> aaof G c (cl_cn k)

| aaof_tm {G e t}
    : oft G e t
      -> (forall s, subof aof nil s G -> thalt nil (subst s e))
      -> aaof G e (cl_tm t)

| aaof_md {G M T c}
    : ofm G pure M T
      -> (forall s, subof aof nil s G -> mhalt nil (subst s M))
      -> fstm G M c
      -> aaof G M (cl_md T c)

| aaof_ucn {G c}
    : aaof G c cl_ucn

| aaof_utm {G e}
    : tvalue G e
      -> aaof G e cl_utm

| aaof_umd {G M c}
    : mvalue G M
      -> fstm G M c
      -> aaof G M (cl_umd c)

| aaof_ns {G t}
    : aaof G t cl_ns
.


Lemma aof_aaof :
  forall G m C, aof G m C -> aaof G m C.
Proof.
intros G m C Hof.
#2 induct Hof; try (intros; simp_sub; eauto using aaof; done).

(* tm *)
intros G e t Hof Hval.
apply aaof_tm; auto; [].
intros s Hs.
apply tvalue_thalt; [].
eapply subst_tvalue; eauto using subof_aof_oof; done.

(* md *)
intros G M T c Hof Hval HfstM.
apply aaof_md; auto; [].
intros s Hs.
apply mvalue_mhalt; [].
eapply subst_mvalue; eauto using subof_aof_oof; done.
Qed.


Lemma aaof_sof :
  forall G m C, aaof G m C -> sof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using sof.
Qed.


Lemma aaof_dof :
  forall G m C, aaof G m C -> dof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using dof.
Qed.


Lemma aaof_tof :
  forall G m C, aaof G m C -> tof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using tof.
Qed.


Lemma subof_aaof_sof :
  forall G1 s G2,
    subof aaof G1 s G2
    -> subof sof G1 s G2.
Proof.
eauto using subof_mono, aaof_sof.
Qed.


Lemma subof_aaof_dof :
  forall G1 s G2,
    subof aaof G1 s G2
    -> subof dof G1 s G2.
Proof.
eauto using subof_mono, aaof_dof.
Qed.


Lemma subof_aaof_tof :
  forall G1 s G2,
    subof aaof G1 s G2
    -> subof tof G1 s G2.
Proof.
eauto using subof_mono, aaof_tof.
Qed.


Lemma subst_aaof :
  forall G1 G2 s m C,
    subof aof G1 s G2
    -> aaof G2 m C
    -> aaof G1 (subst s m) (subst s C).
Proof.
intros G1 G2 s m C Hs H.
revert G1 s Hs.
induct H; try (intros; simp_sub; eauto 8 using aaof, subst_tvalue, subst_mvalue, subst_fstm, subof_aof_oof, subof_aof_tof; done).

(* cn *)
intros G2 c k Hof G1 s Hs.
simp_sub.
apply aaof_cn; eauto using subst_ofc, subof_aof_sof; done.

(* tm *)
intros G e t Hofe Hhaltish G1 s Hs.
simp_sub.
apply aaof_tm; eauto using subst_oft, subof_aof_dof; [].
intros s' Hs'.
simp_sub.
apply Hhaltish; [].
eapply subof_aof_trans; eauto; done.

(* md *)
intros G M T c HofM Hhaltish HfstM G1 s Hs.
simp_sub.
apply aaof_md; eauto using subst_ofm, subst_fstm, subof_aof_dof, subof_aof_tof; [].
intros s' Hs'.
simp_sub.
apply Hhaltish; [].
eapply subof_aof_trans; eauto; done.
Qed.


Lemma subof_aof_aaof_trans :
  forall G1 G2 G3 s s',
    subof aof G1 s G2
    -> subof aaof G2 s' G3
    -> subof aaof G1 (compose s' s) G3.
Proof.
apply subof_subof_trans; eauto using aof_aaof, subst_aaof.
Qed. 


Lemma subof_aaof_dopple :
  forall G s,
    wf G
    -> subof aaof nil s G
    -> exists s',
         subof aof nil s' G
         /\ subrel (deq ceqt ceqm) nil s s' G.
Proof.
intros G s Hwf Hs.
revert s Hs.
induct Hwf.

(* nil *)
intros s Hs.
invert Hs; [].
intros n Htrunc <-.
invert Htrunc; [].
intros <- _.
exists id.
split; eauto using subof_sh, subrel_sh with sequence; done.

(* cn *)
intros G k _ IH Hofk s Hs.
so (subof_nil_cons_invert _#4 Hs) as (c & s' & ->).
invert Hs; [].
intros Hs' Haaof.
simp_sub in Haaof.
so (IH _ Hs') as (r & Hr & Hsr).
invert Haaof; [].
intros Hofc.
exists (dot c r).
split.
  apply subof_dot; auto; [].
  simp_sub.
  apply aof_cn; [].
  eapply ofc_subsume; eauto; [].
  apply subk_refl; [].
  eapply seq_ofk; eauto using wf_nil, subrel_deq_seq; done.

  apply subrel_dot; auto; [].
  simp_sub.
  apply deq_cn; auto using eqc_refl; done.

(* tm *)
intros G t _ IH Hoft s Hs.
so (subof_nil_cons_invert _#4 Hs) as (e & s' & ->).
invert Hs; [].
intros Hs' Haaof.
simp_sub in Haaof.
so (IH _ Hs') as (r & Hr & Hsr).
invert Haaof; [].
intros Hofe Hhaltish.
so (Hhaltish id (subof_id _ _)) as Hhalt.
simp_sub in Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
exists (dot v r).
split.
  apply subof_dot; auto; [].
  simp_sub.
  apply aof_tm; auto; [].
  #2 eapply oft_equiv.
    eapply preservations_tm; eauto using wf_nil; done.

    eapply seq_ofc_type; eauto using wf_nil, subrel_deq_seq; done.

  apply subrel_dot; auto; [].
  simp_sub.
  apply deq_tm; [].
  eapply tsteps_ceqt; eauto; done.

(* md *)
intros G k T _ IH HofT Hfst s Hs.
so (subof_nil_cons_invert _#4 Hs) as (M & s' & ->).
invert Hs; [].
intros Hs' HaaofM.
so (subof_nil_cons_invert _#4 Hs') as (c & s'' & ->).
invert Hs'; [].
intros Hs'' Haaofc.
so (IH _ Hs'') as (r & Hr & Hsr).
simp_sub in HaaofM; simp_sub in Haaofc.
invert Haaofc; [].
intros Hofc.
invert HaaofM; [].
clear HaaofM Haaofc Hs Hs'.
intros HofM Hhaltish HfstM.
so (Hhaltish id (subof_id _ _)) as Hhalt.
simp_sub in Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
so (preservations_fst _#7 wf_nil HofM Hsteps (subst_fsts _#3 Hfst) HfstM) as (c' & HfstV & Heqc).
exists (dot V (dot c' r)).
split.
  #2 apply subof_dot.
    apply subof_dot; auto; [].
    simp_sub.
    apply aof_cn; [].
    #2 eapply ofc_subsume.
      exact (eqc_reg _#4 wf_nil Heqc anderl).

      apply subk_refl; [].
      so (ofs_fst _#3 HofT Hfst) as Hofk.
      eapply seq_ofk; eauto using wf_nil, subrel_deq_seq; done.

    simp_sub.
    apply aof_md; auto; [].
    #2 eapply ofm_subsume.
      eapply preservations_md; eauto using wf_nil; done.

      apply subs_refl; [].
      eapply seq_ofs; eauto using wf_nil, subrel_deq_seq; done.

  #2 apply subrel_dot.
    apply subrel_dot; auto; [].
    simp_sub.
    apply deq_cn; auto; done.

    simp_sub.
    apply deq_md; auto; [].
    eapply msteps_ceqm; eauto; done.
Qed.


Lemma subst_aaof_ceqt :
  forall G1 s G2 e e' t,
    subof aaof G1 s G2
    -> wf G2
    -> ceqt G2 e e' t
    -> ceqt G1 (subst s e) (subst s e') (subst s t).
Proof.
intros G1 s G2 e e' t Hs Hwf Hceq.
so (ceqt_reg _#4 Hwf Hceq) as (Hofe & Hofe').
so (oft_reg _#3 Hwf Hofe) as Hoft.
apply closed_instances_ceqt; eauto using subst_oft, subof_aaof_dof; [].
intros s' Hs'.
simp_sub.
so (subof_aof_aaof_trans _#5 Hs' Hs) as Hss.
so (subof_aaof_dopple _ _ Hwf Hss) as (r & Hr & Hsr).
#2 apply (ceqt_trans _ _ (subst r e)).
  eapply funct_ceq_tm; eauto using wf_nil; done.
#2 apply (ceqt_trans _ _ (subst r e')).
  #2 apply ceqt_comp; eapply tcc_equiv.
    eapply subst_ceqt; eauto; done.

    apply eqc_symm; [].
    eapply seq_ofc_type; eauto using wf_nil, subrel_deq_seq; done.
apply ceqt_symm; [].
eapply funct_ceq_tm; eauto using wf_nil; done.
Qed.


Lemma subst_aaof_ceqm :
  forall G1 s G2 p M M' T,
    subof aaof G1 s G2
    -> wf G2
    -> ceqm G2 p M M' T
    -> ceqm G1 p (subst s M) (subst s M') (subst s T).
Proof.
intros G1 s G2 p M M' T Hs Hwf Hceq.
so (ceqm_reg _#5 Hwf Hceq) as (HofM & HofM').
so (ofm_reg _#4 Hwf HofM) as HofT.
#2 apply closed_instances_ceqm; eauto using subst_ofm, subof_aaof_dof.
  intros ->.
  so (ceqm_fst _#4 Hwf Hceq) as (k & c & c' & Hfst & HfstM & HfstM' & Heqc).
  exists (subst s k), (subst s c), (subst s c').
  do2 3 split; eauto using subst_fsts, subst_fstm, subst_eqc, subof_aaof_dof, subof_dof_tof, subof_aaof_sof; done.

  intros s' Hs'.
  simp_sub.
  so (subof_aof_aaof_trans _#5 Hs' Hs) as Hss.
  so (subof_aaof_dopple _ _ Hwf Hss) as (r & Hr & Hsr).
  #2 apply (ceqm_trans _ _ _ (subst r M)).
    eapply funct_ceq_md; eauto using wf_nil; done.
  #2 apply (ceqm_trans _ _ _ (subst r M')).
    #2 apply ceqm_comp; eapply mcc_subsume.
      eapply subst_ceqm; eauto; done.
  
      apply subs_refl; [].
      apply eqs_symm; [].
      eapply seq_ofs; eauto using wf_nil, subrel_deq_seq; done.
  apply ceqm_symm; [].
  eapply funct_ceq_md; eauto using wf_nil; done.
Qed.


Lemma trs_formation_aaof :
  forall G G' s,
    wf G
    -> trs G G' s
    -> subof aaof G s G'.
Proof.
intros G G' s Hwf Htrs.
revert Hwf.
induct Htrs.

(* nil *)
intros _.
apply subof_id; done.

(* cn *)
intros G G' k s Htr IH Hwf.
invert Hwf; [].
intros HwfG Hofk.
so (IH HwfG) as Hs.
so (trctx_formation _ _ (trs_trctx _#3 Htr) HwfG) as HwfG'.
#2 apply subof_dot.
  apply (subof_aof_aaof_trans _ G); eauto using subof_sh with sequence; done.

  simp_sub.
  apply aaof_cn; [].
  #2 eapply ofc_subsume.
    ofvar; done.

    apply subk_refl; [].
    replsub (subst (compose s sh1) k) with (shift1 (subst s k)).
    apply weaken1_eqk; [].
    rewrite -> (trs_subst_ofk _#4 Htr Hofk); [].
    apply eqk_refl; auto; done.

(* tm *)
intros G G' t s Htr IH Hwf.
invert Hwf; [].
intros HwfG Hoft.
so (IH HwfG) as Hs.
so (trctx_formation _ _ (trs_trctx _#3 Htr) HwfG) as HwfG'.
#2 apply subof_dot.
  apply (subof_aof_aaof_trans _ G); eauto using subof_sh with sequence; done.

  simp_sub.
  #2 apply aaof_tm.
    #2 eapply oft_equiv.
      ofvar; done.

      rewrite -> subst_compose; [].
      apply weaken1_eqc_type; [].
      rewrite -> (trs_subst_ofc _#5 Htr Hoft); [].
      apply eqc_refl; auto; done.

    intros s' Hs'.
    apply tvalue_thalt; [].
    eapply subst_tvalue; eauto using subof_aof_oof; [].
    ofvar; done.

(* md *)
intros G G' T k t s e Htr IH Hsplit Hsnd Hwf.
invert Hwf; [].
intros Ta HwfG HofT Hfst Heq.
f_apply (subst (dot nonsense id)) in Heq as Heq'; simp_sub in Heq'.
subst Ta; clear Heq.
so (IH HwfG) as Hs; clear IH.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
#2 apply subof_dot.
  #2 apply subof_dot.
    apply (subof_aof_aaof_trans _ G); eauto using subof_sh with sequence; done.

    simp_sub.
    apply aaof_cn; [].
    #2 eapply ofc_subsume.
      ofvar; done.

      rewrite -> subst_compose; [].
      apply weaken2_subk; [].
      apply subk_refl; [].
      rewrite -> (trs_subst_ofk _#4 Htr Hofk); [].
      apply eqk_refl; auto; done.

  simp_sub.
  #2 apply aaof_tm.
    #3 sor (refine (snd_formation _#7 Hwf _ Hsnd (subst_split _#4 Hsplit) _)) as Hofe.
      ofvar; done.

      ofvar; done.
    simp_sub in Hofe.
    simp_sub in Hofe.
    rewrite -> subst_dot1_sh in Hofe; [].
    force_exact Hofe; [].
    f_equal; [].
    replsub (subst (dot (var 1) (compose s (sh 2))) t) with (shift1 (subst (dot (var 0) (compose s sh1)) t)).
    f_equal; [].
    rewrite <- (subst_under_id (S (length G)) t) at 1; [].
    refine (sid_ofc _#6 _ (split_formation _#4 HofT Hsplit ander)); [].
    rewrite -> under_succ; [].
    apply subrel_sid_bind_cn; [].
    apply trs_sid; eauto; done.

    intros s' Hs'.
    so (subst_snd _#3 s' Hsnd) as Hsndx.
    #2 refine (snd_halt_iff pure _#3 _ Hsndx andel _).
      assert (ofm (G; cl_cn k; cl_md (shift1 T) (var 0)) pure (var 0) (subst (sh 2) T)) as Hof by ofvar.
      exact (subst_ofm _#6 (subof_aof_dof _#3 Hs') Hof).

      apply mvalue_mhalt; [].
      eapply subst_mvalue; eauto using subof_aof_oof; [].
      ofvar; done.
Qed.


Inductive trsc : context -> context -> sub -> Prop :=
| trsc_nil
    : trsc nil nil id

| trsc_cn {G G' k s}
    : trsc G G' s
      -> trsc (G; cl_cn k) (G'; cl_cn k) (dot (var 0) (compose s sh1))

| trsc_tm {G G' t s}
    : trsc G G' s
      -> trsc (G; cl_tm t) (G'; cl_tm t) (dot (var 0) (compose s sh1))

| trsc_md {G G' T k t s M c}
    : trsc G G' s
      -> split T k t
      -> join (subst (sh 2) T) (var 1) (var 0) M
      -> fstm (G'; cl_cn k; cl_tm t) M c
      -> trsc (G; cl_cn k; cl_md (shift1 T) (var 0)) (G'; cl_cn k; cl_tm t)
         (dot M (dot c (compose s (sh 2))))
.


Lemma trsc_trctx :
  forall G s G',
    trsc G G' s
    -> trctx G G'.
Proof.
intros G s G' H.
induct H; eauto using trctx.
Qed.


Lemma trsc_seq_and_formation :
  forall G G' s,
    wf G
    -> trsc G G' s
    -> subrel seq G' (under (length G) id) s G /\ subof aaof G' s G.
Proof.
intros G G' s Hwf Htrs.
revert Hwf.
induct Htrs.

(* nil *)
intros _.
split.
  simpl.
  apply subrel_id; done.

  apply subof_id; done.

(* cn *)
intros G G' k s Htr IH Hwf.
invert Hwf; [].
intros HwfG Hofk.
so (IH HwfG) as (Hss & Hs).
so (trctx_formation _ _ (trsc_trctx _#3 Htr) HwfG) as HwfG'.
split.
  simpl length.
  rewrite -> under_succ; [].
  rewrite <- (subst_under_id (length G) (cl_cn k)) at 1; [].
  eapply subrel_wclo_bind; auto using seq_behaved, seq_wclo, iscl; done.

  #2 apply subof_dot.
    apply (subof_aof_aaof_trans _ G'); eauto using subof_sh with sequence; done.

    simp_sub.
    apply aaof_cn; [].
    #2 eapply ofc_subsume.
      ofvar; done.

      apply subk_refl; [].
      replsub (subst (compose s sh1) k) with (shift1 (subst s k)).
      apply weaken1_eqk; [].
      so (seq_ofk _#5 HwfG' Hss Hofk) as Heqk.
      rewrite -> subst_under_id in Heqk; [].
      exact Heqk.

(* tm *)
intros G G' t s Htr IH Hwf.
invert Hwf; [].
intros HwfG Hoft.
so (IH HwfG) as (Hss & Hs).
so (trctx_formation _ _ (trsc_trctx _#3 Htr) HwfG) as HwfG'.
split.
  simpl length.
  rewrite -> under_succ; [].
  rewrite <- (subst_under_id (length G) (cl_tm t)) at 1; [].
  eapply subrel_wclo_bind; auto using seq_behaved, seq_wclo, iscl; done.

  #2 apply subof_dot.
    apply (subof_aof_aaof_trans _ G'); eauto using subof_sh with sequence; done.

    simp_sub.
    #2 apply aaof_tm.
      #2 eapply oft_equiv.
        ofvar; done.
  
        rewrite -> subst_compose; [].
        apply weaken1_eqc_type; [].
        so (seq_ofc_type _#5 HwfG' Hss Hoft) as Heqc.
        rewrite -> subst_under_id in Heqc; [].
        exact Heqc.

      intros s' Hs'.
      apply tvalue_thalt; [].
      eapply subst_tvalue; eauto using subof_aof_oof; [].
      ofvar; done.

(* md *)
intros G G' T k t s M c Htr IH Hsplit Hjoin HfstM Hwf.
invert Hwf; [].
intros Ta HwfG HofT Hfst Heq.
f_apply (subst (dot nonsense id)) in Heq as Heq'; simp_sub in Heq'.
subst Ta; clear Heq.
so (IH HwfG) as (Hss & Hs); clear IH.
so (trctx_formation _ _ (trsc_trctx _#3 Htr) HwfG) as HwfG'.
so (seq_ofs _#5 HwfG' Hss HofT) as HeqT.
rewrite -> subst_under_id in HeqT; [].
so (eqs_reg _#3 HwfG' HeqT) as (HofT' & _).
so (split_formation _#4 HofT' Hsplit) as (Hofk & Hoft).
so (wf_tm (wf_cn HwfG' Hofk) Hoft) as Hwf'.
#3 sor (refine (join_formation_and_fst_o _#7 Hwf' (weaken2_ofs _#4 HofT') Hjoin (subst_split _#4 Hsplit) _ _)) as H.
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
destruct H as (HofM & ca & HfstMa & Heqc).
so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
split.
  simpl length.
  simp_sub.
  #2 apply subrel_dot.
    #2 apply subrel_dot.
      eapply subrel_seq_weaken; eauto with sequence; done.

      rewrite -> subst_compose; [].
      rewrite -> subst_under_id; [].
      simp_sub.
      apply seq_cn; auto; done.

    simp_sub.
    apply seq_md; done.

  #2 apply subof_dot.
    #2 apply subof_dot.
      apply (subof_aof_aaof_trans _ G'); eauto using subof_sh with sequence; done.

      simp_sub.
      apply aaof_cn; [].
      #2 eapply ofc_subsume.
        exact (eqc_reg _#4 Hwf' Heqc anderl).

        rewrite -> subst_compose; [].
        apply weaken2_subk; [].
        apply subk_refl; [].
        so (split_formation _#4 HofT Hsplit) as (Hofk_G & _).
        so (seq_ofk _#5 HwfG' Hss Hofk_G) as Heqk.
        rewrite -> subst_under_id in Heqk; [].
        exact Heqk.

    simp_sub.
    #2 apply aaof_md; auto.
      eapply ofm_subsume; eauto; [].
      rewrite -> subst_compose; [].
      apply weaken2_subs; [].
      apply subs_refl; auto; done.

      intros s' Hs'.
      so (subst_join _#4 s' Hjoin) as Hjoinx.
      #3 refine (join_halt_iff _#6 _ _ Hjoinx andel _).
        simp_sub; apply subst_split; eauto; done.

        assert (oft (G'; cl_cn k; cl_tm t) (var 0) (shift1 t)) as Hof by ofvar.
        so (subst_oft _#5 (subof_aof_dof _#3 Hs') Hof) as Hof'.
        rewrite <- subst_dot1_sh in Hof'; [].
        simp_sub in Hof'.
        simp_sub.
        exact Hof'.

        apply tvalue_thalt; [].
        eapply subst_tvalue; eauto using subof_aof_oof; [].
        ofvar; done.
Qed.


Lemma trsc_seq :
  forall G G' s,
    wf G
    -> trsc G G' s
    -> subrel seq G' (under (length G) id) s G.
Proof.
intros G G' s Hwf Htrs.
exact (trsc_seq_and_formation _#3 Hwf Htrs andel).
Qed.


Lemma trsc_formation :
  forall G G' s,
    wf G
    -> trsc G G' s
    -> subof aaof G' s G.
Proof.
intros G G' s Hwf Htrs.
exact (trsc_seq_and_formation _#3 Hwf Htrs ander).
Qed.



Lemma trsc_subst_ofk :
  forall G G' s k,
    wf G
    -> trsc G G' s
    -> ofk G k
    -> eqk G' k (subst s k).
Proof.
intros G G' s k Hwf Htrs Hofk.
so (trsc_seq _#3 Hwf Htrs) as Hss.
so (trctx_formation _ _ (trsc_trctx _#3 Htrs) Hwf) as Hwf'.
so (seq_ofk _#5 Hwf' Hss Hofk) as Heqk.
rewrite -> subst_under_id in Heqk; [].
exact Heqk.
Qed.


Lemma trsc_subst_ofc :
  forall G G' s c k,
    wf G
    -> trsc G G' s
    -> ofc G c k
    -> eqc G' c (subst s c) k.
Proof.
intros G G' s c k Hwf Htrs Hofc.
so (trsc_seq _#3 Hwf Htrs) as Hss.
so (trctx_formation _ _ (trsc_trctx _#3 Htrs) Hwf) as Hwf'.
so (seq_ofc _#6 Hwf' Hss Hofc) as Heqc.
rewrite -> !subst_under_id in Heqc; [].
exact Heqc.
Qed.


Lemma trsc_subst_ofs :
  forall G G' s T,
    wf G
    -> trsc G G' s
    -> ofs G T
    -> eqs G' T (subst s T).
Proof.
intros G G' s T Hwf Htrs Hofs.
so (trsc_seq _#3 Hwf Htrs) as Hss.
so (trctx_formation _ _ (trsc_trctx _#3 Htrs) Hwf) as Hwf'.
so (seq_ofs _#5 Hwf' Hss Hofs) as Heqs.
rewrite -> subst_under_id in Heqs; [].
exact Heqs.
Qed.


Lemma trctx_fun :
  forall G G1 G2,
    trctx G G1
    -> trctx G G2
    -> G1 = G2.
Proof.
intros G G1 G2 H Htr.
revert G2 Htr.
induct H.

(* nil *)
intros G2 Htr.
invert Htr; auto; done.

(* cn *)
intros G G1 k _ IH G2 Htr.
invert Htr; [].
intros G2' Htr' <-.
f_equal; [].
apply IH; auto; done.

(* tm *)
intros G G1 t _ IH G2 Htr.
invert Htr; [].
intros G2' Htr' <-.
f_equal; [].
apply IH; auto; done.

(* md *)
intros G G1 T k t _ IH Hsplit G2 Htr.
invert Htr; [].
intros G2' Ta t' Htr' Hsplit' Heq <-.
f_apply (subst (dot nonsense id)) in Heq as Heq'; simp_sub in Heq'.
subst Ta; clear Heq.
so (split_fun _#5 Hsplit Hsplit') as (_ & <-).
f_equal; f_equal; [].
apply IH; auto; done.
Qed.


Lemma can_trs :
  forall G,
    wf G
    -> exists G' s, trs G G' s.
Proof.
intros G H.
induct H.

(* nil *)
eauto using trs; done.

(* cn *)
intros G k _ (G' & s & Htrs) _.
eauto using trs; done.

(* tm *)
intros G t _ (G' & s & Htrs) _.
eauto using trs; done.

(* md *)
intros G k T Hwf (G' & s & Htrs) HofT Hfst.
so (can_split _ (ofs_hsg _ _ HofT)) as (ka & t & Hsplit).
so (split_fsts_eq _#4 Hsplit Hfst); subst ka.
so (can_snd _ (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (e & Hsnd).
eauto using trs; done.
Qed.


Lemma can_trsc :
  forall G,
    wf G
    -> exists G' s, trsc G G' s.
Proof.
intros G H.
induct H.

(* nil *)
eauto using trsc; done.

(* cn *)
intros G k _ (G' & s & Htrs) _.
eauto using trsc; done.

(* tm *)
intros G t _ (G' & s & Htrs) _.
eauto using trsc; done.

(* md *)
intros G k T Hwf (G' & s & Htrs) HofT Hfst.
so (can_split _ (ofs_hsg _ _ HofT)) as (ka & t & Hsplit).
so (split_fsts_eq _#4 Hsplit Hfst); subst ka.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (M & Hjoin).
so (trctx_formation _ _ (trsc_trctx _#3 Htrs) Hwf) as Hwf'.
so (eqs_reg _#3 Hwf' (trsc_subst_ofs _#4 Hwf Htrs HofT)) as (HofT' & _).
so (join_vars_formation _#5 Hwf' HofT' Hsplit Hjoin) as HofM.
so (split_formation _#4 HofT' Hsplit) as (Hofk' & Hoft').
so (can_fstm _#3 (wf_tm (wf_cn Hwf' Hofk') Hoft') HofM) as (c & HfstM).
eauto using trsc; done.
Qed.


Lemma trctx_trs :
  forall G G', 
    wf G
    -> trctx G G' 
    -> exists s, trs G G' s.
Proof.
intros G G' Hwf HtrG.
so (can_trs _ Hwf) as (G'' & s & Htrs).
so (trctx_fun _#3 HtrG (trs_trctx _#3 Htrs)); subst G''.
eauto.
Qed.


Lemma trs_trsc_eq :
  forall G G1 G2 s1 s2,
    trs G G1 s1
    -> trsc G G2 s2
    -> G1 = G2.
Proof.
intros G G1 G2 s1 s2 Htrs Htrsc.
exact (trctx_fun _#3 (trs_trctx _#3 Htrs) (trsc_trctx _#3 Htrsc)).
Qed.


Lemma trsc_reverse :
  forall G G' s,
    wf G
    -> trsc G G' s
    -> exists s', trs G G' s'.
Proof.
intros G G' s Hwf Htrsc.
so (can_trs _ Hwf) as (Ga' & s' & Htrs).
so (trs_trsc_eq _#5 Htrs Htrsc); subst Ga'.
eauto; done.
Qed.


Lemma trsc_trs_inverse :
  forall G G' s s',
    wf G
    -> trs G G' s
    -> trsc G G' s'
    -> subrel (deq ceqt ceqm) G' (under (length G') id) (compose s s') G'.
Proof.
intros G G' s s' Hwf Htrs Htrsc.
revert G' s s' Htrs Htrsc.
induct Hwf.

(* nil *)
intros G' s s' Htrs Htrsc.
invert Htrs; invert Htrsc; [].
intros <- <- _ <-.
simpl.
apply subrel_id; done.

(* cn *)
intros G k Hwf IH Hofk Gx' sx sx' Htrsx Htrscx.
invert Htrsx; clear Htrsx; [].
intros G' s Htrs <- <-.
invert Htrscx; clear Htrscx; [].
intros s' Htrsc _ <-.
simpl length.
rewrite -> under_succ; [].
simp_sub.
#2 apply subrel_dot.
  rewrite <- compose_assoc; [].
  apply (subrel_deq_ceq_weaken _ G'); eauto using trctx_formation, trs_trctx with sequence; done.

  rewrite -> subst_compose; [].
  rewrite -> subst_under_id; [].
  simp_sub.
  #2 apply deq_cn.
    ofvar; done.

    apply eqc_refl; ofvar; done.

(* tm *)
intros G t Hwf IH Hoft Gx' sx sx' Htrsx Htrscx.
invert Htrsx; clear Htrsx; [].
intros G' s Htrs <- <-.
invert Htrscx; clear Htrscx; [].
intros s' Htrsc _ <-.
simpl length.
rewrite -> under_succ; [].
simp_sub.
#2 apply subrel_dot.
  rewrite <- compose_assoc; [].
  apply (subrel_deq_ceq_weaken _ G'); eauto using trctx_formation, trs_trctx with sequence; done.

  rewrite -> subst_compose; [].
  rewrite -> subst_under_id; [].
  simp_sub.
  apply deq_tm; [].
  apply ceqt_refl; [].
  ofvar; done.

(* md *)
intros G k T Hwf IH HofT Hfst Gx' sx sx' Htrsx Htrscx.
invert Htrsx; clear Htrsx; [].
intros G' Ta t s e Htrs Hsplit Hsnd Heq <- <-.
f_apply (subst (dot nonsense id)) in Heq as Heq'; simp_sub in Heq'.
subst Ta; clear Heq.
invert Htrscx; clear Htrscx; [].
intros Ta s' M c Htrsc _ Hjoin HfstM Heq _ <-.
f_apply (subst (dot nonsense id)) in Heq as Heq'; simp_sub in Heq'.
subst Ta; clear Heq.
so (trctx_formation _ _ (trs_trctx _#3 Htrs) Hwf) as Hwf'.
so (eqs_reg _#3 Hwf' (trsc_subst_ofs _#4 Hwf Htrsc HofT)) as (HofT' & _).
so (split_formation _#4 HofT' Hsplit) as (Hofk & Hoft).
simpl length.
rewrite -> !under_succ.
simp_sub.
#2 apply subrel_dot.
  #2 apply subrel_dot.
    rewrite <- compose_assoc; [].
    apply (subrel_deq_ceq_weaken _ G'); eauto using trctx_formation, trs_trctx with sequence; done.

    rewrite -> subst_compose; [].
    rewrite -> subst_under_id; [].
    simp_sub.
    #2 apply deq_cn.
      ofvar; done.

      #5 eapply join_fst; eauto.
        eauto using wf; done.

        apply weaken2_ofs; auto; done.

        apply subst_split; eauto; done.

        ofvar; done.

        simp_sub.
        simp_sub.
        rewrite -> subst_dot1_sh; [].
        ofvar; done.

  simp_sub.
  apply deq_tm; [].
  repl (subst (dot (var 1) (compose (under (length G') id) (sh 2))) t) with (subst (sh 1) (subst (under (S (length G')) id) t)).
    rewrite <- subst_compose; [].
    simp_sub.
    reflexivity.
  rewrite -> subst_under_id; [].
  repl (subst (sh 1) t) with (subst1 (var 1) (subst (under 1 (sh 2)) t)).
    simp_sub.
    simp_sub.
    rewrite -> subst_dot1_sh; [].
    reflexivity.
  so (subst_snd _#3 (dot M (dot c (sh 2))) Hsnd) as Hsndx; simp_sub in Hsndx.
  so (subst_snd _#3 (dot M (dot c (compose s' (sh 2)))) Hsnd) as Hsndx'; simp_sub in Hsndx'.
  #2 apply (ceqt_trans _ _ (subst (dot M (dot c (sh 2))) e)).
    #4 eapply snd_join_inverse; eauto.
      apply weaken2_ofs; auto; done.
  
      apply subst_split; eauto; done.
  
      ofvar; done.
  
      simp_sub.
      simp_sub.
      rewrite -> subst_dot1_sh; [].
      ofvar; done.

    #2 apply ceqt_comp; apply (tcc_equiv _#7 (subst1 c (subst (under 1 (sh 2)) t))).
      #3 eapply snd_eqs; eauto.
        rewrite -> subst_compose; [].
        refine (subst_eqs _#5 (weaken2 _#4) _); [].
        eapply (trsc_subst_ofs G); eauto; done.
  
        apply join_vars_formation; auto; done.
  
        apply subst_split; eauto; done.

      #2 eapply funct_cn_type; eauto using wf.
        eapply eqc_symm; [].
        #4 eapply join_fst; eauto using wf.
          apply weaken2_ofs; eauto; done.

          apply subst_split; eauto; done.

          ofvar; done.

          simp_sub.
          simp_sub.
          rewrite -> subst_dot1_sh; [].
          ofvar; done.

        simp_sub.
        eapply subst_ofc_type; eauto; [].
        prove_subof; done.
Qed.


Lemma trsc_subst_ofc_type :
  forall G G' s c,
    wf G
    -> trsc G G' s
    -> ofc G c kd_type
    -> eqc G' c (subst s c) kd_type.
Proof.
intros G G' s c Hwf Htrs Hofc.
replsub kd_type with (subst s kd_type).
eapply trsc_subst_ofc; eauto; done.
Qed.


Lemma trp_correct_conv :
  forall G G' s T k t M c e e',
    wf G
    -> trsc G G' s
    -> trp G M T c e
    -> split T k t
    -> snd T (subst s M) e'
    -> ceqt G' e e' (subst1 c t).
Proof.
intros G G' s' T k t M c e e' Hwf Htrsc Htrp Hsplit Hsnd.
apply ceqt_with_wf; intro Hwf'.
so (trsc_reverse _ _ _ Hwf Htrsc) as (s & Htrs).
so (split_fsts _#3 Hsplit) as Hfst.
so (trp_fstm _#5 Htrp) as HfstM.
so (trp_fst _#6 Hwf Htrp Hfst) as Hofc.
so (trp_formation _#8 Hwf (trs_trctx _#3 Htrs) Htrp Hsplit) as Hofe.
so (trp_ofm _#5 Hwf Htrp) as HofM.
so (ofm_reg _#4 Hwf HofM) as HofT.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
assert (eqc G' (subst (dot (subst s' c) s') t) (subst1 c t) kd_type) as Heqsubst.
  replsub (subst (dot (subst s' c) s') t) with (subst s' (subst1 c t)).
  apply eqc_symm; [].
  eapply (trsc_subst_ofc_type G); eauto; [].
  eapply subst1_ofc_type; eauto; done.
so (can_join _ c (subst s e) (ofs_hsg _ _ HofT)) as (M' & Hjoin).
so (funct_ceq_tm _#6 Hwf' Hwf' (trsc_trs_inverse _#4 Hwf Htrs Htrsc) Hofe) as Hceqe.
rewrite -> !subst_under_id in Hceqe.
so (trp_correct _#8 Htrs Htrp Hjoin) as HceqM.
so (subst_aaof_ceqm _#7 (trsc_formation _#3 Hwf Htrsc) Hwf HceqM) as HceqMs.
so (can_snd _ (subst s' M) (subst_hsg s' _ (ofs_hsg _ _ HofT))) as (f & Hsndf).
so (can_snd _ (subst s' M') (subst_hsg s' _ (ofs_hsg _ _ HofT))) as (f' & Hsndf').
#3 sor (refine (snd_ceq _#9 Hwf' _ HceqMs _ Hsndf Hsndf')) as Hceqf.
  apply subst_split; eauto; done.

  eapply subst_fstm; eauto using trsc_formation, subof_aaof_tof; done.
simp_sub in Hceqf.
so (subst_join _#4 s' Hjoin) as Hjoinx; simp_sub in Hjoinx.
#2 apply (ceqt_trans _ _ f').
  apply (ceqt_trans _ _ (subst (compose s s') e)); auto; [].
  #2 apply ceqt_comp; apply (tcc_equiv _#7 (subst1 (subst s' c) (subst (under 1 s') t))).
    eapply (snd_join_inverse _ (subst s' T)); eauto using subst_split, subst_ofs, subst_ofc, trsc_formation, subof_aaof_sof; [].
    simp_sub.
    rewrite -> subst_compose; [].
    replsub (subst (dot (subst s' c) s') t) with (subst s' (subst1 c t)).
    eapply subst_oft; eauto using trsc_formation, subof_aaof_dof; [].
    rewrite <- (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc Hoft)); [].
    eapply subst_oft; eauto using trs_formation; done.

    simp_sub; assumption.
#2 apply (ceqt_trans _ _ f).
  apply ceqt_symm; [].
  apply ceqt_comp; eapply tcc_equiv; eauto; done.
apply ceqt_comp; eapply tcc_equiv; eauto; [].
replsub (subst (dot (subst s' c) s') t) with (subst1 (subst s' c) (subst (under 1 s') t)).
eapply (snd_eqs _ (subst s' T) T); eauto using subst_split, subst_ofm, subst_fstm, trsc_formation, subof_aaof_dof, subof_aaof_tof; [].
apply eqs_symm; [].
eapply (trsc_subst_ofs G); eauto; done.
Qed.


Lemma ceqt_exists_eta :
  forall k t e,
    oft nil e (cn_exists k t)
    -> ceqt nil e (tm_unpack e (tm_pack (var 1) (var 0) (subst (sh 2) k) (subst (under 1 (sh 2)) t))) (cn_exists k t).
Proof.
intros k t e Hofe.
so (ofc_exists_invert _#4 (oft_reg _#3 wf_nil Hofe)) as (Hofk & Hoft).
assert (oft nil
          (tm_unpack e
             (tm_pack (var 1) (var 0) (subst (sh 2) k) (subst (under 1 (sh 2)) t)))
          (cn_exists k t)) as Hof.
  eapply oft_unpack; eauto with static; [].
  simp_sub.
  #3 eapply oft_pack.
    ofvar; done.

    simp_sub; simp_sub; rewrite -> subst_dot1_sh; ofvar; done.

    eapply subst_ofc_type; eauto; prove_subof; done.
#2 apply (ceqt_conditional (thalt nil e)); eauto.
  intro Hhalt.
  eapply Inversion.unpack_halt_inversion'; eauto; done.
intro Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
invert (Progress.canon_tm _ _ Hofv Consistent.similar_exists Hval); [].
intros c v' k' t' <-.
invert Hval; [].
intros Hval'.
apply ceqt_symm; [].
#2 apply (ceqt_trans _ _ (tm_pack c v' k t)).
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    eapply (star_map _ _ _ _ (fun z => tm_unpack z _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply tstep_unpack2; auto; done.
  simp_sub.
  apply star_refl; done.

  #2 apply (ceqt_trans _ _ (tm_pack c v' k' t')).
    so (Inversion.pack_oft_inversion_exists _#7 wf_nil Hofv) as (_ & _ & Hofc & Hofv').
    apply teme_ceqt; eauto using teme with static; done.

    apply ceqt_symm; [].
    apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqt_exists_eta' :
  forall G k t e,
    oft G e (cn_exists k t)
    -> ceqt G e (tm_unpack e (tm_pack (var 1) (var 0) (subst (sh 2) k) (subst (under 1 (sh 2)) t))) (cn_exists k t).
Proof.
intros G k t e Hof.
apply ceqt_with_wf; intro Hwf.
so (ofc_exists_invert _#4 (oft_reg _#3 Hwf Hof)) as (Hofk & Hoft).
#2 apply closed_instances_ceqt; auto.
  #2 eapply oft_unpack; eauto.
    simp_sub.
    #3 apply oft_pack.
      ofvar; done.

      simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
      ofvar; done.

      eapply subst_ofc_type; eauto; prove_subof; done.

    apply ofc_exists; auto; done.

  intros s Hs.
  simp_sub.
  simp_sub.
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as Hof'; simp_sub in Hof'.
  so (ceqt_exists_eta _#3 Hof') as Hceq.
  simp_sub in Hceq.
  exact Hceq.
Qed.


Lemma tri_correct_conv :
  forall G G' s T k t M e e',
    wf G
    -> trsc G G' s
    -> tri G M T e
    -> split T k t
    -> snd (subst (sh 2) T) (var 0) e'
    -> ceqt G' e (tm_letm (subst s M) (tm_pack (var 1) e' (subst (sh 2) k) (subst (under 1 (sh 2)) t))) (cn_exists k t).
Proof.
intros G G' s' T k t M e e' Hwf Htrsc Htri Hsplit Hsnd.
apply ceqt_with_wf; intro Hwf'.
so (trsc_reverse _ _ _ Hwf Htrsc) as (s & Htrs).
so (split_fsts _#3 Hsplit) as Hfst.
so (tri_formation _#7 Hwf (trs_trctx _#3 Htrs) Htri Hsplit) as Hofe.
so (tri_ofm _#4 Hwf Htri) as HofM.
so (ofm_reg _#4 Hwf HofM) as HofT.
so (eqs_reg _#3 Hwf' (trsc_subst_ofs _#4 Hwf Htrsc HofT)) as (HofT' & _).
so (split_formation _#4 HofT' Hsplit) as (Hofk & Hoft).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (N & Hjoin).
so (join_vars_fst_o _#5 Hwf' HofT' Hsplit Hjoin) as (c & HfstN & Heqc).
so (tri_correct _#7 Htrs Htri Hjoin) as HceqM.
so (subst_aaof_ceqm _#7 (trsc_formation _#3 Hwf Htrsc) Hwf HceqM) as HceqMs.
simp_sub in HceqMs.
assert (ceqm G' impure (subst s' M) (md_unpack e N T) T) as HceqMs'.
  #2 eapply ceqm_trans.
    apply ceqm_comp; eapply mcc_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply (trsc_subst_ofs G); eauto; done.

    apply ceqm_symm; [].
    #3 apply ceqm_comp; eapply mcc_unpack.
      so (funct_ceq_tm _#6 Hwf' Hwf' (trsc_trs_inverse _#4 Hwf Htrs Htrsc) Hofe) as Hceq.
      rewrite -> subst_under_id in Hceq; [].
      simp_sub in Hceq.
      exact Hceq.

      so (subst_join _#4 (dot (var 0) (dot (var 1) (compose s' (sh 2)))) Hjoin) as Hjoinx; simp_sub in Hjoinx.
      apply ceqm_forget_pure; [].
      #4 eapply join_eqs; eauto.
        rewrite -> subst_compose; [].
        refine (subst_eqs _#5 (weaken2 _#4) _); [].
        apply (trsc_subst_ofs G); auto; done.

        apply subst_split; eauto; done.

        ofvar; done.

        simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
        ofvar; done.
        
      apply (trsc_subst_ofs G); auto; done.
apply ceqt_symm; [].
#2 eapply ceqt_trans.
  #4 apply ceqt_comp; eapply tcc_letm.
    exact HceqMs'.

    exact (split_fsts _#3 Hsplit).

    apply ceqt_refl; [].
    simp_sub.
    #3 apply oft_pack.
      ofvar; done.

      #4 eapply snd_formation; eauto.
        apply wf_md; auto; done.

        ofvar; done.

        apply subst_split; eauto; done.

        ofvar; done.

      eapply subst_ofc_type; eauto; prove_subof; done.

    apply ofc_exists; eauto; done.
#2 apply reqt_ceqt.
  #4 eapply oft_letm.
    eapply ofm_unpack; eauto; [].
    apply ofm_forget; [].
    eapply join_vars_formation; eauto; done.
    
    eapply split_fsts; eauto; done.

    simp_sub.
    #3 apply oft_pack.
      ofvar; done.

      #3 eapply snd_formation; eauto using wf.
        ofvar; done.

        apply subst_split; eauto; done.

        ofvar; done.

      eapply subst_ofc_type; eauto; prove_subof; done.

    apply ofc_exists; auto; done.
#2 eapply reqt_trans.
  apply reqt_commute_letm_unpack; done.
#2 eapply reqt_trans.
  #2 eapply reqt_unpack_bind.
    apply ceqt_refl; eauto; done.
  #2 eapply reqt_trans.
    #2 apply reqt_letm_beta_instances.
      intros r Hr.
      so (subst_join _#4 r Hjoin) as Hjoinx; simp_sub in Hjoinx.
      #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit) _ Hjoinx andel _).
        simp_sub.
        cut (oft nil (subst r (var 0)) (subst r (shift1 t))).
          intro H.
          rewrite <- subst_dot1_sh in H.
          simp_sub in H.
          exact H.
        eapply subst_oft; eauto using subof_aof_dof; [].
        ofvar; done.

        replsub (project r 0) with (subst r (var 0)).
        apply tvalue_thalt; [].
        eapply subst_tvalue; eauto using subof_aof_oof; [].
        ofvar; done.

      exact HfstN.
  simp_sub.
  simp_sub.
  #2 eapply reqt_trans.
    apply reqt_pack; [].
    eapply ceqt_reqt; [].
    apply ceqt_symm; [].
    #4 refine (snd_join_inverse _#8 _ (subst_split _#4 Hsplit) _ _ Hjoin _).
      apply weaken2_ofs; eauto; done.

      ofvar; done.

      simp_sub.
      simp_sub.
      rewrite -> subst_dot1_sh; [].
      ofvar; done.

      so (subst_snd _#3 (dot N (dot (shift1 c) (sh 2))) Hsnd) as Hsndx; simp_sub in Hsndx.
      exact Hsndx.
  #2 eapply reqt_trans.
    #3 eapply reqt_pack_annot.
      eapply eqc_symm; [].
      relquest.
        apply weaken1_eqc; eauto; done.
        
        simp_sub; reflexivity.

      apply eqk_refl; [].
      apply weaken2_ofk; eauto; done.

      apply eqc_refl; [].
      eapply subst_ofc_type; eauto; prove_subof; done.
  simp_sub.
  apply reqt_refl; done.
eapply ceqt_reqt; [].
apply ceqt_symm; [].
apply ceqt_exists_eta'; eauto; done.
Qed.


Lemma trt_correct_conv :
  forall G G' s t e e',
    wf G
    -> trsc G G' s
    -> trt G e t e'
    -> ceqt G' e' (subst s e) t.
Proof.
intros G G' s' t e e' Hwf Htrsc Htrt.
apply ceqt_with_wf; intro Hwf'; [].
so (trsc_reverse _ _ _ Hwf Htrsc) as (s & Htrs).
so (trt_formation _#5 Hwf (trs_trctx _#3 Htrs) Htrt) as Hofe'.
so (trt_reg _#4 Hwf Htrt) as Hoft.
so (trt_correct _#6 Htrs Htrt) as Hceqt.
so (subst_aaof_ceqt _#6 (trsc_formation _#3 Hwf Htrsc) Hwf Hceqt) as Heqts.
simp_sub in Heqts.
apply ceqt_symm; [].
#2 apply (ceqt_trans _ _ (subst (compose s s') e')).
  apply ceqt_comp; eapply tcc_equiv; eauto; [].
  apply eqc_symm; [].
  eapply (trsc_subst_ofc_type G); eauto; done.

  so (funct_ceq_tm _#6 Hwf' Hwf' (trsc_trs_inverse _#4 Hwf Htrs Htrsc) Hofe') as Hceq.
  rewrite -> !subst_under_id in Hceq; [].
  apply ceqt_symm; auto; done.
Qed.


Theorem trp_abstraction_preservation :
  forall G G' M M' T k t c c' e e',
    wf G
    -> trctx G G'
    -> ceqm G pure M M' T
    -> split T k t
    -> trp G M T c e
    -> trp G M' T c' e'
    -> ceqt G' e e' (subst1 c t).
Proof.
intros G G' M M' T k t c c' e e' Hwf HtrG HceqM Hsplit Htrp Htrp'.
apply ceqt_with_wf; intro Hwf'.
so (trp_reg _#5 Hwf Htrp) as HofT.
so (can_trsc _ Hwf) as (Ga' & s & Htrsc).
so (trctx_fun _#3 HtrG (trsc_trctx _#3 Htrsc)); subst Ga'.
so (trsc_subst_ofs _#4 Hwf Htrsc HofT) as HeqT.
so (eqs_reg _#3 Hwf' HeqT) as (HofT' & _).
so (split_formation _#4 HofT' Hsplit) as (Hofk & Hoft).
so (split_fsts _#3 Hsplit) as Hfst.
so (ofm_fst _#6 Hwf (ceqm_reg _#5 Hwf HceqM andel) (trp_fstm _#5 Htrp) Hfst) as Hofc.
so (can_snd _ (subst s M) (ofs_hsg _ _ HofT)) as (f & Hsnd).
so (can_snd _ (subst s M') (ofs_hsg _ _ HofT)) as (f' & Hsnd').
so (trp_correct_conv _#10 Hwf Htrsc Htrp Hsplit Hsnd) as Hceqf.
so (trp_correct_conv _#10 Hwf Htrsc Htrp' Hsplit Hsnd') as Hceqf'.
so (subst_aaof_ceqm _#7 (trsc_formation _#3 Hwf Htrsc) Hwf HceqM) as HceqMs.
#3 sor (refine (snd_ceq _#9 Hwf' Hsplit _ _ Hsnd Hsnd')) as HceqMs'.
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  apply subs_refl; [].
  apply eqs_symm; auto; done.

  eapply subst_fstm; eauto using trsc_formation, subof_aaof_tof, trp_fstm; done.
eapply ceqt_trans; eauto; [].
#2 apply (ceqt_trans _ _ f').
  apply ceqt_comp; eapply tcc_equiv; eauto; [].
  eapply funct_cn_type; eauto; [].
  apply eqc_symm; [].
  apply (trsc_subst_ofc G); eauto; done.

  apply ceqt_symm; [].
  apply ceqt_comp; eapply tcc_equiv; eauto; [].
  eapply funct_cn_type; eauto; [].
  apply eqc_symm; [].
  so (ceqm_fst' _#7 Hwf HceqM Hfst (trp_fstm _#5 Htrp) (trp_fstm _#5 Htrp')) as Heqc.
  #2 apply (@eqc_trans _ _ (subst s c)).
    apply (trsc_subst_ofc G); auto; done.

    #2 apply (@eqc_trans _ _ (subst s c')).
      #2 eapply eqc_subsume.
        eapply subst_eqc; eauto using trsc_formation, subof_aaof_sof; done.

        apply subk_refl; [].
        apply eqk_symm; [].
        apply (trsc_subst_ofk G); auto; [].
        exact (ofs_fst _#3 HofT Hfst).

      apply eqc_symm; [].
      apply (trsc_subst_ofc G); auto; [].
      exact (ofm_fst _#6 Hwf (ceqm_reg _#5 Hwf HceqM ander) (trp_fstm _#5 Htrp') Hfst).
Qed.


Theorem tri_abstraction_preservation :
  forall G G' M M' T k t e e',
    wf G
    -> trctx G G'
    -> ceqm G impure M M' T
    -> split T k t
    -> tri G M T e
    -> tri G M' T e'
    -> ceqt G' e e' (cn_exists k t).
Proof.
intros G G' M M' T k t e e' Hwf HtrG HceqM Hsplit Htri Htri'.
apply ceqt_with_wf; intro Hwf'.
so (tri_reg _#4 Hwf Htri) as HofT.
so (can_trsc _ Hwf) as (Ga' & s & Htrsc).
so (trctx_fun _#3 HtrG (trsc_trctx _#3 Htrsc)); subst Ga'.
so (trsc_subst_ofs _#4 Hwf Htrsc HofT) as HeqT.
so (eqs_reg _#3 Hwf' HeqT) as (HofT' & _).
so (split_formation _#4 HofT' Hsplit) as (Hofk & Hoft).
so (split_fsts _#3 Hsplit) as Hfst.
so (can_snd _ (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (f & Hsnd).
so (tri_correct_conv _#9 Hwf Htrsc Htri Hsplit Hsnd) as Hceqf.
so (tri_correct_conv _#9 Hwf Htrsc Htri' Hsplit Hsnd) as Hceqf'.
eapply ceqt_trans; eauto; [].
apply ceqt_symm; [].
eapply ceqt_trans; eauto; [].
apply ceqt_symm; [].
#4 apply ceqt_comp; eapply tcc_letm.
  #2 apply ceqm_comp; eapply mcc_subsume.
    exact (subst_aaof_ceqm _#7 (trsc_formation _#3 Hwf Htrsc) Hwf HceqM).

    eapply subs_refl; [].
    eapply eqs_symm; eauto; done.

  exact Hfst.

  apply ceqt_refl; [].
  simp_sub.
  #3 apply oft_pack.
    ofvar; done.

    #3 eapply snd_formation; eauto using wf.
      ofvar; done.

      apply subst_split; eauto; done.

      ofvar; done.

    eapply subst_ofc_type; eauto; prove_subof; done.

  apply ofc_exists; auto; done.  
Qed.


Theorem trt_abstraction_preservation :
  forall G G' e e' t f f',
    wf G
    -> trctx G G'
    -> ceqt G e e' t
    -> trt G e t f
    -> trt G e' t f'
    -> ceqt G' f f' t.
Proof.
intros G G' e e' t f f' Hwf HtrG Hceqe Htrt Htrt'.
apply ceqt_with_wf; intro Hwf'.
so (trt_reg _#4 Hwf Htrt) as Hoft.
so (can_trsc _ Hwf) as (Ga' & s & Htrsc).
so (trctx_fun _#3 HtrG (trsc_trctx _#3 Htrsc)); subst Ga'.
so (trt_correct_conv _#6 Hwf Htrsc Htrt) as Hceqf.
so (trt_correct_conv _#6 Hwf Htrsc Htrt') as Hceqf'.
eapply ceqt_trans; eauto; [].
apply ceqt_symm; [].
eapply ceqt_trans; eauto; [].
apply ceqt_symm; [].
#2 apply ceqt_comp; eapply tcc_equiv.
  exact (subst_aaof_ceqt _#6 (trsc_formation _#3 Hwf Htrsc) Hwf Hceqe).

  apply eqc_symm; [].
  apply (trsc_subst_ofc G); auto; done.
Qed.


Theorem trp_abstraction_preservation_conv :
  forall G G' M M' T k t c c' e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> trp G M T c e
    -> trp G M' T c' e'
    -> eqc G' c c' k
    -> ceqt G' e e' (subst1 c t)
    -> ceqm G pure M M' T.
Proof.
intros G G' M M' T k t c c' e e' Hwf HtrG Hsplit Htrp Htrp' Heqc Hceqe.
so (can_trs _ Hwf) as (G'a & s & Htrs).
so (trctx_fun _#3 HtrG (trs_trctx _#3 Htrs)); subst G'a.
so (trctx_formation _ _ HtrG Hwf) as Hwf'.
so (split_fsts _#3 Hsplit) as Hfst.
so (trp_ofm _#5 Hwf Htrp) as HofM.
so (trp_ofm _#5 Hwf Htrp') as HofM'.
so (ofm_reg _#4 Hwf HofM) as HofT.
so (trp_fstm _#5 Htrp) as HfstM.
so (trp_fstm _#5 Htrp') as HfstM'.
so (ofm_fst _#6 Hwf HofM HfstM Hfst) as Hofc.
so (ofm_fst _#6 Hwf HofM' HfstM' Hfst) as Hofc'.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (can_join _ c (subst s e) (ofs_hsg _ _ HofT)) as (N & Hjoin).
so (trp_correct _#8 Htrs Htrp Hjoin) as HceqM.
so (can_join _ c' (subst s e') (ofs_hsg _ _ HofT)) as (N' & Hjoin').
so (trp_correct _#8 Htrs Htrp' Hjoin') as HceqM'.
eapply (ceqm_trans _ _ _ N); eauto; [].
apply (ceqm_trans _ _ _ N'); eauto using ceqm_symm; [].
#2 eapply join_ceq; eauto.
  so (subst_eqc _#6 (subof_dof_sof _#3 (trs_formation _#3 Hwf Htrs)) Heqc) as Heqc'.
  rewrite -> (trs_subst_ofc _#5 Htrs Hofc) in Heqc'.
  rewrite -> (trs_subst_ofc _#5 Htrs Hofc') in Heqc'.
  rewrite -> (trs_subst_ofk _#4 Htrs Hofk) in Heqc'.
  exact Heqc'.

  so (subst_aaof_ceqt _#6 (trs_formation_aaof _#3 Hwf Htrs) Hwf' Hceqe) as Hceqe'.
  rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc_type _#4 Hofc Hoft)) in Hceqe'; [].
  exact Hceqe'.
Qed.


Theorem tri_abstraction_preservation_conv :
  forall G G' M M' T k t e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> tri G M T e
    -> tri G M' T e'
    -> ceqt G' e e' (cn_exists k t)
    -> ceqm G impure M M' T.
Proof.
intros G G' M M' T k t e e' Hwf HtrG Hsplit Htri Htri' Hceqe.
so (can_trs _ Hwf) as (G'a & s & Htrs).
so (trctx_fun _#3 HtrG (trs_trctx _#3 Htrs)); subst G'a.
so (trctx_formation _ _ HtrG Hwf) as Hwf'.
so (split_fsts _#3 Hsplit) as Hfst.
so (tri_ofm _#4 Hwf Htri) as HofM.
so (tri_ofm _#4 Hwf Htri') as HofM'.
so (ofm_reg _#4 Hwf HofM) as HofT.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (N & Hjoin).
so (tri_correct _#7 Htrs Htri Hjoin) as HceqM.
so (tri_correct _#7 Htrs Htri' Hjoin) as HceqM'.
so (join_vars_formation _#5 Hwf HofT Hsplit Hjoin) as HofN.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
eapply ceqm_trans; eauto; [].
apply ceqm_symm; [].
eapply ceqm_trans; eauto; [].
apply ceqm_symm; [].
#3 apply ceqm_comp; apply (@mcc_unpack _#6 _ k t).
  so (subst_aaof_ceqt _#6 (trs_formation_aaof _#3 Hwf Htrs) Hwf' Hceqe) as Hceqe'.
  rewrite -> (trs_subst_ofc _#5 Htrs (ofc_exists Hofk Hoft)) in Hceqe'; [].
  exact Hceqe'.

  apply ceqm_refl; [].
  apply ofm_forget; auto; done.

  apply eqs_refl; auto; done.
Qed.


Theorem trt_abstraction_preservation_conv :
  forall G G' e e' t f f',
    wf G
    -> trctx G G'
    -> trt G e t f
    -> trt G e' t f'
    -> ceqt G' f f' t
    -> ceqt G e e' t.
Proof.
intros G G' e e' t f f' Hwf HtrG Htre Htre' Hceqf.
so (trctx_trs _ _ Hwf HtrG) as (s & Htrs).
#2 eapply ceqt_trans.
  eapply trt_correct; eauto; done.
apply ceqt_symm; [].
#2 eapply ceqt_trans.
  eapply trt_correct; eauto; done.
apply ceqt_symm; [].
so (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htre)) as Hoft.
rewrite <- (trs_subst_ofc _#5 Htrs Hoft); [].
eapply (subst_aaof_ceqt _ _ G'); eauto using trctx_formation, trs_formation_aaof; done.
Qed.
  

Corollary trp_coherent :
  forall G G' M T k t c c' e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> trp G M T c e
    -> trp G M T c' e'
    -> ceqt G' e e' (subst1 c t).
Proof.
intros G G' M T k t c c' e e' Hwf HtrG Hsplit Htrp Htrp'.
eapply trp_abstraction_preservation; eauto; [].
apply ceqm_refl; [].
eapply trp_ofm; eauto; done.
Qed.


Corollary tri_coherent :
  forall G G' M T k t e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> tri G M T e
    -> tri G M T e'
    -> ceqt G' e e' (cn_exists k t).
Proof.
intros G G' M T k t e e' Hwf HtrG Hsplit Htri Htri'.
eapply tri_abstraction_preservation; eauto; [].
apply ceqm_refl; [].
eapply tri_ofm; eauto; done.
Qed.


Corollary trt_coherent :
  forall G G' e t f f',
    wf G
    -> trctx G G'
    -> trt G e t f
    -> trt G e t f'
    -> ceqt G' f f' t.
Proof.
intros G G' e t f f' Hwf HtrG Htrt Htrt'.
eapply trt_abstraction_preservation; eauto; [].
apply ceqt_refl; [].
eapply trt_oft; eauto; done.
Qed.
