
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Regularity.
Require Import Inversion.
Require Import Closure.
Require Import Congruence.
Require Import Contextual.
Require Import Candidate.
Require Import Logical.
Require Import ModelRegular.
Require Import ModelRegularSg.
Require Import Fundamental.
Require Import Preservation.
Require Import ProveSubof.
Require Progress.
Require Singleton.
Require Sound.


(* Ceqv *)

Lemma ceqt_svalue_ceqv :
  forall G e e' t,
    ceqt G e e' t
    -> svalue e
    -> svalue e'
    -> ceqv G e e' t.
Proof.
intros G e e' t Heqt Hval Hval'.
rewrite -> ceqv_is_leqv; [].
rewrite -> ceqt_is_leqt in Heqt; [].
apply leqt_svalue_leqv; auto; done.
Qed.


Lemma ceqt_plam {G k k' t e e'} :
  eqk G k k'
  -> ceqt (G; cl_cn k) e e' t
  -> svalue e
  -> svalue e'
  -> ceqt G (tm_plam k e) (tm_plam k' e') (cn_all k t).
Proof.
intros Heqk Heqe Hval Hval'.
apply ceqt_comp; apply tcc_plam; auto; [].
apply ceqt_svalue_ceqv; auto; done.
Qed.


(* Halting lemmas *)

Lemma subst_thalt :
  forall G1 G2 s e,
    subof oof G1 s G2
    -> thalt G2 e
    -> thalt G1 (subst s e).
Proof.
intros G1 G2 s M Hsubof Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
exists (subst s V).
split.
  eapply star_map; eauto using subst_tstep; done.

  eapply subst_tvalue; eauto; done.
Qed.


Lemma subst_mhalt :
  forall G1 G2 s M,
    subof oof G1 s G2
    -> mhalt G2 M
    -> mhalt G1 (subst s M).
Proof.
intros G1 G2 s M Hsubof Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
exists (subst s V).
split.
  eapply star_map; eauto using subst_mstep; done.

  eapply subst_mvalue; eauto; done.
Qed.


Lemma letm_val_halt_iff :
  forall p T M v,
    ofm nil p M T
    -> tvalue (nil; cl_ucn; cl_umd (var 0)) v
    -> (mhalt nil M <-> thalt nil (tm_letm M v)).
Proof.
intros p T M v HofM Hhaltv.
split.
  intros Hhalt.
  destruct Hhalt as (V & Hsteps & HvalV).
  so (can_fstm_value _#4 wf_nil (preservations_md _#5 wf_nil HofM Hsteps) HvalV) as (c & HfstV).
  exists (subst (dot V (dot c id)) v).
  split.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun x => tm_letm x _)); eauto with dynamic; done.
  
      #2 eapply star_step.
        eapply tstep_letm2; eauto; done.
  
        simp_sub; apply star_refl; done.

    eapply subst_tvalue; eauto; [].
    prove_subof; auto; done.

  intro Hhalt.
  exact (Inversion.letm_halt_inversion' _ _ Hhalt).
Qed.


Lemma mlett_val_halt_iff :
  forall e V,
    mvalue (nil; cl_utm) V
    -> (thalt nil e <-> mhalt nil (md_lett e V)).
Proof.
intros e V HvalV.
split.
  intros Hhalt.
  destruct Hhalt as (v & Hsteps & Hvalv).
  exists (subst (dot v id) V).
  split.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun x => md_lett x _)); eauto with dynamic; done.
  
      #2 eapply star_step.
        eapply mstep_lett2; eauto; done.
  
        simp_sub; apply star_refl; done.

    eapply subst_mvalue; eauto; [].
    prove_subof; auto; done.

  intro Hhalt.
  exact (Inversion.mlett_halt_inversion' _ _ Hhalt).
Qed.


Lemma ext_halt_iff :
  forall t M,
    ofm nil impure M (sg_dyn t)
    -> (mhalt nil M <-> thalt nil (tm_ext M)).
Proof.
intros t M Hof.
split.
  intros Hhalt.
  destruct Hhalt as (V & Hsteps & HvalV).
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil Hof Hsteps) HvalV); [].
  intros v <-.
  invert HvalV; [].
  intros Hvalv.
  exists v.
  split; auto; [].
  #2 eapply star_trans.
    eapply star_map; eauto with dynamic; done.

    apply star_one; [].
    apply tstep_ext2; auto; done.

  intro Hhalt.
  eapply Inversion.ext_halt_inversion'; eauto; done.
Qed.


Lemma thalt_pair :
  forall G e1 e2,
    thalt G e1
    -> thalt G e2
    -> thalt G (tm_pair e1 e2).
Proof.
intros G e1 e2 Hhalt1 Hhalt2.
destruct Hhalt1 as (V1 & Hsteps1 & Hval1).
destruct Hhalt2 as (V2 & Hsteps2 & Hval2).
exists (tm_pair V1 V2).
split; auto with dynamic; [].
#2 eapply star_trans.
  apply (star_map' _ _ (fun z => tm_pair z _)); eauto with dynamic; done.

  eapply star_map'; eauto with dynamic; done.
Qed.


Lemma thalt_pi1 :
  forall e t1 t2,
    oft nil e (cn_prod t1 t2)
    -> thalt nil e
    -> thalt nil (tm_pi1 e).
Proof.
intros e t1 t2 Hof Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hof Hsteps) Consistent.similar_prod Hval); [].
intros v1 v2 <-.
invert Hval; [].
intros Hval1 Hval2.
exists v1.
split; auto; [].
#2 eapply star_trans.
  eapply star_map'; eauto with dynamic; done.

  apply star_one; apply tstep_pi12; auto; done.
Qed.


Lemma thalt_pi2 :
  forall e t1 t2,
    oft nil e (cn_prod t1 t2)
    -> thalt nil e
    -> thalt nil (tm_pi2 e).
Proof.
intros e t1 t2 Hof Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hof Hsteps) Consistent.similar_prod Hval); [].
intros v1 v2 <-.
invert Hval; [].
intros Hval1 Hval2.
exists v2.
split; auto; [].
#2 eapply star_trans.
  eapply star_map'; eauto with dynamic; done.

  apply star_one; apply tstep_pi22; auto; done.
Qed.


Lemma mhalt_pair :
  forall G M1 M2,
    mhalt G M1 
    -> mhalt G M2
    -> mhalt G (md_pair M1 M2).
Proof.
intros G M1 M2 Hhalt1 Hhalt2.
destruct Hhalt1 as (V1 & Hsteps1 & Hval1).
destruct Hhalt2 as (V2 & Hsteps2 & Hval2).
exists (md_pair V1 V2).
split; auto with dynamic; [].
#2 eapply star_trans.
  eapply (star_map' _ _ (fun z => md_pair z _)); eauto with dynamic; done.

  eapply (star_map' _ _ (fun z => md_pair _ z)); eauto with dynamic; done.
Qed.


Lemma mhalt_pi1 :
  forall T1 T2 M,
    ofm nil pure M (sg_sigma T1 T2)
    -> mhalt nil M
    -> mhalt nil (md_pi1 M).
Proof.
intros T1 T2 M Hof Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
invert (Progress.canon_md _ _ _ (preservations_md _#5 wf_nil Hof Hsteps) Hval); [].
intros V1 V2 <-.
invert Hval; [].
intros Hval1 Hval2.
exists V1.
split; auto; [].
#2 eapply star_trans.
  eapply star_map'; eauto with dynamic; done.

  apply star_one; apply mstep_pi12; auto; done.
Qed.


Lemma mhalt_pi2 :
  forall T1 T2 M,
    ofm nil pure M (sg_sigma T1 T2)
    -> mhalt nil M
    -> mhalt nil (md_pi2 M).
Proof.
intros T1 T2 M Hof Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
invert (Progress.canon_md _ _ _ (preservations_md _#5 wf_nil Hof Hsteps) Hval); [].
intros V1 V2 <-.
invert Hval; [].
intros Hval1 Hval2.
exists V2.
split; auto; [].
#2 eapply star_trans.
  eapply star_map'; eauto with dynamic; done.

  apply star_one; apply mstep_pi22; auto; done.
Qed.


(* Closed instances *)

Lemma closed_instances_ceqt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> (forall s, subof aof nil s G -> ceqt nil (subst s e) (subst s e') (subst s t))
    -> ceqt G e e' t.
Proof.
intros G e e' t Hofe Hofe' Hci.
rewrite -> ceqt_is_leqt; [].
intro Hwf.
do2 2 split; auto; [].
intros E HintE.
so (leqt_refl _#3 Hofe Hwf anderr _ HintE) as Htev.
#4 eapply tpclosed_tev_ceqt; eauto.
  eapply intt_tvrel; eauto using fintctx_stat; [].
  exact (oft_reg _#3 Hwf Hofe).

  eapply intt_tpclosed; eauto using fintctx_stat; [].
  exact (oft_reg _#3 Hwf Hofe).

  apply ceqt_refl; [].
  eapply subst_oft; eauto; [].
  eapply left_subof_dof; eauto; done.

  apply Hci; [].
  eapply right_subof_aof; eauto; done.
Qed.


Lemma closed_instances_ceqm_impure :
  forall G M M' T,
    ofm G impure M T
    -> ofm G impure M' T
    -> (forall s, subof aof nil s G -> ceqm nil impure (subst s M) (subst s M') (subst s T))
    -> ceqm G impure M M' T.
Proof.
intros G M M' T HofM HofM' Hci.
rewrite -> ceqm_is_leqm; [].
intro Hwf.
do2 2 split; auto; [].
intros E HintE.
so (leqm_refl _#4 HofM Hwf anderr _ HintE) as Hmiev.
#2 eapply miev_ceqm; eauto.
  apply ceqm_refl; [].
  eapply subst_ofm; eauto; [].
  eapply left_subof_dof; eauto; done.

  apply Hci; [].
  eapply right_subof_aof; eauto; done.
Qed.


Lemma closed_instances_ceqm_pure :
  forall G M M' T c c' k,
    ofm G pure M T
    -> ofm G pure M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> (forall s, subof aof nil s G -> ceqm nil pure (subst s M) (subst s M') (subst s T))
    -> ceqm G pure M M' T.
Proof.
intros G M M' T c c' k HofM HofM' Hfst HfstM HfstM' Heqc Hci.
rewrite -> ceqm_is_leqm; [].
intro Hwf.
do2 2 split; auto; [].
split.
  intros ka ca ca' Hfsta HfstMa HfstMa'.
  so (fsts_fun _#3 Hfst Hfsta); subst ka; clear Hfsta.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  assumption.

  clear c c' HfstM HfstM' Heqc.
  intros E c HintE HfstM.
  so (ofm_reg _#4 Hwf HofM) as HofT.
  so (leqm_refl _#4 HofM Hwf anderrr _ _ HintE HfstM) as Hmpuev.
  #4 eapply mpuev_ceqm; eauto.
    intros V1 V2 Hint.
    so (ints_reg _#5 Hint) as (HofV1 & HofV2).
    so (ints_val _#6 Hwf HofT (fintctx_stat _ _ Hwf HintE) Hint) as (Hval1 & Hval2).
    auto with static; done.

    intros V1 V2 V1' V2' Heqm1 Heqm2 HvalV1' HvalV2' Hint.
    so (Preservation.can_fstm _#3 wf_nil (ceqm_reg _#5 wf_nil Heqm1 andel)) as (c1 & HfstV1).
    so (Preservation.can_fstm _#3 wf_nil (ceqm_reg _#5 wf_nil Heqm2 andel)) as (c2 & HfstV2).
    so (ints_fst _#9 Hwf HofT (fintctx_stat _ _ Hwf HintE) Hfst HfstV1 HfstV2 Hint).
    eapply (ints_ceqm _#3 V1 V2 V1' V2'); eauto using fintctx_stat; done.

    apply ceqm_refl; [].
    eapply subst_ofm; eauto; [].
    apply left_subof_dof; auto; done.

    apply Hci; [].
    apply right_subof_aof; auto; done.
Qed.


Lemma closed_instances_ceqm :
  forall G p M M' T,
    ofm G p M T
    -> ofm G p M' T
    -> (p = pure
        -> exists k c c',
             fsts T k
             /\ fstm G M c
             /\ fstm G M' c'
             /\ eqc G c c' k)
    -> (forall s, subof aof nil s G -> ceqm nil p (subst s M) (subst s M') (subst s T))
    -> ceqm G p M M' T.
Proof.
intros G p M M' T Hof Hof' Hfsteq Hceq.
destruct p.
  so (Hfsteq (eq_refl _)) as (k & c & c' & Hfst & HfstM & HfstM' & Heqc).
  eapply closed_instances_ceqm_pure; eauto; done.

  eapply closed_instances_ceqm_impure; eauto; done.
Qed.


(* Conditional equivalence *)

Lemma ceqt_conditional :
  forall (P : Prop) e1 e2 t,
    oft nil e1 t
    -> oft nil e2 t
    -> (thalt nil e1 -> P)
    -> (thalt nil e2 -> P)
    -> (P -> ceqt nil e1 e2 t)
    -> ceqt nil e1 e2 t.
Proof.
intros P e1 e2 t Hofe1 Hofe2 Hhaltif1 Hhaltif2 Hceqif.
rewrite -> ceqt_is_leqt; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simp_sub.
#2 do2 3 split; auto.
  split.
    intro Hhalt1.
    so (Hceqif (Hhaltif1 Hhalt1)) as Hceq.
    eapply ceqt_halt; eauto; done.
    
    intro Hhalt2.
    so (Hceqif (Hhaltif2 Hhalt2)) as Hceq.
    eapply ceqt_halt; eauto using ceqt_symm; done.

  intros v v' Hsteps Hsteps' Hval Hval'.
  lapply Hceqif.
    intro Hceq.
    rewrite -> ceqt_is_leqt in Hceq; [].
    so (Hceq wf_nil anderr _ fintctx_nil) as Htev.
    simp_sub in Htev.
    exact (Htev anderl _ _ Hsteps Hsteps' Hval Hval').

    apply Hhaltif1; [].
    exists v; auto; done.
Qed.


Lemma ceqm_conditional_pure :
  forall (P : Prop) M1 M2 T c1 c2 k,
    ofm nil pure M1 T
    -> ofm nil pure M2 T
    -> fstm nil M1 c1
    -> fstm nil M2 c2
    -> fsts T k
    -> eqc nil c1 c2 k
    -> (mhalt nil M1 -> P)
    -> (mhalt nil M2 -> P)
    -> (P -> ceqm nil pure M1 M2 T)
    -> ceqm nil pure M1 M2 T.
Proof.
intros P M1 M2 T c1 c2 k HofM1 HofM2 HfstM1 HfstM2 Hfst Heqc Hhaltimp1 Hhaltimp2 Hceqif.
rewrite -> ceqm_is_leqm; [].
intros _.
#2 do2 3 split; auto.
  intros ka c1a c2a Hfsta HfstM1a HfstM2a.
  so (fsts_fun _#3 Hfst Hfsta); subst ka.
  so (fstm_fun _#4 HfstM1 HfstM1a); subst c1a.
  so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
  assumption.

  intros E c1a HintE HfstM1a.
  so (fstm_fun _#4 HfstM1 HfstM1a); subst c1a.
  clear HfstM1a.
  invert HintE; clear HintE; [].
  intros <-.
  simp_sub.
  #2 do2 3 split; auto.
    split.
      intro Hhalt.
      so (Hceqif (Hhaltimp1 Hhalt)).
      eapply ceqm_halt; eauto; [].
      apply ceqm_comp; apply mcc_forget; auto; done.
  
      intro Hhalt.
      so (Hceqif (Hhaltimp2 Hhalt)).
      eapply ceqm_halt; eauto; [].
      apply ceqm_symm; [].
      apply ceqm_comp; apply mcc_forget; auto; done.

    intros V1 V2 Hsteps Hsteps' HvalV HvalV'.
    lapply Hceqif.
      intro Hceq.
      rewrite -> ceqm_is_leqm in Hceq; [].
      so (Hceq wf_nil anderrr _ _ fintctx_nil HfstM1) as H.
      simp_sub in H.
      so (H anderl _ _ Hsteps Hsteps' HvalV HvalV').
      assumption.

      apply Hhaltimp1; [].
      exists V1; split; auto; done.
Qed.


Lemma ceqm_conditional_impure :
  forall (P : Prop) M1 M2 T,
    ofm nil impure M1 T
    -> ofm nil impure M2 T
    -> (mhalt nil M1 -> P)
    -> (mhalt nil M2 -> P)
    -> (P -> ceqm nil impure M1 M2 T)
    -> ceqm nil impure M1 M2 T.
Proof.
intros P M1 M2 T HofM1 HofM2 Hhaltimp1 Hhaltimp2 Hceqif.
rewrite -> ceqm_is_leqm; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simp_sub.
#2 do2 3 split; auto.
  split.
    intro Hhalt.
    so (Hceqif (Hhaltimp1 Hhalt)).
    eapply ceqm_halt; eauto; done.

    intro Hhalt.
    so (Hceqif (Hhaltimp2 Hhalt)).
    eapply ceqm_halt; eauto using ceqm_symm; done.

  intros V V' Hsteps Hsteps' Hval Hval'.
  lapply Hceqif.
    intro Hceq.
    rewrite -> ceqm_is_leqm in Hceq; [].
    so (Hceq wf_nil anderr _ fintctx_nil) as H.
      simp_sub in H.
      so (H anderl _ _ Hsteps Hsteps' Hval Hval').
      assumption.

      apply Hhaltimp1; [].
      exists V; split; auto; done.
Qed.


Lemma ceqm_conditional :
  forall (P : Prop) p M1 M2 T,
    ofm nil p M1 T
    -> ofm nil p M2 T
    -> (p = pure
        -> exists k c1 c2,
             fsts T k
             /\ fstm nil M1 c1
             /\ fstm nil M2 c2
             /\ eqc nil c1 c2 k)
    -> (mhalt nil M1 -> P)
    -> (mhalt nil M2 -> P)
    -> (P -> ceqm nil p M1 M2 T)
    -> ceqm nil p M1 M2 T.
Proof.
intros P p M1 M2 T HofM1 HofM2 Hfsteq HPif1 HPif2 Hceqif.
destruct p.
  so (Hfsteq (eq_refl _)) as (k & c1 & c2 & Hfst & HfstM1 & HfstM2 & Heqc).
  eapply ceqm_conditional_pure; eauto; done.

  eapply ceqm_conditional_impure; eauto; done.
Qed.


Lemma ceqt_closed_instances_conditional :
  forall (P : sub -> Prop) G e1 e2 t,
    oft G e1 t
    -> oft G e2 t
    -> (forall s,
          subof aof nil s G
          -> (thalt nil (subst s e1) -> P s)
             /\ (thalt nil (subst s e2) -> P s)
             /\ (P s -> ceqt nil (subst s e1) (subst s e2) (subst s t)))
    -> ceqt G e1 e2 t.
Proof.
intros P G e1 e2 t Hof1 Hof2 Hprop.
apply closed_instances_ceqt; auto; [].
intros s Hs.
so (Hprop s Hs) as (HP1 & HP2 & Hceqif).
apply (ceqt_conditional (P s)); eauto using subst_oft, subof_aof_dof.
Qed.


Lemma ceqm_closed_instances_conditional :
  forall (P : sub -> Prop) G p M1 M2 T,
    ofm G p M1 T
    -> ofm G p M2 T
    -> (p = pure
        -> exists k c1 c2,
             fsts T k
             /\ fstm G M1 c1
             /\ fstm G M2 c2
             /\ eqc G c1 c2 k)
    -> (forall s,
          subof aof nil s G
          -> (mhalt nil (subst s M1) -> P s)
             /\ (mhalt nil (subst s M2) -> P s)
             /\ (P s -> ceqm nil p (subst s M1) (subst s M2) (subst s T)))
    -> ceqm G p M1 M2 T.
Proof.
intros P G p M1 M2 T Hof1 Hof2 Hfsteq Hprop.
apply closed_instances_ceqm; auto; [].
intros s Hs.
so (Hprop s Hs) as (HP1 & HP2 & Hceqif).
apply (ceqm_conditional (P s)); eauto using subst_ofm, subof_aof_dof; [].
intros ->.
so (Hfsteq (eq_refl _)) as (k & c1 & c2 & Hfst & HfstM1 & HfstM2 & Heqc).
repeat eexists; eauto using subst_fsts, subst_fstm, subst_eqc, subof_aof_tof, subof_aof_sof; done.
Qed.


(* Extensionality *)

Lemma ceqt_unit_ext :
  forall e e',
    oft nil e cn_unit
    -> oft nil e' cn_unit
    -> (thalt nil e <-> thalt nil e')
    -> ceqt nil e e' cn_unit.
Proof.
intros e e' Hofe Hofe' Hhaltiff.
rewrite -> ceqt_is_leqt; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simpl; simp_sub.
do2 3 split; auto; [].
intros v v' Hsteps Hsteps' Hval Hval'.
rewrite -> intt_unit_unroll; [].
invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe Hsteps) Consistent.similar_unit Hval); [].
intros <-.
invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe' Hsteps') Consistent.similar_unit Hval'); [].
intros <-.
apply unit_interp_; done.
Qed.


Lemma ceqt_arrow_ext :
  forall t1 t2 e e',
    oft nil e (cn_arrow t1 t2)
    -> oft nil e' (cn_arrow t1 t2)
    -> (thalt nil e <-> thalt nil e')
    -> (forall v,
          oft nil v t1
          -> tvalue nil v
          -> ceqt nil (tm_app e v) (tm_app e' v) t2)
    -> ceqt nil e e' (cn_arrow t1 t2).
Proof.
intros t1 t2 e e' Hofe Hofe' Hhaltiff Hext.
rewrite -> ceqt_is_leqt; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simp_sub.
do2 3 split; auto; [].
intros v v' Hsteps Hsteps' Hvalv Hvalv'.
so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
so (preservations_tm _#4 wf_nil Hofe' Hsteps') as Hofv'.
rewrite -> intt_arrow_unroll; [].
apply arrow_interp_; simp_sub; auto; [].
intros w w' Hintw.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hofe)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
clear l Hofcr.
so (intt_tvrel _#3 wf_nil Hoft1 intctx_nil _ _ Hintw) as H; simp_sub in H.
destruct H as (Hofw & Hofw' & Hvalw & Hvalw').
so (leqt_refl _#3 Hofe wf_nil anderr nil fintctx_nil) as Hrele.
simp_sub in Hrele.
rewrite -> intt_arrow_unroll in Hrele.
so (Hrele anderl v v Hsteps Hsteps Hvalv Hvalv) as Hrelv.
invert Hrelv; [].
intros _ _ _ _ Hrelapp.
so (Hrelapp _ _ Hintw) as Hrelvw.
simp_sub in Hrelvw.
#4 eapply tpclosed_tev_ceqt; eauto.
  so (intt_tvrel _#3 wf_nil Hoft2 intctx_nil) as H.
  simp_sub in H.
  exact H.

  so (intt_tpclosed _#3 wf_nil Hoft2 intctx_nil) as H.
  simp_sub in H.
  exact H.

  apply ceqt_refl; eauto with static; done.

  #2 apply (ceqt_trans _ _ (tm_app e w')).
    apply ceqt_symm; [].
    apply tsteps_ceqt; eauto with static; [].
    eapply (star_map' _ _ (fun z => tm_app z _)); eauto with dynamic; done.

    #2 apply (ceqt_trans _ _ (tm_app e' w')).
      apply Hext; auto; done.

      apply tsteps_ceqt; eauto with static; [].
      eapply (star_map' _ _ (fun z => tm_app z _)); eauto with dynamic; done.
Qed.


Lemma ceqt_all_ext :
  forall k t e e',
    oft nil e (cn_all k t)
    -> oft nil e' (cn_all k t)
    -> (thalt nil e <-> thalt nil e')
    -> (forall c,
          ofc nil c k
          -> ceqt nil (tm_papp e c) (tm_papp e' c) (subst1 c t))
    -> ceqt nil e e' (cn_all k t).
Proof.
intros k t e e' Hofe Hofe' Hhaltiff Hext.
rewrite -> ceqt_is_leqt; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simp_sub.
do2 3 split; auto; [].
intros v v' Hsteps Hsteps' Hvalv Hvalv'.
so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
so (preservations_tm _#4 wf_nil Hofe' Hsteps') as Hofv'.
rewrite -> intt_all_unroll; [].
apply all_interp_; simp_sub; auto; [].
intros c c' Q HintQ.
simpl.
so (intk_reg _#6 HintQ) as H; simp_sub in H.
destruct H as (Hofc & Hofc').
so (leqt_refl _#3 Hofe wf_nil anderr nil fintctx_nil) as Hrele.
simp_sub in Hrele.
rewrite -> intt_all_unroll in Hrele; [].
so (Hrele anderl v v Hsteps Hsteps Hvalv Hvalv) as Hrelv.
invert Hrelv; [].
intros _ _ _ _ Hrelapp.
so (Hrelapp _ _ _ HintQ) as Hrelvc.
simpl in Hrelvc.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hofe)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
clear l Hofcr.
#4 eapply tpclosed_tev_ceqt; eauto.
  apply (intt_tvrel (nil; cl_cn k)); auto with static; [].
  apply intctx_cn; simp_sub; auto using intctx with static; done.

  apply (intt_tpclosed (nil; cl_cn k)); auto with static; [].
  apply intctx_cn; simp_sub; auto using intctx with static; done.

  apply ceqt_refl; [].
  eapply oft_papp; eauto; done.

  #2 apply (ceqt_trans _ _ (tm_papp e c')).
    apply ceqt_symm; [].
    #2 apply tsteps_ceqt.
      eapply oft_papp; eauto; done.
      
      eapply (star_map' _ _ (fun z => tm_papp z _)); eauto with dynamic; done.

    #2 apply (ceqt_trans _ _ (tm_papp e' c')).
      apply Hext; auto; done.

      #2 apply tsteps_ceqt.
        eapply oft_papp; eauto; done.

        eapply (star_map' _ _ (fun z => tm_papp z _)); eauto with dynamic; done.
Qed.


Lemma ceqt_prod_ext :
  forall t1 t2 e e',
    oft nil e (cn_prod t1 t2)
    -> oft nil e' (cn_prod t1 t2)
    -> ceqt nil (tm_pi1 e) (tm_pi1 e') t1
    -> ceqt nil (tm_pi2 e) (tm_pi2 e') t2
    -> ceqt nil e e' (cn_prod t1 t2).
Proof.
intros t1 t2 e e' Hofe Hofe' Hext1 Hext2.
rewrite -> ceqt_is_leqt; [].
intros _.
do2 2 split; auto; [].
intros E HintE.
invert HintE; clear HintE; [].
intros <-.
simp_sub.
#2 do2 3 split; auto.
  split.
    intro Hhalt.
    apply pi1_halt_inversion'; [].
    eapply ceqt_halt; eauto; [].
    eapply thalt_pi1; eauto; done.

    intro Hhalt.
    apply pi1_halt_inversion'; [].
    eapply ceqt_halt; eauto using ceqt_symm; [].
    eapply thalt_pi1; eauto; done.
intros v v' Hsteps Hsteps' Hval Hval'.
so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
so (preservations_tm _#4 wf_nil Hofe' Hsteps') as Hofv'.
invert (Progress.canon_tm _ _ Hofv Consistent.similar_prod Hval); [].
intros v1 v2 <-.
invert (Progress.canon_tm _ _ Hofv' Consistent.similar_prod Hval'); [].
intros v1' v2' <-.
invert Hval; [].
intros Hval1 Hval2.
invert Hval'; [].
intros Hval1' Hval2'.
rewrite -> intt_prod_unroll; [].
so (leqt_refl _#3 Hofv wf_nil anderr nil fintctx_nil anderl) as H; simp_sub in H.
so (H _ _ (star_refl _) (star_refl _) Hval Hval) as Hrelv; clear H.
rewrite -> intt_prod_unroll in Hrelv; [].
invert Hrelv; [].
intros _ _ _ _ Hint1 Hint2.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hofe)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
clear l Hofcr.
so (pair_oft_inversion_prod _#5 wf_nil Hofv) as (Hofv1 & Hofv2).
#2 apply prod_interp_; simp_sub; auto with static.
  #2 eapply (intt_ceqt nil); eauto using wf_nil, intctx.
    simp_sub; apply ceqt_refl; auto; done.

    simp_sub.
    #2 apply (ceqt_trans _ _ (tm_pi1 e)).
      apply ceqt_symm; [].
      eapply tsteps_ceqt; eauto with static; [].
      #2 eapply star_trans.
        eapply star_map'; eauto with dynamic; done.

        apply star_one; auto with dynamic; done.

      eapply ceqt_trans; eauto; [].
      eapply tsteps_ceqt; eauto with static; [].
      #2 eapply star_trans.
        eapply star_map'; eauto with dynamic; done.

        apply star_one; auto with dynamic; done.

  #2 eapply (intt_ceqt nil); eauto using wf_nil, intctx.
    simp_sub; apply ceqt_refl; auto; done.

    simp_sub.
    #2 apply (ceqt_trans _ _ (tm_pi2 e)).
      apply ceqt_symm; [].
      eapply tsteps_ceqt; eauto with static; [].
      #2 eapply star_trans.
        eapply star_map'; eauto with dynamic; done.

        apply star_one; auto with dynamic; done.

      eapply ceqt_trans; eauto; [].
      eapply tsteps_ceqt; eauto with static; [].
      #2 eapply star_trans.
        eapply star_map'; eauto with dynamic; done.

        apply star_one; auto with dynamic; done.
Qed.


Lemma ceqm_one_ext :
  forall p M M',
    ofm nil p M sg_one
    -> ofm nil p M' sg_one
    -> (mhalt nil M <-> mhalt nil M')
    -> ceqm nil p M M' sg_one.
Proof.
intros p M M' HofM HofM' Hhaltiff.
rewrite -> ceqm_is_leqm; [].
intros _.
do2 2 split; auto; [].
#2 destruct p.
  split.
    intros k c c' Hfst HfstM HfstM'.
    invert Hfst; [].
    intros <-.
    apply eqc_extone; eauto using wf_nil, ofm_fst; done.
  
    intros E c HintE HfstM.
    invert HintE; clear HintE; [].
    intros <-.
    simpl; simp_sub.
    do2 3 split; auto; [].
    intros V V' Hsteps Hsteps' Hval Hval'.
    invert (Progress.canon_md _#3 (Preservation.preservations_md _#5 wf_nil HofM Hsteps) Hval); [].
    invert (Progress.canon_md _#3 (Preservation.preservations_md _#5 wf_nil HofM' Hsteps') Hval'); [].
    intros <- <-.
    so (intc_fun _#5 wf_nil (ofm_fst _#6 wf_nil HofM HfstM fsts_one) intctx_nil) as Hintk.
    simpl in Hintk.
    invert Hintk; [].
    intros _ _ Heq _.
    rewrite <- Heq; [].
    apply sg_one_interp_; done.
  
  intros E HintE.
  invert HintE; clear HintE; [].
  intros <-.
  simpl; simp_sub.
  do2 3 split; auto; [].
  intros V V' Hsteps Hsteps' Hval Hval'.
  invert (Progress.canon_md _#3 (Preservation.preservations_md _#5 wf_nil HofM Hsteps) Hval); [].
  invert (Progress.canon_md _#3 (Preservation.preservations_md _#5 wf_nil HofM' Hsteps') Hval'); [].
  intros <- <-.
  #2 apply mpst_expand.
    intros N N' (Q & Hint).
    invert Hint; [].
    intros <- <- <-.
    do2 3 split; auto with static dynamic; done.

    exists (Cand sk_one tt).
    apply sg_one_interp_; done.
Qed.


Lemma ceqm_stat_ext :
  forall k M M' c c',
    ofm nil pure M (sg_stat k)
    -> ofm nil pure M' (sg_stat k)
    -> (mhalt nil M <-> mhalt nil M')
    -> fstm nil M c
    -> fstm nil M' c'
    -> eqc nil c c' k
    -> ceqm nil pure M M' (sg_stat k).
Proof.
intros k M M' c c' HofM HofM' Hhaltiff HfstM HfstM' Heqc.
rewrite -> ceqm_is_leqm; [].
intros _.
#2 do2 3 split; auto.
  intros ka ca ca' Hfst HfstMa HfstMa'.
  so (fsts_fun _#3 fsts_stat Hfst); subst ka; clear Hfst.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  assumption.

  intros E ca HintE HfstMa.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  invert HintE; clear HintE; [].
  intros <-.
  simpl; simp_sub.
  do2 3 split; auto; [].
  intros V V' Hsteps Hsteps' HvalV HvalV'.
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil HofM Hsteps) HvalV); [].
  intros b <-.
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil HofM' Hsteps') HvalV'); [].
  intros b' <-.
  apply sg_stat_interp_; [].
  so (intc_fun _#5 wf_nil (eqc_reg _#4 wf_nil Heqc andel) intctx_nil) as Hint.
  simpl in Hint; simp_sub in Hint.
  eapply (intk_defeq nil); eauto with static.
    exact (eqc_reg _#4 wf_nil Heqc anderr).

    exact intctx_nil.

    simpl; simp_sub.
    so (preservations_fst _#7 wf_nil HofM Hsteps fsts_stat HfstM) as (ba & Hfstb & Heqb).
    so (fstm_fun _#4 Hfstb fstm_stat); subst ba.
    assumption.

    simpl; simp_sub.
    eapply eqc_trans; eauto; [].
    so (preservations_fst _#7 wf_nil HofM' Hsteps' fsts_stat HfstM') as (ba & Hfstb & Heqb).
    so (fstm_fun _#4 Hfstb fstm_stat); subst ba.
    assumption.
Qed.



Lemma ceqm_dyn_ext :
  forall t M M',
    ofm nil pure M (sg_dyn t)
    -> ofm nil pure M' (sg_dyn t)
    -> ceqt nil (tm_ext M) (tm_ext M') t
    -> ceqm nil pure M M' (sg_dyn t).
Proof.
intros t M M' HofM HofM' Hceqt.
rewrite -> ceqm_is_leqm; [].
intros _.
#2 do2 3 split; auto.
  intros k c c' Hfst HfstM HfstM'.
  so (fsts_fun _#3 Hfst fsts_dyn); subst k.
  so (ofm_fst _#6 wf_nil HofM HfstM fsts_dyn).
  so (ofm_fst _#6 wf_nil HofM' HfstM' fsts_dyn).
  apply eqc_extone; auto; done.

  intros E c HintE HfstM.
  invert HintE; clear HintE; [].
  intros <-.
  simpl; simp_sub.
  #2 do2 3 split; auto.
    transitivity (thalt nil (tm_ext M)).
      eapply ext_halt_iff; eauto with static; done.

      transitivity (thalt nil (tm_ext M')).
        eapply ceqt_halt_iff; eauto; done.

        symmetry.
        eapply ext_halt_iff; eauto with static; done.

    intros V V' Hsteps Hsteps' HvalV HvalV'.
    invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil HofM Hsteps) HvalV); [].
    intros v <-.
    invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil HofM' Hsteps') HvalV'); [].
    intros v' <-.
    so (intc_fun _#5 wf_nil (ofm_fst _#6 wf_nil HofM HfstM fsts_dyn) intctx_nil) as Hintk.
    simpl in Hintk; simp_sub in Hintk.
    invert Hintk; [].
    intros _ _ Heq _.
    rewrite <- Heq; [].
    #3 apply sg_dyn_interp_.
      simpl; simp_sub.
      clear HofM'.
      eapply preservations_md; eauto with static; done.

      simpl; simp_sub.
      clear HofM.
      eapply preservations_md; eauto with static; done.

      so (ceqt_implies_leqt _#4 Hceqt wf_nil anderr _ fintctx_nil) as Htev.
      simp_sub in Htev.
      invert HvalV; [].
      intro Hvalv.
      invert HvalV'; [].
      intro Hvalv'.
      #2 apply (Htev anderl v v'); auto.
        #2 eapply star_trans.
          eapply star_map; eauto with dynamic; done.

          apply star_one; [].
          apply tstep_ext2; auto; done.

        #2 eapply star_trans.
          eapply star_map; eauto with dynamic; done.

          apply star_one; [].
          apply tstep_ext2; auto; done.
Qed.


Lemma ceqm_pigen_ext :
  forall T1 T2 M M',
    ofm nil pure M (sg_pigen T1 T2)
    -> ofm nil pure M' (sg_pigen T1 T2)
    -> (mhalt nil M <-> mhalt nil M')
    -> (forall V c,
          ofm nil pure V T1
          -> fstm nil V c
          -> mvalue nil V
          -> ceqm nil impure (md_appgn M V) (md_appgn M' V) (subst1 c T2))
    -> ceqm nil pure M M' (sg_pigen T1 T2).
Proof.
intros T1 T2 M M' HofM HofM' Hhaltiff Hext.
rewrite -> ceqm_is_leqm; [].
intros _.
#2 do2 3 split; auto.
  intros k c c' Hfst HfstM HfstM'.
  invert Hfst; [].
  intros <-.
  #2 apply eqc_extone.
    clear HofM'.
    eapply ofm_fst; eauto with static; done.

    clear HofM.
    eapply ofm_fst; eauto with static; done.

  intros E c HintE HfstM.
  invert HintE; clear HintE; [].
  intros <-.
  simp_sub.
  do2 3 split; auto; [].
  intros V V' Hsteps Hsteps' HvalV HvalV'.
  simpl.
  so (intc_fun _#5 wf_nil (ofm_fst _#6 wf_nil HofM HfstM fsts_pigen) intctx_nil) as Hintk.
  simp_sub in Hintk.
  invert Hintk; [].
  intros _ _ Heq _.
  rewrite <- Heq; [].
  so (preservations_md _#5 wf_nil HofM Hsteps) as HofV.
  so (preservations_md _#5 wf_nil HofM' Hsteps') as HofV'.
  #3 apply sg_pigen_interp_; auto.
    simp_sub.
    apply ofm_forget; assumption.

    simp_sub.
    apply ofm_forget; assumption.

    intros W W' Q HintW.
    so (ints_reg _#5 HintW) as H.
    simp_sub in H.
    destruct H as (HofW & HofW').
    so (ints_val _#6 wf_nil (ofm_reg _#4 wf_nil HofW) intctx_nil HintW) as (_ & HvalW').
    so (can_fstm _#3 wf_nil HofW) as (b & HfstW).
    so (can_fstm _#3 wf_nil HofW') as (b' & HfstW').
    exists b, b'; do2 2 split; auto; [].
    simpl; simp_sub.
    so (leqm_refl _#4 HofM) as HleqM.
    so (HleqM wf_nil anderrr _ _ fintctx_nil HfstM anderl) as H.
    simp_sub in H.
    so (H V V Hsteps Hsteps HvalV HvalV) as HintV.
    clear HleqM H.
    simpl in HintV.
    invert HintV; [].
    intros _ _ _ _ Hfun _.
    clear HintV.
    so (Hfun W W' Q HintW) as (ba & ba' & HfstWa & HfstWa' & Hintapp).
    clear Hfun.
    so (fstm_fun _#4 HfstW HfstWa); subst ba; clear HfstWa.
    so (fstm_fun _#4 HfstW' HfstWa'); subst ba'; clear HfstWa'.
    simpl in Hintapp; simp_sub in Hintapp.
    #2 eapply miev_ceqm; eauto.
      apply ceqm_refl; eauto with static; done.

      #2 apply (ceqm_trans _#3 (md_appgn M W')).
        apply ceqm_symm; [].
        apply ceqm_comp; [].
        #2 eapply mcc_appgn; eauto.
          apply msteps_ceqm; eauto with static; done.

          apply ceqm_refl; eauto; done.

        #2 apply (ceqm_trans _#3 (md_appgn M' W')).
          apply Hext; auto; done.

          apply ceqm_comp; [].
          #2 eapply mcc_appgn; eauto.
            apply msteps_ceqm; eauto with static; done.

            apply ceqm_refl; eauto; done.
Qed.


Lemma ceqm_piapp_ext :
  forall T1 T2 M M' c c' k,
    ofm nil pure M (sg_piapp T1 T2)
    -> ofm nil pure M' (sg_piapp T1 T2)
    -> fstm nil M c
    -> fstm nil M' c'
    -> fsts (sg_piapp T1 T2) k
    -> (mhalt nil M <-> mhalt nil M')
    -> eqc nil c c' k
    -> (forall V d,
          ofm nil pure V T1
          -> fstm nil V d
          -> mvalue nil V
          -> ceqm nil pure (md_appap M V) (md_appap M' V) (subst1 d T2))
    -> ceqm nil pure M M' (sg_piapp T1 T2).
Proof.
intros T1 T2 M M' c c' k HofM HofM' HfstM HfstM' Hfstpi Hhaltiff Heqc Hceqif.
rewrite -> ceqm_is_leqm; [].
intros _.
#2 do2 3 split; auto.
  intros ka ca ca' Hfstpia HfstMa HfstMa'.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  so (fsts_fun _#3 Hfstpi Hfstpia); subst ka; clear Hfstpia.
  invert Hfstpi; clear Hfstpi; [].
  intros k1 k2 Hfst1 Hfst2 <-.
  assumption.

  intros E ca HintE HfstMa.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  invert Hfstpi; clear Hfstpi; [].
  intros k1 k2 Hfst1 Hfst2 <-.
  invert (eqc_reg _#4 wf_nil Heqc anderr); [].
  intros Hofk1 Hofk2.
  invert HintE; clear HintE; [].
  intros <-.
  simp_sub.
  do2 3 split; auto; [].
  intros V V' Hsteps Hsteps' HvalV HvalV'.
  simpl.
  so (intc_fun _#5 wf_nil (eqc_reg _#4 wf_nil Heqc andel) intctx_nil) as Hintk.
  simp_sub in Hintk.
  invert Hintk; [].
  intros F _ _ _ _ Heq _.
  rewrite <- Heq; [].
  so (preservations_fst _#7 wf_nil HofM Hsteps (fsts_piapp Hfst1 Hfst2) HfstM) as (d & HfstV & Heqd).
  so (preservations_fst _#7 wf_nil HofM' Hsteps' (fsts_piapp Hfst1 Hfst2) HfstM') as (d' & HfstV' & Heqd').
  #4 eapply sg_piapp_interp_; eauto.
    simp_sub.
    clear dependent M'.
    eapply preservations_md; eauto with static; done.

    simp_sub.
    clear dependent M.
    eapply preservations_md; eauto with static; done.

    unfold intks.
    #3 apply (intk_defeq nil _ nil nil c c); eauto using wf_nil, ofk_pi, intctx_nil.
      rewrite -> !Heq; auto; done.

      simp_sub.
      exact Heqd.

      simp_sub.
      apply (@eqc_trans _ _ c'); auto; done.
    
      intros W W' Q HintW.
      so (leqm_refl _#4 HofM wf_nil anderrr _ _ fintctx_nil HfstM) as H.
      simp_sub in H.
      so (H anderl V V Hsteps Hsteps HvalV HvalV) as HintV.
      clear H.
      simpl in HintV.
      invert HintV; clear HintV; [].
      intros _ _ k1a k2a G _ _ _ _ _ _ Hfst1a Hfst2a _ Hfun Heq'.
      so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
      so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
      exploit (cand_inject (sk_pi (simp k1) (simp k2)) F G).
        etransitivity; eauto; done.
      subst G.
      clear Heq'.
      so (Hfun _ _ _ HintW) as (b & b' & HfstW & HfstW' & HintVW).
      clear Hfun.
      simpl in HintVW; simp_sub in HintVW.
      exists b, b'.
      do2 2 split; auto; [].
      simpl; simp_sub.
      invert (ofm_reg _#4 wf_nil HofM); [].
      intros k1a HofT1 Hfst1a HofT2.
      so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
      so (ints_reg _#5 HintW) as (HofW & HofW').
      simp_sub in HofW; simp_sub in HofW'.
      so (ofm_fst _#6 wf_nil HofW HfstW Hfst1) as Hofb.
      so (ofm_fst _#6 wf_nil HofW' HfstW' Hfst1) as Hofb'.
      assert (intctx (nil; cl_cn k1) (nil; Sta b b' Q) (nil; Sta b b' Q)) as Hintctx.
        #4 apply intctx_cn.
          exact intctx_nil.

          simp_sub.
          apply eqc_refl; assumption.

          simp_sub.
          apply eqc_refl; assumption.

          eapply ints_fst; eauto using wf, intctx_nil; done.
      #4 eapply mpuev_ceqm; eauto.
        intros U U' HintU.
        so (ints_reg _#5 HintU) as (HofU & HofU').
        simpl in HofU; simp_sub in HofU.
        simpl in HofU'; simp_sub in HofU'.
        sor (refine (ints_val (nil; cl_cn k1) _#5 _ _ _ HintU)) as (Hval & Hval'); auto using wf; [].
        do2 3 split; auto using ofm_forget; done.

        intros U1 U2 U1' U2' Hceq1 Hceq2 HvalU1' HvalU2' HintU.
        eapply ints_ceqm'; eauto; [].
        auto with static; done.

        apply ceqm_refl; [].
        eapply ofm_appap; eauto; [].
        clear HofM'.
        eapply preservations_md; eauto using wf_nil; done.

        #2 apply (ceqm_trans _#3 (md_appap M W')).
          apply ceqm_comp; [].
          #3 apply (@mcc_appap _#6 _ _ T1).
            apply ceqm_symm; [].
            eapply msteps_ceqm; eauto; done.

            apply ceqm_refl; assumption.

            assumption.

          #2 apply (ceqm_trans _#3 (md_appap M' W')).
            apply Hceqif; auto; [].
            exact (ints_val _#6 wf_nil HofT1 intctx_nil HintW ander).

            apply ceqm_comp; [].
            #3 apply (@mcc_appap _#6 _ _ T1).
              eapply msteps_ceqm; eauto; done.

              apply ceqm_refl; assumption.

              assumption.
Qed.


Lemma ceqm_sigma_ext :
  forall T1 T2 M M' c,
    ofm nil pure M (sg_sigma T1 T2)
    -> ofm nil pure M' (sg_sigma T1 T2)
    -> fstm nil M c
    -> ceqm nil pure (md_pi1 M) (md_pi1 M') T1
    -> ceqm nil pure (md_pi2 M) (md_pi2 M') (subst1 (cn_pi1 c) T2)
    -> ceqm nil pure M M' (sg_sigma T1 T2).
Proof.
intros T1 T2 M M' c HofM HofM' HfstM Hceq1 Hceq2.
invert (ofm_reg _#4 wf_nil HofM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fstm _#3 wf_nil HofM') as (c' & HfstM').
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
rewrite -> ceqm_is_leqm; [].
assert (eqc nil c c' (kd_sigma k1 k2)) as Heqc.
  so (ceqm_fst _#4 wf_nil Hceq1) as (k1a & c1 & c1' & Hfst1a & HfstM1 & HfstM1' & Heqc1).
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  invert HfstM1; clear HfstM1; [].
  intros ca HfstMa <-.
  invert HfstM1'; clear HfstM1'; [].
  intros ca' HfstMa' <-.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  so (ceqm_fst _#4 wf_nil Hceq2) as (l & c2 & c2' & Hfst2x & HfstM2 & HfstM2' & Heqc2).
  so (fsts_fun _#3 (subst_fsts _#3 Hfst2) Hfst2x); subst l; clear Hfst2x.
  invert HfstM2; clear HfstM2; [].
  intros ca HfstMa <-.
  invert HfstM2'; clear HfstM2'; [].
  intros ca' HfstMa' <-.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  apply eqc_extsigma; auto; [].
  eapply ofs_fst; eauto; done.
intros _.
#2 do2 3 split; auto.
  intros l ca ca' Hfst HfstMa HfstMa'.
  so (fsts_fun _#3 Hfst (fsts_sigma Hfst1 Hfst2)); subst l; clear Hfst.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  so (fstm_fun _#4 HfstM' HfstMa'); subst ca'; clear HfstMa'.
  exact Heqc.

  intros E ca Hint HfstMa.
  so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
  invert Hint; clear Hint; [].
  intros <-.
  simpl; simp_sub.
  #2 do2 3 split; auto.
    transitivity (mhalt nil (md_pi1 M)).
      split.
        intro HhaltM.
        destruct HhaltM as (V & HstepsV & HvalV).
        so (preservations_md _#5 wf_nil HofM HstepsV) as HofV.
        invert (Progress.canon_md _#3 HofV HvalV); [].
        intros V1 V2 <-.
        exists V1.
        invert HvalV; [].
        intros HvalV1 HvalV2.
        split; auto; [].
        #2 eapply star_trans.
          eapply star_map'; eauto with dynamic; done.

          apply star_one; [].
          apply mstep_pi12; auto; done.

        intros Hhaltpi.
        eapply mpi1_halt_inversion'; eauto; done.

      transitivity (mhalt nil (md_pi1 M')).
        eapply ceqm_halt_iff; [].
        apply ceqm_comp; apply mcc_forget; [].
        exact Hceq1.

        split.
          intros Hhaltpi.
          eapply mpi1_halt_inversion'; eauto; done.

          intro HhaltM.
          destruct HhaltM as (V & HstepsV & HvalV).
          so (preservations_md _#5 wf_nil HofM' HstepsV) as HofV.
          invert (Progress.canon_md _#3 HofV HvalV); [].
          intros V1 V2 <-.
          exists V1.
          invert HvalV; [].
          intros HvalV1 HvalV2.
          split; auto; [].
          #2 eapply star_trans.
            eapply star_map'; eauto with dynamic; done.
  
            apply star_one; [].
            apply mstep_pi12; auto; done.

    intros V V' HstepsM HstepsM' HvalV HvalV'.  
    so (ofm_fst _#6 wf_nil HofM HfstM (fsts_sigma Hfst1 Hfst2)) as Hofc.
    so (intc_fun _#5 wf_nil Hofc intctx_nil) as Hintk.
    simp_sub in Hintk.
    invert Hintk; clear Hintk; [].
    intros P _ _ _ _ _ Heq _.
    rewrite <- Heq; [].
    so (preservations_md _#5 wf_nil HofM HstepsM) as HofV.
    invert (Progress.canon_md _#3 HofV HvalV); [].
    intros V1 V2 <-.
    so (preservations_md _#5 wf_nil HofM' HstepsM') as HofV'.
    invert (Progress.canon_md _#3 HofV' HvalV'); [].
    intros V1' V2' <-.
    invert HvalV; clear HvalV; [].
    intros HvalV1 HvalV2.
    invert HvalV'; clear HvalV'; [].
    intros HvalV1' HvalV2'.
    so (preservations_fst _#7 wf_nil HofM HstepsM (fsts_sigma Hfst1 Hfst2) HfstM) as (d & HfstV & Heqd).
    invert HfstV; clear HfstV; [].
    intros d1 d2 HfstV1 HfstV2 <-.
    so (preservations_fst _#7 wf_nil HofM' HstepsM' (fsts_sigma Hfst1 Hfst2) HfstM') as (d' & HfstV' & Heqd').
    invert HfstV'; clear HfstV'; [].
    intros d1' d2' HfstV1' HfstV2' <-.
    so (leqm_refl _#4 HofM wf_nil anderrr _ _ fintctx_nil HfstM anderl) as H.
    simp_sub in H.
    so (H _ _ HstepsM HstepsM (mvalue_pair HvalV1 HvalV2) (mvalue_pair HvalV1 HvalV2)) as HintV.
    clear H.
    simpl in HintV.
    invert HintV; clear HintV; [].
    intros d1a d1aa k1a k2a Pa _ _ _ _ _ _ HfstV1a HfstV1aa Hfst1a Hfst2a HintV1 HintV2 Heq'.
    so (fstm_fun _#4 HfstV1 HfstV1a); subst d1a; clear HfstV1a.
    so (fstm_fun _#4 HfstV1 HfstV1aa); subst d1aa; clear HfstV1aa.
    so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
    so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
    exploit (cand_inject (sk_sigma (simp k1) (simp k2)) P Pa) as HeqP.
      transitivity (intc c nil); auto; done.
    subst Pa.
    clear Heq'.
    #4 eapply sg_sigma_interp_; eauto.
      simp_sub; apply ofm_forget; auto; done.

      simp_sub; apply ofm_forget; auto; done.

      #2 eapply ints_ceqm'; eauto using wf_nil, intctx_nil.
        simp_sub.
        apply ceqm_refl; [].
        exact (pair_ofm_inversion_1 _#6 wf_nil HofV).

        simp_sub.
        #2 apply (ceqm_trans _#3 (md_pi1 M)).
          apply ceqm_symm; [].
          apply msteps_ceqm; eauto with static; [].
          #2 eapply star_trans.
            eapply star_map'; eauto with dynamic; done.

            apply star_one; apply mstep_pi12; auto; done.

          #2 apply (ceqm_trans _#3 (md_pi1 M')).
            exact Hceq1.

            apply msteps_ceqm; eauto with static; [].
            #2 eapply star_trans.
              eapply star_map'; eauto with dynamic; done.
  
              apply star_one; apply mstep_pi12; auto; done.
            
      so (ofm_fst _#6 wf_nil (pair_ofm_inversion_1 _#6 wf_nil HofV) HfstV1 Hfst1) as Hofd1.
      so (ofm_fst _#6 wf_nil (pair_ofm_inversion_1 _#6 wf_nil HofV') HfstV1' Hfst1) as Hofd1'.
      so (ofc_ofcr _#3 (eqc_reg _#4 wf_nil Heqd anderl)) as (l & Hofcr).
      invert Hofcr; [].
      intros _ l2 _ Hofd2 _ _.
      clear l Hofcr.
      assert (ceqm nil pure V2 V2' (subst1 d1 T2)) as Hceq2'.
        apply ceqm_comp; [].
        #2 apply (mcc_subsume _#8 (subst1 (cn_pi1 c) T2)).
          #2 refine (anti_msteps_ceqm_msteps_trans _#7 _ Hceq2 _).
            #2 eapply star_trans.
              eapply star_map'; eauto with dynamic; done.

              apply star_one; apply mstep_pi22; auto; done.

            #2 eapply star_trans.
              eapply star_map'; eauto with dynamic; done.

              apply star_one; apply mstep_pi22; auto; done.

          apply subs_refl; [].
          eapply funct_sg; eauto using wf_nil; [].
          #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair d1 d2))).
            eapply eqc_pi1; eauto; done.

            eapply eqc_beta1; eauto; done.
      #4 rewrite <- (ints_fun (nil; cl_cn k1) _ (nil; Sta d1 d1 (Cand (simp k1) (fst P)))).
        #3 eapply (ints_ceqm' (nil; cl_cn k1)); eauto.
          so (ofs_fst _#3 HofT1 Hfst1).
          auto using wf; done.

          #4 apply intctx_cn.
            exact intctx_nil.

            apply eqc_refl; [].
            simp_sub; assumption.

            apply eqc_refl; [].
            simp_sub; assumption.

            exact (ints_fst _#9 wf_nil HofT1 intctx_nil Hfst1 HfstV1 HfstV1 HintV1).

          simpl; simp_sub.
          apply ceqm_refl; [].
          exact (ceqm_reg _#5 wf_nil Hceq2' andel).

        so (ofs_fst _#3 HofT1 Hfst1).
        auto using wf; done.

        exact HofT2.

        #4 apply intctx_cn.
          exact intctx_nil.

          simp_sub.
          apply eqc_refl; [].
          apply (ofm_fst _ pure V1 T1); auto using wf; [].
          eapply pair_ofm_inversion_1; eauto using wf; done.

          simp_sub.
          so (ofc_ofcr _#3 (eqc_reg _#4 wf_nil Heqd' anderl)) as (l & Hofcr).
          invert Hofcr; [].
          intros _ l2' _ Hofd2' _ _.
          clear l Hofcr.
          #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair d1 d2))).
            eauto with static; done.

            #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair d1' d2'))).
              eapply eqc_pi1; eauto using eqc_symm, eqc_trans; done.

              eauto with static; done.

          exact (ints_fst _#9 wf_nil HofT1 intctx_nil Hfst1 HfstV1 HfstV1 HintV1).
Qed.


Lemma ceqm_sigma_ext' :
  forall T1 T2 M M' k1 c,
    fsts T1 k1
    -> fstm nil M c
    -> ofs (nil; cl_cn k1) T2
    -> ceqm nil pure (md_pi1 M) (md_pi1 M') T1
    -> ceqm nil pure (md_pi2 M) (md_pi2 M') (subst1 (cn_pi1 c) T2)
    -> ceqm nil pure M M' (sg_sigma T1 T2).
Proof.
intros T1 T2 M M' k1 c Hfst1 HfstM HofT2 Hceq1 Hceq2.
so (ofm_reg _#4 wf_nil (ceqm_reg _#5 wf_nil Hceq1 andel)) as HofT1.
assert (ofm nil pure M (sg_sigma T1 T2)) as HofM.
  #2 eapply Singleton.ofm_extsigma_dep; eauto using wf_nil.
    exact (ceqm_reg _#5 wf_nil Hceq1 andel).

    exact (ceqm_reg _#5 wf_nil Hceq2 andel).
so (can_fstm _#3 wf_nil (ceqm_reg _#5 wf_nil Hceq1 ander)) as (d & H).
invert H; [].
intros c' HfstM' _.
clear d H.
assert (ofm nil pure M' (sg_sigma T1 T2)) as HofM'.
  #2 eapply Singleton.ofm_extsigma_dep; eauto using wf_nil.
    exact (ceqm_reg _#5 wf_nil Hceq1 ander).

    #2 eapply ofm_subsume.
      exact (ceqm_reg _#5 wf_nil Hceq2 ander).

      apply subs_refl; [].
      eapply funct_sg; eauto using wf_nil; [].
      apply (ceqm_fst' _ (md_pi1 M) (md_pi1 M') T1); auto with static; done.
eapply ceqm_sigma_ext; eauto; done.
Qed.


(* Extensionality for open terms -- mostly not used *)

Lemma ceqt_unit_ext_open :
  forall G e e',
    oft G e cn_unit
    -> oft G e' cn_unit
    -> (forall s, subof aof nil s G -> (thalt nil (subst s e) <-> thalt nil (subst s e')))
    -> ceqt G e e' cn_unit.
Proof.
intros G e e' Hofe Hofe' Hhaltiff.
apply closed_instances_ceqt; auto; [].
intros s Hs.
simp_sub.
#2 apply ceqt_unit_ext; auto.
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as H; simp_sub in H.
  exact H.

  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe') as H; simp_sub in H.
  exact H.
Qed.


Lemma ceqm_one_ext_open :
  forall G p M M',
    ofm G p M sg_one
    -> ofm G p M' sg_one
    -> (forall s, subof aof nil s G -> (mhalt nil (subst s M) <-> mhalt nil (subst s M')))
    -> ceqm G p M M' sg_one.
Proof.
intros G p M M' HofM HofM' Hhaltiff.
apply ceqm_with_wf; [].
intro Hwf.
destruct p.
  so (can_fstm _ _ _ Hwf HofM) as (c & HfstM).
  so (can_fstm _ _ _ Hwf HofM') as (c' & HfstM').
  eapply closed_instances_ceqm_pure; eauto using ofm_fst with static; [].
  intros s Hs.
  #2 eapply ceqm_one_ext; eauto.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM) as H.
    simp_sub in H.
    assumption.

    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM') as H.
    simp_sub in H.
    assumption.

  eapply closed_instances_ceqm_impure; eauto; [].
  intros s Hs.
  #2 eapply ceqm_one_ext; eauto.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM) as H.
    simp_sub in H.
    assumption.

    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM') as H.
    simp_sub in H.
    assumption.
Qed.


Lemma ceqm_stat_ext_open :
  forall G k M M' c c',
    ofm G pure M (sg_stat k)
    -> ofm G pure M' (sg_stat k)
    -> (forall s, subof aof nil s G -> (mhalt nil (subst s M) <-> mhalt nil (subst s M')))
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> ceqm G pure M M' (sg_stat k).
Proof.
intros G k M M' c c' HofM HofM' Hhaltiff HfstM HfstM' Heqc.
apply ceqm_with_wf; [].
intro Hwf.
eapply closed_instances_ceqm_pure; eauto with static; [].
intros s Hs.
simp_sub.
#2 eapply ceqm_stat_ext; eauto using subst_fstm, subst_eqc, subof_aof_tof, subof_aof_sof.
  relquest.
    eapply subst_ofm; eauto using subof_aof_dof; done.

    simp_sub; reflexivity.

  relquest.
    eapply subst_ofm; eauto using subof_aof_dof; done.

    simp_sub; reflexivity.
Qed.


Lemma ceqm_dyn_ext_open :
  forall G t M M',
    ofm G pure M (sg_dyn t)
    -> ofm G pure M' (sg_dyn t)
    -> ceqt G (tm_ext M) (tm_ext M') t
    -> ceqm G pure M M' (sg_dyn t).
Proof.
intros G t M M' HofM HofM' Hceq.
apply ceqm_with_wf; [].
intro Hwf.
so (can_fstm _ _ _ Hwf HofM) as (c & HfstM).
so (can_fstm _ _ _ Hwf HofM') as (c' & HfstM').
#2 eapply closed_instances_ceqm_pure; eauto using fsts_dyn.
  #2 apply eqc_extone.
    clear HofM'.
    eapply ofm_fst; eauto with static; done.

    clear HofM.
    eapply ofm_fst; eauto with static; done.

  intros s Hs.
  simp_sub.
  #3 eapply ceqm_dyn_ext; eauto.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM) as H.
    simp_sub in H.
    assumption.

    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM') as H.
    simp_sub in H.
    assumption.

    so (subst_ceqt _#6 Hs Hwf Hceq) as H.
    simp_sub in H.
    assumption.
Qed.


(* Beta principles *)

Lemma fstm_pure :
  forall G p M T c,
    ofm G p M T
    -> fstm G M c
    -> ofm G pure M T.
Proof.
intros G p M T c HofM Hfst.
revert c Hfst.
induct HofM;
try (intros; eauto with static; done).

(* appgn *)
intros G T1 T2 M1 M2 c2 _ _ _ _ _ c Hfst.
invert Hfst; done.

(* appap *)
intros G p T1 T2 M1 M2 c2 _ IH1 HofM2 _ Hfst2 c Hfst.
invert Hfst; [].
intros c1 _ Hfst1 _ _.
so (IH1 _ Hfst1) as HofM1.
eapply ofm_appap; eauto; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 c Hfst.
invert Hfst; [].
intros c1 Hfst1 c2 Hfst2 _.
apply ofm_pair; eauto; done.

(* unpack *)
intros G k t T e M _ _ _ _ c Hfst.
invert Hfst; done.

(* lett *)
intros G p t T e M Hofe _ IH c Hfst.
invert Hfst; [].
intros HfstM; [].
eapply ofm_lett; eauto; [].
eapply IH; [].
rewrite <- (subst_dot0_sh M); [].
eapply subst_fstm; eauto; [].
prove_subof; done.

(* letm *)
intros G T1 T2 k1 M1 M2 _ _ _ _ _ _ c Hfst.
invert Hfst; done.

(* seal *)
intros G T M _ _ c Hfst.
invert Hfst; done.
Qed.


Lemma ceqt_app_beta :
  forall G t e1 e2 t',
    oft G (tm_app (tm_lam t e1) e2) t'
    -> thalt G e2
    -> ceqt G (tm_app (tm_lam t e1) e2) (subst1 e2 e1) t'.
Proof.
intros G t1 e1 e2 t2 Hof Hhalt.
apply ceqt_with_wf; intro Hwf; [].
so (app_oft_inversion _#4 Hwf Hof) as (t1' & Hof' & Hof2).
so (lam_oft_inversion _#4 Hwf Hof') as (t2' & Heqc & Hof1).
so (Consistent.arrow_injective _#5 Hwf Heqc) as (Heqt1 & Heqt2).
clear Hof'.
destruct Hhalt as (v & Hsteps & Hval).
#2 apply (ceqt_trans _ _ (subst1 v e1)).
  eapply tsteps_ceqt; eauto; [].
  #2 eapply star_trans.
    apply (star_map' _ _ (fun x => tm_app _ x)); eauto with dynamic; done.

    apply star_one; [].
    apply tstep_app3; auto; done.

  apply ceqt_comp; [].
  eapply tcc_equiv.
    #2 apply (funct_ceq_tm _ (G; cl_tm t1)); eauto.
      apply wf_tm; auto; [].
      exact (eqc_reg _#4 Hwf Heqt1 anderl).
  
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply deq_tm; [].
      apply ceqt_symm; [].
      eapply tsteps_ceqt; eauto; [].
      eapply oft_equiv; eauto with static; done.

    simp_sub.
    apply eqc_symm; auto; done.
Qed.    


Lemma ceqt_pi1_beta :
  forall G t e1 e2,
    oft G (tm_pi1 (tm_pair e1 e2)) t
    -> thalt G e1
    -> thalt G e2
    -> ceqt G (tm_pi1 (tm_pair e1 e2)) e1 t.
Proof.
intros G t1 e1 e2 Hof Hhalt1 Hhalt2.
apply ceqt_with_wf; intro Hwf; [].
so (pi1_oft_inversion _#3 Hwf Hof) as (t2 & Hof').
so (pair_oft_inversion_prod _#5 Hwf Hof') as (Hof1 & Hof2).
clear Hof'.
destruct Hhalt1 as (v1 & Hsteps1 & Hval1).
destruct Hhalt2 as (v2 & Hsteps2 & Hval2).
#2 eapply ceqt_trans.
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    apply (star_map' _ _ (fun z => tm_pi1 (tm_pair z _))); eauto with dynamic; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => tm_pi1 (tm_pair _ z))); eauto with dynamic; done.

      apply star_one; apply tstep_pi12; auto; done.

  apply ceqt_symm; [].
  apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqt_pi2_beta :
  forall G t e1 e2,
    oft G (tm_pi2 (tm_pair e1 e2)) t
    -> thalt G e1
    -> thalt G e2
    -> ceqt G (tm_pi2 (tm_pair e1 e2)) e2 t.
Proof.
intros G t2 e1 e2 Hof Hhalt1 Hhalt2.
apply ceqt_with_wf; intro Hwf; [].
so (pi2_oft_inversion _#3 Hwf Hof) as (t1 & Hof').
so (pair_oft_inversion_prod _#5 Hwf Hof') as (Hof1 & Hof2).
clear Hof'.
destruct Hhalt1 as (v1 & Hsteps1 & Hval1).
destruct Hhalt2 as (v2 & Hsteps2 & Hval2).
#2 eapply ceqt_trans.
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    apply (star_map' _ _ (fun z => tm_pi2 (tm_pair z _))); eauto with dynamic; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => tm_pi2 (tm_pair _ z))); eauto with dynamic; done.

      apply star_one; apply tstep_pi22; auto; done.

  apply ceqt_symm; [].
  apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqt_ext_beta :
  forall G e t,
    oft G e t
    -> ceqt G (tm_ext (md_dyn e)) e t.
Proof.
intros G e t Hofe.
apply ceqt_with_wf.
intros Hwf.
apply closed_instances_ceqt; auto with static; [].
intros s Hs.
simp_sub.
so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as Hofes.
#2 apply (ceqt_conditional (thalt nil (subst s e))); eauto with static.
  intro Hhalt.
  exact (Inversion.dyn_halt_inversion' _ (Inversion.ext_halt_inversion' _ Hhalt)).

  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  #2 apply (ceqt_trans _ _ v).
    apply tsteps_ceqt; auto with static; [].
    #2 eapply star_trans.
      apply (star_map' _ _ (fun x => tm_ext (md_dyn x))); eauto with dynamic; done.

      apply star_one; [].
      apply tstep_ext2; auto; done.

  apply ceqt_symm; [].
  apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqt_lett_beta :
  forall G e1 e2 t,
    oft G (tm_lett e1 e2) t
    -> thalt G e1
    -> ceqt G (tm_lett e1 e2) (subst1 e1 e2) t.
Proof.
intros G e1 e2 t2 Hof Hhalt.
apply ceqt_with_wf; intro Hwf; [].
so (lett_oft_inversion _#4 Hwf Hof) as (t1 & Hof1 & Hof2).
destruct Hhalt as (v & Hsteps & Hvalv).
#2 apply (ceqt_trans _#2 (subst1 v e2)).
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => tm_lett x _)); eauto with dynamic; done.

    apply star_one; [].
    apply tstep_lett2; eauto; done.

  apply ceqt_symm; [].
  replsub t2 with (subst1 e1 (shift1 t2)).
  so (oft_reg _#3 Hwf Hof1) as Hoft1.
  apply (funct_ceq_tm _ (G; cl_tm t1)); auto using wf; [].
  apply subrel_dot; auto using subrel_id; [].
  simp_sub.
  apply deq_tm; [].
  apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqt_letm_beta :
  forall G M e t c,
    oft G (tm_letm M e) t
    -> mhalt G M
    -> fstm G M c
    -> ceqt G (tm_letm M e) (subst (dot M (dot c id)) e) t.
Proof.
intros G M e t c Hof Hhalt HfstM.
apply ceqt_with_wf; [].
intro Hwf.
so (Inversion.letm_oft_inversion _#4 Hwf Hof) as (T & k & HofM & Hfst & Hofe).
destruct Hhalt as (V & Hsteps & HvalV).
so (preservations_md _#5 Hwf HofM Hsteps) as HofV.
so (preservations_fst _#7 Hwf HofM Hsteps Hfst HfstM) as (b & HfstV & Heqc).
#2 apply (ceqt_trans _ _ (subst (dot V (dot b id)) e)).
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => tm_letm x _)); eauto with dynamic; done.

    apply star_one; [].
    apply tstep_letm2; auto; done.

  apply ceqt_symm; [].
  replsub t with (subst (dot M (dot c id)) (subst (sh 2) t)).
  so (ofm_reg _#4 Hwf HofM) as HofT.
  apply (funct_ceq_tm _ (G; cl_cn k; cl_md (shift1 T) (var 0))); auto using wf; [].
  #2 apply subrel_dot.
    #2 apply subrel_dot.
      apply subrel_id; done.

      simp_sub.
      apply deq_cn; auto; [].
      exact (eqc_reg _#4 Hwf Heqc andel).

    simp_sub.
    apply deq_md; auto; [].
    apply msteps_ceqm; auto; [].
    eapply fstm_pure; eauto; done.
Qed.


Lemma ceqt_letm_beta_nofst :
  forall G V e t,
    oft G (tm_letm V e) t
    -> mvalue G V
    -> (subst (dot (var 0) (dot nonsense (sh 2))) e = e)
    -> ceqt G (tm_letm V e) (subst (dot V (dot nonsense id)) e) t.
Proof.
intros G V e t Hof Hval Hnotfree.
apply ceqt_with_wf; [].
intro Hwf.
so (letm_oft_inversion _#4 Hwf Hof) as (T & _ & HofV & _).
so (can_fstm_value _#4 Hwf HofV Hval) as (c & HfstV).
apply tsteps_ceqt; auto; [].
#2 eapply star_step.
  eapply tstep_letm2; eauto; done.

  replsub (subst (dot V (dot nonsense id)) e) with (subst (dot V (dot c id)) (subst (dot (var 0) (dot nonsense (sh 2))) e)).
  rewrite -> Hnotfree; [].
  apply star_refl; done.  
Qed.


Lemma ceqm_appgn_beta :
  forall G p T M1 M2 U c,
    ofm G p (md_appgn (md_lamgn T M1) M2) U
    -> mhalt G M2
    -> fstm G M2 c
    -> ceqm G p (md_appgn (md_lamgn T M1) M2) (subst (dot M2 (dot c id)) M1) U.
Proof.
intros G p T1 M1 M2 U c Hof Hhalt HfstM2.
apply ceqm_with_wf; intro Hwf; [].
so (appgn_ofm_inversion _#5 Hwf Hof) as (U1 & U2 & ca & Hof' & HofM2 & HfstM2a & Hsubs & Heqp).
subst p.
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamgn_ofm_inversion _#5 Hwf Hof') as (T2 & k1 & Hsubs' & HfstT1 & HofM1).
clear Hof'.
so (Declarative.subs_pigen_invert'' _#5 Hwf Hsubs') as (l1 & Hsubs1 & HfstU1 & Hsubs2).
so (subs_fst _#5 Hwf Hsubs1 HfstU1 HfstT1) as Hsubk.
so (subs_reg _#3 Hwf Hsubs1) as (_ & HofT1).
destruct Hhalt as (V & Hsteps & Hval).
so (preservations_fst _#7 Hwf HofM2 Hsteps HfstU1 HfstM2) as (b & HfstV & Heqc).
#2 apply (ceqm_trans _ _ _ (subst (dot V (dot b id)) M1)).
  apply msteps_ceqm; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => md_appgn _ x)); eauto with dynamic; done.

    apply star_one; [].
    apply mstep_appgn3; auto; done.

  apply ceqm_symm; [].
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; eapply mcc_subsume.
    refine (funct_ceq_md _#7 _ _ _ HofM1); auto using wf_md; [].
    #4 prove_subrel.
      eapply ofc_subsume; eauto; [].
      exact (eqc_reg _#4 Hwf Heqc andel).

      eapply eqc_subsume; eauto; done.

      apply msteps_ceqm; eauto with static; done.

      assumption.

    simp_sub.
    eapply subst1_subs; eauto; [].
    exact (eqc_reg _#4 Hwf Heqc andel).
Qed.


Lemma ceqm_appap_beta :
  forall G p T M1 M2 U c,
    ofm G p (md_appap (md_lamap T M1) M2) U
    -> mhalt G M2
    -> fstm G M2 c
    -> ceqm G p (md_appap (md_lamap T M1) M2) (subst (dot M2 (dot c id)) M1) U.
Proof.
intros G p T1 M1 M2 U c Hof Hhalt HfstM2.
apply ceqm_with_wf; intro Hwf; [].
so (appap_ofm_inversion _#5 Hwf Hof) as (U1 & U2 & ca & Hof' & HofM2 & HfstM2a & Hsubs).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamap_ofm_inversion _#5 Hwf Hof') as (T2 & k1 & Hsubs' & HfstT1 & HofM1).
clear Hof'.
so (Declarative.subs_piapp_invert'' _#5 Hwf Hsubs') as (l1 & Hsubs1 & HfstU1 & Hsubs2).
so (subs_fst _#5 Hwf Hsubs1 HfstU1 HfstT1) as Hsubk.
so (subs_reg _#3 Hwf Hsubs1) as (_ & HofT1).
destruct Hhalt as (V & Hsteps & Hval).
so (preservations_fst _#7 Hwf HofM2 Hsteps HfstU1 HfstM2) as (b & HfstV & Heqc).
#2 apply (ceqm_trans _ _ _ (subst (dot V (dot b id)) M1)).
  apply msteps_ceqm; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => md_appap _ x)); eauto with dynamic; done.

    apply star_one; [].
    apply mstep_appap3; auto; done.

  apply ceqm_symm; [].
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; eapply mcc_subsume.
    refine (funct_ceq_md _#7 _ _ _ (ofm_forget_pure HofM1)); auto using wf_md; [].
    #4 prove_subrel.
      eapply ofc_subsume; eauto; [].
      exact (eqc_reg _#4 Hwf Heqc andel).

      eapply eqc_subsume; eauto; done.

      apply msteps_ceqm; eauto with static; done.

      assumption.

    simp_sub.
    eapply subst1_subs; eauto; [].
    exact (eqc_reg _#4 Hwf Heqc andel).
Qed.


Lemma ceqm_appap_beta_fstm_eqc :
  forall G p T M1 M2 U k c c',
    ofm G p (md_appap (md_lamap T M1) M2) U
    -> mhalt G M2
    -> fstm G M2 c
    -> fsts T k
    -> eqc G c c' k
    -> ceqm G p (md_appap (md_lamap T M1) M2) (subst (dot M2 (dot c' id)) M1) U.
Proof.
intros G p T1 M1 M2 T k1 c c' Hof HhaltM2 HfstM2 Hfst1 Heqc.
apply ceqm_with_wf; [].
intro Hwf.
#2 eapply ceqm_trans.
  eapply ceqm_appap_beta; eauto; done.
so (appap_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & ca & Hoflam & HofM2 & HfstM2a & Hsubs).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamap_ofm_inversion _#5 Hwf Hoflam) as (T2 & k1a & Hsubspi & Hfst1a & HofM1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (Declarative.subs_piapp_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
apply ceqm_comp; eapply mcc_subsume; eauto; [].
#2 apply ceqm_comp; apply (mcc_subsume _#8 (subst1 c T2)).
  replsub (subst1 c T2) with (subst (dot M2 (dot c id)) (shift1 T2)).
  apply ceqm_forget_pure; [].
  #2 refine (funct_ceq_md _#7 Hwf _ _ HofM1).
    apply wf_md; auto using wf; [].
    exact (subs_reg _#3 Hwf Hsubs1 ander).

    #2 prove_subrel; auto.
      exact (eqc_reg _#4 Hwf Heqc andel).

      apply ceqm_refl; [].
      eapply ofm_subsume; eauto; done.

  eapply subst1_subs; eauto; [].
  exact (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1').
Qed.


Lemma ceqm_pair_dep :
  forall G p T1 T2 M1 M1' M2 M2' k1 c1,
    wf G
    -> fsts T1 k1
    -> fstm G M1 c1
    -> ceqm G pure M1 M1' T1
    -> ceqm G p M2 M2' (subst1 c1 T2)
    -> ofs (G; cl_cn k1) T2
    -> ceqm G p (md_pair M1 M2) (md_pair M1' M2') (sg_sigma T1 T2).
Proof.
intros G p T1 T2 M1 M1' M2 M2' k1 c1 Hwf Hfst1 HfstM1 Hceq1 Hceq2 HofT2.
so (ceqm_reg _#5 Hwf Hceq1) as (HofM1 & _).
so (ofm_reg _#4 Hwf HofM1) as HofT1.
so (ofm_fst _#6 Hwf HofM1 HfstM1 Hfst1) as Hofc1.
apply ceqm_comp; [].
#2 apply (mcc_subsume _#8 (sg_sigma (Singleton.singles c1 T1) (shift1 (subst1 c1 T2)))).
  apply ceqm_comp; [].
  apply mcc_pair; auto; [].
  apply ceqm_forget_pure; [].
  apply ceqm_self_respecting; auto; done.

  #3 eapply subs_sigma; eauto.
    eapply Singleton.singles_subs; eauto; done.

    apply Singleton.singles_fst; eauto; done.

    simp_sub.
    apply subs_refl; [].
    rewrite <- (subst_dot0_sh T2) at 2; [].
    #2 eapply seq_ofs; eauto.
      apply wf_cn; auto; [].
      apply Singleton.single_ofk; auto; done.

      prove_subrel; [].
      apply eqc_symm; [].
      apply Singleton.single_eqc_weak; auto; done.
Qed.


Lemma ceqm_pi1_beta_instances :
  forall G T1 T2 M1 M2,
    ofm G pure M1 T1
    -> ofm G pure M2 T2
    -> (forall s, subof aof nil s G -> mhalt nil (subst s M2))
    -> ceqm G pure (md_pi1 (md_pair M1 M2)) M1 T1.
Proof.
intros G T1 T2 M1 M2 HofM1 HofM2 HhaltM2.
apply ceqm_with_wf; intro Hwf; [].
so (can_fstm _#3 Hwf HofM1) as (c1 & HfstM1).
so (can_fstm _#3 Hwf HofM2) as (c2 & HfstM2).
so (ofm_fst_o _#5 Hwf HofM1 HfstM1) as (k1 & Hfst1 & Hofc1).
so (ofm_fst_o _#5 Hwf HofM2 HfstM2) as (k2 & Hfst2 & Hofc2).
eapply closed_instances_ceqm_pure; eauto with static; [].
intros s Hs.
simp_sub.
assert (ofm nil pure (md_pi1 (md_pair (subst s M1) (subst s M2))) (subst s T1)) as Hof.
  eapply ofm_pi1; [].
  #2 eapply ofm_pair.
    eapply subst_ofm; eauto using subof_aof_dof; done.

    eapply subst_ofm; eauto using subof_aof_dof; done.
#9 eapply (ceqm_conditional_pure (mhalt nil (subst s M1))).
  exact Hof.

  eapply subst_ofm; eauto using subof_aof_dof; done.

  eapply fstm_pi1; [].
  eapply fstm_pair; eauto using subst_fstm, subof_aof_tof; done.

  eapply subst_fstm; eauto using subof_aof_tof; done.

  apply subst_fsts; eauto; done.

  eapply eqc_beta1; eauto using subst_ofc, subof_aof_sof; done.

  intro H.
  so (mpi1_halt_inversion' _ H) as H1.
  so (mpair_halt_inversion' _ _ H1) as (H2 & _).
  exact H2.

  auto; done.
intro HhaltM1x.
destruct HhaltM1x as (V1 & Hsteps1 & Hval1).
destruct (HhaltM2 s Hs) as (V2 & Hsteps2 & Hval2).
#2 eapply ceqm_trans.
  apply msteps_ceqm; auto.
  #2 eapply star_trans.
    apply (star_map' _ _ (fun z => md_pi1 (md_pair z _))); eauto with dynamic; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => md_pi1 (md_pair _ z))); eauto with dynamic; done.

      apply star_one; apply mstep_pi12; auto; done.

  apply ceqm_symm; [].
  apply msteps_ceqm; eauto using subst_ofm, subof_aof_dof; done.
Qed.


(* Could easily scrap the M1 halts assumption. *)
Lemma ceqm_pi1_beta :
  forall G T1 T2 M1 M2,
    ofm G pure M1 T1
    -> ofm G pure M2 T2
    -> mhalt G M1
    -> mhalt G M2
    -> ceqm G pure (md_pi1 (md_pair M1 M2)) M1 T1.
Proof.
intros G T1 T2 M1 M2 HofM1 HofM2 _ HhaltM2.
eapply ceqm_pi1_beta_instances; eauto; [].
intros s Hs.
eapply subst_mhalt; eauto using subof_aof_oof; done.
Qed.


Lemma ceqm_pi1_beta' :
  forall G T M1 M2,
    ofm G pure (md_pi1 (md_pair M1 M2)) T
    -> mhalt G M1
    -> mhalt G M2
    -> ceqm G pure (md_pi1 (md_pair M1 M2)) M1 T.
Proof.
intros G T1 M1 M2 Hof HhaltM1 HhaltM2.
apply ceqm_with_wf; intro Hwf; [].
so (pi1_ofm_inversion _#4 Hwf Hof) as (T2 & Hof').
so (pair_ofm_inversion_sigma_pure _#5 Hwf Hof') as (c1 & Hof1 & HfstM1 & Hof2).
eapply ceqm_pi1_beta; eauto.
Qed.


Lemma ceqm_pi2_beta_instances :
  forall G T1 T2 M1 M2,
    ofm G pure M1 T1
    -> ofm G pure M2 T2
    -> (forall s, subof aof nil s G -> mhalt nil (subst s M1))
    -> ceqm G pure (md_pi2 (md_pair M1 M2)) M2 T2.
Proof.
intros G T1 T2 M1 M2 HofM1 HofM2 HhaltM1.
apply ceqm_with_wf; intro Hwf; [].
so (can_fstm _#3 Hwf HofM1) as (c1 & HfstM1).
so (can_fstm _#3 Hwf HofM2) as (c2 & HfstM2).
so (ofm_fst_o _#5 Hwf HofM1 HfstM1) as (k1 & Hfst1 & Hofc1).
so (ofm_fst_o _#5 Hwf HofM2 HfstM2) as (k2 & Hfst2 & Hofc2).
#2 eapply closed_instances_ceqm_pure; eauto with static.
  relquest.
    eapply ofm_pi2; eauto with static; done.

    simp_sub; reflexivity.
intros s Hs.
simp_sub.
assert (ofm nil pure (md_pi2 (md_pair (subst s M1) (subst s M2))) (subst s T2)) as Hof.
  relquest.
    #2 eapply ofm_pi2.
      #2 eapply ofm_pair.
        eapply subst_ofm; eauto using subof_aof_dof; done.
    
        eapply subst_ofm; eauto using subof_aof_dof; done.

      eapply fstm_pair; eapply subst_fstm; eauto using subof_aof_tof; done.

    simp_sub; reflexivity.
#9 eapply (ceqm_conditional_pure (mhalt nil (subst s M2))).
  exact Hof.

  eapply subst_ofm; eauto using subof_aof_dof; done.

  eapply fstm_pi2; [].
  eapply fstm_pair; eauto using subst_fstm, subof_aof_tof; done.

  eapply subst_fstm; eauto using subof_aof_tof; done.

  apply subst_fsts; eauto; done.

  eapply eqc_beta2; eauto using subst_ofc, subof_aof_sof; done.

  intro H.
  so (mpi2_halt_inversion' _ H) as H1.
  so (mpair_halt_inversion' _ _ H1) as (_ & H2).
  exact H2.

  auto; done.
intro HhaltM2x.
destruct (HhaltM1 s Hs) as (V1 & Hsteps1 & Hval1).
destruct HhaltM2x as (V2 & Hsteps2 & Hval2).
#2 eapply ceqm_trans.
  apply msteps_ceqm; auto.
  #2 eapply star_trans.
    apply (star_map' _ _ (fun z => md_pi2 (md_pair z _))); eauto with dynamic; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => md_pi2 (md_pair _ z))); eauto with dynamic; done.

      apply star_one; apply mstep_pi22; auto; done.

  apply ceqm_symm; [].
  apply msteps_ceqm; eauto using subst_ofm, subof_aof_dof; done.
Qed.

(* Could easily scrap the M2 halts assumption. *)
Lemma ceqm_pi2_beta :
  forall G T1 T2 M1 M2,
    ofm G pure M1 T1
    -> ofm G pure M2 T2
    -> mhalt G M1
    -> mhalt G M2
    -> ceqm G pure (md_pi2 (md_pair M1 M2)) M2 T2.
Proof.
intros G T1 T2 M1 M2 HofM1 HofM2 HhaltM1 _.
eapply ceqm_pi2_beta_instances; eauto; [].
intros s Hs.
eapply subst_mhalt; eauto using subof_aof_oof; done.
Qed.


Lemma ceqm_pi2_beta' :
  forall G T M1 M2,
    ofm G pure (md_pi2 (md_pair M1 M2)) T
    -> mhalt G M1
    -> mhalt G M2
    -> ceqm G pure (md_pi2 (md_pair M1 M2)) M2 T.
Proof.
intros G T M1 M2 Hof HhaltM1 HhaltM2.
apply ceqm_with_wf; intro Hwf; [].
so (pi2_ofm_inversion _#4 Hwf Hof) as (T1 & T2 & cc & Hof' & Hfstpair & Hsubs).
invert Hfstpair; clear Hfstpair; [].
intros c1 c2 HfstM1 HfstM2 <-.
so (pair_ofm_inversion_sigma_pure _#5 Hwf Hof') as (c1a & HofM1 & HfstM1a & HofM2).
so (fstm_fun _#4 HfstM1 HfstM1a); subst c1a; clear HfstM1a.
apply ceqm_comp; [].
#2 eapply mcc_subsume.
  eapply ceqm_pi2_beta; eauto; done.

  eapply subs_trans; eauto; [].
  apply subs_refl; [].
  invert (ofm_reg _#4 Hwf Hof'); [].
  intros k1 HofT1 Hfst1 HofT2.
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  eapply funct_sg; eauto; [].
  apply eqc_symm; [].
  eapply eqc_beta1; eauto using ofm_fst, subst_fsts; done.
Qed.


Lemma ceqm_unpack_beta :
  forall G p c e k t M T T',
    ofm G p (md_unpack (tm_pack c e k t) M T) T'
    -> thalt G e
    -> ceqm G p (md_unpack (tm_pack c e k t) M T) (subst (dot e (dot c id)) M) T'.
Proof.
intros G p c e k t M T T' Hof Hhalt.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k' & t' & Hof' & HofM & Hsubs & Heqp).
subst p.
so (pack_oft_inversion _#6 Hwf Hof') as (Heqc & Hofc & Hofe).
so (Consistent.exists_injective _#5 Hwf (eqc_symm Heqc)) as (Heqk & Heqt).
clear Hof' Heqc.
destruct Hhalt as (v & Hsteps & Hval).
so (preservations_tm _#4 Hwf Hofe Hsteps) as Hofv.
#2 apply (ceqm_trans _#3 (subst (dot v (dot c id)) M)).
  eapply msteps_ceqm; eauto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => md_unpack (tm_pack _ x _ _) _ _)); eauto with dynamic; done.

    apply star_one; [].
    apply mstep_unpack2; auto; done.

  apply ceqm_symm; [].
  apply ceqm_comp; [].
  #2 eapply mcc_subsume.
    #2 apply (funct_ceq_md _ (G; cl_cn k'; cl_tm t')); eauto.
      #2 apply wf_tm.
        apply wf_cn; auto; [].
        exact (eqk_reg _#3 Hwf Heqk ander).

        so (subst_ofc_type _#4 (subof_conv_0 _#3 (eqk_symm Heqk)) (eqc_reg _#4 (wf_cn Hwf (ofc_reg _#3 Hwf Hofc)) Heqt anderl)) as H.
        simp_sub in H.
        exact H.

      #2 apply subrel_dot.
        apply subrel_dot; auto using subrel_id; [].
        simp_sub.
        #2 apply deq_cn.
          eapply ofc_subsume; eauto with static; done.

          apply eqc_refl; [].
          eapply ofc_subsume; eauto with static; done.

        simp_sub.
        apply deq_tm; [].
        eapply tsteps_ceqt; eauto; [].
        eapply oft_equiv; eauto; [].
        eapply subst1_eqc_type; eauto; done.

    simp_sub.
    assumption.
Qed.


Lemma ceqm_lett_beta :
  forall G p e M T,
    ofm G p (md_lett e M) T
    -> thalt G e
    -> ceqm G p (md_lett e M) (subst1 e M) T.
Proof.
intros G p e M T Hof Hhalt.
apply ceqm_with_wf; [].
intro Hwf.
so (Inversion.lett_ofm_inversion _#5 Hwf Hof) as (t & Hofe & HofM).
destruct Hhalt as (v & Hsteps & Hvalv).
#2 apply (ceqm_trans _#3 (subst1 v M)).
  apply msteps_ceqm; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun x => md_lett x _)); eauto with dynamic; done.

    apply star_one; [].
    apply mstep_lett2; eauto; done.

  apply ceqm_symm; [].
  replsub T with (subst1 e (shift1 T)).
  so (oft_reg _#3 Hwf Hofe) as Hoft.
  apply (funct_ceq_md _ (G; cl_tm t)); auto using wf; [].
  apply subrel_dot; auto using subrel_id; [].
  simp_sub.
  apply deq_tm; [].
  apply tsteps_ceqt; auto; done.
Qed.


Lemma ceqm_letm_beta_instances :
  forall G p M1 M2 T U c1,
    ofm G p (md_letm M1 M2 T) U
    -> (forall s, subof aof nil s G -> mhalt nil (subst s M1))
    -> fstm G M1 c1
    -> ceqm G p (md_letm M1 M2 T) (subst (dot M1 (dot c1 id)) M2) U.
Proof.
intros G p M1 M2 T2 U c1 Hof Hhalt HfstM1.
apply ceqm_with_wf; intro Hwf; [].
so (letm_ofm_inversion _#6 Hwf Hof) as (T1 & k1 & HofM1 & Hfst1 & HofM2 & Hsubs & ->).
so (ofm_reg _#4 Hwf HofM1) as HofT1.
so (ofm_fst _#6 Hwf HofM1 HfstM1 Hfst1) as Hofc1.
so (fstm_pure _#5 HofM1 HfstM1) as HofM1'.
assert (ofm G impure (subst (dot M1 (dot c1 id)) M2) U) as Hofr.
  eapply ofm_subsume; eauto; [].
  relquest.
    eapply subst_ofm; eauto; prove_subof; eauto; done.

    simp_sub; reflexivity.
apply closed_instances_ceqm_impure; auto; [].
intros s Hs.
simp_sub.
so (subst_ofm _#6 (subof_aof_dof _#3 Hs) HofM1') as HofM1x.
destruct (Hhalt s Hs) as (V & Hsteps & Hval).
so (preservations_fst _#7 wf_nil HofM1x Hsteps (subst_fsts _#3 Hfst1) (subst_fstm _#5 (subof_aof_tof _#3 Hs) HfstM1)) as (c1' & HfstV & Heqc).
#2 eapply ceqm_trans.
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H; exact H.
  #2 eapply star_trans.
    apply (star_map' _ _ (fun z => md_letm z _ _)); eauto with dynamic; done.
  #2 eapply star_step.
    eapply mstep_letm2; eauto; done.
  simp_sub.
  apply star_refl; done.
apply ceqm_symm; [].
#2 apply ceqm_comp; eapply mcc_subsume.
  refine (funct_ceq_md _#7 wf_nil (wf_md Hwf HofT1 Hfst1) _ HofM2); [].
  #4 apply subrel_deq_ceq_md; auto.
    exact wf_nil.

    apply subrel_deq_ceq_refl; apply subof_aof_dof; auto; done.

    apply msteps_ceqm; auto; done.

    eapply subst_fstm; eauto using subof_aof_tof; done.

  simp_sub.
  eapply subst_subs; eauto using subof_aof_sof; done.
Qed.


(* Commuting conversions *)

Lemma ceqt_commute_letm_unpack :
  forall G e1 M T e2 k t l u,
    oft G e1 (cn_exists k t)
    -> ofm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
    -> ofs G T
    -> fsts T l
    -> oft (G; cl_cn l; cl_md (shift1 T) (var 0)) e2 (subst (sh 2) u)
    -> ceqt G (tm_letm (md_unpack e1 M T) e2) (tm_unpack e1 (tm_letm M (subst (under 2 (sh 2)) e2))) u.
Proof.
cut (forall e1 M T e2 k t l u,
       oft nil e1 (cn_exists k t)
       -> ofm (nil; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
       -> ofs nil T
       -> fsts T l
       -> oft (nil; cl_cn l; cl_md (shift1 T) (var 0)) e2 (subst (sh 2) u)
       -> ofc nil u kd_type
       -> ceqt nil (tm_letm (md_unpack e1 M T) e2) (tm_unpack e1 (tm_letm M (subst (under 2 (sh 2)) e2))) u).
  intro Hprop.
  intros G e1 M T e2 k t l u Hofe1 HofM HofT Hfst Hofe2.  
  apply ceqt_with_wf; intro Hwf; [].
  so (oft_reg _#3 (wf_md Hwf HofT Hfst) Hofe2) as Hofuw.
  #2 sor (refine (subst_ofc G _ (dot nonsense (dot (Declarative.inh l) id)) _ _ _ Hofuw)) as Hofu.
    prove_subof; [].
    apply Declarative.inh_ofc; [].
    eapply ofs_fst; eauto; done.
  simp_sub in Hofu; clear Hofuw.
  #3 apply closed_instances_ceqt.
    eapply oft_letm; eauto; [].
    eapply ofm_unpack; eauto; done.

    eapply oft_unpack; eauto; [].
    #3 eapply oft_letm; eauto.
      eapply subst_fsts; eauto; done.

      simp_sub.
      relquest.
        eapply subst_oft; eauto; prove_subof; done.

        simp_sub; reflexivity.

      eapply subst_ofc_type; eauto; prove_subof; done.

    intros s Hs.
    simp_sub.
    simp_sub.
    #2 eapply rel_quest_1.
      #6 eapply Hprop.
        so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe1) as H; simp_sub in H.
        exact H.

        relquest.
          eapply subst_ofm; eauto; [].
          prove_subof; [].
          eapply subof_dof_trans; eauto using subof_aof_dof; [].
          prove_subof; done.

          simp_sub; reflexivity.

        eapply subst_ofs; eauto using subof_aof_sof; done.

        eapply subst_fsts; eauto using subof_aof_tof; done.

        relquest.
          eapply subst_oft; eauto; [].
          prove_subof; [].
          eapply subof_dof_trans; eauto using subof_aof_dof; [].
          prove_subof; done.

          simp_sub; reflexivity.

          eapply subst_ofc_type; eauto using subof_aof_sof; done.

      simp_sub; reflexivity.
intros e1 M T e2 k t l u Hofe1 HofM HofT HfstT Hofe2 Hofu.
assert (oft nil (tm_letm (md_unpack e1 M T) e2) u) as Hofleft.
  eapply oft_letm; eauto; [].
  eapply ofm_unpack; eauto; done.
assert (oft nil (tm_unpack e1 (tm_letm M (subst (under 2 (sh 2)) e2))) u) as Hofright.
  eapply oft_unpack; eauto; [].
  #3 eapply oft_letm; eauto.
    eapply subst_fsts; eauto; done.

    relquest.
      eapply subst_oft; eauto; [].
      rewrite -> !under_succ; [].
      simp_sub.
      prove_subof; done.

      simp_sub; reflexivity.

    eapply subst_ofc_type; eauto; prove_subof; done.
#3 apply (ceqt_conditional (thalt nil e1)); auto.
  intro Hhalt.
  eapply munpack_halt_inversion'; [].
  eapply letm_halt_inversion'; eauto; done.

  intro Hhalt.
  eapply unpack_halt_inversion'; eauto; done.

  intro Hhalt.
  destruct Hhalt as (v1 & Hsteps1 & Hval1).
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe1 Hsteps1) Consistent.similar_exists Hval1); [].
  intros c v k' t' <-.
  invert Hval1; clear Hval1; [].
  intros Hvalv.
  #2 eapply ceqt_trans.
    apply tsteps_ceqt; auto; [].
    eapply (star_map _#4 (fun z => tm_letm z _)); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    apply star_one; apply mstep_unpack2; auto; done.
  apply ceqt_symm; [].
  apply tsteps_ceqt; auto; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun z => tm_unpack z _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply tstep_unpack2; auto; done.
  simp_sub.
  simp_sub.
  rewrite -> subst_dot01_sh; [].
  apply star_refl; done.
Qed.


Lemma ceqt_commute_ext_unpack :
  forall G e M T t,
    oft G (tm_ext (md_unpack e M T)) t
    -> ceqt G (tm_ext (md_unpack e M T)) (tm_unpack e (tm_ext M)) t.
Proof.
intros G e M T u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (oft_reg _#3 Hwf Hof) as Hofu.
so (ext_oft_inversion _#3 Hwf Hof) as Hof'.
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & HofM & Hsubs & _).
assert (oft G (tm_unpack e (tm_ext M)) u) as Hofr.
  eapply oft_unpack; eauto; [].
  apply oft_ext; [].
  eapply ofm_subsume; eauto; [].
  relquest.
    apply weaken2_subs; eauto; done.

    simp_sub; reflexivity.
apply (ceqt_closed_instances_conditional (fun s => thalt nil (subst s e))); auto; [].
intros s Hs.
do2 2 split.  
  intros Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 (ext_halt_inversion' _ Hhalt)).

  intros Hhalt.
  simp_sub in Hhalt.
  exact (unpack_halt_inversion' _ _ Hhalt).

  intros Hhalt.
  simp_sub.
  destruct Hhalt as (v & Hsteps & Hval).
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k' t' <-.
  invert Hval; [].
  intros Hval'.
  #2 eapply ceqt_trans.
    #2 apply tsteps_ceqt.
      so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H; exact H.
    eapply (star_map _#5); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.

    apply ceqt_symm; [].
    #2 apply tsteps_ceqt.
      so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H; exact H.
    #2 eapply star_trans.
      eapply (star_map' _ _ (fun z => tm_unpack z _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply tstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.
Qed.


Lemma ceqm_commute_appgn_unpack_fstm_eqc :
  forall G p e M1 T1 T2 M2 T k1 c2 c2',
    ofm G p (md_appgn (md_unpack e M1 (sg_pigen T1 T2)) M2) T
    -> fsts T1 k1
    -> fstm G M2 c2
    -> eqc G c2 c2' k1
    -> ceqm G p
       (md_appgn (md_unpack e M1 (sg_pigen T1 T2)) M2) 
       (md_unpack e (md_appgn M1 (subst (sh 2) M2)) (subst1 c2' T2)) T.
Proof.
intros G p e M1 T1 T2 M2 U k1 c2 c2' Hof Hfst1 HfstM2 Heqc.
apply ceqm_with_wf; intro Hwf; [].
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & c2a & Hof' & HofM2 & HfstM2a & HsubsU & ->).
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a; clear HfstM2a.
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & HofM1 & Hsubspi & _).
so (Declarative.subs_pigen_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
invert (subs_reg _#3 Hwf Hsubspi andel); [].
intros k1a _ Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
simp_sub in HofM1.
assert (ofm G impure (md_unpack e (md_appgn M1 (subst (sh 2) M2)) (subst1 c2' T2)) U) as Hofr.
  eapply ofm_subsume; eauto; [].
  #2 eapply ofm_subsume.
    #2 eapply ofm_unpack; eauto.
      #2 eapply ofm_subsume.
        #3 eapply ofm_appgn.
          exact HofM1.
            
          #2 eapply ofm_subsume.
            eapply subst_ofm; eauto; prove_subof; done.

            exact (subst_subs _#5 (weaken2 _#4) Hsubs1).

          eapply subst_fstm; eauto; prove_subof; done.

        simp_sub.
        apply subs_refl; [].
        #2 eapply seq_ofs; eauto.
          so (ofc_exists_invert _#4 (oft_reg _#3 Hwf Hofe)) as (Hofk & Hoft).
          auto using wf; done.
        prove_subrel; [].
        eapply subst_eqc; eauto; prove_subof; done.

      eapply subst1_ofs; eauto; [].
      exact (eqc_reg _#4 Hwf Heqc anderl).

    #2 eapply subs_trans.
      eapply subs_refl; [].
      eapply funct_sg; eauto; [].
      eapply eqc_symm; eauto; done.

      eapply subst_subs; eauto; prove_subof; [].
      exact (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1').
#2 apply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e))); auto.
  intro Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 (appgn_halt_inversion' _ _ Hhalt andel)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 Hhalt).

  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as Hofe'; simp_sub in Hofe'.    
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe' Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k' t' <-.
  invert Hval; [].
  intros Hval'.
  simp_sub.
  #2 eapply ceqm_trans.
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
    eapply (star_map' _ _ (fun z => md_appgn z _)); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    simp_sub.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl.

    apply ceqm_symm; [].
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H.
      exact H.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.
Qed.


Lemma ceqm_commute_appap_unpack_fstm_eqc :
  forall G p e M1 T1 T2 M2 T k1 c2 c2',
    ofm G p (md_appap (md_unpack e M1 (sg_piapp T1 T2)) M2) T
    -> fsts T1 k1
    -> fstm G M2 c2
    -> eqc G c2 c2' k1
    -> ceqm G p
       (md_appap (md_unpack e M1 (sg_piapp T1 T2)) M2) 
       (md_unpack e (md_appap M1 (subst (sh 2) M2)) (subst1 c2' T2)) T.
Proof.
intros G p e M1 T1 T2 M2 U k1 c2 c2' Hof Hfst1 HfstM2 Heqc.
apply ceqm_with_wf; intro Hwf; [].
so (appap_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & c2a & Hof' & HofM2 & HfstM2a & HsubsU).
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a; clear HfstM2a.
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & HofM1 & Hsubspi & ->).
so (Declarative.subs_piapp_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
invert (subs_reg _#3 Hwf Hsubspi andel); [].
intros k1a _ Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
simp_sub in HofM1.
assert (ofm G impure (md_unpack e (md_appap M1 (subst (sh 2) M2)) (subst1 c2' T2)) U) as Hofr.
  eapply ofm_subsume; eauto; [].
  #2 eapply ofm_subsume.
    #2 eapply ofm_unpack; eauto.
      #2 eapply ofm_subsume.
        #3 eapply ofm_appap.
          exact HofM1.
            
          #2 eapply ofm_subsume.
            eapply subst_ofm; eauto; prove_subof; done.

            exact (subst_subs _#5 (weaken2 _#4) Hsubs1).

          eapply subst_fstm; eauto; prove_subof; done.

        simp_sub.
        apply subs_refl; [].
        #2 eapply seq_ofs; eauto.
          so (ofc_exists_invert _#4 (oft_reg _#3 Hwf Hofe)) as (Hofk & Hoft).
          auto using wf; done.
        prove_subrel; [].
        eapply subst_eqc; eauto; prove_subof; done.

      eapply subst1_ofs; eauto; [].
      exact (eqc_reg _#4 Hwf Heqc anderl).

    #2 eapply subs_trans.
      eapply subs_refl; [].
      eapply funct_sg; eauto; [].
      eapply eqc_symm; eauto; done.

      eapply subst_subs; eauto; prove_subof; [].
      exact (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1').
#2 apply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e))); auto.
  intro Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 (appap_halt_inversion' _ _ Hhalt andel)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 Hhalt).

  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as Hofe'; simp_sub in Hofe'.    
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe' Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k' t' <-.
  invert Hval; [].
  intros Hval'.
  simp_sub.
  #2 eapply ceqm_trans.
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
    eapply (star_map' _ _ (fun z => md_appap z _)); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    simp_sub.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl.

    apply ceqm_symm; [].
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H.
      exact H.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.
Qed.


Lemma ceqm_commute_pair_unpack_1 :
  forall G p e M1 M2 T1 T,
    ofm G p (md_pair (md_unpack e M1 T1) M2) T
    -> ceqm G p (md_pair (md_unpack e M1 T1) M2) (md_unpack e (md_pair M1 (subst (sh 2) M2)) T) T.
Proof.
intros G p e M1 M2 T1' U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (pair_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & Hsubs & Hof' & Hof2).
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & Hof1 & Hsubs1 & ->).
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (can_fsts _ _ HofT1') as (k1' & Hfst1').
so (ofm_reg _#4 Hwf Hof2) as HofT2.
assert (ofm G impure (md_unpack e (md_pair M1 (subst (sh 2) M2)) U) U) as Hofr.
  #2 eapply ofm_unpack; eauto.
    #2 eapply ofm_subsume.
      eapply ofm_pair; eauto; [].
      eapply subst_ofm; eauto; prove_subof; done.

      simp_sub.
      #2 eapply subs_trans.
        #5 eapply subs_sigma.
          eapply subst_subs; eauto; prove_subof; done.

          eapply subst_fsts; eauto; prove_subof; done.

          eapply subs_refl; apply eqs_refl; [].
          eapply subst_ofs; eauto; prove_subof; done.

          eapply subst_fsts; eauto; prove_subof; done.

          eapply subst_ofs; eauto; prove_subof; done.

        #2 eapply rel_quest_1.
          eapply subst_subs; eauto; prove_subof; done.

          simp_sub; reflexivity.

    exact (subs_reg _#3 Hwf Hsubs ander).
#2 apply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e))); auto.
  intros Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 (mpair_halt_inversion' _ _ Hhalt andel)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 Hhalt).

  intro Hhalt.
  simp_sub.
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as Hofe'; simp_sub in Hofe'.
  destruct Hhalt as (v & Hsteps & Hval).
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe' Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k' t' <-.
  invert Hval; [].
  intros Hval'.
  #2 eapply ceqm_trans.
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
    apply (star_map' _ _ (fun z => md_pair z _)); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.

    apply ceqm_symm; [].
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H.
      exact H.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.
Qed.


Lemma ceqm_commute_pair_unpack_2 :
  forall G p e M1 M2 T2 T,
    ofm G p (md_pair M1 (md_unpack e M2 T2)) T
    -> ceqm G p (md_pair M1 (md_unpack e M2 T2)) (md_unpack e (md_pair (subst (sh 2) M1) M2) T) T.
Proof.
intros G p e M1 M2 T2' U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (pair_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & Hsubs & Hof1 & Hof').
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & Hof2 & Hsubs2 & ->).
so (ofm_reg _#4 Hwf Hof1) as HofT1.
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (subs_reg _#3 Hwf Hsubs2) as (HofT2' & HofT2).
assert (ofm G impure (md_unpack e (md_pair (subst (sh 2) M1) M2) U) U) as Hofr.
  #2 eapply ofm_unpack; eauto.
    #2 eapply ofm_subsume.
      eapply ofm_pair; eauto; [].
      eapply subst_ofm; eauto; prove_subof; done.

      simp_sub.
      #2 eapply subs_trans.
        #5 eapply subs_sigma.
          eapply subs_refl; apply eqs_refl; [].
          eapply subst_ofs; eauto; prove_subof; done.

          eapply subst_fsts; eauto; prove_subof; done.

          eapply subst_subs; eauto; prove_subof; done.

          eapply subst_fsts; eauto; prove_subof; done.

          eapply subst_ofs; eauto; prove_subof; done.

        #2 eapply rel_quest_1.
          eapply subst_subs; eauto; prove_subof; done.

          simp_sub; reflexivity.

    exact (subs_reg _#3 Hwf Hsubs ander).
#2 apply (ceqm_closed_instances_conditional (fun s => mhalt nil (subst s M1) /\ thalt nil (subst s e))); auto.
  intros Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  split.
    exact (mpair_halt_inversion' _ _ Hhalt andel).

    exact (munpack_halt_inversion' _#3 (mpair_halt_inversion' _ _ Hhalt ander)).

  intro Hhalt.
  simp_sub in Hhalt.
  destruct Hhalt as (V & Hsteps & HvalV).
  so (star_expon _#4 Hsteps) as (i & Hsteps').
  so (munpack_halt_inversion _#5 Hsteps' HvalV) as (i1 & i2 & c & v & k' & t' & _ & _ & Hvalv & Hsteps1 & Hsteps2).
  split.
    simp_sub in Hsteps2.
    refine (mpair_halt_inversion' _ _ _ andel); [].
    eexists; eauto using expon_star; done.

    eexists; split.
      eapply expon_star; eauto; done.

      apply tvalue_pack; auto; done.

  intros (HhaltM1, Hhalt).
  simp_sub.
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) as Hofe'; simp_sub in Hofe'.
  destruct Hhalt as (v & Hsteps & Hval).
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe' Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k' t' <-.
  invert Hval; [].
  intros Hval'.
  destruct HhaltM1 as (V & HstepsM1 & HvalV).
  #2 eapply ceqm_trans.
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => md_pair z _)); eauto with dynamic; done.
    apply (star_map' _ _ (fun z => md_pair _ z)); eauto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.

    apply ceqm_symm; [].
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H.
      exact H.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply mstep_unpack2; auto; done.
    simp_sub.
    apply (star_map' _ _ (fun z => md_pair z _)); eauto with dynamic; done.
Qed.


Lemma ceqm_commute_pair_unpack_both :
  forall G p e1 M1 e2 M2 T1 T2 U,
    ofm G p (md_pair (md_unpack e1 M1 T1) (md_unpack e2 M2 T2)) U
    -> ceqm G p 
       (md_pair (md_unpack e1 M1 T1) (md_unpack e2 M2 T2)) 
       (md_unpack e1 
          (md_unpack (subst (sh 2) e2) 
             (md_pair
                (subst (dot (var 2) (dot (var 3) (sh 4))) M1)
                (subst (dot (var 0) (dot (var 1) (sh 4))) M2))
             (subst (sh 2) U))
          U)
       U.
Proof.
intros G p e1 M1 e2 M2 T1 T2 U Hof.
#2 eapply ceqm_trans_formation.
  eapply ceqm_commute_pair_unpack_1; eauto; done.
intro Hof'.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof') as (k1 & t1 & Hofe1 & Hof'' & Hsubs & ->).
apply ceqm_comp; eapply mcc_subsume; eauto; [].
#3 apply ceqm_comp; eapply mcc_unpack.
  apply ceqt_refl; eauto; done.

  simp_sub.
  #2 eapply rel_quest_1.
    eapply ceqm_commute_pair_unpack_2; eauto; done.

    simp_sub.
    rewrite <- (subst_dot01_sh M1) at 1; [].
    simp_sub.
    reflexivity.

  apply eqs_refl; [].
  exact (subs_reg _#3 Hwf Hsubs andel).
Qed.



Lemma ceqm_commute_unpack_unpack :
  forall G p e1 e2 M T U,
    ofm G p (md_unpack (tm_unpack e1 e2) M T) U
    -> ceqm G p (md_unpack (tm_unpack e1 e2) M T) (md_unpack e1 (md_unpack e2 (subst (dot (var 0) (dot (var 1) (sh 4))) M) (subst (sh 2) T)) T) U.
Proof.
intros G p e1 e2 M T U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k & t & Hof' & HofM & Hsubs & ->).
so (unpack_oft_inversion _#4 Hwf Hof') as (k' & t' & Hofe1 & Hofe2).
simp_sub in Hofe2.
assert (ofm G impure (md_unpack e1 (md_unpack e2 (subst (dot (var 0) (dot (var 1) (sh 4))) M) (subst (sh 2) T)) T) U) as Hofr.
  so (subs_reg _#3 Hwf Hsubs) as (HofT & _).
  eapply ofm_subsume; eauto; [].
  eapply ofm_unpack; eauto; [].
  #2 eapply ofm_unpack; eauto.
    relquest.
      eapply subst_ofm; eauto; prove_subof; done.

      simp_sub; reflexivity.

    apply weaken2_ofs; auto; done.
#2 apply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e1))); eauto.
  intros Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intros Hhalt.
  simp_sub in Hhalt.
  exact (unpack_halt_inversion' _ _ (munpack_halt_inversion' _#3 Hhalt)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 Hhalt).

  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe1) as Hofe1'; simp_sub in Hofe1'.
  invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil Hofe1' Hsteps) Consistent.similar_exists Hval); [].
  intros c v' k'' t'' <-.
  invert Hval; [].
  intros Hval'.
  simp_sub.
  #2 eapply ceqm_trans.
    #2 apply msteps_ceqm.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
    eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; [].
    #2 eapply star_trans.
      apply (star_map' _ _ (fun z => tm_unpack z _)); eauto with dynamic; done.
    #2 eapply star_step.
      apply tstep_unpack2; auto; done.
    simp_sub.
    apply star_refl; done.
  apply ceqm_symm; [].
  #2 eapply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H.
    exact H.
  #2 eapply star_trans.
    eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply mstep_unpack2; auto; done.
  simp_sub.
  apply star_refl; done.
Qed.


Lemma ceqm_commute_unpack_lett :
  forall G p e1 e2 M T,
    ofm G p (md_unpack (tm_lett e1 e2) M T) T
    -> ceqm G p (md_unpack (tm_lett e1 e2) M T) (md_lett e1 (md_unpack e2 (subst (dot (var 0) (dot (var 1) (sh 3))) M) (shift1 T))) T.
Proof.
intros G p e1 e2 M T Hof.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k & t & Hof' & HofM & _ & ->).
so (lett_oft_inversion _#4 Hwf Hof') as (u & Hofe1 & Hofe2).
so (ofm_reg _#4 Hwf Hof) as HofT.
simp_sub in Hofe2.
assert (ofm G impure (md_lett e1 (md_unpack e2 (subst (dot (var 0) (dot (var 1) (sh 3))) M) (shift1 T))) T) as Hofr.
  eapply ofm_lett; eauto; [].
  #2 eapply ofm_unpack; eauto.
    relquest.
      eapply subst_ofm; eauto; prove_subof; done.

      simp_sub; reflexivity.

    apply weaken1_ofs; auto; done.
#2 eapply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e1))); eauto.
  intros Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  exact (lett_halt_inversion' _ _ (munpack_halt_inversion' _#3 Hhalt)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (mlett_halt_inversion' _ _ Hhalt).
intro Hhalt.
simp_sub.
destruct Hhalt as (v & Hsteps & Hval).
#2 eapply ceqm_trans.
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H; exact H.
  eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; [].
  #2 eapply star_trans.
    eapply (star_map' _ _ (fun z => tm_lett z _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply tstep_lett2; auto; done.
  simp_sub.
  apply star_refl; done.

  apply ceqm_symm; [].
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H; exact H.
  #2 eapply star_trans.
    eapply (star_map _#4 (fun z => md_lett z _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply mstep_lett2; auto; done.
  simp_sub.
  apply star_refl; done.
Qed.


Lemma ceqm_commute_unpack_letm :
  forall G T1 T2 k t k1 M1 e M2,
    ofm G impure M1 T1
    -> fsts T1 k1
    -> oft (G; cl_cn k1; cl_md (shift1 T1) (var 0)) e (subst (sh 2) (cn_exists k t))
    -> ofm (G; cl_cn k; cl_tm t) impure M2 (subst (sh 2) T2)
    -> ceqm G impure (md_unpack (tm_letm M1 e) M2 T2) (md_letm M1 (md_unpack e (subst (under 2 (sh 2)) M2) (subst (sh 2) T2)) T2) T2.
Proof.
cut (forall T2 M1 e M2,
       ofm nil impure (md_unpack (tm_letm M1 e) M2 T2) T2
       -> ofm nil impure (md_letm M1 (md_unpack e (subst (under 2 (sh 2)) M2) (subst (sh 2) T2)) T2) T2
       -> ceqm nil impure (md_unpack (tm_letm M1 e) M2 T2) (md_letm M1 (md_unpack e (subst (under 2 (sh 2)) M2) (subst (sh 2) T2)) T2) T2).
  intro Hprop.
  intros G T1 T2 k t k1 M1 e M2 HofM1 Hfst1 Hofe HofM2.
  simp_sub in Hofe.
  apply ceqm_with_wf; intro Hwf; [].
  so (ofm_reg _#4 Hwf HofM1) as HofT1.
  so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
  so (oft_reg _#3 (wf_md Hwf HofT1 Hfst1) Hofe) as Hofexpre.
  #2 sor (refine (subst_ofc_type G _ (dot nonsense (dot (Declarative.inh k1) id)) _ _ Hofexpre)) as Hofex.
    prove_subof; [].
    apply Declarative.inh_ofc; auto; done.
  simp_sub in Hofex.
  clear Hofexpre.
  so (ofc_ofcr _#3 Hofex) as (l & Hofex').
  invert Hofex'; [].
  intros Hofk Hoft _.
  clear l Hofex'.
  so (ofm_reg _#4 (wf_tm (wf_cn Hwf Hofk) Hoft) HofM2) as HofT2pre.
  #2 sor (refine (subst_ofs G _ (dot nonsense (dot (Declarative.inh k) id)) _ _ HofT2pre)) as HofT2.
    prove_subof; [].
    apply Declarative.inh_ofc; auto; done.
  simp_sub in HofT2.
  clear HofT2pre.
  assert (ofm G impure (md_unpack (tm_letm M1 e) M2 T2) T2) as Hof1.
    eapply ofm_unpack; eauto with static; done.
  assert (ofm G impure (md_letm M1 (md_unpack e (subst (under 2 (sh 2)) M2) (subst (sh 2) T2)) T2) T2) as Hof2.
    eapply ofm_letm; eauto; [].
    #2 eapply ofm_unpack; eauto.
      relquest.
        eapply subst_ofm; eauto; [].
        rewrite -> !under_succ; [].
        simp_sub.
        prove_subof; done.
  
        simp_sub; reflexivity.
  
      apply weaken2_ofs; auto; done.
  apply closed_instances_ceqm_impure; eauto; [].
  intros s Hs.
  simp_sub.
  #3 exploit (Hprop (subst s T2) (subst s M1) (subst (under 2 s) e) (subst (under 2 s) M2)) as H.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof1) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof2) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    simp_sub in H.
    simp_sub.
    exact H.
intros T2 M1 e M2 Hof1 Hof2.
so (letm_ofm_inversion _#6 wf_nil Hof2) as (T1 & _ & HofM1 & _).
#3 apply (ceqm_conditional_impure (mhalt nil M1)); auto.
  intro Hhalt.
  so (munpack_halt_inversion' _#3 Hhalt) as Hhalt'.
  so (letm_halt_inversion' _ _ Hhalt') as Hhalt''.
  assumption.

  intro Hhalt.
  so (mletm_halt_inversion' _#3 Hhalt) as Hhalt'.
  assumption.
intro HhaltM1.
destruct HhaltM1 as (V1 & HstepsM1 & HvalV1).
so (can_fstm_value _#4 wf_nil (preservations_md _#5 wf_nil HofM1 HstepsM1) HvalV1) as (c1 & HfstV1).
#2 eapply ceqm_trans_formation.
  apply msteps_ceqm; auto; [].
  #2 eapply star_trans.
    apply (star_map' _ _ (fun x => md_unpack (tm_letm x _) _ _)); eauto with dynamic; done.

    apply star_one.
    eapply mstep_unpack1; [].
    eapply tstep_letm2; eauto; done.
intro Hof1'; clear Hof1.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  apply msteps_ceqm; auto; [].
  #2 eapply star_trans.
    apply (star_map' _ _ (fun x => md_letm x _ _)); eauto with dynamic; done.

    apply star_one.
    eapply mstep_letm2; eauto; done.
apply ceqm_symm; [].
simp_sub.
simp_sub.
rewrite -> (subst_dot01_sh M2); [].
apply ceqm_refl; auto; done.
Qed.


Lemma ceqm_commute_letm_unpack :
  forall G p e M1 M2 T1 T2,
    ofm G p (md_letm (md_unpack e M1 T1) M2 T2) T2
    -> ceqm G p (md_letm (md_unpack e M1 T1) M2 T2) (md_unpack e (md_letm M1 (subst (dot (var 0) (dot (var 1) (sh 4))) M2) (subst (sh 2) T2)) T2) T2.
Proof.
intros G p e M1 M2 T1 T2 Hof.
apply ceqm_with_wf; intro Hwf; [].
so (letm_ofm_inversion _#6 Hwf Hof) as (T1' & k1' & Hof' & Hfst1' & HofM2 & _ & ->).
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe & HofM1 & Hsubs & _).
so (ofm_reg _#4 Hwf Hof) as HofT2.
so (subs_reg _#3 Hwf Hsubs) as (HofT1 & _).
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (subs_fst _#5 Hwf Hsubs Hfst1 Hfst1') as Hsubk.
assert (ofm G impure (md_unpack e (md_letm M1 (subst (dot (var 0) (dot (var 1) (sh 4))) M2) (subst (sh 2) T2)) T2) T2) as Hofr.
  eapply ofm_unpack; eauto; [].
  eapply ofm_letm; eauto using subst_fsts, weaken2_ofs; [].
  simp_sub.
  relquest.
    #2 eapply subst_ofm; eauto; prove_subof.
      #2 eapply ofc_subsume.
        ofvar; done.

        eapply subst_subk; eauto; prove_subof; done.

      #2 eapply ofm_subsume.
        ofvar; done.

        eapply subst_subs; eauto; prove_subof; done.

    simp_sub; reflexivity.
#2 eapply (ceqm_closed_instances_conditional (fun s => thalt nil (subst s e))); eauto.
  intros Heq.
  discriminate Heq.
intros s Hs.
do2 2 split.
  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 (mletm_halt_inversion' _#3 Hhalt)).

  intro Hhalt.
  simp_sub in Hhalt.
  exact (munpack_halt_inversion' _#3 Hhalt).
intro Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
invert (Progress.canon_tm _ _ (preservations_tm _#4 wf_nil (subst_oft _#5 (subof_aof_dof _#3 Hs) Hofe) Hsteps) Consistent.similar_exists Hval); [].
intros c v' k' t' <-.
invert Hval; [].
intros Hval'.
simp_sub.
#2 eapply ceqm_trans.
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H; exact H.
  eapply (star_map _#4 (fun z => md_letm z _ _)); eauto with dynamic; [].
  #2 eapply star_trans.
    eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply mstep_unpack2; auto; done.
  simp_sub.
  apply star_refl; done.

  apply ceqm_symm; [].
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hofr) as H; simp_sub in H; exact H.
  #2 eapply star_trans.
    eapply (star_map _#4 (fun z => md_unpack z _ _)); eauto with dynamic; done.
  #2 eapply star_step.
    apply mstep_unpack2; auto; done.
  simp_sub.
  apply star_refl; done.
Qed.


(* Eta principles *)

Lemma ceqt_prod_eta :
  forall t1 t2 e,
    oft nil e (cn_prod t1 t2)
    -> ceqt nil e (tm_pair (tm_pi1 e) (tm_pi2 e)) (cn_prod t1 t2).
Proof.
intros t1 t2 e Hofe.
#2 apply (ceqt_conditional (thalt nil e)); eauto with static.
  intro Hhalt.
  exact (pi1_halt_inversion' _ (pair_halt_inversion' _ _ Hhalt andel)).
intro Hhalt.
#2 apply ceqt_prod_ext; eauto with static.
  apply ceqt_symm; [].
  apply ceqt_pi1_beta; eauto using thalt_pi1, thalt_pi2 with static; done.

  apply ceqt_symm; [].
  apply ceqt_pi2_beta; eauto using thalt_pi1, thalt_pi2 with static; done.
Qed.


Lemma ceqm_sigma_eta :
  forall T1 T2 M,
    ofm nil pure M (sg_sigma T1 T2)
    -> ceqm nil pure M (md_pair (md_pi1 M) (md_pi2 M)) (sg_sigma T1 T2).
Proof.
intros T1 T2 M HofM.
invert (ofm_reg _#4 wf_nil HofM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fstm _#3 wf_nil HofM) as (c & HfstM).
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (ofm_fst _#6 wf_nil HofM HfstM (fsts_sigma Hfst1 Hfst2)) as Hofc.
assert (ofm nil pure (md_pair (md_pi1 M) (md_pi2 M)) (sg_sigma T1 T2)) as Hofeta.
  eapply Singleton.ofm_pair_dep; eauto with static; done.
#5 eapply (ceqm_conditional_pure (mhalt nil M)); eauto.
  eapply fstm_pair; eauto using fstm; done.

  eapply fsts_sigma; eauto; done.

  apply Sound.eqc_sigmaeta; eauto using ofs_fst; done.

  intro Hhalt.
  so (mpair_halt_inversion' _ _ Hhalt) as (Hhalt' & _).
  exact (mpi1_halt_inversion' _ Hhalt').
intro Hhalt.
#2 eapply ceqm_sigma_ext; eauto.
  apply ceqm_symm; [].
  #2 eapply ceqm_pi1_beta; eauto with static.
    eapply mhalt_pi1; eauto using wf; done.

    eapply mhalt_pi2; eauto using wf; done.

  apply ceqm_symm; [].
  eapply ceqm_pi2_beta; eauto with static.
    eapply mhalt_pi1; eauto using wf; done.

    eapply mhalt_pi2; eauto using wf; done.
Qed.


Lemma ceqm_letm_eta :
  forall G M T,
    ofm G impure M T
    -> ceqm G impure (md_letm M (var 0) T) M T.
Proof.
cut (forall M T,
       ofm nil impure M T
       -> ceqm nil impure (md_letm M (var 0) T) M T).
  intro Hprop.
  intros G M T HofM.
  apply ceqm_with_wf; intro Hwf; [].
  so (ofm_reg _#4 Hwf HofM) as HofT.
  so (can_fsts _ _ HofT) as (k & Hfst).
  #2 apply closed_instances_ceqm_impure; auto.
    eapply ofm_letm; eauto; [].
    apply ofm_forget; [].
    ofvar; done.
  
    intros s Hs.
    simp_sub.
    apply Hprop; [].
    eapply subst_ofm; eauto using subof_aof_dof; done.
intros M T HofM.
so (ofm_reg _#4 wf_nil HofM) as HofT.
so (can_fsts _ _ HofT) as (k & Hfst).
assert (ofm nil impure (md_letm M (var 0) T) T) as Hof'.
  eapply ofm_letm; eauto; [].
  apply ofm_forget; [].
  ofvar; done.
#2 apply (ceqm_conditional_impure (mhalt nil M)); auto.
  intro Hhalt.
  eapply mletm_halt_inversion'; eauto; done.

  intro Hhalt.
  destruct Hhalt as (V & Hsteps & Hval).
  so (preservations_md _#5 wf_nil HofM Hsteps) as HofV.
  so (can_fstm_value _#4 wf_nil HofV Hval) as (c & HfstV).
  #2 eapply ceqm_trans.
    apply msteps_ceqm; eauto; [].
    #2 eapply star_trans.
      apply (star_map' _ _ (fun x => md_letm x _ _)); eauto with dynamic; done.

      apply star_one; [].
      eapply mstep_letm2; eauto; done.

    simp_sub.
    apply ceqm_symm; [].
    apply msteps_ceqm; eauto; done.
Qed.
