
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Functionality.
Require Import Regularity.
Require Import Congruence.


Ltac ofvar :=
  eapply rel_quest;
  [(match goal with
    | |- ofc _ _ _ => eapply ofc_var; auto with sequence; done
    | |- oft _ _ _ => eapply oft_var; auto with sequence; done
    | |- ofm _ _ _ _ => eapply ofm_var; auto with sequence; done
    | |- fstm _ _ _ => eapply fstm_var; auto with sequence; done
    | |- tvalue _ _ => eapply tvalue_var; auto with sequence; done
    | |- mvalue _ _ => eapply mvalue_var; auto with sequence; done
    end) | simp_sub; reflexivity].


Ltac open_class :=
  (match goal with
   | |- sof _ _ (cl_cn _) => apply sof_cn
   | |- sof _ _ (cl_tm _) => apply sof_tm
   | |- sof _ _ (cl_md _ _) => apply sof_md
   | |- sof _ _ cl_ucn => apply sof_ucn
   | |- sof _ _ cl_utm => apply sof_utm
   | |- sof _ _ (cl_umd _) => apply sof_umd
   | |- sof _ _ cl_ns => apply sof_ns
   | |- dof _ _ (cl_cn _) => apply dof_cn
   | |- dof _ _ (cl_tm _) => apply dof_tm
   | |- dof _ _ (cl_md _ _) => apply dof_md
   | |- dof _ _ cl_ucn => apply dof_ucn
   | |- dof _ _ cl_utm => apply dof_utm
   | |- dof _ _ (cl_umd _) => apply dof_umd
   | |- dof _ _ cl_ns => apply dof_ns
   | |- tof _ _ (cl_cn _) => apply tof_cn
   | |- tof _ _ (cl_tm _) => apply tof_tm
   | |- tof _ _ (cl_md _ _) => apply tof_md
   | |- tof _ _ cl_ucn => apply tof_ucn
   | |- tof _ _ cl_utm => apply tof_utm
   | |- tof _ _ cl_ucn => apply tof_ucn
   | |- tof _ _ cl_utm => apply tof_utm
   | |- tof _ _ (cl_umd _) => apply tof_umd
   | |- tof _ _ (cl_umd _) => apply tof_umd
   | |- tof _ _ cl_ns => apply tof_ns
   | |- oof _ _ (cl_cn _) => apply oof_cn
   | |- oof _ _ (cl_tm _) => apply oof_tm
   | |- oof _ _ (cl_md _ _) => apply oof_md
   | |- oof _ _ cl_ucn => apply oof_ucn
   | |- oof _ _ cl_utm => apply oof_utm
   | |- oof _ _ (cl_umd _) => apply oof_umd
   | |- oof _ _ cl_ns => apply oof_ns
   | |- aof _ _ (cl_cn _) => apply aof_cn
   | |- aof _ _ (cl_tm _) => apply aof_tm
   | |- aof _ _ (cl_md _ _) => apply aof_md
   | |- aof _ _ cl_ucn => apply aof_ucn
   | |- aof _ _ cl_utm => apply aof_utm
   | |- aof _ _ (cl_umd _) => apply aof_umd
   | |- aof _ _ cl_ns => apply aof_ns
   | |- seqs _ _ _ (cl_cn _) => apply seqs_cn
   | |- seqs _ _ _ (cl_tm _) => apply seqs_tm
   | |- seqs _ _ _ (cl_md _ _) => apply seqs_md
   | |- seqs _ _ _ cl_ucn => apply seqs_ucn
   | |- seqs _ _ _ cl_utm => apply seqs_utm
   | |- seqs _ _ _ (cl_umd _) => apply seq_umd
   | |- seqs _ _ _ cl_ns => apply seqs_ns
   | |- seq _ _ _ (cl_cn _) => apply seq_cn
   | |- seq _ _ _ (cl_tm _) => apply seq_tm
   | |- seq _ _ _ (cl_md _ _) => apply seq_md
   | |- seq _ _ _ cl_ucn => apply seq_ucn
   | |- seq _ _ _ cl_utm => apply seq_utm
   | |- seq _ _ _ (cl_umd _) => apply seq_umd
   | |- seq _ _ _ cl_ns => apply seq_ns
   | |- sid _ _ _ (cl_cn _) => apply sid_cn
   | |- sid _ _ _ (cl_tm _) => apply sid_tm
   | |- sid _ _ _ (cl_md _ _) => apply sid_md
   | |- sid _ _ _ cl_ucn => apply sid_ucn
   | |- sid _ _ _ cl_utm => apply sid_utm
   | |- sid _ _ _ (cl_umd _) => apply sid_umd
   | |- sid _ _ _ cl_ns => apply sid_ns
   | |- deq _ _ _ _ _ (cl_cn _) => apply deq_cn
   | |- deq _ _ _ _ _ (cl_tm _) => apply deq_tm
   | |- deq _ _ _ _ _ (cl_md _ _) => apply deq_md
   | |- deq _ _ _ _ _ cl_ucn => apply deq_ucn
   | |- deq _ _ _ _ _ cl_utm => apply deq_utm
   | |- deq _ _ _ _ _ (cl_umd _) => apply deq_umd
   | |- deq _ _ _ _ _ cl_ns => apply deq_ns
   end).


Ltac prove_subof :=
  unfold sh1, id;
  (match goal with
   | |- subof _ _ (dot _ _) _ =>
       apply subof_dot;
       [ prove_subof
       | simp_sub; open_class; try ofvar
       ]

   | |- subof _ _ (sh _) _ =>
       apply subof_sh; repeat (apply truncate_S); apply truncate_0; done

   | |- _ => idtac
   end).


Ltac prove_subrel :=
  unfold sh1, id;
  (match goal with
   | |- subrel _ _ (dot _ _) (dot _ _) _ =>
       apply subrel_dot;
       [ prove_subrel
       | simp_sub; open_class; try ofvar
       ]

   | |- subrel _ _ (sh _) (sh _) _ =>
       apply subrel_sh; repeat (apply truncate_S); apply truncate_0; done

   | |- _ => idtac
   end).
