
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import ProveSubof.
Require Import Inversion.
Require Import Regularity.
Require Import Preservation.
Require Import Congruence.
Require Import Contextual.
Require Import Equivalence.


Definition reqt G e e' :=
  forall t,
    oft G e t
    -> ceqt G e e' t.


Definition reqm G M M' :=
  forall p T,
    ofm G p M T
    -> ceqm G p M M' T.


Lemma reqt_with_wf :
  forall G e e',
    (wf G -> reqt G e e')
    -> reqt G e e'.
Proof.
intros G e e' Hr.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
exact (Hr Hwf t Hof).
Qed.


Lemma reqt_ceqt :
  forall G e e' t,
    oft G e t
    -> reqt G e e'
    -> ceqt G e e' t.
Proof.
intros; auto.
Qed.


Lemma reqm_ceqm :
  forall G p M M' T,
    ofm G p M T
    -> reqm G M M'
    -> ceqm G p M M' T.
Proof.
intros; auto.
Qed.


Lemma reqt_refl :
  forall G e, reqt G e e.
Proof.
intros G e.
intros t Hof.
apply ceqt_refl; assumption.
Qed.


Lemma reqm_refl :
  forall G M, reqm G M M.
Proof.
intros G M.
intros p T Hof.
apply ceqm_refl; assumption.
Qed.


Lemma reqt_trans :
  forall G e1 e2 e3,
    reqt G e1 e2
    -> reqt G e2 e3
    -> reqt G e1 e3.
Proof.
intros G e1 e2 e3 H12 H23.
intros t Hof1.
apply ceqt_with_wf; intro Hwf; [].
so (H12 t Hof1) as Hc12.
so (H23 t (ceqt_reg _#4 Hwf Hc12 ander)) as Hc23.
eapply ceqt_trans; eauto; done.
Qed.


Lemma reqm_trans :
  forall G M1 M2 M3,
    reqm G M1 M2
    -> reqm G M2 M3
    -> reqm G M1 M3.
Proof.
intros G M1 M2 M3 H12 H23.
intros p T Hof1.
apply ceqm_with_wf; intro Hwf; [].
so (H12 p T Hof1) as Hc12.
so (H23 p T (ceqm_reg _#5 Hwf Hc12 ander)) as Hc23.
eapply ceqm_trans; eauto; done.
Qed.


Lemma tsteps_reqt :
  forall G e e',
    star (tstep G) e e'
    -> reqt G e e'.
Proof.
intros G e e' Hsteps.
intros t Hof.
eapply tsteps_ceqt; eauto.
Qed.


Lemma msteps_reqm :
  forall G M M',
    star (mstep G) M M'
    -> reqm G M M'.
Proof.
intros G M M' Hsteps.
intros p T Hof.
eapply msteps_ceqm; eauto.
Qed.


Lemma reqt_lam :
  forall G t e e',
    reqt (G; cl_tm t) e e'
    -> reqt G (tm_lam t e) (tm_lam t e').
Proof.
intros G t1 e e' Hr.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (lam_oft_inversion _#4 Hwf Hof) as (t2 & Heqc & Hofe).
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc anderl)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 _ _; clear l Hofcr.
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_comp; eapply tcc_lam; auto using eqc_refl; done.
Qed.


Lemma reqt_app :
  forall G e1 e1' e2 e2',
    reqt G e1 e1'
    -> reqt G e2 e2'
    -> reqt G (tm_app e1 e2) (tm_app e1' e2').
Proof.
intros G e1 e1' e2 e2' Hr1 Hr2.
intros t2 Hof.
apply ceqt_with_wf; intro Hwf; [].
so (app_oft_inversion _#4 Hwf Hof) as (t1 & Hof1 & Hof2).
apply ceqt_comp; [].
#2 eapply tcc_app.
  eapply Hr1; eauto; done.

  eapply Hr2; eauto; done.
Qed.


Lemma reqt_app_beta :
  forall G t e1 e2,
    thalt G e2
    -> reqt G (tm_app (tm_lam t e1) e2) (subst1 e2 e1).
Proof.
intros G t e1 e2 Hhalt.
intros u Hof.
eapply ceqt_app_beta; eauto.
Qed.


Lemma reqt_app_beta_instances_conditional :
  forall G t e1 e2,
    (forall s,
       subof aof nil s G 
       -> thalt nil (subst (dot (subst s e2) s) e1)
       -> thalt nil (subst s e2))
    -> reqt G (tm_app (tm_lam t e1) e2) (subst1 e2 e1).
Proof.
intros G t1 e1 e2 Hhaltish.
intros t2 Hof.
apply ceqt_with_wf; intro Hwf; [].
so (app_oft_inversion _#4 Hwf Hof) as (t1' & Hoflam & Hofe2).
so (lam_oft_inversion _#4 Hwf Hoflam) as (t2' & Heqc & Hofe1).
so (Consistent.arrow_injective _#5 Hwf Heqc) as (Heqt1 & Heqt2).
so (eqc_reg _#4 Hwf Heqt1) as (_ & Hoft1 & _).
#3 apply closed_instances_ceqt.
  assumption.

  replsub t2 with (subst1 e2 (shift1 t2)).
  eapply subst1_oft_tm; eauto; [].
  eapply oft_conv_0t; eauto using eqc_symm; [].
  eapply oft_equiv; eauto; [].
  apply eqc_symm; [].
  apply weaken1_eqc_type; auto; done.
intros s Hs.
simp_sub.
#5 apply (ceqt_conditional (thalt nil (subst s e2))).
  #2 eapply oft_app.
    #2 apply oft_lam.
      eapply subst_ofc_type; eauto using subof_aof_sof; done.

      simp_sub.
      #2 eapply oft_equiv.
        eapply subst_oft; eauto; [].
        prove_subof; [].
        eapply subof_dof_trans; eauto using subof_aof_dof; prove_subof; done.

        simp_sub.
        apply eqc_symm; [].
        eapply subst_eqc_type; eauto; prove_subof; [].
        eapply subof_sof_trans; eauto using subof_aof_sof; prove_subof; done.

    #2 eapply oft_equiv.
      eapply subst_oft; eauto using subof_aof_dof; done.

      eapply subst_eqc_type; eauto using subof_aof_sof; done.

  #2 eapply oft_equiv.
    eapply subst_oft; eauto; prove_subof; eauto using subof_aof_dof; [].
    #2 eapply oft_equiv.
      eapply subst_oft; eauto using subof_aof_dof; done.

      eapply subst_eqc_type; eauto using subof_aof_sof; done.

    simp_sub.
    apply eqc_symm; [].
    eapply subst_eqc_type; eauto using subof_aof_sof; done.

  intro Hhalt.
  eapply app_halt_inversion'; eauto; done.

  apply Hhaltish; auto; done.
intro Hhalt.
#2 eapply rel_quest_1.
  apply ceqt_app_beta; auto; [].
  so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H.
  simp_sub in H.
  exact H.

  simp_sub; reflexivity.
Qed.


Lemma reqt_app_beta_instances :
  forall G t e1 e2,
    (forall s, subof aof nil s G -> thalt nil (subst s e2))
    -> reqt G (tm_app (tm_lam t e1) e2) (subst1 e2 e1).
Proof.
intros G t e1 e2 Hhaltish.
apply reqt_app_beta_instances_conditional; auto.
Qed.


Lemma reqt_pair :
  forall G e1 e1' e2 e2',
    reqt G e1 e1'
    -> reqt G e2 e2'
    -> reqt G (tm_pair e1 e2) (tm_pair e1' e2').
Proof.
intros G e1 e1' e2 e2' Hr1 Hr2.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
so (pair_oft_inversion _#4 Hwf Hof) as (t1 & t2 & Heqt & Hof1 & Hof2).
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_comp; apply tcc_pair; eauto; done.
Qed.


(* We could dispense with the e1 halts assumption. *)
Lemma reqt_pi1_beta :
  forall G e1 e2,
    thalt G e1
    -> thalt G e2
    -> reqt G (tm_pi1 (tm_pair e1 e2)) e1.
Proof.
intros G e1 e2 Hhalt1 Hhalt2.
intros t Hof.
eapply ceqt_pi1_beta; eauto; done.
Qed.


Lemma reqt_pi2_beta :
  forall G e1 e2,
    thalt G e1
    -> thalt G e2
    -> reqt G (tm_pi2 (tm_pair e1 e2)) e2.
Proof.
intros G e1 e2 Hhalt1 Hhalt2.
intros t Hof.
eapply ceqt_pi2_beta; eauto; done.
Qed.


Lemma reqt_plam :
  forall G k e e',
    reqt (G; cl_cn k) e e'
    -> svalue e'
    -> reqt G (tm_plam k e) (tm_plam k e').
Proof.
intros G k e e' Hr Hvale'.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
so (plam_oft_inversion _#4 Hwf Hof) as (t' & Heqc & Hofe & Hvale).
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc anderl)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk _ _; clear l Hofcr.
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_plam; auto using eqk_refl; done.
Qed.


Lemma reqt_papp :
  forall G e e' c,
    reqt G e e'
    -> reqt G (tm_papp e c) (tm_papp e' c).
Proof.
intros G e e' c Hr.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (papp_oft_inversion _#4 Hwf Hof) as (k & t & Hofe & Hofc & Hequ).
so (Hr _ Hofe) as Hceq.
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_comp; eapply tcc_papp; eauto with static; done.
Qed.


Lemma reqt_papp_beta :
  forall G k e c,
    reqt G (tm_papp (tm_plam k e) c) (subst1 c e).
Proof.
intros G k e c.
apply tsteps_reqt; [].
apply star_one; [].
apply tstep_papp2; done.
Qed.


Lemma reqt_pack :
  forall G c e e' k t,
    reqt G e e'
    -> reqt G (tm_pack c e k t) (tm_pack c e' k t).
Proof.
intros G c e e' k t Hr.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (pack_oft_inversion _#6 Hwf Hof) as (Hequ & Hofc & Hofe).
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Hequ anderl)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_comp; apply tcc_pack; eauto with static; done.
Qed.


Lemma reqt_pack_annot :
  forall G c c' e k k' t t',
    eqc G c c' k
    -> eqk G k k'
    -> eqc (G; cl_cn k) t t' kd_type
    -> reqt G (tm_pack c e k t) (tm_pack c' e k' t').
Proof.
intros G c c' e k k' t t' Heqc Heqk Heqt.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (pack_oft_inversion _#6 Hwf Hof) as (Hequ & _ & Hofe).
apply ceqt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
apply ceqt_comp; apply tcc_pack; auto; [].
apply ceqt_refl; auto; done.
Qed.


Lemma reqt_unpack :
  forall G e1 e1' e2,
    reqt G e1 e1'
    -> reqt G (tm_unpack e1 e2) (tm_unpack e1' e2).
Proof.
intros G e1 e1' e2 Hr.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (unpack_oft_inversion _#4 Hwf Hof) as (k & t & Hof1 & Hof2).
#3 apply ceqt_comp; eapply tcc_unpack.
  apply Hr; eauto; done.

  apply ceqt_refl; auto; done.

  exact (oft_reg _#3 Hwf Hof).
Qed.


Lemma reqt_unpack_bind :
  forall G e1 e1' e2 e2' k t,
    ceqt G e1 e1' (cn_exists k t)
    -> reqt (G; cl_cn k; cl_tm t) e2 e2'
    -> reqt G (tm_unpack e1 e2) (tm_unpack e1' e2').
Proof.
intros G e1 e1' e2 e2' k t Hceq1 Hreq2.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (unpack_oft_inversion _#4 Hwf Hof) as (k' & t' & Hofe1 & Hofe2).
so (Principal.oft_unique _#4 Hwf (ceqt_reg _#4 Hwf Hceq1 andel) Hofe1) as Heqc.
so (Consistent.exists_injective _#5 Hwf Heqc) as (Heqk & Heqt).
#3 apply ceqt_comp; eapply tcc_unpack.
  exact Hceq1.

  apply Hreq2; [].
  rewrite <- (subst_under_id 2 e2); [].
  relquest.
    eapply subst_oft; eauto; [].
    rewrite -> !under_succ; [].
    simp_sub.
    #2 prove_subof.
      #2 eapply ofc_subsume.
        ofvar; done.

        apply weaken2_subk; apply subk_refl; auto; done.

      #2 eapply oft_equiv.
        ofvar; done.

        rewrite -> subst_dot1_sh; [].
        apply weaken1_eqc_type; auto; done.

    simp_sub; reflexivity.

  exact (oft_reg _#3 Hwf Hof).
Qed.


Lemma reqt_letm :
  forall G M M' e,
    reqm G M M'
    -> reqt G (tm_letm M e) (tm_letm M' e).
Proof.
intros G M M' e Hr.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
so (letm_oft_inversion _#4 Hwf Hof) as (T & k & HofM & Hfst & Hofe).
#2 apply ceqt_comp; eapply tcc_letm; eauto.
  apply ceqt_refl; auto; done.

  exact (oft_reg _#3 Hwf Hof).
Qed.


Lemma reqt_letm_beta :
  forall G M e c,
    mhalt G M
    -> fstm G M c
    -> reqt G (tm_letm M e) (subst (dot M (dot c id)) e).
Proof.
intros G M e c Hhalt Hfst.
intros t Hof.
eapply ceqt_letm_beta; eauto.
Qed.


Lemma reqt_letm_beta_instances :
  forall G M e c,
    (forall s, subof aof nil s G -> mhalt nil (subst s M))
    -> fstm G M c
    -> reqt G (tm_letm M e) (subst (dot M (dot c id)) e).
Proof.
intros G M e c Hhalt HfstM.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
so (letm_oft_inversion _#4 Hwf Hof) as (T & k & HofM & Hfst & Hofe).
#2 apply closed_instances_ceqt; auto.
  relquest.
    eapply subst_oft; eauto; [].
    #2 prove_subof; eauto.
      eapply ofm_fst; eauto; done.

      eapply fstm_pure; eauto; done.

    simp_sub; reflexivity.
intros s Hs.
simp_sub.
#2 eapply rel_quest_1.
  #2 eapply ceqt_letm_beta; auto.
    so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
    exact H.

    eapply subst_fstm; eauto using subof_aof_tof; done.

  simp_sub; reflexivity.
Qed.


Lemma reqt_letm_beta_nofst :
  forall G V e,
    mvalue G V
    -> (subst (dot (var 0) (dot nonsense (sh 2))) e = e)
    -> reqt G (tm_letm V e) (subst (dot V (dot nonsense id)) e).
Proof.
intros G V e Hval Hnotfree.
intros t Hof.
eapply ceqt_letm_beta_nofst; eauto; done.
Qed.


Lemma reqt_commute_letm_unpack :
  forall G e1 M T e2,
    reqt G (tm_letm (md_unpack e1 M T) e2) (tm_unpack e1 (tm_letm M (subst (under 2 (sh 2)) e2))).
Proof.
intros G e1 M T e2.
intros u Hof.
apply ceqt_with_wf; intro Hwf; [].
so (letm_oft_inversion _#4 Hwf Hof) as (T' & l' & Hof' & Hfst' & Hofe2).
so (unpack_ofm_inversion _#6 Hwf Hof') as (k & t & Hofe1 & HofM & Hsubs & _).
so (subs_reg _#3 Hwf Hsubs) as (HofT & _).
so (can_fsts _ _ HofT) as (l & Hfst).
so (subs_fst _#5 Hwf Hsubs Hfst Hfst') as Hsubk.
eapply ceqt_commute_letm_unpack; eauto; [].
rewrite <- (subst_dot01_sh e2); [].
rewrite <- (subst_dot01_sh (subst (sh 2) u)); [].
eapply subst_oft; eauto; [].
#2 prove_subof.
  #2 eapply ofc_subsume.
    ofvar; done.

    apply weaken2_subk; auto; done.

  #2 eapply ofm_subsume.
    ofvar; done.

    apply weaken2_subs; auto; done.
Qed.


Lemma reqm_lamgn :
  forall G T k M M',
    fsts T k
    -> reqm (G; cl_cn k; cl_md (shift1 T) (var 0)) M M'
    -> reqm G (md_lamgn T M) (md_lamgn T M').
Proof.
intros G T1 k1 M M' Hfst Hr.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (lamgn_ofm_inversion _#5 Hwf Hof) as (T2 & k1a & Hsubs & Hfsta & HofM).
so (fsts_fun _#3 Hfst Hfsta); subst k1a; clear Hfsta.
so (Hr _ _ HofM) as Hceq.
apply ceqm_comp; eapply mcc_subsume; eauto; [].
apply ceqm_forget_pure; [].
apply ceqm_comp; eapply mcc_lamgn; eauto; [].
invert (subs_reg _#3 Hwf Hsubs andel); [].
intros _ HofT1 _ _.
apply eqs_refl; auto; done.
Qed.


Lemma reqm_appgn :
  forall G M1 M1' M2 M2',
    reqm G M1 M1'
    -> reqm G M2 M2'
    -> reqm G (md_appgn M1 M2) (md_appgn M1' M2').
Proof.
intros G M1 M1' M2 M2' Hr1 Hr2.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c & HofM1 & HofM2 & HfstM2 & Hsubs & Heqp).
subst p.
apply ceqm_comp; eapply mcc_subsume; eauto; [].
apply ceqm_comp; eapply mcc_appgn; eauto.
Qed.


Lemma reqm_appgn_beta :
  forall G T M1 M2 c,
    mhalt G M2
    -> fstm G M2 c
    -> reqm G (md_appgn (md_lamgn T M1) M2) (subst (dot M2 (dot c id)) M1).
Proof.
intros G T M1 M2 c Hhalt Hfst.
intros p U Hof.
eapply ceqm_appgn_beta; eauto.
Qed.


Lemma reqm_appgn_beta_instances_fstm_eqc :
  forall G T M1 M2 k c c',
    (forall s, subof aof nil s G -> mhalt nil (subst s M2))
    -> fstm G M2 c
    -> fsts T k
    -> eqc G c c' k
    -> reqm G (md_appgn (md_lamgn T M1) M2) (subst (dot M2 (dot c' id)) M1).
Proof.
intros G T1 M1 M2 k c c' HhaltM2 HfstM2 Hfst1 Heqc.
intros p T Hof.
apply ceqm_with_wf; intro Hwf; [].
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & ca & Hoflam & HofM2 & HfstM2a & Hsubs & ->).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamgn_ofm_inversion _#5 Hwf Hoflam) as (T2 & k1a & Hsubspi & Hfst1a & HofM1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (Declarative.subs_pigen_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
assert (ceqm G impure (subst (dot M2 (dot c id)) M1) (subst (dot M2 (dot c' id)) M1) T) as Hceq.
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; apply (mcc_subsume _#8 (subst1 c T2)).
    replsub (subst1 c T2) with (subst (dot M2 (dot c id)) (shift1 T2)).
    #2 refine (funct_ceq_md _#7 Hwf _ _ HofM1).
      apply wf_md; auto using wf; [].
      exact (subs_reg _#3 Hwf Hsubs1 ander).
  
      #2 prove_subrel; auto.
        exact (eqc_reg _#4 Hwf Heqc andel).
  
        apply ceqm_refl; [].
        eapply ofm_subsume; eauto; done.
  
    eapply subst1_subs; eauto; [].
    exact (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1').
eapply ceqm_trans; eauto; [].
#2 apply closed_instances_ceqm_impure; auto.
  exact (ceqm_reg _#5 Hwf Hceq andel).
clear Hceq.
intros s Hs.
simp_sub.
#2 eapply rel_quest_1.
  #3 eapply ceqm_appgn_beta.
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
    exact H.

    apply HhaltM2; auto; done.

    eapply subst_fstm; eauto using subof_aof_tof; done.

  simp_sub.
  reflexivity.
Qed.


Lemma reqm_appgn_beta_fstm_eqc :
  forall G T M1 M2 k c c',
    mhalt G M2
    -> fstm G M2 c
    -> fsts T k
    -> eqc G c c' k
    -> reqm G (md_appgn (md_lamgn T M1) M2) (subst (dot M2 (dot c' id)) M1).
Proof.
intros G T M1 M2 k c c' Hhalt HfstM2 Hfst Heqc.
eapply reqm_appgn_beta_instances_fstm_eqc; eauto; [].
intros s Hs.
eapply subst_mhalt; eauto using subof_aof_oof; done.
Qed.


Lemma reqm_appap :
  forall G M1 M1' M2 M2',
    reqm G M1 M1'
    -> reqm G M2 M2'
    -> reqm G (md_appap M1 M2) (md_appap M1' M2').
Proof.
intros G M1 M1' M2 M2' Hr1 Hr2.
intros p T Hof.
apply ceqm_with_wf; intro Hwf; [].
so (appap_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c2 & HofM1 & HofM2 & HfstM2 & Hsubs).
apply ceqm_comp; eapply mcc_subsume; eauto; [].
apply ceqm_comp; eapply mcc_appap; eauto.
Qed.


Lemma reqm_appap_beta :
  forall G T M1 M2 c,
    mhalt G M2
    -> fstm G M2 c
    -> reqm G (md_appap (md_lamap T M1) M2) (subst (dot M2 (dot c id)) M1).
Proof.
intros G T M1 M2 c2 Hhalt HfstM2.
intros p U Hof.
eapply ceqm_appap_beta; eauto.
Qed.


Lemma reqm_appap_beta_instances_fstm_eqc :
  forall G T M1 M2 k c c',
    (forall s, subof aof nil s G -> mhalt nil (subst s M2))
    -> fstm G M2 c
    -> fsts T k
    -> eqc G c c' k
    -> reqm G (md_appap (md_lamap T M1) M2) (subst (dot M2 (dot c' id)) M1).
Proof.
intros G T1 M1 M2 k1 c c' HhaltM2 HfstM2 Hfst1 Heqc.
intros p T Hof.
apply ceqm_with_wf; intro Hwf; [].
apply ceqm_forget_pure; [].
so (appap_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & ca & Hoflam & HofM2 & HfstM2a & Hsubs).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamap_ofm_inversion _#5 Hwf Hoflam) as (T2 & k1a & Hsubspi & Hfst1a & HofM1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (Declarative.subs_piapp_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
assert (ceqm G pure (subst (dot M2 (dot c id)) M1) (subst (dot M2 (dot c' id)) M1) T) as Hceq.
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; apply (mcc_subsume _#8 (subst1 c T2)).
    replsub (subst1 c T2) with (subst (dot M2 (dot c id)) (shift1 T2)).
    #2 refine (funct_ceq_md _#7 Hwf _ _ HofM1).
      apply wf_md; auto using wf; [].
      exact (subs_reg _#3 Hwf Hsubs1 ander).
  
      #2 prove_subrel; auto.
        exact (eqc_reg _#4 Hwf Heqc andel).
  
        apply ceqm_refl; [].
        eapply ofm_subsume; eauto; done.
  
    eapply subst1_subs; eauto; [].
    exact (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1').
eapply ceqm_trans; eauto; [].
so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (l & HfstT).
so (can_fstm_shift _#5 (wf_md Hwf (subs_reg _#3 Hwf Hsubs1 ander) Hfst1) HofM1) as (d & HfstM1).
so (ofs_fst _#3 (subs_reg _#3 Hwf Hsubs1 andel) Hfst1') as Hofk1'.
so (subs_reg _#3 (wf_cn Hwf Hofk1') Hsubs2) as (HofT2 & HofT2').
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (can_fsts _ _ HofT2') as (k2' & Hfst2').
so (ofm_fst _#6 (wf_md Hwf (subs_reg _#3 Hwf Hsubs1 ander) Hfst1) HofM1 HfstM1 (subst_fsts _#3 Hfst2)) as Hofds.
so (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1') as Hofc.
so (subs_fst _#5 Hwf Hsubs1 Hfst1' Hfst1) as Hsubk1.
assert (ofm G pure (md_appap (md_lamap T1 M1) M2) T) as Hof'.
  eapply ofm_subsume; eauto; [].
  eapply ofm_appap; eauto; [].
  eapply ofm_subsume; eauto; [].
  eapply ofm_lamap; eauto; [].
  exact (subs_reg _#3 Hwf Hsubs1 ander).
#5 eapply closed_instances_ceqm_pure; eauto.
  exact (ceqm_reg _#5 Hwf Hceq andel).

  eapply fstm_appap; eauto; [].
  eapply fstm_lamap; eauto; done.

  eapply subst_fstm; eauto; [].
  prove_subof; auto; done.

  simp_sub.
  #2 eapply eqc_subsume.
    eapply eqc_beta; eauto using ofc_subsume; [].
    replsub d with (subst1 nonsense (shift1 d)).
    eapply subst_ofc; eauto; prove_subof; done.

    simp_sub.
    so (subs_fst _#5 Hwf Hsubs (subst_fsts _#3 Hfst2') HfstT) as Hsubk.
    eapply subk_trans; eauto; [].
    so (subs_fst _#5 (wf_cn Hwf Hofk1') Hsubs2 Hfst2 Hfst2') as Hsubk2.
    eapply subst_subk; eauto; prove_subof; auto; done.

  clear Hceq.
  intros s Hs.
  simp_sub.
  #2 eapply rel_quest_1.
    #3 eapply ceqm_appap_beta.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof') as H; simp_sub in H.
      exact H.
  
      apply HhaltM2; auto; done.
  
      eapply subst_fstm; eauto using subof_aof_tof; done.
  
    simp_sub.
    reflexivity.
Qed.


Lemma reqm_unpack :
  forall G e e' M T,
    reqt G e e'
    -> reqm G (md_unpack e M T) (md_unpack e' M T).
Proof.
intros G e e' M T Hr.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k & t & Hofe & HofM & Hsubs & Heqp).
subst p.
apply ceqm_comp; eapply mcc_subsume; eauto; [].
#2 apply ceqm_comp; eapply mcc_unpack; eauto.
  apply ceqm_refl; auto; done.

  apply eqs_refl; exact (subs_reg _#3 Hwf Hsubs andel).
Qed.


Lemma reqm_unpack_bind :
  forall G e e' M M' T k t,
    ceqt G e e' (cn_exists k t)
    -> reqm (G; cl_cn k; cl_tm t) M M'
    -> reqm G (md_unpack e M T) (md_unpack e' M' T).
Proof.
intros G e e' M M' T k t Hceq1 Hreq2.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k' & t' & Hofe & HofM & Hsubs & ->).
so (Principal.oft_unique _#4 Hwf (ceqt_reg _#4 Hwf Hceq1 andel) Hofe) as Heqc.
so (Consistent.exists_injective _#5 Hwf Heqc) as (Heqk & Heqt).
apply ceqm_comp; eapply mcc_subsume; eauto; [].
#3 apply ceqm_comp; eapply mcc_unpack.
  exact Hceq1.

  apply Hreq2; [].
  rewrite <- (subst_under_id 2 M); [].
  relquest.
    eapply subst_ofm; eauto; [].
    rewrite -> !under_succ; [].
    simp_sub.
    #2 prove_subof.
      #2 eapply ofc_subsume.
        ofvar; done.

        apply weaken2_subk; apply subk_refl; auto; done.

      #2 eapply oft_equiv.
        ofvar; done.

        rewrite -> subst_dot1_sh; [].
        apply weaken1_eqc_type; auto; done.

    simp_sub; reflexivity.

  apply eqs_refl; [].
  exact (subs_reg _#3 Hwf Hsubs andel).
Qed.


Lemma reqm_unpack_beta :
  forall G c e k t M T,
    thalt G e
    -> reqm G (md_unpack (tm_pack c e k t) M T) (subst (dot e (dot c id)) M).
Proof.
intros G c e k t M T Hhalt.
intros p U Hof.
eapply ceqm_unpack_beta; eauto.
Qed.


Lemma reqm_lett :
  forall G e e' M,
    reqt G e e'
    -> reqm G (md_lett e M) (md_lett e' M).
Proof.
intros G e e' M Hr.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (lett_ofm_inversion _#5 Hwf Hof) as (t & Hofe & HofM).
apply ceqm_comp; eapply mcc_lett; eauto; [].
apply ceqm_refl; assumption.
Qed.


Lemma reqm_lett_beta :
  forall G e M,
    thalt G e
    -> reqm G (md_lett e M) (subst1 e M).
Proof.
intros G e M Hhalt.
intros p T Hof.
apply ceqm_lett_beta; auto.
Qed.


Lemma reqm_letm :
  forall G M1 M1' M2 T,
    reqm G M1 M1'
    -> reqm G (md_letm M1 M2 T) (md_letm M1' M2 T).
Proof.
intros G M1 M1' M2 T Hr.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (letm_ofm_inversion _#6 Hwf Hof) as (T1 & k1 & HofM1 & HfstT1 & HofM2 & Hsubs & Heqp).
subst p.
apply ceqm_comp; eapply mcc_subsume; eauto; [].
#2 apply ceqm_comp; eapply mcc_letm; eauto.
  apply ceqm_refl; auto; done.

  apply eqs_refl; [].
  exact (subs_reg _#3 Hwf Hsubs andel).
Qed.


Lemma reqm_letm_beta_instances :
  forall G M1 M2 T c1,
    (forall s, subof aof nil s G -> mhalt nil (subst s M1))
    -> fstm G M1 c1
    -> reqm G (md_letm M1 M2 T) (subst (dot M1 (dot c1 id)) M2).
Proof.
intros G M1 M2 T c1 Hhalt HfstM1.
intros p U Hof.
eapply ceqm_letm_beta_instances; eauto; done.
Qed.


Lemma reqm_commute_appgn_unpack_fstm_eqc :
  forall G e M1 T1 T2 M2 k1 c2 c2',
    fsts T1 k1
    -> fstm G M2 c2
    -> eqc G c2 c2' k1
    -> reqm G 
       (md_appgn (md_unpack e M1 (sg_pigen T1 T2)) M2) 
       (md_unpack e (md_appgn M1 (subst (sh 2) M2)) (subst1 c2' T2)).
Proof.
intros G e M1 T1 T2 M2 k1 c2 c2' Hfst1 HfstM2 Heqc.
intros p U Hof.
eapply ceqm_commute_appgn_unpack_fstm_eqc; eauto; done.
Qed.


Lemma reqm_commute_appap_unpack_fstm_eqc :
  forall G e M1 T1 T2 M2 k1 c2 c2',
    fsts T1 k1
    -> fstm G M2 c2
    -> eqc G c2 c2' k1
    -> reqm G 
       (md_appap (md_unpack e M1 (sg_piapp T1 T2)) M2) 
       (md_unpack e (md_appap M1 (subst (sh 2) M2)) (subst1 c2' T2)).
Proof.
intros G e M1 T1 T2 M2 k1 c2 c2' Hfst1 HfstM2 Heqc.
intros p U Hof.
eapply ceqm_commute_appap_unpack_fstm_eqc; eauto; done.
Qed.


Lemma reqm_commute_unpack_unpack :
  forall G e1 e2 M T,
    reqm G (md_unpack (tm_unpack e1 e2) M T) (md_unpack e1 (md_unpack e2 (subst (dot (var 0) (dot (var 1) (sh 4))) M) (subst (sh 2) T)) T).
Proof.
intros G e1 e2 M T.
intros p U Hof.
eapply ceqm_commute_unpack_unpack; eauto; done.
Qed.


Lemma reqm_commute_unpack_letm :
  forall G M1 e M2 T,
    reqm G (md_unpack (tm_letm M1 e) M2 T) (md_letm M1 (md_unpack e (subst (under 2 (sh 2)) M2) (subst (sh 2) T)) T).
Proof.
intros G M1 e M2 T.
intros p U Hof.
apply ceqm_with_wf; intro Hwf; [].
so (unpack_ofm_inversion _#6 Hwf Hof) as (k & c & Hof' & HofM2 & Hsubs & Heqp).
so (letm_oft_inversion _#4 Hwf Hof') as (S & l & HofM1 & HfstS & Hofe).
subst p.
apply ceqm_comp; eapply mcc_subsume; eauto; [].
eapply ceqm_commute_unpack_letm; eauto.
Qed.


Lemma ceqt_reqt :
  forall G e e' t,
    ceqt G e e' t
    -> reqt G e e'.
Proof.
intros G e1 e2 t Hceq.
intros t' Hof'.
apply ceqt_with_wf; intro Hwf; [].
apply ceqt_comp; [].
eapply tcc_equiv; eauto; [].
so (ceqt_reg _#4 Hwf Hceq) as (Hof & _).
eapply Principal.oft_unique; eauto; done.
Qed.


(* Conditional reductive equality *)

Definition keqt (G : context) (e e' : (* tm *) term) (P : sub -> Prop) :=
  forall t,
    wf G
    -> oft G e t
    -> oft G e' t
       /\ (forall s,
             subof aof nil s G
             -> (thalt nil (subst s e) \/ (P s /\ thalt nil (subst s e')))
             -> ceqt nil (subst s e) (subst s e') (subst s t)).


Definition keqm (G : context) (M M' : (* md *) term) (P : sub -> Prop) :=
  forall p T,
    wf G
    -> ofm G p M T
    -> ofm G p M' T
       /\ (p = pure
           -> exists c c' k,
                fstm G M c
                /\ fstm G M' c'
                /\ fsts T k
                /\ eqc G c c' k)
       /\ (forall s,
             subof aof nil s G
             -> (mhalt nil (subst s M) \/ (P s /\ mhalt nil (subst s M')))
             -> ceqm nil p (subst s M) (subst s M') (subst s T)).


Definition andp (P Q : sub -> Prop) := fun s => P s /\ Q s.
Definition truep : sub -> Prop := fun s => True.


Lemma keqt_reqt :
  forall G e e',
    keqt G e e' truep
    -> reqt G e e'.
Proof.
intros G e e' Hkeq.
intros t Hof.
apply ceqt_with_wf; intro Hwf; [].
so (Hkeq t Hwf Hof) as (Hofe' & Hceqif).
apply (ceqt_closed_instances_conditional (fun s => thalt nil (subst s e))); auto; [].
intros s Hs.
do2 2 split.
  auto; done.

  intro Hhalt.
  lapply (Hceqif s Hs).
    intros Hceq.
    exact (ceqt_halt _#3 (ceqt_symm _#4 Hceq) Hhalt).

    right.
    unfold truep.
    split; auto; done.

  intros Hhalt.
  apply Hceqif; auto; done.
Qed.


Lemma keqm_reqm :
  forall G M M',
    keqm G M M' truep
    -> reqm G M M'.
Proof.
intros G M M' Hseq.
intros p T HofM.
apply ceqm_with_wf; intro Hwf; [].
so (Hseq _ _ Hwf HofM) as (HofM' & Hfststuff & Hceqif).
clear Hseq.
assert (forall s,
          subof aof nil s G 
          -> mhalt nil (subst s M') 
          -> mhalt nil (subst s M)) as Hhaltif.
  intros s Hs Hhalt.
  so (Hceqif _ Hs (or_intror (conj I Hhalt))) as Hceq.
  exact (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_symm _#5 Hceq)) Hhalt).
destruct p.
  (* pure *)
  so (Hfststuff (eq_refl _)) as (c & c' & k & HfstM & HfstM' & Hfst & Heqc).
  clear Hfststuff.
  eapply closed_instances_ceqm_pure; eauto; [].
  intros s Hs.
  eapply ceqm_conditional_pure; eauto using subst_ofm, subst_eqc, subst_fstm, subst_fsts, subof_aof_sof, subof_aof_dof, subof_aof_tof; [].
  intro; apply Hhaltif; auto; done.

  (* impure *)
  clear Hfststuff.
  eapply closed_instances_ceqm_impure; eauto; [].
  intros s Hs.
  eapply ceqm_conditional_impure; eauto using subst_ofm, subof_aof_dof; [].
  intro; apply Hhaltif; auto; done.
Qed.


Lemma reqt_keqt :
  forall G e e',
    reqt G e e'
    -> keqt G e e' truep.
Proof.
intros G e e' Hr.
intros t Hwf Hof.
so (Hr t Hof) as Hceq.
split.
  exact (ceqt_reg _#4 Hwf Hceq ander).

  intros s Hs _.
  eapply subst_ceqt; eauto; done.
Qed.


Lemma reqm_keqm :
  forall G M M',
    reqm G M M'
    -> keqm G M M' truep.
Proof.
intros G M M' Hr.
intros p T Hwf Hof.
so (Hr p T Hof) as Hceq.
do2 2 split.
  exact (ceqm_reg _#5 Hwf Hceq ander).

  intros ->.
  so (ceqm_fst _#4 Hwf Hceq) as (k & c & c' & Hfst & HfstM & HfstM' & Heqc).
  repeat eexists; eauto; done.

  intros s Hs _.
  eapply subst_ceqm; eauto; done.
Qed.


Lemma keqt_ceqt :
  forall G e e' t,
    oft G e t
    -> keqt G e e' truep
    -> ceqt G e e' t.
Proof.
eauto using keqt_reqt, reqt_ceqt.
Qed.


Lemma keqm_ceqm :
  forall G p M M' T,
    ofm G p M T
    -> keqm G M M' truep
    -> ceqm G p M M' T.
Proof.
eauto using reqm_ceqm, keqm_reqm.
Qed.


Lemma ceqt_keqt :
  forall G e e' t,
    ceqt G e e' t
    -> keqt G e e' truep.
Proof.
intros G e e' t Hceq.
apply reqt_keqt; [].
eapply ceqt_reqt; eauto; done.
Qed.


Lemma keqt_discharge :
  forall G P e e',
    keqt G e e' P
    -> (forall s,
          subof aof nil s G
          -> thalt nil (subst s e')
          -> P s)
    -> keqt G e e' truep.
Proof.
intros G P e e' Hkeq HPif.
intros t Hwf Hof.
so (Hkeq t Hwf Hof) as (Hofe' & Hceqif).
split; auto; [].
intros s Hs Hhalts.
apply Hceqif; auto; [].
#2 destruct Hhalts as [| (_, ?)].
  left; auto; done.

  right; auto; done.
Qed.


Lemma keqm_partial_discharge :
  forall G (P Q : sub -> Prop) M M',
    keqm G M M' P
    -> (forall s,
          subof aof nil s G
          -> Q s
          -> mhalt nil (subst s M')
          -> P s)
    -> keqm G M M' Q.
Proof.
intros G P Q M M' Hkeq Himp.
intros p T Hwf Hof.
so (Hkeq p T Hwf Hof) as (Hof' & Hfsteq & Hceqif).
do2 2 split; auto; [].
intros s Hs Hhalts.
eapply Hceqif; eauto; [].
destruct Hhalts as [? | (?, ?)]; auto; done.
Qed.


Lemma keqm_discharge :
  forall G P M M',
    keqm G M M' P
    -> (forall s,
          subof aof nil s G
          -> mhalt nil (subst s M')
          -> P s)
    -> keqm G M M' truep.
Proof.
intros G P M M' Hkeq Himp.
eapply keqm_partial_discharge; eauto; [].
intros; eauto; done.
Qed.


Lemma keqm_middle_ceqm :
  forall G (P Q : sub -> Prop) p M1 M2 M3 M4 T,
    ofm G p M1 T
    -> ofm G p M4 T
    -> keqm G M1 M2 P
    -> keqm G M4 M3 Q
    -> (ofm G p M2 T
        -> ofm G p M3 T
        -> ceqm G p M2 M3 T)
    -> (forall s,
          subof aof nil s G
          -> mhalt nil (subst s M4)
          -> P s)
    -> (forall s,
          subof aof nil s G
          -> mhalt nil (subst s M1)
          -> Q s)
    -> ceqm G p M1 M4 T.
Proof.
intros G P Q p M1 M2 M3 M4 T HofM1 HofM4 Hk12 Hk43 Hceq23if Himplp Himplq.
apply ceqm_with_wf; intro Hwf; [].
so (Hk12 _ _ Hwf HofM1) as (HofM2 & Hfsteq12 & Hceq12if).
so (Hk43 _ _ Hwf HofM4) as (HofM3 & Hfsteq43 & Hceq43if).
so (Hceq23if HofM2 HofM3) as Hceq23.
clear Hk12 Hk43.
#2 eapply (ceqm_closed_instances_conditional (fun s => mhalt nil (subst s M1))); eauto.
  intros ->.
  so (Hfsteq12 (eq_refl _)) as (c1 & c2 & k & HfstM1 & HfstM2 & Hfst & Heqc12).
  so (Hfsteq43 (eq_refl _)) as (c4 & c3 & ka & HfstM4 & HfstM3 & Hfsta & Heqc43).
  so (fsts_fun _#3 Hfst Hfsta); subst ka; clear Hfsta.
  so (ceqm_fst' _#7 Hwf Hceq23 Hfst HfstM2 HfstM3) as Heqc.
  exists k, c1, c4.
  do2 3 split; eauto with static; done.

  intros s Hs.
  so (subst_ceqm _#7 Hs Hwf Hceq23) as Hceq23'.
  #2 do2 2 split; auto.
    intro Hhalt4.
    so (Hceq43if _ Hs (or_introl Hhalt4)) as Hceq43.
    lapply (Hceq12if _ Hs).
      intros Hceq12.
      exact (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_trans _#6 Hceq43 (ceqm_symm _#5 (ceqm_trans _#6 Hceq12 Hceq23')))) Hhalt4).

      right.
      split.
        eapply Himplp; eauto; done.

        exact (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_trans _#6 Hceq43 (ceqm_symm _#5 Hceq23'))) Hhalt4).

    intros Hhalt1.
    so (Hceq12if _ Hs (or_introl Hhalt1)) as Hceq12.
    eapply ceqm_trans; eauto; [].
    eapply ceqm_trans; eauto; [].
    apply ceqm_symm; [].
    apply Hceq43if; auto; [].
    right.
    split.
      eapply Himplq; eauto; done.

      exact (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_trans _#6 Hceq12 Hceq23')) Hhalt1).
Qed.


Lemma keqm_with_wf :
  forall G M M' P,
    (wf G -> keqm G M M' P)
    -> keqm G M M' P.
Proof.
intros G M M' P H.
intros p T Hwf.
exact (H Hwf p T Hwf).
Qed.


Lemma keqt_refl :
  forall G e,
    keqt G e e truep.
Proof.
intros G e.
intros t Hwf Hof.
split; auto; [].
intros s Hs _.
apply ceqt_refl; [].
eapply subst_oft; eauto using subof_aof_dof; done.
Qed.


Lemma keqm_refl :
  forall G M,
    keqm G M M truep.
Proof.
intros G M.
intros p T Hwf Hof.
#2 do2 2 split; auto.
  intros ->.
  so (can_fstm _#3 Hwf Hof) as (c & HfstM).
  so (ofm_fst_o _#5 Hwf Hof HfstM) as (k & Hfst & Hofc).
  do 3 eexists; do2 3 split; eauto using eqc_refl; done.

  intros s Hs _.
  apply ceqm_refl; [].
  eapply subst_ofm; eauto using subof_aof_dof; done.
Qed.    


Lemma keqt_trans :
  forall G P1 P2 e1 e2 e3,
    keqt G e1 e2 P1
    -> keqt G e2 e3 P2
    -> keqt G e1 e3 (andp P1 P2).
Proof.
intros G P1 P2 e1 e2 e3 Hk12 Hk23.
intros t Hwf Hof1.
so (Hk12 t Hwf Hof1) as (Hof2 & Hceq12if).
so (Hk23 t Hwf Hof2) as (Hof3 & Hceq23if).
split; auto; [].
  intros s Hs Hhalt.
  lapply (Hceq12if _ Hs).
    intro Hceq12.

    lapply (Hceq23if _ Hs).
      intro Hceq23.
      eapply ceqt_trans; eauto; done.

      #2 destruct Hhalt as [Hhalt1 | ((HP, HP'), Hhalt3)].
        left.
        exact (ceqt_halt _#3 Hceq12 Hhalt1).

        right.
        split; auto; done.

    #2 destruct Hhalt as [Hhalt1 | ((HP, HP'), Hhalt3)].
      left; assumption.

      right; split; auto; [].
      lapply (Hceq23if _ Hs).
        intro Hceq23.
        exact (ceqt_halt _#3 (ceqt_symm _#4 Hceq23) Hhalt3).

        right; split; auto; done.
Qed.


Lemma keqm_trans :
  forall G P1 P2 M1 M2 M3,
    keqm G M1 M2 P1
    -> keqm G M2 M3 P2
    -> keqm G M1 M3 (andp P1 P2).
Proof.
intros G P P' M1 M2 M3 H12 H23.
intros p T Hwf Hof1.
so (H12 p T Hwf Hof1) as (Hof2 & Hfst12 & Hceq12if).
so (H23 p T Hwf Hof2) as (Hof3 & Hfst23 & Hceq23if).
#2 do2 2 split; auto.
  intros ->.
  so (Hfst12 (eq_refl _)) as (c1 & c2 & k & HfstM1 & HfstM2 & Hfst & Heqc12).
  so (Hfst23 (eq_refl _)) as (c2a & c3 & ka & HfstM2a & HfstM3 & Hfsta & Heqc23).
  so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
  so (fsts_fun _#3 Hfst Hfsta); subst ka.
  do 3 eexists; do2 3 split; eauto using eqc_trans; done.

  intros s Hs Hhalt.
  lapply (Hceq12if _ Hs).
    intro Hceq12.

    lapply (Hceq23if _ Hs).
      intro Hceq23.
      eapply ceqm_trans; eauto; done.

      #2 destruct Hhalt as [Hhalt1 | ((HP, HP'), Hhalt3)].
        left.
        exact (ceqm_halt _#3 (ceqm_forget_impure _#5 Hceq12) Hhalt1).

        right.
        split; auto; done.

    #2 destruct Hhalt as [Hhalt1 | ((HP, HP'), Hhalt3)].
      left; assumption.

      right; split; auto; [].
      lapply (Hceq23if _ Hs).
        intro Hceq23.
        exact (ceqm_halt _#3 (ceqm_symm _#5 (ceqm_forget_impure _#5 Hceq23)) Hhalt3).

        right; split; auto; done.
Qed.


Lemma keqt_app :
  forall G P1 P2 e1 e1' e2 e2',
    keqt G e1 e1' P1
    -> keqt G e2 e2' P2
    -> keqt G (tm_app e1 e2) (tm_app e1' e2') (andp P1 P2).
Proof.
intros G P1 P2 e1 e1' e2 e2' Hkeq1 Hkeq2.
intros t2 Hwf Hof.
so (app_oft_inversion _#4 Hwf Hof) as (t1 & Hof1 & Hof2).
so (Hkeq1 _ Hwf Hof1) as (Hof1' & Hceqif1).
so (Hkeq2 _ Hwf Hof2) as (Hof2' & Hceqif2).
split; eauto using oft_app; [].
intros s Hs Hhalts.
simp_sub.
unfold andp in Hhalts.
lapply (Hceqif1 _ Hs).
  intros Hceq1.
  lapply (Hceqif2 _ Hs).
    intros Hceq2.
    apply ceqt_comp; eapply tcc_app; eauto; done.

    clear Hceq1.
    destruct Hhalts as [Hhalt | ((HP1, HP2), Hhalt)].
      left.
      simp_sub in Hhalt.
      exact (app_halt_inversion' _ _ Hhalt ander).

      right.
      split; auto; [].
      simp_sub in Hhalt.
      exact (app_halt_inversion' _ _ Hhalt ander).

  destruct Hhalts as [Hhalt | ((HP1, HP2), Hhalt)].
    left.
    simp_sub in Hhalt.
    exact (app_halt_inversion' _ _ Hhalt andel).

    right.
    split; auto; [].
    simp_sub in Hhalt.
    exact (app_halt_inversion' _ _ Hhalt andel).
Qed.


Lemma keqt_app_beta :
  forall G t e1 e2,
    keqt G (tm_app (tm_lam t e1) e2) (subst1 e2 e1) (fun s => thalt nil (subst s e2)).
Proof.
intros G t1' e1 e2.
intros t2 Hwf Hof.
so (app_oft_inversion _#4 Hwf Hof) as (t1 & Hoflam & Hof2).
so (lam_oft_inversion_arrow _#5 Hwf Hoflam) as (Heqc & Hof1).
split.
  relquest.
    eapply subst_oft; eauto; prove_subof; eauto with static; done.

    simp_sub; reflexivity.

  intros s Hs Hhalt.
  simp_sub.
  #2 eapply rel_quest_1.
    #2 apply ceqt_app_beta.
      #2 eapply oft_app.
        #2 apply oft_lam.
          so (eqc_reg _#4 Hwf Heqc andel).
          eapply subst_ofc_type; eauto using subof_aof_sof; done.

          relquest.
            #2 eapply subst_oft; eauto; prove_subof.
              eapply subof_dof_trans; eauto using weaken1, subof_aof_dof; done.

              #2 eapply oft_equiv.
                ofvar; done.

                eapply subst_eqc_type; eauto using subof_sof_trans, subof_aof_sof, weaken1; done.

            simp_sub; reflexivity.

        #2 eapply oft_equiv.
          eapply subst_oft; eauto using subof_aof_dof; done.

          apply eqc_symm; [].
          eapply subst_eqc_type; eauto using subof_aof_sof; done.

      destruct Hhalt as [Hhalt' | (?, _)]; auto; [].
      simp_sub in Hhalt'.
      exact (app_halt_inversion' _ _ Hhalt' ander).

    simp_sub; reflexivity.
Qed.


Lemma keqt_pi1_beta :
  forall G e1 e2,
    keqt G (tm_pi1 (tm_pair e1 e2)) e1 (fun s => thalt nil (subst s e2)).
Proof.
intros G e1 e2.
intros t Hwf Hof.
so (pi1_oft_inversion _#3 Hwf Hof) as (t' & Hof').
so (pair_oft_inversion_prod _#5 Hwf Hof') as (Hof1 & Hof2).
split; auto; [].
intros s Hs Hhalt.
simp_sub.
simp_sub in Hhalt.
assert (thalt nil (subst s e1) /\ thalt nil (subst s e2)) as (Hhalt1 & Hhalt2).
  destruct Hhalt as [Hhalt' | (?, ?)].
  so (pair_halt_inversion' _ _ (pi1_halt_inversion' _ Hhalt')) as (Hhalt1 & Hhalt2).
  auto; done.

  auto; done.
eapply ceqt_pi1_beta; eauto; [].
so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
exact H.
Qed.


Lemma keqt_pi2_beta :
  forall G e1 e2,
    keqt G (tm_pi2 (tm_pair e1 e2)) e2 (fun s => thalt nil (subst s e1)).
Proof.
intros G e1 e2.
intros t Hwf Hof.
so (pi2_oft_inversion _#3 Hwf Hof) as (t' & Hof').
so (pair_oft_inversion_prod _#5 Hwf Hof') as (Hof1 & Hof2).
split; auto; [].
intros s Hs Hhalt.
simp_sub.
simp_sub in Hhalt.
assert (thalt nil (subst s e1) /\ thalt nil (subst s e2)) as (Hhalt1 & Hhalt2).
  destruct Hhalt as [Hhalt' | (?, ?)].
  so (pair_halt_inversion' _ _ (pi2_halt_inversion' _ Hhalt')) as (Hhalt1 & Hhalt2).
  auto; done.

  auto; done.
eapply ceqt_pi2_beta; eauto; [].
so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
exact H.
Qed.


Lemma keqt_lett_beta :
  forall G e1 e2,
    keqt G (tm_lett e1 e2) (subst1 e1 e2) (fun s => thalt nil (subst s e1)).
Proof.
intros G e1 e2.
intros t2 Hwf Hof.
so (lett_oft_inversion _#4 Hwf Hof) as (t1 & Hof1 & Hof2).
split.
  relquest.
    eapply subst_oft; eauto; prove_subof; auto; done.

    simp_sub; reflexivity.

  intros s Hs Hhalts.
  simp_sub.
  #2 eapply rel_quest_1.
    #2 apply ceqt_lett_beta.
      so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.

      destruct Hhalts as [Hhalt | (? & _)]; auto; [].
      simp_sub in Hhalt.
      exact (lett_halt_inversion' _ _ Hhalt).

    simp_sub; reflexivity.
Qed.


Lemma keqm_pi1 :
  forall G P M M',
    keqm G M M' P
    -> keqm G (md_pi1 M) (md_pi1 M') P.
Proof.
intros G P M M' Hkeq.
intros p T1 Hwf Hof.
so (pi1_ofm_inversion _#4 Hwf Hof) as (T2 & HofM).
so (Hkeq _ _ Hwf HofM) as (HofM' & Hfsteq & Hceqif).
do2 2 split.
  apply ofm_forget_pure; [].
  eapply ofm_pi1; eauto; done.

  intros ->.
  so (Hfsteq (eq_refl _)) as (c & c' & k & HfstM & HfstM' & Hfstsigma & Heqc).
  invert Hfstsigma; [].
  intros k1 K2 Hfst1 hfst2 <-.
  exists (cn_pi1 c), (cn_pi1 c'), k1.
  do2 3 split; eauto with static; done.

  intros s Hs Hhalt.
  lapply (Hceqif _ Hs).
    intros Hceq.
    simp_sub in Hceq.
    simp_sub.
    apply ceqm_forget_pure; [].
    apply ceqm_comp; eapply mcc_pi1; eauto; done.

    destruct Hhalt as [H | (?, H)].
      left.
      exact (mpi1_halt_inversion' _ H).

      right.
      split; auto; [].
      simp_sub in H.
      exact (mpi1_halt_inversion' _ H).
Qed.


Lemma keqm_pi2 :
  forall G P M M',
    keqm G M M' P
    -> keqm G (md_pi2 M) (md_pi2 M') P.
Proof.
intros G P M M' Hkeq.
intros p T Hwf Hof.
so (pi2_ofm_inversion _#4 Hwf Hof) as (T1 & T2 & c & HofM & HfstM & Hsubs).
so (Hkeq _ _ Hwf HofM) as (HofM' & Hfsteq & Hceqif).
so (Hfsteq (eq_refl _)) as (ca & c' & k & HfstMa & HfstM' & Hfstsigma & Heqc); clear Hfsteq.
so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
invert Hfstsigma; clear Hfstsigma; [].
intros k1 k2 Hfst1 Hfst2 <-.
invert (ofm_reg _#4 Hwf HofM); [].
intros k1a HofT1 Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
do2 2 split.
  apply ofm_forget_pure; [].
  eapply ofm_subsume; eauto; [].
  #2 eapply ofm_subsume.
    eapply ofm_pi2; eauto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply funct_sg; eauto; [].
    eapply eqc_pi1; eauto; done.

  intros ->.
  so (can_fsts _ _ (subs_reg _#3 Hwf Hsubs ander)) as (k & HfstT).
  exists (cn_pi2 c), (cn_pi2 c'), k.
  do2 3 split; auto using fstm_pi2; [].
  #2 eapply eqc_subsume.
    eapply eqc_pi2; eauto; done.

    exact (subs_fst _#5 Hwf Hsubs (subst_fsts _#3 Hfst2) HfstT).

  intros s Hs Hhalt.
  lapply (Hceqif _ Hs).
    intros Hceq.
    simp_sub in Hceq.
    simp_sub.
    apply ceqm_forget_pure; [].
    #2 apply ceqm_comp; eapply mcc_subsume.
      apply ceqm_comp; eapply mcc_pi2; eauto using subst_fstm, subof_aof_tof; done.

      simp_sub.
      so (subst_subs _#5 (subof_aof_sof _#3 Hs) Hsubs) as H.
      simp_sub in H.
      exact H.

    destruct Hhalt as [H | (?, H)].
      left.
      exact (mpi2_halt_inversion' _ H).

      right.
      split; auto; [].
      simp_sub in H.
      exact (mpi2_halt_inversion' _ H).
Qed.


Lemma keqm_pi1_beta :
  forall G M1 M2,
    keqm G (md_pi1 (md_pair M1 M2)) M1 (fun s => mhalt nil (subst s M2)).
Proof.
intros G M1 M2.
intros p T1 Hwf Hof.
so (pi1_ofm_inversion _#4 Hwf Hof) as (T2 & Hof').
so (pair_ofm_inversion_sigma_pure _#5 Hwf Hof') as (c1 & Hof1 & HfstM1 & Hof2).
do2 2 split.
  apply ofm_forget_pure; [].
  exact Hof1.

  intros ->.
  so (can_fstm _#3 Hwf Hof2) as (c2 & HfstM2).
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof1)) as (k1 & Hfst1).
  so (ofm_fst_o _#5 Hwf Hof2 HfstM2) as (l & Hfst & Hofc2).
  exists (cn_pi1 (cn_pair c1 c2)), c1, k1.
  do2 3 split; eauto using fstm; [].
  #2 eapply eqc_beta1.
    exact (ofm_fst _#6 Hwf Hof1 HfstM1 Hfst1).

    exact Hofc2.

  intros s Hs Hhalt.
  simp_sub.
  simp_sub in Hhalt.
  assert (mhalt nil (subst s M1) /\ mhalt nil (subst s M2)) as (Hhalt1 & Hhalt2).
    destruct Hhalt as [Hhalt' | (?, ?)].
    so (mpair_halt_inversion' _ _ (mpi1_halt_inversion' _ Hhalt')) as (Hhalt1 & Hhalt2).
    auto; done.
  
    auto; done.
  apply ceqm_forget_pure; [].
  #2 eapply ceqm_pi1_beta; auto.
    eapply subst_ofm; eauto using subof_aof_dof; done.

    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof2) as H; simp_sub in H.
    exact H.
Qed.


Lemma keqm_pi2_beta :
  forall G M1 M2,
    keqm G (md_pi2 (md_pair M1 M2)) M2 (fun s => mhalt nil (subst s M1)).
Proof.
intros G M1 M2.
intros p T Hwf Hof.
so (pi2_ofm_inversion _#4 Hwf Hof) as (T1 & T2 & c & Hof' & Hfstpair & Hsubs).
invert Hfstpair; clear Hfstpair; [].
intros c1 c2 HfstM1 HfstM2 <-.
so (pair_ofm_inversion_sigma_pure _#5 Hwf Hof') as (c1a & Hof1 & HfstM1a & Hof2).
so (fstm_fun _#4 HfstM1 HfstM1a); subst c1a; clear HfstM1a.
invert (ofm_reg _#4 Hwf Hof'); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (ofm_fst _#6 Hwf Hof1 HfstM1 Hfst1) as Hofc1.
so (ofm_fst _#6 Hwf Hof2 HfstM2 (subst_fsts _#3 Hfst2)) as Hofc2.
assert (subs G (subst1 c1 T2) T) as Hsubs'.
  eapply subs_trans; eauto; [].
  apply subs_refl; [].
  eapply funct_sg; eauto; [].
  apply eqc_symm; [].
  eapply eqc_beta1; eauto; done.
do2 2 split.
  apply ofm_forget_pure; [].
  eapply ofm_subsume; eauto; done.

  intros ->.
  so (can_fsts _ _ (subs_reg _#3 Hwf Hsubs ander)) as (l & HfstT).
  exists (cn_pi2 (cn_pair c1 c2)), c2, l.
  do2 3 split; eauto using fstm; [].
  #2 eapply eqc_subsume.
    eapply eqc_beta2; eauto; done.

    exact (subs_fst _#5 Hwf Hsubs' (subst_fsts _#3 Hfst2) HfstT).

  intros s Hs Hhalt.
  simp_sub.
  simp_sub in Hhalt.
  assert (mhalt nil (subst s M1) /\ mhalt nil (subst s M2)) as (Hhalt1 & Hhalt2).
    destruct Hhalt as [Hhalt' | (?, ?)].
    so (mpair_halt_inversion' _ _ (mpi2_halt_inversion' _ Hhalt')) as (Hhalt1 & Hhalt2).
    auto; done.
  
    auto; done.
  apply ceqm_forget_pure; [].
  #2 apply ceqm_comp; eapply mcc_subsume.
    #2 eapply ceqm_pi2_beta; auto.
      eapply subst_ofm; eauto using subof_aof_dof; done.
  
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof2) as H; simp_sub in H.
      exact H.

    so (subst_subs _#5 (subof_aof_sof _#3 Hs) Hsubs') as H; simp_sub in H.
    exact H.
Qed.


Lemma keqm_appgn :
  forall G P1 P2 M1 M1' M2 M2',
    keqm G M1 M1' P1
    -> keqm G M2 M2' P2
    -> keqm G (md_appgn M1 M2) (md_appgn M1' M2') (andp P1 P2).
Proof.
intros G P1 P2 M1 M1' M2 M2' Hs1 Hs2.
intros p T Hwf Hof.
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c2 & HofM1 & HofM2 & HfstM2 & Hsubs & ->).
so (Hs1 _ _ Hwf HofM1) as (HofM1' & Hfsteq1 & Hceqif1).
so (Hs2 _ _ Hwf HofM2) as (HofM2' & Hfsteq2 & Hceqif2).
so (Hfsteq2 (eq_refl _)) as (c2a & c2' & k1 & HfstM2a & HfstM2' & Hfst1 & Heqc2).
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
invert (ofm_reg _#4 Hwf HofM1); [].
intros k1a HofT1 Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
do2 2 split.
  eapply ofm_subsume; eauto; [].
  #2 eapply ofm_subsume.
    eapply ofm_appgn; eauto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply funct_sg; eauto; done.

  intro Heq.
  discriminate Heq.

  intros s Hs Hhalts.
  simp_sub in Hhalts.
  simp_sub.
  lapply (Hceqif1 _ Hs).
    intro Hceq1.
    lapply (Hceqif2 _ Hs).
      intro Hceq2.
      #2 apply ceqm_comp; eapply mcc_subsume.
        #3 apply ceqm_comp; eapply mcc_appgn.
          simp_sub in Hceq1.
          exact Hceq1.
    
          exact Hceq2.
    
          eapply subst_fstm; eauto using subof_aof_tof; done.
    
        simp_sub.
        replsub (subst (dot (subst s c2) s) T2) with (subst s (subst1 c2 T2)).
        eapply subst_subs; eauto using subof_aof_sof; done.

      #2 destruct Hhalts as [Hhaltapp | ((HP1, HP2), Hhaltapp)].
        so (appgn_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
        left; auto; done.

        so (appgn_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
        right; auto; done.

    #2 destruct Hhalts as [Hhaltapp | ((HP1, HP2), Hhaltapp)].
      so (appgn_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
      left; auto; done.

      so (appgn_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
      right; auto; done.
Qed.


Lemma keqm_appgn_beta_fstm_eqc :
  forall G T M1 M2 k c2 c2',
    fstm G M2 c2
    -> fsts T k
    -> eqc G c2 c2' k
    -> keqm G (md_appgn (md_lamgn T M1) M2) (subst (dot M2 (dot c2' id)) M1) (fun s => mhalt nil (subst s M2)).
Proof.
intros G T1 M1 M2 k1 c c' HfstM2 Hfst1 Heqc.
intros p T Hwf Hof.
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & ca & Hoflam & HofM2 & HfstM2a & Hsubs & ->).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamgn_ofm_inversion _#5 Hwf Hoflam) as (T2 & k1a & Hsubspi & Hfst1a & HofM1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (Declarative.subs_pigen_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (l & HfstT).
so (ofs_fst _#3 (subs_reg _#3 Hwf Hsubs1 andel) Hfst1') as Hofk1'.
so (subs_reg _#3 (wf_cn Hwf Hofk1') Hsubs2) as (HofT2 & HofT2').
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (can_fsts _ _ HofT2') as (k2' & Hfst2').
so (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1') as Hofc.
so (subs_fst _#5 Hwf Hsubs1 Hfst1' Hfst1) as Hsubk1.
assert (ceqm G impure (subst (dot M2 (dot c id)) M1) (subst (dot M2 (dot c' id)) M1) T) as Hceq.
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; apply (mcc_subsume _#8 (subst1 c T2)).
    replsub (subst1 c T2) with (subst (dot M2 (dot c id)) (shift1 T2)).
    #2 refine (funct_ceq_md _#7 Hwf _ _ HofM1).
      apply wf_md; auto using wf; [].
      exact (subs_reg _#3 Hwf Hsubs1 ander).
  
      #2 prove_subrel; auto.
        exact (eqc_reg _#4 Hwf Heqc andel).
  
        apply ceqm_refl; [].
        eapply ofm_subsume; eauto; done.
  
    eapply subst1_subs; eauto; done.
do2 2 split.
  exact (ceqm_reg _#5 Hwf Hceq ander).

  intros Heq.
  discriminate Heq.

  intros s Hs Hhalts.
  simp_sub in Hhalts.
  assert (mhalt nil (subst s M2)) as Hhalt2.
    destruct Hhalts as [Hhaltapp | (?, _)]; auto; [].
    exact (appgn_halt_inversion' _ _ Hhaltapp ander).
  #2 eapply ceqm_trans.
    simp_sub.
    #3 eapply ceqm_appgn_beta.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
      exact H.
  
      apply Hhalt2; auto; done.
  
      eapply subst_fstm; eauto using subof_aof_tof; done.
  
    simp_sub.
    so (subst_ceqm _#7 Hs Hwf Hceq) as H; simp_sub in H.
    exact H.
Qed.


Lemma keqm_appap :
  forall G P1 P2 M1 M1' M2 M2',
    keqm G M1 M1' P1
    -> keqm G M2 M2' P2
    -> keqm G (md_appap M1 M2) (md_appap M1' M2') (andp P1 P2).
Proof.
intros G P1 P2 M1 M1' M2 M2' Hs1 Hs2.
intros p T Hwf Hof.
so (appap_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c2 & HofM1 & HofM2 & HfstM2 & Hsubs).
so (Hs1 _ _ Hwf HofM1) as (HofM1' & Hfsteq1 & Hceqif1).
so (Hs2 _ _ Hwf HofM2) as (HofM2' & Hfsteq2 & Hceqif2).
so (Hfsteq2 (eq_refl _)) as (c2a & c2' & k1 & HfstM2a & HfstM2' & Hfst1 & Heqc2).
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
invert (ofm_reg _#4 Hwf HofM1); [].
intros k1a HofT1 Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
do2 2 split.
  eapply ofm_subsume; eauto; [].
  #2 eapply ofm_subsume.
    eapply ofm_appap; eauto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply funct_sg; eauto; done.

  intros ->.
  so (Hfsteq1 (eq_refl _)) as (c1 & c1' & l & HfstM1 & HfstM1' & Hfstpi & Hceq1).
  invert Hfstpi; clear Hfstpi; [].
  intros k1a k2 Hfst1a Hfst2 <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  so (can_fsts _ _ (subs_reg _#3 Hwf Hsubs ander)) as (k & Hfst).
  so (subs_fst _#5 Hwf Hsubs (subst_fsts _#3 Hfst2) Hfst) as Hsubk.
  do 3 eexists; do2 3 split; eauto with static; done.

  intros s Hs Hhalts.
  simp_sub in Hhalts.
  simp_sub.
  lapply (Hceqif1 _ Hs).
    intro Hceq1.
    lapply (Hceqif2 _ Hs).
      intro Hceq2.
      #2 apply ceqm_comp; eapply mcc_subsume.
        #3 apply ceqm_comp; eapply mcc_appap.
          simp_sub in Hceq1.
          exact Hceq1.
    
          exact Hceq2.
    
          eapply subst_fstm; eauto using subof_aof_tof; done.
    
        simp_sub.
        replsub (subst (dot (subst s c2) s) T2) with (subst s (subst1 c2 T2)).
        eapply subst_subs; eauto using subof_aof_sof; done.

      #2 destruct Hhalts as [Hhaltapp | ((HP1, HP2), Hhaltapp)].
        so (appap_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
        left; auto; done.

        so (appap_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
        right; auto; done.

    #2 destruct Hhalts as [Hhaltapp | ((HP1, HP2), Hhaltapp)].
      so (appap_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
      left; auto; done.

      so (appap_halt_inversion' _ _ Hhaltapp) as (Hhalt1 & Hhalt2).
      right; auto; done.
Qed.


Lemma keqm_appap_beta_fstm_eqc :
  forall G T M1 M2 k c2 c2',
    fstm G M2 c2
    -> fsts T k
    -> eqc G c2 c2' k
    -> keqm G (md_appap (md_lamap T M1) M2) (subst (dot M2 (dot c2' id)) M1) (fun s => mhalt nil (subst s M2)).
Proof.
intros G T1 M1 M2 k1 c c' HfstM2 Hfst1 Heqc.
intros p T Hwf Hof.
so (appap_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & ca & Hoflam & HofM2 & HfstM2a & Hsubs).
so (fstm_fun _#4 HfstM2 HfstM2a); subst ca; clear HfstM2a.
so (lamap_ofm_inversion _#5 Hwf Hoflam) as (T2 & k1a & Hsubspi & Hfst1a & HofM1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (Declarative.subs_piapp_invert'' _#5 Hwf Hsubspi) as (k1' & Hsubs1 & Hfst1' & Hsubs2).
so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (l & HfstT).
so (can_fstm_shift _#5 (wf_md Hwf (subs_reg _#3 Hwf Hsubs1 ander) Hfst1) HofM1) as (d & HfstM1).
so (ofs_fst _#3 (subs_reg _#3 Hwf Hsubs1 andel) Hfst1') as Hofk1'.
so (subs_reg _#3 (wf_cn Hwf Hofk1') Hsubs2) as (HofT2 & HofT2').
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (can_fsts _ _ HofT2') as (k2' & Hfst2').
so (ofm_fst _#6 (wf_md Hwf (subs_reg _#3 Hwf Hsubs1 ander) Hfst1) HofM1 HfstM1 (subst_fsts _#3 Hfst2)) as Hofds.
so (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1') as Hofc.
so (subs_fst _#5 Hwf Hsubs1 Hfst1' Hfst1) as Hsubk1.
assert (ceqm G pure (subst (dot M2 (dot c id)) M1) (subst (dot M2 (dot c' id)) M1) T) as Hceq.
  apply ceqm_comp; eapply mcc_subsume; eauto; [].
  #2 apply ceqm_comp; apply (mcc_subsume _#8 (subst1 c T2)).
    replsub (subst1 c T2) with (subst (dot M2 (dot c id)) (shift1 T2)).
    #2 refine (funct_ceq_md _#7 Hwf _ _ HofM1).
      apply wf_md; auto using wf; [].
      exact (subs_reg _#3 Hwf Hsubs1 ander).
  
      #2 prove_subrel; auto.
        exact (eqc_reg _#4 Hwf Heqc andel).
  
        apply ceqm_refl; [].
        eapply ofm_subsume; eauto; done.
  
    eapply subst1_subs; eauto; done.
do2 2 split.
  exact (ofm_forget_pure (ceqm_reg _#5 Hwf Hceq ander)).

  intros ->.
  so (ceqm_fst _#4 Hwf Hceq) as (la & b & b' & HfstTa & Hfstb & Hfstb' & Heqb).
  so (fsts_fun _#3 HfstT HfstTa); subst la; clear HfstTa.
  do 3 eexists; do2 3 split.
    eapply fstm_appap; eauto with static; done.

    exact Hfstb'.
    
    exact HfstT.

    eapply eqc_trans; eauto; [].
    clear b' Hfstb' Heqb.
    #2 sor (refine (subst_fstm G _ (dot M2 (dot c id)) _ _ _ HfstM1)) as HfstM1x.
      prove_subof; auto; done.
    so (fstm_fun _#4 Hfstb HfstM1x); subst b.
    clear HfstM1x Hfstb.
    simp_sub.
    #2 eapply eqc_subsume.
      #2 eapply eqc_beta.
        so (subst_ofc _#5 (unshift_md _#3) Hofds) as Hofd.
        simp_sub in Hofd.
        exact Hofd.

        exact (eqc_reg _#4 Hwf Heqc andel).

      #2 eapply subk_trans.
        so (subs_fst _#5 (wf_cn Hwf Hofk1') Hsubs2 Hfst2 Hfst2') as Hsubk2.
        eapply subst1_subk; eauto; done.

        exact (subs_fst _#5 Hwf Hsubs (subst_fsts _#3 Hfst2') HfstT).

  intros s Hs Hhalts.
  simp_sub in Hhalts.
  assert (mhalt nil (subst s M2)) as Hhalt2.
    destruct Hhalts as [Hhaltapp | (?, _)]; auto; [].
    exact (appap_halt_inversion' _ _ Hhaltapp ander).
  #2 eapply ceqm_trans.
    assert (ofm G pure (md_appap (md_lamap T1 M1) M2) T) as Hof'.
      eapply ofm_subsume; eauto; [].
      eapply ofm_appap; eauto; [].
      eapply ofm_subsume; eauto; [].
      eapply ofm_lamap; eauto; [].
      exact (subs_reg _#3 Hwf Hsubs1 ander).
    simp_sub.
    #3 eapply ceqm_appap_beta.
      so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof') as H; simp_sub in H.
      apply ofm_forget_pure; [].
      exact H.
  
      apply Hhalt2; auto; done.
  
      eapply subst_fstm; eauto using subof_aof_tof; done.
  
    simp_sub.
    so (subst_ceqm _#7 Hs Hwf Hceq) as H; simp_sub in H.
    apply ceqm_forget_pure; [].
    exact H.
Qed.


Lemma keqm_unpack_beta :
  forall G c e k t M T,
    keqm G (md_unpack (tm_pack c e k t) M T) (subst (dot e (dot c id)) M) (fun s => thalt nil (subst s e)).
Proof.
intros G c e k t M T.
intros p U Hwf Hof.
so (unpack_ofm_inversion _#6 Hwf Hof) as (k' & t' & Hof' & HofM & Hsubs & ->).
so (pack_oft_inversion_exists _#7 Hwf Hof') as (Heqk & Heqt & Hofc & Hofe).
do2 2 split.
  eapply ofm_subsume; eauto; [].
  relquest.
    eapply subst_ofm; eauto; prove_subof; eauto with static; done.

    simp_sub; reflexivity.

  intro Heq.
  discriminate Heq.

  intros s Hs Hhalts.
  assert (thalt nil (subst s e)) as Hhalt.
    destruct Hhalts as [Hhalt' | (?, _)]; auto; [].
    simp_sub in Hhalt'.
    exact (pack_halt_inversion' _#4 (munpack_halt_inversion' _#3 Hhalt')).
  simp_sub.
  #2 eapply rel_quest_1.
    apply ceqm_unpack_beta; auto; [].
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
    exact H.

    simp_sub; reflexivity.
Qed.


Lemma keqm_lett :
  forall G P e e' M,
    keqt G e e' P
    -> keqm G (md_lett e M) (md_lett e' M) P.
Proof.
intros G P e e' M Hkeq.
intros p T Hwf Hof.
so (lett_ofm_inversion _#5 Hwf Hof) as (t & Hofe & HofM).
so (oft_reg _#3 Hwf Hofe) as Hoft.
so (Hkeq t Hwf Hofe) as (Hofe' & Hceqif).
do2 2 split.
  eapply ofm_lett; eauto; done.

  intros ->.
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (k & Hfst).
  so (can_fstm _#3 (wf_tm Hwf Hoft) HofM) as (c & HfstM).
  so (subst_fstm _#5 (subof_utm_tm_tof _ _) HfstM) as HfstM'.
  simp_sub in HfstM'.
  so (ofm_fst _#6 (wf_tm Hwf Hoft) HofM HfstM (subst_fsts _#3 Hfst)) as Hofc.
  so (subord_tm_cn _#4 Hofc) as Heq.
  assert (c = shift1 (subst1 nonsense c)) as Heq'.
    simp_sub; auto; done.
  exists (subst1 nonsense c), (subst1 nonsense c), k.
  do2 3 split.
    apply fstm_lett; [].
    rewrite <- Heq'; [].
    exact HfstM'.

    apply fstm_lett; [].
    rewrite <- Heq'; [].
    exact HfstM'.

    exact Hfst.

    apply eqc_refl; [].
    relquest.
      eapply subst_ofc; eauto; prove_subof; done.

      simp_sub; reflexivity.

  intros s Hs Hhalt.
  simp_sub.
  lapply (Hceqif _ Hs).
    intro Hceq.
    apply ceqm_comp; eapply mcc_lett; eauto; [].
    apply ceqm_refl; [].
    relquest.
      eapply subst_ofm; eauto; prove_subof; [].
      eapply subof_dof_trans; eauto using weaken1, subof_aof_dof; done.

      simp_sub; reflexivity.

    #2 destruct Hhalt as [Hhalt' | (HP, Hhalt')].
      left.
      simp_sub in Hhalt'.
      exact (mlett_halt_inversion' _ _ Hhalt').

      right.
      simp_sub in Hhalt'.
      split; auto; [].
      exact (mlett_halt_inversion' _ _ Hhalt').
Qed.


Lemma keqm_lett_beta :
  forall G e M,
    keqm G (md_lett e M) (subst1 e M) (fun s => thalt nil (subst s e)).
Proof.
intros G e M.
intros p T Hwf Hof.
so (lett_ofm_inversion _#5 Hwf Hof) as (t & Hofe & HofM).
do2 2 split.
  relquest.
    eapply subst_ofm; eauto; prove_subof; auto; done.

    simp_sub; reflexivity.

  intros ->.
  so (oft_reg _#3 Hwf Hofe) as Hoft.
  so (ofm_reg _#4 Hwf Hof) as HofT.
  so (can_fsts _ _ HofT) as (k & Hfst).
  so (can_fstm _#3 (wf_tm Hwf Hoft) HofM) as (c & HfstM).
  so (ofm_fst _#6 (wf_tm Hwf Hoft) HofM HfstM (subst_fsts _#3 Hfst)) as Hofc.
  so (subord_tm_cn _#4 Hofc) as Heq.
  assert (c = shift1 (subst1 nonsense c)) as Heq'.
    simp_sub; auto; done.
  rewrite -> Heq' in HfstM.
  rewrite -> Heq' in Hofc.
  clear Heq Heq'.
  forget (subst1 nonsense c) as d.
  so (subst_ofc _#5 (unshift_tm _ _) Hofc) as Hofd; simp_sub in Hofd.
  exists d, d, k.
  #2 do2 3 split; auto with static.
    apply fstm_lett; [].
    rewrite <- (subst_dot0_sh M); [].
    relquest.
      eapply subst_fstm; eauto; prove_subof; done.

      simp_sub; reflexivity.

    relquest.
      eapply subst_fstm; eauto; prove_subof; done.

      simp_sub; reflexivity.

  intros s Hs Hhalt.
  simp_sub.
  assert (thalt nil (subst s e)) as Hhalte.
    destruct Hhalt as [Hhalt' | (?, _)]; auto; [].
    simp_sub in Hhalt'.
    exact (mlett_halt_inversion' _ _ Hhalt').
  #2 eapply rel_quest_1.
    eapply ceqm_lett_beta; eauto; [].
    so (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) as H; simp_sub in H.
    exact H.

    simp_sub; reflexivity.
Qed.
