
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Determinism.
Require Import TSubst.
Require Import ProveSubof.
Require Import Formation.
Require Import Functionality.
Require Import Regularity.
Require Import Inversion.
Require Import Singleton.
Require Import Preservation.
Require Import Equivalence.
Require Import Translation.
Require Progress.



Lemma can_split :
  forall T,
    hsg T
    -> exists k t, split T k t.
Proof.
intros T HT.
#3 induct HT; eauto using split.
intros T1 T2 _ (k & t & H) _ (k' & t' & H'); eauto using split; done.
intros T1 T2 _ (k & t & H) _ (k' & t' & H'); eauto using split; done.
intros T1 T2 _ (k & t & H) _ (k' & t' & H'); eauto using split; done.
Qed.


Lemma split_fsts :
  forall T k t,
    split T k t
    -> fsts T k.
Proof.
intros T k t Hsplit.
induct Hsplit; intros; auto using fsts.
Qed.


Lemma split_fsts_eq :
  forall T k t k',
    split T k t
    -> fsts T k'
    -> k = k'.
Proof.
intros T k t k' Hsplit Hfsts.
so (split_fsts _#3 Hsplit) as Hfsts'.
eapply fsts_fun; eauto; done.
Qed.  


Lemma split_fun1 :
  forall T k k' t t',
    split T k t
    -> split T k' t'
    -> k = k'.
Proof.
intros T k k' t t' Hsplit Hsplit'.
exact (fsts_fun _#3 (split_fsts _#3 Hsplit) (split_fsts _#3 Hsplit')).
Qed.


Lemma split_fun2 :
  forall T k k' t t',
    split T k t
    -> split T k' t'
    -> t = t'.
Proof.
intros T k k' t t' Hsplit Hsplit'.
revert k' t' Hsplit'.
#3 induct Hsplit;
try (intros; invert Hsplit'; intros; subst; auto; done).

(* pigen *)
intros T1 T2 k1 k2 t1 t2 Hsplit1 IH1 Hsplit2 IH2 k' t' Hsplit'.
invert Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fun1 _#5 Hsplit1 Hsplit1'); subst k1'.
so (split_fun1 _#5 Hsplit2 Hsplit2'); subst k2'.
so (IH1 _ _ Hsplit1'); subst t1'.
so (IH2 _ _ Hsplit2'); subst t2'.
reflexivity.

(* piapp *)
intros T1 T2 k1 k2 t1 t2 Hsplit1 IH1 Hsplit2 IH2 k' t' Hsplit'.
invert Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fun1 _#5 Hsplit1 Hsplit1'); subst k1'.
so (split_fun1 _#5 Hsplit2 Hsplit2'); subst k2'.
so (IH1 _ _ Hsplit1'); subst t1'.
so (IH2 _ _ Hsplit2'); subst t2'.
reflexivity.

(* sigma *)
intros T1 T2 k1 k2 t1 t2 _ IH1 _ IH2 k' t' Hsplit'.
invert Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (IH1 _ _ Hsplit1'); subst t1'.
so (IH2 _ _ Hsplit2'); subst t2'.
reflexivity.
Qed.


Lemma split_fun :
  forall T k k' t t',
    split T k t
    -> split T k' t'
    -> k = k' /\ t = t'.
Proof.
intros; eauto using split_fun1, split_fun2.
Qed.


Lemma split_formation :
  forall G T k t,
    ofs G T
    -> split T k t
    -> ofk G k
       /\ ofc (G; cl_cn k) t kd_type.
Proof.
intros G T k t HofT Hsplit.
revert k t Hsplit.
induct HofT.

(* one *)
intros G k t Hsplit.
invert Hsplit; [].
intros <- <-.
auto with static; done.

(* stat *)
intros G k Hofk k' t Hsplit.
invert Hsplit; [].
intros <- <-.
auto with static; done.

(* dyn *)
intros G t Hoft k t' Hsplit.
invert Hsplit; [].
intros <- <-.
split; auto using weaken1_ofc_type with static; done.

(* pigen *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 k t Hsplit.
invert Hsplit; clear Hsplit; [].
intros k1' k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1'.
so (IH1 _ _ Hsplit1) as (Hofk1 & Hoft1).
so (IH2 _ _ Hsplit2) as (Hofk2 & Hoft2).
split; auto with static; [].
#2 apply ofc_all.
  apply weaken1_ofk; auto; done.

  #2 apply ofc_arrow.
    eapply subst_ofc_type; eauto; [].
    apply subof_bind_sof_cn; apply subof_sh; auto with sequence; done.

    #2 apply ofc_exists.
      eapply subst_ofk; eauto; [].
      apply subof_bind_sof_cn; apply subof_sh; auto with sequence; done.

      eapply subst_ofc_type; eauto; [].
      apply subof_bind_sof_cn; [].
      apply subof_bind_sof_cn; apply subof_sh; auto with sequence; done.

(* piapp *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 k t Hsplit.
invert Hsplit; clear Hsplit; [].
intros k1' k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1'.
so (IH1 _ _ Hsplit1) as (Hofk1 & Hoft1).
so (IH2 _ _ Hsplit2) as (Hofk2 & Hoft2).
split; auto with static; [].
#2 apply ofc_all.
  apply weaken1_ofk; auto; done.

  #2 apply ofc_arrow.
    eapply subst_ofc_type; eauto; [].
    apply subof_bind_sof_cn; apply subof_sh; auto with sequence; done.

    eapply subst_ofc_type; eauto; [].
    #2 apply subof_dot.
      apply subof_dot; auto with static sequence; [].
      simp_sub.
      apply sof_cn; [].
      #2 eapply rel_quest.
        eapply ofc_var; eauto with sequence; done.

        simp_sub; reflexivity.

      simp_sub.
      apply sof_cn; [].
      #2 eapply rel_quest.
        #2 eapply ofc_app.
          #2 eapply rel_quest.
            eapply ofc_var; eauto with sequence; done.

            simp_sub.
            reflexivity.

          #2 eapply rel_quest.
            eapply ofc_var; eauto with sequence; done.

            simp_sub.
            reflexivity.

        simp_sub.
        reflexivity.

(* sigma *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 k t Hsplit.
invert Hsplit; clear Hsplit; [].
intros k1' k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1'.
so (IH1 _ _ Hsplit1) as (Hofk1 & Hoft1).
so (IH2 _ _ Hsplit2) as (Hofk2 & Hoft2).
split; auto with static; [].
#2 apply ofc_prod.
  eapply subst_ofc_type; eauto; [].
  #2 apply subof_dot.
    apply subof_sh; auto with sequence; done.

    simp_sub.
    apply sof_cn; [].
    eapply ofc_pi1; [].
    #2 eapply rel_quest.
      eapply ofc_var; eauto with sequence; done.

      simp_sub; reflexivity.

  eapply subst_ofc_type; eauto; [].
  #2 apply subof_dot.
    #2 apply subof_dot.
      apply subof_sh; auto with sequence; done.

      simp_sub.
      apply sof_cn; [].
      eapply ofc_pi1; [].
      #2 eapply rel_quest.
        eapply ofc_var; eauto with sequence; done.

        simp_sub; reflexivity.

    simp_sub.
    apply sof_cn; [].
    #2 eapply rel_quest.
      eapply ofc_pi2; [].
      #2 eapply rel_quest.
        eapply ofc_var; eauto with sequence; done.

        simp_sub.
        reflexivity.

      simp_sub.
      reflexivity.
Qed.


Lemma subst_split :
  forall T k t s,
    split T k t
    -> split (subst s T) (subst s k) (subst (under 1 s) t).
Proof.
intros T k t s Hsplit.
revert s.
induct Hsplit;
intros; simp_sub; (eapply rel_quest; [eauto using split | simp_sub; reflexivity]).
Qed.


Lemma split_eqs :
  forall G T T' k k' t t',
    wf G
    -> eqs G T T'
    -> split T k t
    -> split T' k' t'
    -> eqk G k k' /\ eqc (G; cl_cn k) t t' kd_type.
Proof.
intros G T T' k k' t t' Hwf Heqs Hsplit Hsplit'.
revert k k' t t' Hwf Hsplit Hsplit'.
induct Heqs.

(* refl *)
intros G T HofT k k' t t' Hwf Hsplit Hsplit'.
so (split_fun _#5 Hsplit Hsplit') as (<- & <-).
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
split; auto with static; done.

(* symm *)
intros G T T' _ IH k k' t t' Hwf Hsplit Hsplit'.
so (IH _#4 Hwf Hsplit' Hsplit) as (Heqk & Heqt).
split; auto with static; [].
so (subst_eqc _#6 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqt) as Heqt'.
simp_sub in Heqt'.
apply eqc_symm; auto; done.

(* trans *)
intros G T1 T2 T3 HeqT12 IH1 _ IH2 k1 k3 t1 t3 Hwf Hsplit1 Hsplit3.
so (can_split _ (ofs_hsg _ _ (eqs_reg _#3 Hwf HeqT12 ander))) as (k2 & t2 & Hsplit2).
so (IH1 _#4 Hwf Hsplit1 Hsplit2) as (Heqk12 & Heqt12).
so (IH2 _#4 Hwf Hsplit2 Hsplit3) as (Heqk23 & Heqt23).
split; eauto using eqk_trans; [].
eapply eqc_trans; eauto; [].
so (subst_eqc _#6 (subof_conv_0 _#3 Heqk12) Heqt23) as Heqt23'; simp_sub in Heqt23'.
exact Heqt23'.

(* stat *)
intros G k k' Heqk ka ka' t t' Hwf Hsplit Hsplit'.
invert Hsplit; [].
intros <- <-.
invert Hsplit'; [].
intros <- <-.
split; auto with static; done.

(* dyn *)
intros G t t' Heqt k k' ta ta' Hwf Hsplit Hsplit'.
invert Hsplit; [].
intros <- <-.
invert Hsplit'; [].
intros <- <-.
split; auto using weaken1_eqc_type with static; done.

(* pigen *)
intros G T1 T1' k1 T2 T2' _ IH1 Hfst1 _ IH2 k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (IH1 _#4 Hwf Hsplit1 Hsplit1') as (Heqk1 & Heqt1).
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (IH2 _#4 (wf_cn Hwf Hofk1) Hsplit2 Hsplit2') as (Heqk2 & Heqt2).
split; auto with static; [].
cut (eqc G (cn_all k1 (cn_arrow t1 (cn_exists k2 t2))) (cn_all k1' (cn_arrow t1' (cn_exists k2' t2'))) kd_type).
  intros Heqc.
  so (subst_eqc _#6 (weaken1 _ _ (cl_cn kd_one)) Heqc) as H; simp_sub in H.
  exact H.
eauto with static; done.

(* piapp *)
intros G T1 T1' k1 T2 T2' _ IH1 Hfst1 _ IH2 k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (IH1 _#4 Hwf Hsplit1 Hsplit1') as (Heqk1 & Heqt1).
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (IH2 _#4 (wf_cn Hwf Hofk1) Hsplit2 Hsplit2') as (Heqk2 & Heqt2).
split; auto with static; [].
#2 apply eqc_all.
  eapply subst_eqk; eauto using weaken1; done.

  #2 apply eqc_arrow.
    eapply subst_eqc_type; eauto; [].
    simp_sub.
    prove_subof; done.

    eapply subst_eqc_type; eauto; [].
    prove_subof; [].
    relquest.
      eapply ofc_app; ofvar; done.
      
      simp_sub; reflexivity.

(* sigma *)
intros G T1 T1' k1 T2 T2' _ IH1 Hfst1 _ IH2 k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (IH1 _#4 Hwf Hsplit1 Hsplit1') as (Heqk1 & Heqt1).
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (IH2 _#4 (wf_cn Hwf Hofk1) Hsplit2 Hsplit2') as (Heqk2 & Heqt2).
split; auto with static; [].
#2 apply eqc_prod.
  eapply subst_eqc_type; eauto; [].
  prove_subof; [].
  eapply ofc_pi1; ofvar; done.

  eapply subst_eqc_type; eauto; [].
  #2 prove_subof.
    eapply ofc_pi1; ofvar; done.

    relquest.
      eapply ofc_pi2; ofvar; done.

      simp_sub; reflexivity.
Qed.


Inductive snd : (* sg *) term -> (* md *) term -> (* tm *) term -> Prop :=
| snd_one {M}
    : snd sg_one M (tm_letm M tm_star)

| snd_stat {k M}
    : snd (sg_stat k) M (tm_letm M tm_star)

| snd_dyn {t M}
    : snd (sg_dyn t) M (tm_ext M)

| snd_pigen {T1 T2 M k1 t1 k2 t2 N e}
    : split T1 k1 t1
      -> split T2 k2 t2
      -> join (subst (sh 4) T1) (var 1) (var 0) N
      -> snd (subst (dot (var 3) (sh 6)) T2) (var 0) e
      -> snd (sg_pigen T1 T2) M
         (tm_letm M
            (tm_plam 
               (subst (sh 2) k1)
               (tm_lam
                  (subst (under 1 (sh 2)) t1)
                  (tm_letm
                     (md_appgn (var 2) N)
                     (tm_pack 
                        (var 1)
                        e
                        (subst (dot (var 3) (sh 6)) k2)
                        (subst (dot (var 0) (dot (var 4) (sh 7))) t2))))))

| snd_piapp {T1 T2 M k1 t1 N e}
    : split T1 k1 t1
      -> join (subst (sh 4) T1) (var 1) (var 0) N
      -> snd (subst (dot (var 1) (sh 4)) T2) (md_appap (var 2) N) e
      -> snd (sg_piapp T1 T2) M
         (tm_letm M
            (tm_plam
               (subst (sh 2) k1)
               (tm_lam (subst (under 1 (sh 2)) t1) e)))

| snd_sigma {T1 T2 M e1 e2}
    : snd (subst (sh 2) T1) (md_pi1 (var 0)) e1
      -> snd (subst (dot (cn_pi1 (var 1)) (sh 2)) T2) (md_pi2 (var 0)) e2
      -> snd (sg_sigma T1 T2) M (tm_letm M (tm_pair e1 e2))
                

with join : (* sg *) term -> (* cn *) term -> (* tm *) term -> (* md *) term -> Prop :=
| join_one {c e}
    : join sg_one c e (md_lett e md_star)

| join_stat {k c e}
    : join (sg_stat k) c e (md_lett e (md_stat (shift1 c)))

| join_dyn {t c e}
    : join (sg_dyn t) c e (md_dyn e)

| join_pigen {T1 T2 c e f M}
    : snd (subst (sh 3) T1) (var 0) f
      -> join (subst (dot (var 3) (sh 5)) T2) (var 1) (var 0) M
      -> join (sg_pigen T1 T2) c e
         (md_lett e
            (md_lamgn
               (shift1 T1)
               (md_unpack
                  (tm_app (tm_papp (var 2) (var 1)) f)
                  M
                  (subst (dot (var 1) (sh 3)) T2))))

| join_piapp {T1 T2 c e f M}
    : snd (subst (sh 3) T1) (var 0) f
      -> join
           (subst (dot (var 1) (sh 3)) T2)
           (cn_app (subst (sh 3) c) (var 1))
           (tm_app (tm_papp (var 2) (var 1)) f)
           M
      -> join (sg_piapp T1 T2) c e
         (md_lett e (md_lamap (shift1 T1) M))

| join_sigma {T1 T2 c e M1 M2}
    : join (shift1 T1) (cn_pi1 (shift1 c)) (tm_pi1 (var 0)) M1
      -> join (subst (dot (cn_pi1 (shift1 c)) sh1) T2) (cn_pi2 (shift1 c)) (tm_pi2 (var 0)) M2
      -> join (sg_sigma T1 T2) c e (md_lett e (md_pair M1 M2))
.


Scheme snd_mut_ind := Minimality for snd Sort Prop
with   join_mut_ind := Minimality for join Sort Prop.
Combined Scheme snd_join_ind from snd_mut_ind, join_mut_ind.


Lemma sg_ind_subst :
  forall (P : (* sg *) term -> Prop),
    P sg_one
    -> (forall k, P (sg_stat k))
    -> (forall t, P (sg_dyn t))
    -> (forall T1 T2,
          (forall s, P (subst s T1))
          -> (forall s, P (subst s T2))
          -> P (sg_pigen T1 T2))
    -> (forall T1 T2,
          (forall s, P (subst s T1))
          -> (forall s, P (subst s T2))
          -> P (sg_piapp T1 T2))
    -> (forall T1 T2,
          (forall s, P (subst s T1))
          -> (forall s, P (subst s T2))
          -> P (sg_sigma T1 T2))
    -> (forall T, hsg T -> P T).
Proof.
intros P H1 H2 H3 H4 H5 H6 T HT.
cut (forall s, P (subst s T)).
  intro H.
  so (H id) as H'.
  simp_sub in H'.
  exact H'.
induct HT;
intros; simp_sub; auto; [apply H4 | apply H5 | apply H6]; intros; simp_sub; auto.
Qed.


Lemma sg_ind_subst_hsg :
  forall (P : (* sg *) term -> Prop),
    P sg_one
    -> (forall k, P (sg_stat k))
    -> (forall t, P (sg_dyn t))
    -> (forall T1 T2,
          hsg T1
          -> (forall s, P (subst s T1))
          -> hsg T2
          -> (forall s, P (subst s T2))
          -> P (sg_pigen T1 T2))
    -> (forall T1 T2,
          hsg T1
          -> (forall s, P (subst s T1))
          -> hsg T2
          -> (forall s, P (subst s T2))
          -> P (sg_piapp T1 T2))
    -> (forall T1 T2,
          hsg T1
          -> (forall s, P (subst s T1))
          -> hsg T2
          -> (forall s, P (subst s T2))
          -> P (sg_sigma T1 T2))
    -> (forall T, hsg T -> P T).
Proof.
intros P H1 H2 H3 H4 H5 H6 T HT.
cut (forall s, P (subst s T)).
  intro H.
  so (H id) as H'.
  simp_sub in H'.
  exact H'.
induct HT;
intros; simp_sub; auto; [apply H4 | apply H5 | apply H6]; intros; simp_sub; auto using subst_hsg.
Qed.


Lemma can_snd_join :
  (forall T M,
     hsg T
     -> exists e, snd T M e)
  /\
  (forall T c e,
     hsg T
     -> exists M, join T c e M).
Proof.
#4 exploit
  (sg_ind_subst_hsg
     (fun T =>
        (forall M, exists e, snd T M e)
        /\ (forall c e, exists M, join T c e M))) as Hprop;
try (intros; split; intros; eauto using snd, join; done).

(* pigen *)
intros T1 T2 HT1 IH1 HT2 IH2.
split.
  intros M.
  so (can_split _ HT1) as (k1 & t1 & Hsplit1).
  so (can_split _ HT2) as (k2 & t2 & Hsplit2).
  so (IH1 (sh 4) ander (var 1) (var 0)) as (N & Hjoin).
  so (IH2 (dot (var 3) (sh 6)) andel (var 0)) as (e & Hsnd).
  eexists; eapply snd_pigen; eauto; done.

  intros c e.
  so (IH1 (sh 3) andel (var 0)) as (f & Hsnd).
  so (IH2 (dot (var 3) (sh 5)) ander (var 1) (var 0)) as (M & Hjoin).
  eexists; eapply join_pigen; eauto; done.

(* piapp *)
intros T1 T2 HT1 IH1 HT2 IH2.
split.
  intros M.
  so (can_split _ HT1) as (k1 & t1 & Hsplit1).
  so (IH1 (sh 4) ander (var 1) (var 0)) as (N & Hjoin).
  so (IH2 (dot (var 1) (sh 4)) andel (md_appap (var 2) N)) as (e & Hsnd).
  eexists; eapply snd_piapp; eauto; done.

  intros c e.
  so (IH1 (sh 3) andel (var 0)) as (f & Hsnd).
  so (IH2 (dot (var 1) (sh 3)) ander (cn_app (subst (sh 3) c) (var 1)) (tm_app (tm_papp (var 2) (var 1)) f)) as (M & Hjoin).
  eexists; eapply join_piapp; eauto; done.

(* sigma *)
intros T1 T2 HT1 IH1 HT2 IH2.
split.
  intros M.
  so (IH1 (sh 2) andel (md_pi1 (var 0))) as (e1 & Hsnd1).
  so (IH2 (dot (cn_pi1 (var 1)) (sh 2)) andel (md_pi2 (var 0))) as (e2 & Hsnd2).
  eexists; eapply snd_sigma; eauto; done.

  intros c e.
  so (IH1 sh1 ander (cn_pi1 (shift1 c)) (tm_pi1 (var 0))) as (M1 & Hjoin1).
  so (IH2 (dot (cn_pi1 (shift1 c)) sh1) ander (cn_pi2 (shift1 c)) (tm_pi2 (var 0))) as (M2 & Hjoin2).
  eexists; eapply join_sigma; eauto; done.

(* epilogue *)
split.
  intros T M HT.
  exact (Hprop _ HT andel M).

  intros T c e HT.
  exact (Hprop _ HT ander c e).
Qed.    


Lemma can_snd :
  forall T M,
    hsg T
    -> exists e, snd T M e.
Proof.
exact (can_snd_join andel).
Qed.


Lemma can_join :
  forall T c e,
    hsg T
    -> exists M, join T c e M.
Proof.
exact (can_snd_join ander).
Qed.


Lemma snd_and_join_formation :
  (forall G T M e k t c,
     wf G
     -> ofm G pure M T
     -> snd T M e
     -> split T k t
     -> fstm G M c
     -> oft G e (subst1 c t))
  /\
  (forall G T c e M k t,
     wf G
     -> ofs G T
     -> join T c e M
     -> split T k t
     -> ofc G c k
     -> oft G e (subst1 c t)
     -> ofm G pure M T)
  /\
  (forall G T c e M k t,
     wf G
     -> ofs G T
     -> join T c e M
     -> split T k t
     -> ofc G c k
     -> oft G e (subst1 c t)
     -> exists c', fstm G M c' /\ eqc G c c' k).
Proof.
exploit (sg_ind_subst
           (fun T => 
              forall G k t,
                wf G
                -> ofs G T
                -> split T k t
                -> (forall M e c,
                      ofm G pure M T
                      -> snd T M e
                      -> fstm G M c
                      -> oft G e (subst1 c t))
                   /\
                   (forall c e M,
                      ofc G c k
                      -> oft G e (subst1 c t)
                      -> join T c e M
                      -> ofm G pure M T
                         /\ (exists c', fstm G M c' /\ eqc G c c' k)))) as Hprop.

(* one *)
intros G k t Hwf HofT Hsplit.
invert Hsplit; [].
intros <- <-.
split.
  intros M e c HofM Hsnd HfstM.
  invert Hsnd; clear Hsnd; [].
  intros <-.
  simp_sub.
  eauto with static; done.

  intros c e M Hofc Hofe Hjoin.
  invert Hjoin; clear Hjoin; [].
  intros <-.
  split.
    eauto with static; done.

    exists cn_star.
    split.
      apply fstm_lett; [].
      simp_sub.
      apply fstm_star; done.

      apply eqc_extone; auto with static; done.

(* stat *)
intros ka G k' t Hwf HofT Hsplit.
invert HofT; clear HofT; [].
intro Hofk.
invert Hsplit; clear Hsplit; [].
intros <- <-.
split.
  intros M e c HofM Hsnd HfstM.
  invert Hsnd; clear Hsnd; [].
  intros <-.
  simp_sub.
  eauto with static; done.

  intros c e M Hofc Hofe Hjoin.
  invert Hjoin; clear Hjoin; [].
  intros <-.
  split.
    eapply ofm_lett; eauto; [].
    simp_sub.
    apply ofm_stat; [].
    apply weaken1_ofc; auto; done.

    exists c.
    split; auto with static; done.

(* dyn *)
intros ta G k t' Hwf HofT Hsplit.
invert Hsplit; clear Hsplit; [].
intros <- <-.
split.
  intros M e c HofM Hsnd Hfstm.  
  invert Hsnd; clear Hsnd; [].
  intros <-.
  simp_sub.
  auto with static; done.

  intros c e M Hofc Hofe Hjoin.
  invert Hjoin; clear Hjoin; [].
  intros <-.
  simp_sub in Hofe.
  split.
    auto with static; done.

    exists cn_star.
    auto with static; done.

(* pigen *)
intros T1 T2 IH1 IH2 G k t Hwf HofT Hsplit.
invert HofT; clear HofT; [].
intros k1 HofT1 Hfst1 HofT2.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (split_fsts _#3 Hsplit2) as Hfst2.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
split.
  intros M e c HofM HsndM HfstM.
  invert HsndM; clear HsndM; [].
  intros k1a t1a k2a t2a N d Hsplit1a Hsplit2a Hjoin Hsnd <-.
  destruct (split_fun _#5 Hsplit1 Hsplit1a); subst k1a t1a.
  destruct (split_fun _#5 Hsplit2 Hsplit2a); subst k2a t2a.
  clear Hsplit1a Hsplit2a.
  assert (wf (G; cl_cn kd_one; cl_md (shift1 (sg_pigen T1 T2)) (var 0); cl_cn (subst (sh 2) k1); cl_tm (subst (under 1 (sh 2)) t1))) as Hwf'.
    #2 apply wf_tm.
      #2 apply wf_cn.
        apply wf_md; eauto with static; done.

        eapply subst_ofk; eauto; prove_subof; done.

      eapply subst_ofc_type; eauto; [].
      simp_sub; prove_subof; done.
  #5 sor (refine (IH1 _#4 Hwf' _ _ ander _#5 Hjoin)) as (HofN & (b & HfstN & Heqb)); clear IH1 Hjoin.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eauto; done.

    ofvar; done.
    
    ofvar; done.
  assert (wf (G; cl_cn kd_one; cl_md (shift1 (sg_pigen T1 T2)) (var 0); cl_cn (subst (sh 2) k1); cl_tm (subst (under 1 (sh 2)) t1); cl_cn (subst (dot (var 1) (sh 4)) k2); cl_md (shift1 (subst (dot (var 1) (sh 4)) T2)) (var 0))) as Hwf''.
    #2 apply wf_md; auto.
      eapply subst_ofs; eauto; prove_subof; done.

      apply subst_fsts; eauto; done.
  #5 sor (refine (IH2 _#4 Hwf'' _ _ andel _#4 Hsnd _)) as Hofd; clear IH2 Hsnd.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eauto; done.

    ofvar; done.

    ofvar; done.
  #4 eapply oft_letm.
    eapply ofm_forget; eauto; done.

    apply fsts_pigen; eauto; done.

    simp_sub.
    #2 apply oft_plam; auto using svalue.
      eapply subst_ofk; eauto; prove_subof; done.

      #2 apply oft_lam.
        eapply subst_ofc_type; eauto; prove_subof; done.

        #4 eapply oft_letm.
          #2 eapply (@ofm_subsume _ _ _ (subst (dot (var 1) (sh 4)) T2)).
            #3 eapply ofm_appgn.
              apply ofm_forget; ofvar; done.
  
              exact HofN.
  
              exact HfstN.
              
            apply subs_refl; [].
            simp_sub.
            eapply seq_ofs; eauto; [].
            prove_subrel; [].
            apply eqc_symm; [].
            exact Heqb.

          simp_sub.
          apply subst_fsts; eauto; done.

          simp_sub.
          #3 apply oft_pack.
            ofvar; done.

            simp_sub.
            simp_sub in Hofd.
            exact Hofd.

            eapply subst_ofc_type; eauto; prove_subof; done.

          simp_sub.
          #2 apply ofc_exists.
            eapply subst_ofk; eauto; prove_subof; done.

            eapply subst_ofc_type; eauto; prove_subof; done.

    simp_sub.
    simp_sub.
    apply ofc_all; auto; [].
    apply ofc_arrow; auto; [].
    apply ofc_exists; auto; [].
    rewrite -> subst_dot01_sh; [].
    exact Hoft2.

  intros c e M Hofc Hofe Hjoin.      
  invert Hjoin; clear Hjoin; [].
  intros f N Hsnd Hjoin <-.
  assert (wf (G; cl_tm (cn_all k1 (cn_arrow t1 (cn_exists k2 (subst (dot (var 0) (dot (var 1) (compose (sh 3) (dot (var 0) (dot (var 1) (dot (subst (sh 2) c) (sh 2))))))) t2)))); cl_cn (shift1 k1); cl_md (shift1 (shift1 T1)) (var 0))) as Hwf'.
    #3 apply wf_md.
      apply wf_tm; auto; [].
      apply ofc_all; auto; [].
      apply ofc_arrow; auto; [].
      apply ofc_exists; auto; [].
      simp_sub.
      rewrite -> subst_dot01_sh; [].
      exact Hoft2.

      apply weaken1_ofs; auto; done.

      apply subst_fsts; auto; done.
  #5 sor (refine (IH1 _#4 Hwf' _ _ andel _#4 Hsnd _)) as Hoff; clear IH1 Hsnd.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eauto; done.

    ofvar; done.

    ofvar; done.
  assert (wf (G; cl_tm (cn_all k1 (cn_arrow t1 (cn_exists k2 (subst (dot (var 0) (dot (var 1) (compose (sh 3) (dot (var 0) (dot (var 1) (dot (subst (sh 2) c) (sh 2))))))) t2)))); cl_cn (shift1 k1); cl_md (shift1 (shift1 T1)) (var 0); cl_cn (subst (dot (var 1) (sh 3)) k2); cl_tm (subst (dot (var 0) (dot (var 2) (sh 4))) t2))) as Hwf''.
    #2 apply wf_tm.
      apply wf_cn; auto; [].
      eapply subst_ofk; eauto; prove_subof; done.

      eapply subst_ofc_type; eauto; prove_subof; done.
  clear Hwf'.
  #5 sor (refine (IH2 _#4 Hwf'' _ _ ander _#5 Hjoin)) as (HofN & (b & HfstN & Heqb)); clear IH2 Hjoin.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eauto; done.

    ofvar; done.

    ofvar; done.
  split.
    #2 eapply ofm_lett.
      exact Hofe.

      simp_sub.
      #3 eapply ofm_lamgn.
        apply weaken1_ofs; auto; done.

        apply subst_fsts; eauto; done.

        relquest.
          #3 eapply ofm_unpack.
            relquest.
              #2 eapply oft_app.
                relquest.
                  #2 eapply oft_papp.
                    ofvar; done.

                    ofvar; done.

                  simp_sub; reflexivity.

                simp_sub in Hoff.
                simp_sub.
                exact Hoff.
                
              simp_sub; reflexivity.
              
            apply ofm_forget; [].
            simp_sub; simp_sub in HofN.
            exact HofN.

            eapply subst_ofs; eauto; prove_subof; done.

          simp_sub; reflexivity.

    exists cn_star.
    split.
      apply fstm_lett; [].
      simp_sub; apply fstm_lamgn; done.

      apply eqc_extone; auto with static; done.

(* piapp *)
intros T1 T2 IH1 IH2 G k t Hwf HofT Hsplit.
invert HofT; clear HofT; [].
intros k1 HofT1 Hfst1 HofT2.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (split_fsts _#3 Hsplit2) as Hfst2.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
split.
  intros M e c HofM HsndM HfstM.
  invert HsndM; clear HsndM; [].
  intros k1a t1a N d Hsplit1a Hjoin Hsnd <-.
  destruct (split_fun _#5 Hsplit1 Hsplit1a); subst k1a t1a.
  clear Hsplit1a.
  so (ofm_fst _#6 Hwf HofM HfstM (fsts_piapp Hfst1 Hfst2)) as Hofc.
  assert (wf (G; cl_cn (single c (kd_pi k1 k2)); cl_md (shift1 (singles c (sg_piapp T1 T2))) (var 0); cl_cn (subst (sh 2) k1); cl_tm (subst (under 1 (sh 2)) t1))) as Hwf'.
    #2 apply wf_tm.
      #2 apply wf_cn.
        apply wf_md; auto.
          eapply singles_ofs; eauto with static; done.

          apply singles_fst; auto with static; done.

        apply weaken2_ofk; auto; done.

      eapply subst_ofc_type; eauto; [].
      simp_sub; prove_subof; done.
  #5 sor (refine (IH1 _#4 Hwf' _ _ ander _#5 Hjoin)) as (HofN & (b & HfstN & Heqb)); clear IH1.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eassumption.

    ofvar; done.

    ofvar; done.
  so (eqc_reg _#4 Hwf' Heqb) as (_ & Hofb & _).
  #5 sor (refine (IH2 _#4 Hwf' _ _ andel _#4 Hsnd _)) as Hofd; clear IH2.
    eapply subst_ofs; eauto; prove_subof; done.

    apply subst_split; eassumption.

    #2 eapply ofm_subsume.
      eapply ofm_appap; eauto; [].
      simpl singles.
      simp_sub.
      ofvar; done.

      simp_sub.
      #2 apply (@subs_trans _ _ (subst (dot b (sh 4)) T2)).
        rewrite -> subst_singles; [| exact (ofs_hsg _ _ HofT2)].
        #2 eapply singles_subs; eauto using subst_fsts.
          eapply subst_ofs; eauto; [].
          #2 apply subof_dot.
            apply subof_sh; auto with sequence; done.

            simp_sub.
            apply sof_cn; [].
            simp_sub in Hofb.
            exact Hofb.

          simp_sub.
          relquest.
            #2 eapply ofc_app.
              relquest.
                eapply subst_ofc; eauto; [].
                apply subof_sh; auto with sequence; done.

                simp_sub; reflexivity.

              simp_sub in Hofb.
              exact Hofb.
              
            simp_sub; reflexivity.

        apply subs_refl; [].
        apply eqs_symm; [].
        eapply seq_ofs; eauto; [].
        #2 apply subrel_dot.
          apply subrel_sh; auto with sequence; done.
  
          simp_sub.
          apply seq_cn; [].
          simp_sub in Heqb.
          exact Heqb.

    eapply fstm_appap; eauto; [].
    ofvar; done.
  #4 eapply (@oft_letm _ (singles c (sg_piapp T1 T2))).
    apply ofm_forget; [].
    apply singles_ofm; auto; done.

    apply singles_fst; eauto with static; done.

    simp_sub.
    #2 apply oft_plam; auto using svalue.
      apply weaken2_ofk; auto; done.

      #2 apply oft_lam.
        eapply subst_ofc_type; eauto; [].
        prove_subof; done.

        simp_sub.
        simp_sub in Hofd.
        eapply oft_equiv; eauto; [].
        eapply seq_ofc_type; eauto; [].
        #2 apply subrel_dot.
          #2 apply subrel_dot.
            apply subrel_sh; auto with sequence; done.

            simp_sub.
            apply seq_cn; [].
            apply eqc_refl; [].
            ofvar; done.
        
          simp_sub.
          apply seq_cn; [].
          apply eqc_symm; [].
          relquest.
            eapply eqc_app; eauto; [].
            so (single_eqc_weak _#3 Hwf Hofc) as Heqvar.
            replsub (subst (sh 4) c) with (subst (sh 3) (shift1 c)).
            replsub (var 3) with (subst (sh 3) (var 0)).
            apply eqc_symm; [].
            relquest.
              eapply subst_eqc; eauto; [].
              apply subof_sh; auto with sequence; done.

              simp_sub.
              reflexivity.

            simp_sub.
            reflexivity.

    so (split_formation _#4 (ofs_piapp HofT1 Hfst1 HofT2) (split_piapp Hsplit1 Hsplit2)) as (_ & Hof).
    so (subst1_ofc_type _#4 (ofm_fst _#6 Hwf HofM HfstM (fsts_piapp Hfst1 Hfst2)) Hof) as HH.
    simp_sub in HH; simp_sub; exact HH.

  intros c e M Hofc Hofe Hjoin.
  invert Hjoin; clear Hjoin; [].
  intros f N Hsnd Hjoin <-.
  simp_sub in Hofe.
  assert (wf (G;
              cl_tm (cn_all k1
                       (cn_arrow t1
                          (subst (dot (cn_app (shift1 c) (var 0)) (dot (var 0) sh1))
                             t2)));
              cl_cn (shift1 k1);
              cl_md (subst (sh 2) T1) (var 0))) as Hwf'.
    replsub (subst (sh 2) T1) with (shift1 (shift1 T1)).
    #3 apply wf_md.
      apply wf_tm; auto; [].
      so (split_formation _#4 (ofs_piapp HofT1 Hfst1 HofT2) (split_piapp Hsplit1 Hsplit2)) as (_ & Hof).
      so (subst1_ofc_type _#4 Hofc Hof) as Hof'.
      simp_sub in Hof'.
      exact Hof'.

      apply weaken1_ofs; auto; done.

      apply subst_fsts; auto; done.
    #5 sor (refine (IH1 _#4 Hwf' _ _ andel _#4 Hsnd _)) as Hoff; clear IH1.
      eapply subst_ofs; eauto; [].
      prove_subof; done.

      apply subst_split; eauto; done.

      ofvar; done.

      ofvar; done.
    #5 sor (refine (IH2 _#4 Hwf' _ _ ander _#5 Hjoin)) as (HofN & (b & HfstN & Heqb)); clear IH2.
      eapply subst_ofs; eauto; [].
      prove_subof; done.

      apply subst_split; eauto; done.

      relquest.
        #2 eapply ofc_app.
          relquest.
            eapply subst_ofc; eauto; [].
            prove_subof; done.

            simp_sub; reflexivity.

          ofvar; done.

        simp_sub; reflexivity.

      simp_sub.
      #2 eapply oft_app.
        relquest.
          #2 eapply oft_papp.
            ofvar; done.

            ofvar; done.

          simp_sub; reflexivity.

        simp_sub in Hoff.
        exact Hoff.
    split.
      eapply ofm_lett; eauto; [].
      relquest.
        #3 eapply ofm_lamap.
          apply weaken1_ofs; auto; done.

          apply subst_fsts; eauto; done.

          simp_sub.
          replsub (subst (dot (var 1) (sh 3)) T2) with (shift1 (subst (dot (var 0) (sh 2)) T2)) in HofN.
          eassumption.

        simp_sub; reflexivity.

      clear Hoff HofN Hsnd Hjoin.
      so (eqc_reg _#4 Hwf' Heqb) as (_ & Hofb & _).
      assert (b = subst (dot nonsense (dot (var 1) (dot nonsense (sh 3)))) b) as Heeqb.
        rewrite <- (subst_under_id 3 b) at 1; [].
        eapply (sid_ofc (G; cl_utm; cl_cn (shift1 k1); cl_md (shift1 (shift1 T1)) (var 0))); eauto; [].
        simp_sub.
        #2 apply subrel_dot.
          #2 apply subrel_dot.
            #2 apply subrel_dot.
              apply subrel_sh; auto with sequence; done.

              simp_sub.
              apply sid_tm; done.

            simp_sub.
            apply sid_cn; done.

          simp_sub.
          apply sid_md; done.
      exists (cn_lam k1 (subst (dot nonsense (dot (var 0) (dot nonsense sh1))) b)).
      split.
        apply fstm_lett; [].
        simp_sub.
        #2 apply fstm_lamap.
          apply subst_fsts; auto; done.

          simp_sub.
          rewrite <- Heeqb; [].
          rewrite <- (subst_under_id 3 N); [].
          rewrite <- (subst_under_id 3 b); [].
          eapply subst_fstm; eauto; [].
          simp_sub.
          prove_subof; done.
            
        clear HfstN.
        #2 apply (@eqc_extpi _ _ _ k2 k2); auto.
          apply ofc_lam; auto; [].
          replsub k2 with (subst (dot nonsense (dot (var 0) (dot nonsense sh1))) (subst (dot (var 1) (sh 3)) k2)).
          eapply subst_ofc; eauto; [].
          prove_subof; done.

          simp_sub.
          apply eqc_symm; [].
          #2 eapply eqc_trans.
            relquest.
              #2 eapply eqc_beta.
                eapply subst_ofc; eauto; [].
                prove_subof; done.

                ofvar; done.

              simp_sub; reflexivity.
              
            simp_sub.
            apply eqc_symm; [].
            relquest.
              #2 eapply rel_quest_2.
                eapply subst_eqc; eauto; [].
                prove_subof; done.

                simp_sub; reflexivity.

              simp_sub; reflexivity.

(* sigma *)
intros T1 T2 IH1 IH2 G k t Hwf Hofs Hsplit.
invert Hofs; clear Hofs; [].
intros k1 HofT1 Hfst1 HofT2.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1 t2 Hsplit1 Hsplit2 <- <-.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_fsts _#3 Hsplit2) as Hfst2.
split.
  intros M e c HofM Hsnd HfstM.
  so (ofm_fst _#6 Hwf HofM HfstM (fsts_sigma Hfst1 Hfst2)) as Hofc.
  invert Hsnd; clear Hsnd; [].
  intros e1 e2 Hsnd1 Hsnd2 <-.
  assert (wf (G; cl_cn (single c (kd_sigma k1 k2)); cl_md (shift1 (singles c (sg_sigma T1 T2))) (var 0))) as Hwf'.
    #2 apply wf_md; auto.
      eapply singles_ofs; eauto with static; done.

      apply singles_fst; auto with static; done.
  assert (hsg (sg_sigma T1 T2)) as Hpi.
    eauto using hsg, ofs_hsg; done.
  #5 sor (refine (IH1 _#4 Hwf' _ _ andel _#3 _ Hsnd1 _)) as Hofe1; clear IH1.
    apply weaken2_ofs; auto; done.

    apply subst_split; eauto; done.

    eapply ofm_pi1; [].
    #2 eapply ofm_subsume.
      ofvar; done.

      relquest.
        apply weaken2_subs; [].
        eapply singles_subs; eauto with static; done.

        simp_sub; reflexivity.

    eapply fstm_pi1; [].
    ofvar; done.
  #5 sor (refine (IH2 _#4 Hwf' _ _ andel _#3 _ Hsnd2 _)) as Hofe2; clear IH2.
    eapply subst_ofs; eauto; [].
    prove_subof; [].
    eapply ofc_pi1; [].
    #2 eapply ofc_subsume.
      ofvar; done.

      relquest.
        apply weaken2_subk; [].
        apply single_subk; auto; done.

        simp_sub; reflexivity.
      
    apply subst_split; eauto; done.

    relquest.
      #2 eapply ofm_pi2.
        #2 eapply ofm_subsume.
          ofvar; done.

          relquest.
            apply weaken2_subs; [].
            eapply singles_subs; eauto with static; done.

            simp_sub; reflexivity.

        ofvar; done.

      simp_sub; reflexivity.

    eapply fstm_pi2; [].
    ofvar; done.
  #4 eapply (@oft_letm _ (singles c (sg_sigma T1 T2))).
    eapply ofm_forget; [].
    apply singles_ofm; auto; done.

    apply singles_fst; eauto with static; done.

    simp_sub.
    assert (subrel seq
              (G; cl_cn (single c (kd_sigma k1 k2));
               cl_md (shift1 (singles c (sg_sigma T1 T2))) (var 0))
              (dot (cn_pi1 (var 1)) (sh 2)) (dot (cn_pi1 (subst (sh 2) c)) (sh 2))
              (G; cl_cn k1)).
      apply subrel_dot; eauto using subrel_sh with sequence; [].
      simp_sub.
      apply seq_cn; [].
      eapply eqc_pi1; [].
      relquest.
        replsub (var 1) with (shift1 (var 0)).
        replsub (subst (sh 2) c) with (shift1 (shift1 c)).
        apply weaken1_eqc; [].
        apply single_eqc_weak; auto; done.

        simp_sub; reflexivity.
    #2 apply oft_pair.
      simp_sub in Hofe1.
      simp_sub in Hofe1. (* shouldn't need to do this multiple times *)
      eapply oft_equiv; eauto; [].
      eapply seq_ofc_type; eauto; done.

      simp_sub in Hofe2.
      eapply oft_equiv; eauto; [].
      eapply seq_ofc_type; eauto; [].
      apply subrel_dot; auto; [].
      simp_sub.
      apply seq_cn; [].
      relquest.
        eapply eqc_pi2; [].
        relquest.
          replsub (var 1) with (shift1 (var 0)).
          replsub (subst (sh 2) c) with (shift1 (shift1 c)).
          apply weaken1_eqc; [].
          apply single_eqc_weak; auto; done.
  
          simp_sub; reflexivity.

        simp_sub; reflexivity.

    simp_sub.
    so (subst1_ofc _#5 Hofc (split_formation _#4 (ofs_sigma HofT1 Hfst1 HofT2) (split_sigma Hsplit1 Hsplit2) ander)) as HH.
    simp_sub in HH.
    exact HH.

  intros c e M Hofc Hofe Hjoin.
  invert Hjoin; clear Hjoin; [].
  intros M1 M2 Hjoin1 Hjoin2 <-.
  assert (wf (G; cl_tm
                   (cn_prod (subst1 (cn_pi1 c) t1)
                      (subst (dot (cn_pi2 c) (dot (cn_pi1 c) id)) t2)))) as Hwf'.

    apply wf_tm; auto; [].
    #2 apply ofc_prod.
      eapply subst1_ofc_type; eauto with static; done.

      eapply subst_ofc_type; eauto; [].
      #2 apply subof_dot.
        apply subof_dot; eauto using subof_sh with sequence; [].
          simp_sub; apply sof_cn; eauto with static; done.

          simp_sub; apply sof_cn; [].
          eapply ofc_pi2; eauto; done.
  #5 sor (refine (IH1 _#4 Hwf' _ _ ander _#3 _ _ Hjoin1)) as (HofM1 & (b1 & HfstM1 & Heqb1)); clear IH1 Hjoin1.
    apply weaken1_ofs; auto; done.

    apply subst_split; eauto; done.

    eapply ofc_pi1; [].
    relquest.
    apply weaken1_ofc; eauto; done.
    simp_sub; reflexivity.

    eapply oft_pi1; [].
    relquest.
      ofvar; done.

      simp_sub; reflexivity.
  #5 sor (refine (IH2 _#4 Hwf' _ _ ander _#3 _ _ Hjoin2)) as (HofM2 & (b2 & HfstM2 & Heqb2)); clear IH2 Hjoin2.
    eapply subst_ofs; eauto; [].
    prove_subof; [].
    eapply ofc_pi1; [].
    relquest.
      apply weaken1_ofc; eauto; done.

      simp_sub; reflexivity.

    apply subst_split; eauto; done.

    relquest.
      eapply ofc_pi2; [].
      relquest.
        apply weaken1_ofc; eauto; done.

        simp_sub; reflexivity.

      simp_sub; reflexivity.

    eapply oft_pi2; [].
    ofvar; done.
  split.
    eapply ofm_lett; eauto; [].
    simp_sub.
    #2 eapply ofm_pair_dep; eauto using subst_fsts.
      eapply ofm_subsume; eauto; [].
      apply subs_refl; [].
      simp_sub.
      eapply seq_ofs; eauto; [].
      prove_subrel; [].
      exact Heqb1.

      eapply subst_ofs; eauto; [].
      prove_subof; done.

    exists (cn_pair (subst1 nonsense b1) (subst1 nonsense b2)).
    split.
      apply fstm_lett; [].
      simp_sub.
      #2 erewrite <- subord_tm_cn.
        #2 erewrite <- subord_tm_cn.
          rewrite <- (subst_dot0_sh (md_pair M1 M2)), <- (subst_dot0_sh (cn_pair b1 b2)); [].
          eapply subst_fstm; eauto using fstm_pair; [].
          prove_subof; done.

          exact (eqc_reg _#4 Hwf' Heqb2 anderl).

        exact (eqc_reg _#4 Hwf' Heqb1 anderl).

      cut (eqc (G; cl_tm (cn_prod (subst1 (cn_pi1 c) t1) (subst (dot (cn_pi2 c) (dot (cn_pi1 c) id)) t2))) (shift1 c) (cn_pair b1 b2) (shift1 (kd_sigma k1 k2))).
        intro Heqc.
        #2 sor (refine (subst_eqc G _ (dot nonsense id) _#3 _ Heqc)) as Heqc'.
          prove_subof; done.
          
          simp_sub in Heqc'.
          exact Heqc'.
      simp_sub.
      #3 apply eqc_extsigma.
        #2 apply (@eqc_trans _ _ b1).
          exact Heqb1.

          apply eqc_symm; [].
          #2 eapply eqc_beta1.
            exact (eqc_reg _#4 Hwf' Heqb1 anderl).

            exact (eqc_reg _#4 Hwf' Heqb2 anderl).

        #2 apply (@eqc_trans _ _ b2).
          relquest.
            exact Heqb2.

            simp_sub; reflexivity.

          apply eqc_symm; [].
          #2 eapply eqc_beta2.
            exact (eqc_reg _#4 Hwf' Heqb1 anderl).

            relquest.
              exact (eqc_reg _#4 Hwf' Heqb2 anderl).

              simp_sub; reflexivity.

        eapply subst_ofk; eauto; [].
        prove_subof; done.

(* epilogue *)
simpl in Hprop.
do2 2 split.
  intros G T M e k t c Hwf HofM Hsnd Hsplit Hfstm.
  so (ofm_reg _#4 Hwf HofM) as HofT.
  refine (Hprop T (ofs_hsg _ _ HofT) _#6 andel _#6); eauto; done.

  intros G T c e M k t Hwf HofT Hjoin Hsplit Hofc Hofe.
  refine (Hprop T (ofs_hsg _ _ HofT) _#6 ander _#6 andel); eauto; done.

  intros G T c e M k t Hwf HofT Hjoin Hsplit Hofc Hofe.
  refine (Hprop T (ofs_hsg _ _ HofT) _#6 ander _#6 ander); eauto; done.
Qed.


Lemma snd_formation :
  forall G T M e k t c,
    wf G
    -> ofm G pure M T
    -> snd T M e
    -> split T k t
    -> fstm G M c
    -> oft G e (subst1 c t).
Proof.
exact (snd_and_join_formation andel).
Qed.


Lemma join_formation :
  forall G T c e M k t,
    wf G
    -> ofs G T
    -> join T c e M
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> ofm G pure M T.
Proof.
exact (snd_and_join_formation anderl).
Qed.


Lemma join_fst_o :
  forall G T c e M k t,
    wf G
    -> ofs G T
    -> join T c e M
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> exists c', fstm G M c' /\ eqc G c c' k.
Proof.
exact (snd_and_join_formation anderr).
Qed.


Lemma join_formation_and_fst_o :
  forall G T c e M k t,
    wf G
    -> ofs G T
    -> join T c e M
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> ofm G pure M T /\ exists c', fstm G M c' /\ eqc G c c' k.
Proof.
intros; eauto using join_formation, join_fst_o.
Qed.


Lemma join_fst :
  forall G T c e M k t c',
    wf G
    -> ofs G T
    -> join T c e M
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> fstm G M c' 
    -> eqc G c c' k.
Proof.
intros G T c e M k t c' Hwf Hofs Hjoin Hsplit Hofc Hofe HfstM.
so (join_fst_o _#7 Hwf Hofs Hjoin Hsplit Hofc Hofe) as (c'' & HfstM' & Heqc).
so (fstm_fun _#4 HfstM HfstM'); subst c''.
assumption.
Qed.


Lemma join_vars_formation :
  forall G T k t M,
    wf G
    -> ofs G T
    -> split T k t
    -> join (subst (sh 2) T) (var 1) (var 0) M
    -> ofm (G; cl_cn k; cl_tm t) pure M (subst (sh 2) T).
Proof.
intros G T k t M Hwf HofT Hsplit Hjoin.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
#2 eapply join_formation; eauto using wf, subst_split, weaken2_ofs.
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
Qed.


Lemma join_vars_fst_o :
  forall G T k t M,
    wf G
    -> ofs G T
    -> split T k t
    -> join (subst (sh 2) T) (var 1) (var 0) M
    -> exists c,
         fstm (G; cl_cn k; cl_tm t) M (shift1 c)
         /\ eqc (G; cl_cn k) (var 0) c (shift1 k).
Proof.
intros G T k t M Hwf HofT Hsplit Hjoin.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
#3 sor (refine (join_formation_and_fst_o _#7 (wf_tm (wf_cn Hwf Hofk) Hoft) (weaken2_ofs _#4 HofT) Hjoin (subst_split _#4 Hsplit) _ _)) as (HofM & c & Hfstm & Heqc).
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
so (ofm_fst _#6 (wf_tm (wf_cn Hwf Hofk) Hoft) HofM Hfstm (subst_fsts _#3 (split_fsts _#3 Hsplit))) as Hofc.
so (subord_tm_cn _#4 Hofc) as Heq.
exists (subst1 nonsense c).
split.
  simp_sub.
  rewrite <- Heq; [].
  exact Hfstm.

  replsub (var 0) with (subst1 nonsense (var 1)).
  relquest.
    eapply subst_eqc; eauto; prove_subof; done.

    simp_sub; reflexivity.
Qed.


Lemma subst_snd_join :
  (forall T M e s,
     snd T M e
     -> snd (subst s T) (subst s M) (subst s e))
  /\
  (forall T c e M s,
     join T c e M
     -> join (subst s T) (subst s c) (subst s e) (subst s M)).
Proof.
exploit (snd_join_ind
           (fun T M e => forall s, snd (subst s T) (subst s M) (subst s e))
           (fun T c e M => forall s, join (subst s T) (subst s c) (subst s e) (subst s M))) as Hind.

(* one *)
intros M s.           
simp_sub.
relquest.
  apply snd_one; done.

  simp_sub; reflexivity.

(* stat *)
intros k M s.
simp_sub.
relquest.
  apply snd_stat; done.

  simp_sub; reflexivity.

(* dyn *)
intros t M s.
simp_sub.
relquest.
  apply snd_dyn; done.
  
  simp_sub; reflexivity.

(* pigen *)
intros T1 T2 M k1 t1 k2 t2 N e Hsplit1 Hsplit2 _ IH1 _ IH2 s.
simp_sub.
relquest.
  #2 eapply snd_pigen; eauto using subst_split.
    so (IH1 (under 4 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 6 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* piapp *)
intros T1 T2 M k1 t1 N e Hsplit1 _ IH1 _ IH2 s.
simp_sub.
relquest.
  #2 eapply snd_piapp; eauto using subst_split.
    so (IH1 (under 4 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 4 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* sigma *)
intros T1 T2 M e1 e2 _ IH1 _ IH2 s.
simp_sub.
relquest.
  #2 eapply snd_sigma.
    so (IH1 (under 2 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 2 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* one *)
intros c e s.    
simp_sub.
apply join_one; done.

(* stat *)
intros k c e s.
simp_sub.
relquest.
  apply join_stat; done.

  simp_sub; reflexivity.

(* dyn *)
intros t c e s.
simp_sub.
apply join_dyn; done.

(* pigen *)
intros T1 T2 c e f M _ IH1 _ IH2 s.  
simp_sub.
relquest.
  #2 eapply join_pigen.
    so (IH1 (under 3 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 5 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* piapp *)
intros T1 T2 c e f M _ IH1 _ IH2 s.
simp_sub.
relquest.
  #2 eapply join_piapp.
    so (IH1 (under 3 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 3 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* sigma *)
intros T1 T2 c e M1 M2 _ IH1 _ IH2 s.
simp_sub.
relquest.
  #2 eapply join_sigma.
    so (IH1 (under 1 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

    so (IH2 (under 1 s)) as H.
    simp_sub in H.
    simp_sub.
    exact H.

  simp_sub; reflexivity.

(* epilogue *)
split.
  intros; apply (Hind andel); auto; done.

  intros; apply (Hind ander); auto; done.
Qed.


Lemma subst_snd :
  forall T M e s,
    snd T M e
    -> snd (subst s T) (subst s M) (subst s e).
Proof.
exact (subst_snd_join andel).
Qed.


Lemma subst_join :
  forall T c e M s,
    join T c e M
    -> join (subst s T) (subst s c) (subst s e) (subst s M).
Proof.
exact (subst_snd_join ander).
Qed.


Lemma snd_var_formation :
  forall G T k t e,
    wf G
    -> ofs G T
    -> split T k t
    -> snd (subst sh1 T) (var 0) e
    -> oft (G; cl_cn k; cl_md (shift1 T) (var 0)) (subst (dot (var 0) (sh 2)) e) (shift1 t).
Proof.
intros G T k t e Hwf HofT Hsplit Hsnd.
so (split_fsts _#3 Hsplit) as Hfst.
so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd) as Hsndx; simp_sub in Hsndx.
clear Hsnd.
relquest.
  eapply snd_formation; eauto using wf, subst_split; ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; reflexivity.
Qed.


Lemma snd_halt_only_if :
  forall T M e,
    snd T M e
    -> thalt nil e
    -> mhalt nil M.
Proof.
intros T M e Hsnd Hhalt.
revert Hhalt.
case Hsnd; clear codependent Hsnd;
eauto using letm_halt_inversion'; [].
(* dyn *)
intros _ M Hhalt.
eapply ext_halt_inversion'; eauto; done.
Qed.


Lemma snd_vars_halt_only_if :
  forall T e M s,
    snd T (var 0) e
    -> thalt nil (subst (dot M s) e)
    -> mhalt nil M.
Proof.
intros T e M s Hsnd Hhalt.
so (subst_snd _#3 (dot M s) Hsnd) as H; simp_sub in H.
eapply snd_halt_only_if; eauto; done.
Qed.


Lemma snd_halt_iff :
  forall p T M e,
    ofm nil p M T
    -> snd T M e
    -> (mhalt nil M <-> thalt nil e).
Proof.
intros p T M e HofM Hsnd.
split.
  intro Hhalt.
  revert HofM Hhalt.
  replace T with (subst id T); [| simp_sub; reflexivity].
  replace M with (subst id M); [| simp_sub; reflexivity].
  replace e with (subst id e); [| simp_sub; reflexivity].
  generalize id.
  #2 induct Hsnd;
  try (intros; simp_sub;
       refine (letm_val_halt_iff _#4 _ _ andel _); eauto with dynamic; done).
    (* dyn *)
    intros t M s HofM Hhalt.
    destruct Hhalt as (V & Hsteps & HvalV).
    simp_sub.
    so (preservations_md _#5 wf_nil HofM Hsteps) as HofV.
    invert (Progress.canon_md _#3 HofV HvalV); [].
    intros v <-.
    invert HvalV; [].
    intros Hvalv.
    exists v.
    split; auto; [].
    #2 eapply star_trans.
      eapply star_map; eauto with dynamic; done.
  
      apply star_one.
      apply tstep_ext2; auto; done.

    (* sigma *)
    intros T1 T2 M e1 e2 _ IH1 _ IH2 s HofM Hhalt.
    simp_sub.
    destruct Hhalt as (V & Hsteps & HvalV).
    simp_sub in HofM.
    so (preservations_md _#5 wf_nil HofM Hsteps) as HofV.
    so (mvalue_pure _#4 HvalV HofV) as HofV'.
    renameover HofV' into HofV.
    invert (Progress.canon_md _#3 HofV HvalV); [].
    intros V1 V2 <-.
    invert HvalV; [].
    intros HvalV1 HvalV2.
    so (Inversion.pair_ofm_inversion_sigma_pure _#5 wf_nil HofV) as (c1 & HofV1 & HfstV1 & HofV2).
    so (can_fstm _#3 wf_nil HofV2) as (c2 & HfstV2).
    #3 exploit (IH1 (dot (md_pair V1 V2) (dot (cn_pair c1 c2) s))) as Hhalt1.
      simp_sub.
      apply ofm_forget_pure; [].
      eapply ofm_pi1; eauto; done.

      simp_sub.
      exists V1.
      split; auto; [].
      apply star_one; auto with dynamic; done.
    #3 exploit (IH2 (dot (md_pair V1 V2) (dot (cn_pair c1 c2) s))) as Hhalt2.
      simp_sub.
      apply ofm_forget_pure; [].
      relquest.
        eapply ofm_pi2; eauto with static; done.

        simp_sub; reflexivity.

      simp_sub.
      exists V2.
      split; auto; [].
      apply star_one; auto with dynamic; done.
    destruct Hhalt1 as (v1 & Hsteps1 & Hvalv1).
    destruct Hhalt2 as (v2 & Hsteps2 & Hvalv2).
    exists (tm_pair v1 v2).
    split; auto with dynamic; [].
    #2 eapply star_trans.
      eapply (star_map _#4 (fun x => tm_letm x _)); eauto with dynamic; done.

      #2 eapply star_step.
        eapply tstep_letm2; eauto with static; done.

        simp_sub.
        #2 eapply star_trans.
          apply (star_map' _ _ (fun x => tm_pair x _)); eauto with dynamic; done.

          apply (star_map' _ _ (fun x => tm_pair _ x)); auto with dynamic; done.

  clear HofM.
  case Hsnd; clear codependent Hsnd;
  try (intros; eapply Inversion.letm_halt_inversion'; eauto; done); [].
  (* dyn *)
  intros _ M Hhalt.
  exact (Inversion.ext_halt_inversion' _ Hhalt).
Qed.


Lemma snd_halt_val_instances :
  forall G T M e,
    ofm G pure M T
    -> mvalue G M
    -> snd T M e
    -> forall s, subof aof nil s G -> thalt nil (subst s e).
Proof.
intros G T M e Hof Hval Hsnd s Hs.
refine (snd_halt_iff _#4 (subst_ofm _#6 (subof_aof_dof _#3 Hs) Hof) (subst_snd _#4 Hsnd) andel _); [].
apply mvalue_mhalt; [].
eapply subst_mvalue; eauto using subof_aof_oof; done.
Qed.


Lemma join_halt_only_if :
  forall T c e M,
    join T c e M
    -> mhalt nil M
    -> thalt nil e.
Proof.
intros T c e M Hjoin Hhalt.
revert Hhalt.
case Hjoin; clear codependent Hjoin;
eauto using mlett_halt_inversion'; [].
(* dyn *)
intros _ _ e Hhalt.
eapply dyn_halt_inversion'; eauto; done.
Qed.


Lemma join_vars_halt_only_if :
  forall T c M e s,
    join T c (var 0) M
    -> mhalt nil (subst (dot e s) M)
    -> thalt nil e.
Proof.
intros T c M e s Hjoin Hhalt.
so (subst_join _#4 (dot e s) Hjoin) as H; simp_sub in H.
eapply join_halt_only_if; eauto; done.
Qed.


Lemma join_halt_iff :
  forall T k t c e M,
    split T k t
    -> oft nil e (subst1 c t)
    -> join T c e M
    -> (thalt nil e <-> mhalt nil M).
Proof.
intros T k t c e M Hsplit Hofe Hjoin.
split.
  intro Hhalt.
  revert k t Hsplit Hofe Hhalt.
  replace T with (subst id T); [| simp_sub; reflexivity].
  replace c with (subst id c); [| simp_sub; reflexivity].
  replace e with (subst id e); [| simp_sub; reflexivity].
  replace M with (subst id M); [| simp_sub; reflexivity].
  generalize id.
  #2 induct Hjoin;
  try (intros; simp_sub;
       refine (mlett_val_halt_iff _ _ _ andel _); eauto with dynamic; done).
    (* dyn *)
    intros t c e s _ _ _ _ Hhalt.
    simp_sub.
    destruct Hhalt as (v & Hsteps & Hval).
    exists (md_dyn v); split.
      eapply star_map; eauto with dynamic; done.

      auto with dynamic; done.

    (* pair *)
    intros T1 T2 c e M1 M2 _ IH1 _ IH2 s k t Hsplit Hofe Hhalt.
    simp_sub in Hsplit.
    invert Hsplit; clear Hsplit; [].
    intros k1 k2 t1 t2 Hsplit1 Hsplit2 <- <-.
    destruct Hhalt as (v & Hsteps & Hvalv).
    simp_sub in Hofe.
    so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
    clear Hofe.
    invert (Progress.canon_tm _ _ Hofv Consistent.similar_prod Hvalv); [].
    intros v1 v2 <-.
    invert Hvalv; clear Hvalv; [].
    intros Hvalv1 Hvalv2.
    #4 exploit (IH1 (dot (tm_pair v1 v2) s) k1 t1) as Hhalt1.
      simp_sub.
      assumption.

      simp_sub.
      simp_sub in Hofv.
      eapply oft_pi1; eauto; done.

      simp_sub.
      exists v1; split; auto; [].
      apply star_one; auto with dynamic; done.
    #4 exploit (IH2 (dot (tm_pair v1 v2) s) (subst1 (cn_pi1 (subst s c)) k2) (subst (dot (var 0) (dot (cn_pi1 (subst (compose s sh1) c)) sh1)) t2)) as Hhalt2.
      simp_sub.
      so (subst_split _ _ _ (dot (cn_pi1 (subst s c)) id) Hsplit2) as HH.
      simp_sub in HH.
      exact HH.

      simp_sub.
      eapply oft_pi2; eauto; done.
      
      simp_sub.
      exists v2; split; auto; [].
      apply star_one; auto with dynamic; done.
    destruct Hhalt1 as (V1 & Hsteps1 & HvalV1).
    destruct Hhalt2 as (V2 & Hsteps2 & HvalV2).
    exists (md_pair V1 V2).
    split; auto with dynamic; [].
    simp_sub.
    #2 eapply star_trans.
      eapply (star_map _#4 (fun x => md_lett x _)); eauto with dynamic; done.

      #2 eapply star_step.
        apply mstep_lett2; auto with dynamic; done.

        simp_sub.
        #2 eapply star_trans.
          apply (star_map' _ _ (fun x => md_pair x _)); eauto with dynamic; done.

          apply (star_map' _ _ (fun x => md_pair _ x)); auto with dynamic; done.

  clear k t Hsplit Hofe.
  case Hjoin; clear codependent Hjoin;
  try (intros; eapply Inversion.mlett_halt_inversion'; eauto; done); [].
  (* dyn *)
  intros _ _ e Hhalt.
  apply Inversion.dyn_halt_inversion'; auto; done.
Qed.


Lemma join_vars_halt_if :
  forall T k t M e c s,
    split T k t
    -> oft nil e (subst (dot c s) t)
    -> join (subst (sh 2) T) (var 1) (var 0) M
    -> thalt nil e
    -> mhalt nil (subst (dot e (dot c s)) M).
Proof.
intros T k t M e c s Hsplit Hof Hjoin Hhalt.
so (subst_join _#4 (dot e (dot c s)) Hjoin) as Hjoinx; simp_sub in Hjoinx.
refine (join_halt_iff _#6 (subst_split _#4 Hsplit) _ Hjoinx andel Hhalt); [].
simp_sub.
exact Hof.
Qed.


Lemma snd_join_fun :
  (forall T M e e',
     snd T M e
     -> snd T M e'
     -> e = e')
  /\
  (forall T c e M M',
     join T c e M
     -> join T c e M'
     -> M = M').
Proof.
#3 exploit
  (snd_join_ind
     (fun T M e => forall e', snd T M e' -> e = e')
     (fun T c e M => forall M', join T c e M' -> M = M')) as Hind;
try
  (intros;
   repeat
     (match goal with
      | H : snd ?T _ _ |- _ =>
          invert H; []; clear H

      | H : join ?T _ _ _ |- _ =>
          invert H; []; clear H
      end);
   subst;
   eauto;
   intros;
   repeat
     (match goal with
      | H1 : split ?T _ _, H2 : split ?T _ _ |- _ =>
          destruct (split_fun _#5 H1 H2); clear H1 H2
      end);
   subst; eauto; repeat f_equal; eauto; done).

(* piapp *)
intros T1 T2 M k1 t1 N e Hsplit1 _ IH1 _ IH2 ee Hsndee.
invert Hsndee; clear Hsndee; [].
intros k1a t1a N' e' Hsplit1a Hjoin Hsnd <-.
destruct (split_fun _#5 Hsplit1 Hsplit1a); subst k1a t1a.
repeat f_equal; [].
so (IH1 _ Hjoin); subst N'.
auto; done.

(* piapp *)
intros T1 T2 c e f M _ IH1 _ IH2 ee Hsndee.
invert Hsndee; clear Hsndee; [].
intros e' N' Hsnd Hjoin <-.
repeat f_equal; [].
so (IH1 _ Hsnd); subst e'.
auto; done.

(* epilogue *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma snd_fun :
  forall T M e e',
    snd T M e
    -> snd T M e'
    -> e = e'.
Proof.
exact (snd_join_fun andel).
Qed.


Lemma join_fun :
  forall T c e M M',
    join T c e M
    -> join T c e M'
    -> M = M'.
Proof.
exact (snd_join_fun ander).
Qed.
