
Require Import Tactics.
Require Import Sequence.
Require Export TSubstGeneral.
Require Import TSubst.
Require Import Static.
Require Import Dynamic.
Require Import Breduce.
Require Import Regularity.
Require Import Congruence.
Require Import TCongruence.
Require Import BreduceCong.
Require Adequacy.


Lemma reduce_tcompatible_strong :
  tcompatible eqk eqc (fun G e e' t => treduce G e e') (fun G e e' t => treduce G e e').
Proof.
unfold tcompatible.
#6 split;
[intros G e e' t _ H; induct H | intros G e e' t _ H; induct H];
eauto with reduce.

(* lam *)
intros G t1 t1' t2 e e' _ Hred.
apply treduce_lam; [].
rewrite <- (subst_dot0_sh e); rewrite <- (subst_dot0_sh e'); [].
eapply subst_treduce; eauto; [].
apply subof_dot; eauto with static sequence; [].
simp_sub.
apply oof_tm; auto with dynamic sequence; done.

(* plam *)
intros G k k' t e e' _ Hred.
apply treduce_plam; [].
rewrite <- (subst_dot0_sh e); rewrite <- (subst_dot0_sh e'); [].
eapply subst_treduce; eauto; [].
apply subof_dot; eauto with static sequence; [].
simp_sub.
apply oof_cn; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' Hred1 Hred2 _.
apply treduce_unpack1; auto; [].
rewrite <- (subst_dot01_sh e2); rewrite <- (subst_dot01_sh e2'); [].
eapply subst_treduce; eauto; [].
#2 apply subof_dot.
  apply subof_dot; eauto with static sequence; [].
  simp_sub.
  apply oof_cn; done.

  simp_sub.
  apply oof_tm; auto with dynamic sequence; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' Hred1 Hred2.
apply treduce_lett1; auto; [].
rewrite <- (subst_dot0_sh e2); rewrite <- (subst_dot0_sh e2'); [].
eapply subst_treduce; eauto; [].
apply subof_dot; eauto with static sequence; [].
simp_sub.
apply oof_tm; auto with dynamic sequence; done.

(* lam again *)
intros G t1 t1' t2 e e' _ Hred.
apply treduce_lam; [].
rewrite <- (subst_dot0_sh e); rewrite <- (subst_dot0_sh e'); [].
eapply subst_treduce; eauto; [].
apply subof_dot; eauto with static sequence; [].
simp_sub.
apply oof_tm; auto with dynamic sequence; done.

(* plam again *)
intros G k k' t e e' _ Hred.
apply treduce_plam; [].
rewrite <- (subst_dot0_sh e); rewrite <- (subst_dot0_sh e'); [].
eapply subst_treduce; eauto; [].
apply subof_dot; eauto with static sequence; [].
simp_sub.
apply oof_cn; done.
Qed.


Lemma reduce_tcompatible_weak :
  tcompatible samek samec (fun G e e' t => treduce G e e') (fun G e e' t => treduce G e e').
Proof.
#3 apply (tcompatible_mono _ eqk _ eqc).
  intros G k k' H.
  invert H; intros; subst; eauto with static; done.

  intros G c c' k H.
  invert H; intros; subst; eauto with static; done.

  exact reduce_tcompatible_strong.
Qed.


Lemma reduce_tcsubstitutive :
  tcsubstitutive (fun G e e' t => treduce G e e') (fun G e e' t => treduce G e e').
Proof.
split.
  intros.
  eapply subst_treduce; eauto; [].
  eapply subof_mono; try eassumption; eauto using aof_oof; done.

  intros.
  eapply subst_treduce; eauto; [].
  eapply subof_mono; try eassumption; eauto using aof_oof; done.
Qed.


Definition ttreducer := ttreg (fun G e e' t => treduce G e e').
Definition tvreducer := tvreg (fun G e e' t => treduce G e e').


Lemma reducer_tsemicongruence :
  tsemicongruence ttreducer tvreducer.
Proof.
#4 apply reg_tsemicongruence.
  split.
    intros; apply treduce_refl; done.

    intros; apply treduce_refl; done.

  split.
    intros; eapply treduce_trans; eauto; done.

    intros; eapply treduce_trans; eauto; done.

  exact reduce_tcsubstitutive.

  exact reduce_tcompatible_weak.
Qed.


Lemma anti_reducer_tsemicongruence :
  tsemicongruence (ttreg (fun G e e' t => treduce G e' e)) (tvreg (fun G e e' t => treduce G e' e)).
Proof.
#3 apply (tsemicongruence_ext
         (fun G e e' t => ttreg (fun G e e' t => treduce G e e') G e' e t) _
         (fun G e e' t => tvreg (fun G e e' t => treduce G e e') G e' e t)).
  intros G e e' t.
  split; intros; apply ttreg_transpose; assumption.

  intros G e e' t.
  split; intros; apply tvreg_transpose; assumption.

  apply transpose_tsemicongruence; exact reducer_tsemicongruence.
Qed.


Lemma reg_tconsistent :
  forall (Rt : _ -> _ -> _ -> _ -> Prop),
    (forall e e' t,
       oft nil e t
       -> oft nil e' t
       -> target e
       -> target e'
       -> Rt nil e e' t
       -> thalt nil e
       -> thalt nil e')
    -> tconsistent (ttreg Rt).
Proof.
intros Rt Hcon.
intros e e' t HR Hhalt.
destruct HR as (Hofe & Hofe' & Htar & Htar' & HR).
eauto.
Qed.


Lemma reducer_tconsistent :
  tconsistent ttreducer.
Proof.
apply reg_tconsistent; [].
intros e e' t Hof Hof' Htar Htar' Hred Hhalt.
refine (Adequacy.adequacy_treduce e e' _#5 andel _); eauto.
Qed.


Lemma anti_reducer_tconsistent :
  tconsistent (ttreg (fun G e e' t => treduce G e' e)).
Proof.
apply reg_tconsistent; [].
intros e e' t Hof Hof' Htar Htar' Hred Hhalt.
refine (Adequacy.adequacy_treduce e' e _#5 ander _); eauto.
Qed.
