
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Dynamic.
Require Import Regularity.
Require Import Singleton.
Require Import Declarative.
Require Import Breduce.
Require Import Congruence.
Require Import TCongruence.
Require Import BreduceCong.
Require Import TBreduceCong.
Require Import Preservation.
Require Import Contextual.


Definition tcapproxt (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  exists Rt Rv,
    tsemicongruence Rt Rv
    /\ tconsistent Rt
    /\ Rt G e e' t.


Definition tcapproxv (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  exists Rt Rv,
    tsemicongruence Rt Rv
    /\ tconsistent Rt
    /\ Rv G e e' t.


Definition tceqt (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  tcapproxt G e e' t /\ tcapproxt G e' e t.
      

Definition tceqv (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  tcapproxv G e e' t /\ tcapproxv G e' e t.
      

Lemma tcapproxt_univ :
  forall Rt Rv G e e' t,
    tsemicongruence Rt Rv
    -> tconsistent Rt
    -> Rt G e e' t
    -> tcapproxt G e e' t.
Proof.
intros Rt Rv G e e' t Hsc Hcons HR.
exists Rt, Rv.
do2 2 split; trivial; done.
Qed.


Lemma tcapproxv_univ :
  forall Rt Rv G e e' t,
    tsemicongruence Rt Rv
    -> tconsistent Rt
    -> Rv G e e' t
    -> tcapproxv G e e' t.
Proof.
intros Rt Rv G e e' t Hsc Hcons HR.
exists Rt, Rv.
do2 2 split; trivial; done.
Qed.


Lemma tcapproxt_univ_quasi :
  forall Rt Rv G e e' t,
    tquasicongruence Rt Rv
    -> tconsistent Rt
    -> Rt G e e' t
    -> tcapproxt G e e' t.
Proof.
intros Rt Rv G e e' t Hquasi Hcon HR.
#3 apply (tcapproxt_univ (ttc Rt) (ttc Rv)).
  apply tc_tsemicongruence; auto; done.

  apply tc_tconsistent; auto; done.

  eapply ttc_intro; eauto; done.
Qed.


Lemma tcapproxv_univ_quasi :
  forall Rt Rv G e e' t,
    tquasicongruence Rt Rv
    -> tconsistent Rt
    -> Rv G e e' t
    -> tcapproxv G e e' t.
Proof.
intros Rt Rv G e e' t Hquasi Hcon HR.
#3 apply (tcapproxv_univ (ttc Rt) (ttc Rv)).
  apply tc_tsemicongruence; auto; done.

  apply tc_tconsistent; auto; done.

  eapply ttc_intro; eauto; done.
Qed.


Lemma tcapprox_tcregular :
  tcregular tcapproxt tcapproxv.
Proof.
split.
  intros G e e' t Hwf Hap.
  destruct Hap as (Rt & Rv & Hsc & _ & HR).
  apply (Hsc andel andel); auto; done.

  intros G e e' t Hwf Hap.
  destruct Hap as (Rt & Rv & Hsc & _ & HR).
  apply (Hsc andel ander); auto; done.
Qed.


Lemma tcapproxt_reg :
  forall G e e' t,
    wf G
    -> tcapproxt G e e' t
    -> oft G e t /\ oft G e' t /\ target e /\ target e'.
Proof.
exact (tcapprox_tcregular andel).
Qed.


Lemma tcapproxv_reg :
  forall G e e' t,
    wf G
    -> tcapproxv G e e' t
    -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e'.
Proof.
exact (tcapprox_tcregular ander).
Qed.


Lemma treduce_tcapproxt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> target e
    -> target e'
    -> treduce G e e'
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Hof Hof' Htar Htar' Hred.
apply (tcapproxt_univ ttreducer tvreducer); auto using reducer_tsemicongruence, reducer_tconsistent;[].
eapply ttreg_intro; eauto using reduce_csubstitutive.
Qed.


Lemma anti_treduce_tcapproxt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> target e
    -> target e'
    -> treduce G e e'
    -> tcapproxt G e' e t.
Proof.
intros G e e' t Hof Hof' Htar Htar' Hred.
apply (tcapproxt_univ (ttreg (fun G e e' t => treduce G e' e)) (tvreg (fun G e e' t => treduce G e' e))); auto using anti_reducer_tsemicongruence, anti_reducer_tconsistent; [].
eapply ttreg_intro; eauto using transpose_csubstitutive, reduce_csubstitutive.
Qed.


Lemma treduce_tcapproxv :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> svalue e
    -> svalue e'
    -> target e
    -> target e'
    -> treduce G e e'
    -> tcapproxv G e e' t.
Proof.
intros G e e' t Hof Hof' Hval Hval' Htar Htar' Hred.
apply (tcapproxv_univ ttreducer tvreducer); auto using reducer_tsemicongruence, reducer_tconsistent; [].
eapply tvreg_intro; eauto using reduce_csubstitutive.
Qed.


Theorem tcapprox_tconsistent :
  tconsistent tcapproxt.
Proof.
intros e e' t Hap Hhalt.
destruct Hap as (Rt & Rv & _ & Hcons & HR).
exact (Hcons _ _ _ HR Hhalt).
Qed.


Lemma tcapproxt_refl :
  forall G e t, oft G e t -> target e -> tcapproxt G e e t.
Proof.
intros G e t Hof Htar.
apply treduce_tcapproxt; auto with reduce.
Qed.


Lemma tcapproxv_refl :
  forall G e t, oft G e t -> svalue e -> target e -> tcapproxv G e e t.
Proof.
intros G e t Hof Hval Htar.
apply treduce_tcapproxv; auto with reduce.
Qed.


Lemma tcapprox_tcreflexive :
  tcreflexive tcapproxt tcapproxv.
Proof.
split.
  exact tcapproxt_refl.

  exact tcapproxv_refl.
Qed.


Lemma tcapprox_tcsubstitutive :
  tcsubstitutive tcapproxt tcapproxv.
Proof.
split.
  intros G1 s G2 e e' t Hs Htar Hwf (Rt & Rv & Hsc & Hcons & HR).
  exists Rt, Rv.
  do2 2 split; auto; [].
  eapply (Hsc anderrrl andel); eauto; done.

  intros G1 s G2 e e' t Hs Htar Hwf (Rt & Rv & Hsc & Hcons & HR).
  exists Rt, Rv.
  do2 2 split; auto; [].
  eapply (Hsc anderrrl ander); eauto; done.
Qed.


Lemma subst_tcapproxt :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> targetsub s
    -> wf G2
    -> tcapproxt G2 e e' t
    -> tcapproxt G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (tcapprox_tcsubstitutive andel).
Qed.


Lemma subst_tcapproxv :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> targetsub s
    -> wf G2
    -> tcapproxv G2 e e' t
    -> tcapproxv G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (tcapprox_tcsubstitutive ander).
Qed.


Lemma tcapprox_weakly_tcompatible :
  weakly_tcompatible samek samec tcapproxt tcapproxv.
Proof.
unfold tcapproxt, tcapproxv.
split.
  intros G e e' t Hwf Htcc.
  revert Hwf.
  induct Htcc;
  try (intros;
       (match goal with
        | H : exists Rt Rv, _ |- _ =>
          destruct H as (? & ? & Hcong & ? & ?);
          so Hcong as (_ & (? & ?) & _ & (Ht & Hv))
        end);
       do 2 eexists;
       do2 2 split; eauto;
       first [apply Ht | apply Hv];
       eauto using ttcc;
       done).

  intros G e e' t Hwf Hvcc.
  revert Hwf.
  induct Hvcc;
  try (intros;
       (match goal with
        | H : exists Rt Rv, _ |- _ =>
          destruct H as (? & ? & Hcong & ? & ?);
          so Hcong as (_ & (? & ?) & _ & (Ht & Hv))
        end);
       do 2 eexists;
       do2 2 split; eauto;
       first [apply Ht | apply Hv];
       eauto using tvcc;
       done); done.
Qed.


Lemma tc_tcapprox_tsemicongruence :
  tsemicongruence (ttc tcapproxt) (ttc tcapproxv).
Proof.
  apply tc_tsemicongruence_strong; auto using tcapprox_tcregular, tcapprox_tcreflexive, tcapprox_tcsubstitutive, tcapprox_weakly_tcompatible; done.
Qed.


Lemma tcapprox_tctransitive :
  tctransitive tcapproxt tcapproxv.
Proof.
split.
  intros G t e1 e2 e3 H12 H23.
  eapply tcapproxt_univ; eauto using tc_tcapprox_tsemicongruence, tc_tconsistent, tcapprox_tconsistent; [].
  eapply plus_trans; eauto using plus_one; done.

  intros G t e1 e2 e3 H12 H23.
  eapply tcapproxv_univ; eauto using tc_tcapprox_tsemicongruence, tc_tconsistent, tcapprox_tconsistent; [].
  eapply plus_trans; eauto using plus_one; done.
Qed.


Lemma tcapproxt_trans :
  forall G e1 e2 e3 t,
    tcapproxt G e1 e2 t
    -> tcapproxt G e2 e3 t
    -> tcapproxt G e1 e3 t.
Proof.
intros.
eapply (tcapprox_tctransitive andel); eauto.
Qed.


Lemma tcapproxv_trans :
  forall G e1 e2 e3 t,
    tcapproxv G e1 e2 t
    -> tcapproxv G e2 e3 t
    -> tcapproxv G e1 e3 t.
Proof.
intros.
eapply (tcapprox_tctransitive ander); eauto.
Qed.


Lemma tcapprox_tcompatible_weak :
  tcompatible samek samec tcapproxt tcapproxv.
Proof.
split.
  intros G e e' t Hwf Hcc.
  eapply tcapproxt_univ; eauto using tc_tcapprox_tsemicongruence, tc_tconsistent, tcapprox_tconsistent; [].
  apply (tc_tcapprox_tsemicongruence anderr anderr andel); auto; [].
  #2 apply (ttcc_mono samek _ samec _ tcapproxt _ tcapproxv ); auto.
    intros; apply plus_one; auto; done.

    intros; apply plus_one; auto; done.

  intros G e e' t Hwf Hcc.
  eapply tcapproxv_univ; eauto using tc_tcapprox_tsemicongruence, tc_tconsistent, tcapprox_tconsistent; [].
  apply (tc_tcapprox_tsemicongruence anderr anderr ander); auto; [].
  #2 apply (tvcc_mono samek _ samec _ tcapproxt _ tcapproxv); auto.
    intros; apply plus_one; auto; done.

    intros; apply plus_one; auto; done.
Qed.


Lemma tcapproxt_comp_weak_ifwf :
  forall G e e' t,
    wf G
    -> ttcc samek samec tcapproxt tcapproxv G e e' t 
    -> tcapproxt G e e' t.
Proof.
exact (tcapprox_tcompatible_weak andel).
Qed.


Lemma tcapproxv_comp_weak_ifwf :
  forall G e e' t,
    wf G
    -> tvcc samek samec tcapproxt tcapproxv G e e' t 
    -> tcapproxv G e e' t.
Proof.
exact (tcapprox_tcompatible_weak ander).
Qed.


Theorem tcapprox_tsemicongruence :
  tsemicongruence tcapproxt tcapproxv.
Proof.
do2 4 split.
  exact tcapprox_tcregular.

  exact tcapprox_tcreflexive.

  exact tcapprox_tctransitive.

  exact tcapprox_tcsubstitutive.

  exact tcapprox_tcompatible_weak.
Qed.


Lemma tcapproxt_halt :
  forall e e' t,
    tcapproxt nil e e' t
    -> thalt nil e
    -> thalt nil e'.
Proof.
exact tcapprox_tconsistent.
Qed.


Lemma tcapprox_with_wf :
  (forall G e e' t,
     (wf G -> tcapproxt G e e' t)
     -> tcapproxt G e e' t)
  /\
  (forall G e e' t,
     (wf G -> tcapproxv G e e' t)
     -> tcapproxv G e e' t).
Proof.
assert (tcregular
         (fun G e e' t => wf G -> tcapproxt G e e' t)
         (fun G e e' t => wf G -> tcapproxv G e e' t)) as Hreg.
  split.
    intros G e e' t Hwf Hap.
    exact (tcapproxt_reg _#4 Hwf (Hap Hwf)).

    intros G e e' t Hwf Hap.
    exact (tcapproxv_reg _#4 Hwf (Hap Hwf)).
assert (tsemicongruence
         (fun G e e' t => wf G -> tcapproxt G e e' t)
         (fun G e e' t => wf G -> tcapproxv G e e' t)) as Hsc.
  do2 4 split.
    exact Hreg.

    split.
      intros G e t Hof Htar Hwf.
      apply tcapproxt_refl; auto; done.

      intros G e t Hof Hval Htar Hwf.
      apply tcapproxv_refl; auto; done.

    split.
      intros G t e1 e2 e3 H12 H23 Hwf.
      eapply tcapproxt_trans; eauto; done.

      intros G t e1 e2 e3 H12 H23 Hwf.
      eapply tcapproxv_trans; eauto; done.

    split.
      intros G1 s G2 e e' t Hs Htars Hwf Hap _.
      eapply (tcapprox_tcsubstitutive andel); eauto; done.

      intros G1 s G2 e e' t Hs Htars Hwf Hap _.
      eapply (tcapprox_tcsubstitutive ander); eauto; done.

    split.
      intros G e e' t Hwf Hcc _.
      apply tcapproxt_comp_weak_ifwf; auto; [].
      apply (ttcc_mono_wf _#2 (fun G e e' t => wf G -> tcapproxt G e e' t) _ (fun G e e' t => wf G -> tcapproxv G e e' t)); auto using same_tsregular; done.

      intros G e e' t Hwf Hcc _.
      apply tcapproxv_comp_weak_ifwf; auto; [].
      apply (tvcc_mono_wf _#2 (fun G e e' t => wf G -> tcapproxt G e e' t) _ (fun G e e' t => wf G -> tcapproxv G e e' t)); auto using same_tsregular; done.
assert (tconsistent
         (fun G e e' t => wf G -> tcapproxt G e e' t)) as Hcons.
  intros e e' t Hap Hhalt.
  eapply tcapprox_tconsistent; eauto using wf_nil; done.
split.
  intros G e e' t Hap.
  exact (tcapproxt_univ _#6 Hsc Hcons Hap).

  intros G e e' t Hap.
  exact (tcapproxv_univ _#6 Hsc Hcons Hap).
Qed.
      

Lemma tcapproxt_with_wf :
  forall G e e' t,
    (wf G -> tcapproxt G e e' t)
    -> tcapproxt G e e' t.
Proof.
exact (tcapprox_with_wf andel).
Qed.


Lemma tcapproxv_with_wf :
  forall G e e' t,
    (wf G -> tcapproxv G e e' t)
    -> tcapproxv G e e' t.
Proof.
exact (tcapprox_with_wf ander).
Qed.


Lemma weaken1_tcapproxt :
  forall G A e e' t,
    tcapproxt G e e' t
    -> tcapproxt (G; A) (shift1 e) (shift1 e') (shift1 t).
Proof.
intros G A e e' t Hap.
apply tcapproxt_with_wf; [].
intro Hwf.
eapply subst_tcapproxt; eauto using weaken1, wf_tail, targetsub_sh; done.
Qed.


Lemma weaken1_tcapproxv :
  forall G A e e' t,
    tcapproxv G e e' t
    -> tcapproxv (G; A) (shift1 e) (shift1 e') (shift1 t).
Proof.
intros G A e e' t Hap.
apply tcapproxv_with_wf; [].
intro Hwf.
eapply subst_tcapproxv; eauto using weaken1, wf_tail, targetsub_sh; done.
Qed.


Lemma tcapproxt_comp_weak :
  forall G e e' t,
    ttcc samek samec tcapproxt tcapproxv G e e' t 
    -> tcapproxt G e e' t.
Proof.
intros.
apply tcapproxt_with_wf; [].
intro.
apply tcapproxt_comp_weak_ifwf; auto; done.
Qed.


Lemma tcapproxv_comp_weak :
  forall G e e' t,
    tvcc samek samec tcapproxt tcapproxv G e e' t 
    -> tcapproxv G e e' t.
Proof.
intros.
apply tcapproxv_with_wf; [].
intro.
apply tcapproxv_comp_weak_ifwf; auto; done.
Qed.


Lemma tcapproxt_comp :
  forall G e e' t,
    ttcc eqk eqc tcapproxt tcapproxv G e e' t
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Htcc.
apply tcapproxt_with_wf; [].
#5 induct Htcc;
try (intros; apply tcapproxt_comp_weak; eauto using ttcc; done).

(* lam *)
intros G t1 t1' t2 e e' Heqc Hap Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hoft1 & Hoft1' & _).
so (tcapproxt_reg _#4 (wf_tm Hwf Hoft1) Hap) as (_ & Hofe' & Htar & Htar').
so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) Hofe')) as Hoft2.
simp_sub in Hoft2.
#2 apply (tcapproxt_trans _ _ (tm_lam t1 e')).
  apply tcapproxt_comp_weak; apply ttcc_lam; auto with same; done.

  #3 apply treduce_tcapproxt; auto using target.
    apply oft_lam; auto; done.
  
    #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
      apply oft_lam; auto; [].
      eapply oft_conv_0t; eauto; done.
  
      apply eqc_arrow; auto with static; done.
  
    apply treduce_lam; [].
    apply treduce_refl; done.

(* plam *)
intros G k k' t e e' Heqk Hap Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (tcapproxv_reg _#4 (wf_cn Hwf Hofk) Hap) as (_ & Hofe' & _ & Hval' & Htar & Htar').
so (oft_reg _#3 (wf_cn Hwf Hofk) Hofe') as Hoft.
#2 apply (tcapproxt_trans _ _ (tm_plam k e')).
  apply tcapproxt_comp_weak; apply ttcc_plam; auto with same; done.

  #3 apply treduce_tcapproxt; auto using target.
    apply oft_plam; auto; done.

    #2 apply (@oft_equiv _ (cn_all k' t)).
      apply oft_plam; auto; [].
      eapply oft_conv_0c; eauto; done.

      so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft) as Hoft'; simp_sub in Hoft'.
      apply eqc_all; auto with static; done.

  apply treduce_plam; [].
  apply treduce_refl; done.

(* papp *)
intros G k t e e' c c' Hap Heqc Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (tcapproxt_reg _#4 Hwf Hap) as (_ & Hofe' & _ & Htar').
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe')) as (kk & H).
invert H; intros _ Hoft _; clear kk H.
#2 apply (tcapproxt_trans _ _ (tm_papp e' c)).
  apply tcapproxt_comp_weak; eapply ttcc_papp; eauto with same; done.

  #3 apply treduce_tcapproxt; auto using target.
    eapply oft_papp; eauto; done.

    #2 apply (@oft_equiv _ (subst1 c' t)).
      eapply oft_papp; eauto; done.

      apply eqc_symm; [].
      eapply funct_cn_type; eauto; done.

    apply treduce_papp1; [].
    apply treduce_refl; done.

(* pack *)
intros G k k' t t' c c' e e' Heqc Heqk Hap Heqt Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & _).
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (tcapproxt_reg _#4 Hwf Hap) as (_ & Hofe' & Htar & Htar').
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
#2 apply (tcapproxt_trans _ _ (tm_pack c e' k t)).
  apply tcapproxt_comp_weak; eapply ttcc_pack; eauto with same; done.

  #3 apply treduce_tcapproxt; auto using target.
    apply oft_pack; auto; done.

    #2 apply (@oft_equiv _ (cn_exists k' t')).
      #3 apply oft_pack.
        eapply ofc_subsume; eauto with static; done.

        eapply oft_equiv; eauto; [].
        eapply funct_eqc_type; eauto; done.

        so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
        assumption.

      apply eqc_symm; [].
      apply eqc_exists; auto; done.

    apply treduce_pack; [].
    apply treduce_refl; done.

(* fix *)
intros G t t' e e' Heqt Hap Hwf.
so (eqc_reg _#4 Hwf Heqt) as (Hoft & Hoft' & _).
so (tcapproxt_reg _#4 Hwf Hap) as (_ & Hofe' & _ & Htar').
#2 apply (tcapproxt_trans _ _ (tm_fix t e')).
  apply tcapproxt_comp_weak; eapply ttcc_fix; eauto with same; done.

  #3 apply treduce_tcapproxt; auto using target.
    apply oft_fix; auto; done.

    #2 apply (@oft_equiv _ t').
      apply oft_fix; [].
      eapply oft_equiv; eauto with static; done.

      apply eqc_symm; assumption.

    apply treduce_fix1; [].
    apply treduce_refl; done.
Qed.


Lemma tcapproxv_comp :
  forall G e e' t,
    tvcc eqk eqc tcapproxt tcapproxv G e e' t
    -> tcapproxv G e e' t.
Proof.
intros G e e' t Htvcc.
apply tcapproxv_with_wf; [].
#3 induct Htvcc;
try (intros; apply tcapproxv_comp_weak; eauto using tvcc; done).

(* lam *)
intros G t1 t1' t2 e e' Heqc Hap Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hoft1 & Hoft1' & _).
so (tcapproxt_reg _#4 (wf_tm Hwf Hoft1) Hap) as (_ & Hofe' & Htar & Htar').
so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) Hofe')) as Hoft2.
simp_sub in Hoft2.
#2 apply (tcapproxv_trans _ _ (tm_lam t1 e')).
  apply tcapproxv_comp_weak; apply tvcc_lam; auto with same; done.

  #3 apply treduce_tcapproxv; auto using target, svalue.
    apply oft_lam; auto; done.
  
    #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
      apply oft_lam; auto; [].
      eapply oft_conv_0t; eauto; done.
  
      apply eqc_arrow; auto with static; done.
  
    apply treduce_lam; [].
    apply treduce_refl; done.

(* plam *)
intros G k k' t e e' Heqk Hap Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (tcapproxv_reg _#4 (wf_cn Hwf Hofk) Hap) as (_ & Hofe' & _ & Hval' & _ & Htar').
so (oft_reg _#3 (wf_cn Hwf Hofk) Hofe') as Hoft.
#2 apply (tcapproxv_trans _ _ (tm_plam k e')).
  apply tcapproxv_comp_weak; apply tvcc_plam; auto with same; done.

  #3 apply treduce_tcapproxv; auto using svalue, target.
    apply oft_plam; auto; done.

    #2 apply (@oft_equiv _ (cn_all k' t)).
      apply oft_plam; auto; [].
      eapply oft_conv_0c; eauto; done.

      so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft) as Hoft'; simp_sub in Hoft'.
      apply eqc_all; auto with static; done.

  apply treduce_plam; [].
  apply treduce_refl; done.

(* pack *)
intros G k k' t t' c c' e e' Heqc Heqk Hap Heqt Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & _).
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (tcapproxv_reg _#4 Hwf Hap) as (_ & Hofe' & _ & Hval' & _ & Htar').
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
#2 apply (tcapproxv_trans _ _ (tm_pack c e' k t)).
  apply tcapproxv_comp_weak; eapply tvcc_pack; eauto with same; done.

  #3 apply treduce_tcapproxv; auto using svalue, target.
    apply oft_pack; auto; done.

    #2 apply (@oft_equiv _ (cn_exists k' t')).
      #3 apply oft_pack.
        eapply ofc_subsume; eauto with static; done.

        eapply oft_equiv; eauto; [].
        eapply funct_eqc_type; eauto; done.

        so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
        assumption.

      apply eqc_symm; [].
      apply eqc_exists; auto; done.

    apply treduce_pack; [].
    apply treduce_refl; done.
Qed.


Lemma tcapprox_tcompatible_strong :
  tcompatible eqk eqc tcapproxt tcapproxv.
Proof.
split.
  intros; apply tcapproxt_comp; auto; done.

  intros; apply tcapproxv_comp; auto; done.
Qed.


(* Contextual equivalence *)

Lemma tceqt_univ :
  forall Rt Rv G e e' t,
    tcongruence Rt Rv
    -> tconsistent Rt
    -> Rt G e e' t
    -> tceqt G e e' t.
Proof.
intros Rt Rv G e e' t Hcong Hcons HR.
destruct Hcong as (Hsc & (Hsymm & _)).
split.
  eapply tcapproxt_univ; eauto; done.

  eapply tcapproxt_univ; eauto; [].
  apply Hsymm; assumption.
Qed.


Lemma tceqv_univ :
  forall Rt Rv G e e' t,
    tcongruence Rt Rv
    -> tconsistent Rt
    -> Rv G e e' t
    -> tceqv G e e' t.
Proof.
intros Rt Rv G e e' t Hcong Hcons HR.
destruct Hcong as (Hsc & (_ & Hsymm)).
split.
  eapply tcapproxv_univ; eauto; done.

  eapply tcapproxv_univ; eauto; [].
  apply Hsymm; assumption.
Qed.


Lemma tceqt_univ_quasi :
  forall Rt Rv G e e' t,
    tquasicongruence Rt Rv
    -> tconsistent Rt
    -> tconsistent (fun G e e' t => Rt G e' e t)
    -> Rt G e e' t
    -> tceqt G e e' t.
Proof.
intros Rt Rv G e e' t Hquasi Hcon Hcon' HR.
#3 apply (tceqt_univ (tstc Rt) (tstc Rv)).
  apply stc_tcongruence; auto; done.

  apply tstc_tconsistent; auto; done.

  eapply tstc_intro; eauto; done.
Qed.


Lemma tceqv_univ_quasi :
  forall Rt Rv G e e' t,
    tquasicongruence Rt Rv
    -> tconsistent Rt
    -> tconsistent (fun G e e' t => Rt G e' e t)
    -> Rv G e e' t
    -> tceqv G e e' t.
Proof.
intros Rt Rv G e e' t Hquasi Hcon Hcon' HR.
#3 apply (tceqv_univ (tstc Rt) (tstc Rv)).
  apply stc_tcongruence; auto; done.

  apply tstc_tconsistent; auto; done.

  eapply tstc_intro; eauto; done.
Qed.


Lemma tceqt_with_wf :
  forall G e e' t,
    (wf G -> tceqt G e e' t)
    -> tceqt G e e' t.
Proof.
intros G e e' t Heq.
split.
  apply tcapproxt_with_wf; [].
  intro Hwf.
  exact (Heq Hwf andel).

  apply tcapproxt_with_wf; [].
  intro Hwf.
  exact (Heq Hwf ander).
Qed.


Lemma tceqv_with_wf :
  forall G e e' t,
    (wf G -> tceqv G e e' t)
    -> tceqv G e e' t.
Proof.
intros G e e' t Heq.
split.
  apply tcapproxv_with_wf; [].
  intro Hwf.
  exact (Heq Hwf andel).

  apply tcapproxv_with_wf; [].
  intro Hwf.
  exact (Heq Hwf ander).
Qed.


Lemma tceq_tsemicongruence :
  tsemicongruence tceqt tceqv.
Proof.
#2 apply conjunction_tsemicongruence.    
  exact tcapprox_tsemicongruence.

  apply transpose_tsemicongruence; exact tcapprox_tsemicongruence.
Qed.


Theorem tceq_tcongruence :
  tcongruence tceqt tceqv.
Proof.
split.
  exact tceq_tsemicongruence.

  split.
    intros G t e e' (? & ?).
    split; assumption.

    intros G t e e' (? & ?).
    split; assumption.
Qed.


Theorem tceq_tconsistent :
  tconsistent tceqt.
Proof.
intros e e' t Hceq Hhalt.
destruct Hceq.
eapply tcapprox_tconsistent; eauto; done.
Qed.


Lemma tceqt_reg :
  forall G e e' t,
    wf G
    -> tceqt G e e' t
    -> oft G e t /\ oft G e' t /\ target e /\ target e'.
Proof.
exact (tceq_tsemicongruence andel andel).
Qed.


Lemma tceqv_reg :
  forall G e e' t,
    wf G
    -> tceqv G e e' t
    -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e'.
Proof.
exact (tceq_tsemicongruence andel ander).
Qed.


Lemma tceqt_refl :
  forall G e t, oft G e t -> target e -> tceqt G e e t.
Proof.
exact (tceq_tsemicongruence ander andel andel).
Qed.


Lemma tceqv_refl :
  forall G e t, oft G e t -> svalue e -> target e -> tceqv G e e t.
Proof.
exact (tceq_tsemicongruence ander andel ander).
Qed.


Lemma tceqt_symm :
  forall G e e' t, tceqt G e e' t -> tceqt G e' e t.
Proof.
intros G e e' t (? & ?).
split; assumption.
Qed.


Lemma tceqv_symm :
  forall G e e' t, tceqv G e e' t -> tceqv G e' e t.
Proof.
intros G e e' t (? & ?).
split; assumption.
Qed.


Lemma tceqt_trans :
  forall G e1 e2 e3 t,
    tceqt G e1 e2 t
    -> tceqt G e2 e3 t
    -> tceqt G e1 e3 t.
Proof.
intros.
eapply (tceq_tsemicongruence ander ander andel andel); eauto.
Qed.


Lemma tceqv_trans :
  forall G e1 e2 e3 t,
    tceqv G e1 e2 t
    -> tceqv G e2 e3 t
    -> tceqv G e1 e3 t.
Proof.
intros.
eapply (tceq_tsemicongruence ander ander andel ander); eauto.
Qed.


Lemma subst_tceqt :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> targetsub s
    -> wf G2
    -> tceqt G2 e e' t
    -> tceqt G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (tceq_tsemicongruence ander ander ander andel andel).
Qed.


Lemma subst_tceqv :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> targetsub s
    -> wf G2
    -> tceqv G2 e e' t
    -> tceqv G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (tceq_tsemicongruence ander ander ander andel ander).
Qed.


Lemma tceq_tcompatible_strong :
  tcompatible eqk eqc tceqt tceqv.
Proof.
#2 apply conjunction_tcompatible.
  exact tcapprox_tcompatible_strong.

  #7 apply transpose_tcompatible.
    split; auto with static; done.

    exact tcapprox_tcregular.

    split; eauto using subst_eqk, subst_eqc; done.

    exact tcapprox_tcsubstitutive.

    eauto with static; done.

    split; eauto with static; done.

    split.
      intros G e e' t _ Httcc.
      apply tcapproxt_comp; [].
      refine (ttcc_mono _#16 Httcc); eauto with static; done.

      intros G e e' t _ Httcc.
      apply tcapproxv_comp; [].
      refine (tvcc_mono _#16 Httcc); eauto with static; done.
Qed.


Lemma tceqt_comp :
  forall G e e' t,
    ttcc eqk eqc tceqt tceqv G e e' t
    -> tceqt G e e' t.
Proof.
intros.
apply tceqt_with_wf; [].
intro.
apply tceq_tcompatible_strong; auto; done.
Qed.


Lemma tceqv_comp :
  forall G e e' t,
    tvcc eqk eqc tceqt tceqv G e e' t
    -> tceqv G e e' t.
Proof.
intros.
apply tceqv_with_wf; [].
intro.
apply tceq_tcompatible_strong; auto; done.
Qed.


Lemma tceqt_halt :
  forall e e' t,
    tceqt nil e e' t
    -> thalt nil e
    -> thalt nil e'.
Proof.
intros e e' t Htceqt Hhalt.
destruct Htceqt.
eapply tcapproxt_halt; eauto.
Qed.


Lemma tceqt_halt_iff :
  forall e e' t,
    tceqt nil e e' t
    -> (thalt nil e <-> thalt nil e').
Proof.
intros e e' t H.
split; intro; eapply tceqt_halt; eauto using tceqt_symm.
Qed.
  

(* Evaluation implies contextual equivalence *)

Lemma tstep_target :
  forall G e e',
    tstep G e e'
    -> target e
    -> target e'.
Proof.
intros G e e' Hstep Htar.
revert Htar.
#7 induct Hstep; try (intros; invert Htar; intros; eauto using target; done).

(* app beta *)
intros G t e1 v2 _ Htar.
invert Htar; [].
intros Htar' Htar2.
invert Htar'; [].
intros Htar1.
apply subst_target; eauto using targetsub; done.

(* pi1 beta *)
intros G v1 v2 _ _ Htar.
invert Htar; [].
intros Htar'.
invert Htar'; [].
intros Htar1 Htar2.
assumption.

(* pi2 beta *)
intros G v1 v2 _ _ Htar.
invert Htar; [].
intros Htar'.
invert Htar'; [].
intros Htar1 Htar2.
assumption.

(* papp beta *)
intros G k e c Htar.
invert Htar; [].
intros Htar' Htar2.
invert Htar'; [].
intros Htar1.
apply subst_target; eauto using targetsub; done.

(* unpack beta *)
intros G c v k t e _ Htar.
invert Htar; [].
intros Htar' Htare.
invert Htar'; [].
intros Htarc Htarv.
apply subst_target; eauto using targetsub; done.

(* fix beta *)
intros G t e Htar.
invert Htar; [].
intros Htar'.
apply target_app; auto; [].
apply target_lam; [].
apply subst_target; eauto using targetsub; done.

(* lett beta *)
intros G v1 e2 _ Htar.
invert Htar; [].
intros Htar1 Htar2.
apply subst_target; eauto using targetsub; done.
Qed.


Lemma tsteps_target :
  forall G e e',
    star (tstep G) e e'
    -> target e
    -> target e'.
Proof.
intros G e e' H Htar.
revert Htar.
induct H; eauto using tstep_target.
Qed.


Lemma tstep_tcapproxt :
  forall G e e' t,
    oft G e t
    -> target e
    -> tstep G e e'
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Hof Htar Hstep.
apply tcapproxt_with_wf; [].
intro Hwf.
apply treduce_tcapproxt; eauto 6 using subst_oft, subst_tstep, subof_aof_dof, subof_aof_oof, preservation_tm, tstep_treduce, tstep_target.
Qed.


Lemma anti_tstep_tcapproxt :
  forall G e e' t,
    oft G e t
    -> target e
    -> tstep G e e'
    -> tcapproxt G e' e t.
Proof.
intros G e e' t Hof Htar Hstep.
apply tcapproxt_with_wf; [].
intro Hwf.
apply anti_treduce_tcapproxt; eauto 6 using subst_oft, subst_tstep, subof_aof_dof, subof_aof_oof, preservation_tm, tstep_treduce, tstep_target.
Qed.


Lemma tsteps_tcapproxt :
  forall G e e' t,
    oft G e t
    -> target e
    -> star (tstep G) e e'
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Hof Htar Hsteps.
apply tcapproxt_with_wf; [].
intro Hwf.
revert Hof Htar.
induct Hsteps.

(* refl *)
auto using tcapproxt_refl; done.

(* step *)
intros e1 e2 e3 Hstep _ IH Hof Htar.
so (IH (preservation_tm _#4 Hwf Hof Hstep)) as H23.
eapply tcapproxt_trans; eauto using tstep_target; [].
apply tstep_tcapproxt; auto; done.
Qed.


Lemma anti_tsteps_tcapproxt :
  forall G e e' t,
    oft G e t
    -> target e
    -> star (tstep G) e e'
    -> tcapproxt G e' e t.
Proof.
intros G e e' t Hof Htar Hsteps.
apply tcapproxt_with_wf; [].
intros Hwf.
revert Hof Htar.
induct Hsteps.

(* refl *)
intros e Hof Htar.
apply tcapproxt_refl; auto; done.

(* step *)
intros e1 e2 e3 H12 _ IH Hof1 Htar1.
eapply tcapproxt_trans; eauto using anti_tstep_tcapproxt, preservation_tm, tstep_target; done.
Qed.


Lemma anti_tsteps_tcapproxt_trans :
  forall G e1 e2 e3 t,
    star (tstep G) e2 e1
    -> tcapproxt G e2 e3 t
    -> tcapproxt G e1 e3 t.
Proof.
intros G e1 e2 e3 t Hsteps Hap.
apply tcapproxt_with_wf; [].
intro Hwf.
eapply tcapproxt_trans; eauto; [].
so (tcapproxt_reg _#4 Hwf Hap andel) as Hof.
eapply anti_tsteps_tcapproxt; eauto; [].
exact (tcapproxt_reg _#4 Hwf Hap anderrl).
Qed.


Lemma tstep_tceqt :
  forall G e e' t,
    oft G e t
    -> target e
    -> tstep G e e'
    -> tceqt G e e' t.
Proof.
intros G e e' t Hof Htar Hstep.
split.
  apply tstep_tcapproxt; auto; done.

  apply anti_tstep_tcapproxt; auto; done.
Qed.


Lemma tsteps_tceqt :
  forall G e e' t,
    oft G e t
    -> target e
    -> star (tstep G) e e'
    -> tceqt G e e' t.
Proof.
intros G e e' t Hof Htar Hsteps.
apply tceqt_with_wf; [].
intro Hwf.
revert Hof Htar.
induct Hsteps.

(* refl *)
auto using tceqt_refl; done.

(* step *)
intros e1 e2 e3 Hstep _ IH Hof Htar.
so (IH (preservation_tm _#4 Hwf Hof Hstep)) as H23.
eapply tceqt_trans; eauto using tstep_target; [].
apply tstep_tceqt; auto; done.
Qed.


Lemma tceqt_trans_formation :
  forall G e1 e2 e3 t,
    tceqt G e1 e2 t
    -> (oft G e2 t -> target e2 -> tceqt G e2 e3 t)
    -> tceqt G e1 e3 t.
Proof.
intros G e1 e2 e3 t H12 H23.
apply tceqt_with_wf; [].
intro Hwf.
so (H23 (tceqt_reg _#4 Hwf H12 anderl) (tceqt_reg _#4 Hwf H12 anderrr)).
eapply tceqt_trans; eauto; done.
Qed.


Lemma tceqt_tsteps_trans :
  forall G e1 e2 e3 t,
    tceqt G e1 e2 t
    -> star (tstep G) e2 e3
    -> tceqt G e1 e3 t.
Proof.
intros G e1 e2 e3 t H12 Hsteps.
apply tceqt_with_wf; [].
intro Hwf.
eapply tceqt_trans; eauto; [].
#2 apply tsteps_tceqt; auto.
  exact (tceqt_reg _#4 Hwf H12 anderl).

  exact (tceqt_reg _#4 Hwf H12 anderrr).
Qed.
    

Lemma anti_tsteps_tceqt_trans :
  forall G e1 e2 e3 t,
    star (tstep G) e2 e1
    -> tceqt G e2 e3 t
    -> tceqt G e1 e3 t.
Proof.
intros G e1 e2 e3 t H12 H23.
apply tceqt_symm; [].
eapply tceqt_tsteps_trans; eauto using tceqt_symm.
Qed.


Lemma tcapproxv_tcapproxt :
  forall G e e' t,
    tcapproxv G e e' t
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Hap.
apply tcapproxt_with_wf; [].
intros Hwf.
so (tcapproxv_reg _#4 Hwf Hap) as (Hofe & Hofe' & Hval & Hval' & Htar & Htar').
assert (tcapproxt G (tm_papp (tm_plam kd_one (shift1 e)) cn_star) (tm_papp (tm_plam kd_one (shift1 e')) cn_star) t) as Hap'.
  replsub t with (subst1 cn_star (shift1 t)).
  apply tcapproxt_comp; eapply ttcc_papp; eauto with static; [].
  apply tcapproxt_comp; apply ttcc_plam; auto with static; [].
  eapply subst_tcapproxv; eauto using weaken1, targetsub; done.
#2 apply (tcapproxt_trans _ _ (tm_papp (tm_plam kd_one (shift1 e)) cn_star)).
  #3 apply anti_tstep_tcapproxt; auto.
    relquest.
      eapply oft_papp; eauto with static; [].
      eapply oft_plam; eauto using subst_svalue with static; [].
      apply weaken1_oft; eauto; done.

      simp_sub; reflexivity.

    apply target_papp; auto using target; [].
    apply target_plam; [].
    apply subst_target; eauto using targetsub; done.

    relquest.
      apply tstep_papp2; done.

      simp_sub; reflexivity.

  eapply tcapproxt_trans; eauto; [].
  #3 apply tstep_tcapproxt.
    relquest.
      eapply oft_papp; eauto with static; [].
      eapply oft_plam; eauto using subst_svalue with static; [].
      apply weaken1_oft; eauto; done.

      simp_sub; reflexivity.

    apply target_papp; auto using target; [].
    apply target_plam; [].
    apply subst_target; eauto using targetsub; done.

    relquest.
      apply tstep_papp2; done.

      simp_sub; reflexivity.
Qed.


Lemma tceqv_tceqt :
  forall G e e' t,
    tceqv G e e' t
    -> tceqt G e e' t.
Proof.
intros G e e' t (H & H').
split; auto using tcapproxv_tcapproxt.
Qed.
