
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Singleton.


Inductive split : (* sg *) term -> (* kd *) term -> (* cn => cn *) term -> Prop :=
| split_one
    : split sg_one kd_one cn_unit

| split_stat {k}
    : split (sg_stat k) k cn_unit

| split_dyn {t}
    : split (sg_dyn t) kd_one (shift1 t)

| split_pigen {T1 T2 k1 k2 t1 t2}
    : split T1 k1 t1
      -> split T2 k2 t2
      -> split 
         (sg_pigen T1 T2)
         kd_one
         (cn_all 
            (shift1 k1) 
            (cn_arrow 
               (subst (under 1 sh1) t1)
               (cn_exists
                  (subst (under 1 sh1) k2)
                  (subst (under 2 sh1) t2))))

| split_piapp {T1 T2 k1 k2 t1 t2}
    : split T1 k1 t1
      -> split T2 k2 t2
      -> split 
         (sg_piapp T1 T2)
         (kd_pi k1 k2)
         (cn_all
            (shift1 k1)
            (cn_arrow
               (subst (under 1 sh1) t1)
               (subst (dot (cn_app (var 1) (var 0)) (dot (var 0) (sh 2))) t2)))

| split_sigma {T1 T2 k1 k2 t1 t2}
    : split T1 k1 t1
      -> split T2 k2 t2
      -> split 
         (sg_sigma T1 T2)
         (kd_sigma k1 k2)
         (cn_prod 
            (subst (dot (cn_pi1 (var 0)) sh1) t1)
            (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2))
.


Inductive trsubs : context -> (* sg *) term -> (* sg *) term -> (* tm *) term -> Prop :=
| trsubs_refl {G T T' k t}
    : split T k t
      -> eqs G T T'
      -> trsubs G T T' 
         (tm_plam k (tm_lam t (var 0)))

| trsubs_trans {G T1 T2 T3 k1 t1 f1 f2}
    : split T1 k1 t1
      -> trsubs G T1 T2 f1
      -> trsubs G T2 T3 f2
      -> trsubs G T1 T3 
         (tm_plam k1 
            (tm_lam t1 
               (tm_app 
                  (tm_papp (subst (sh 2) f2) (var 1))
                  (tm_app (tm_papp (subst (sh 2) f1) (var 1)) (var 0)))))

| trsubs_stat {G k k'}
    : subk G k k'
      -> trsubs G (sg_stat k) (sg_stat k')
         (tm_plam k (tm_lam cn_unit (var 0)))

| trsubs_pigen {G T1 T1' T2 T2' k1 k1' t1 t1' k2 k2' t2 t2' f1 f2}
    : split T1 k1 t1
      -> split T1' k1' t1'
      -> split T2 k2 t2
      -> split T2' k2' t2'
      -> trsubs G T1' T1 f1
      -> trsubs (G; cl_cn k1') T2 T2' f2
      -> ofs (G; cl_cn k1) T2
      -> trsubs G (sg_pigen T1 T2) (sg_pigen T1' T2')
         (tm_plam kd_one
            (tm_lam (shift1 (cn_all k1 (cn_arrow t1 (cn_exists k2 t2))))
               (tm_plam (subst (sh 2) k1')
                  (tm_lam (subst (under 1 (sh 2)) t1')
                     (tm_unpack 
                        (tm_app 
                           (tm_papp (var 2) (var 1)) 
                           (tm_app (tm_papp (subst (sh 4) f1) (var 1)) (var 0)))
                        (tm_pack
                           (var 1)
                           (tm_app (tm_papp (subst (dot (var 3) (sh 6)) f2) (var 1)) (var 0))
                           (subst (dot (var 3) (sh 6)) k2')
                           (subst (under 1 (dot (var 3) (sh 6))) t2')))))))

| trsubs_piapp {G T1 T1' T2 T2' k1 k1' t1 t1' k2 t2 f1 f2}
    : split T1 k1 t1
      -> split T1' k1' t1'
      -> split T2 k2 t2
      -> trsubs G T1' T1 f1
      -> trsubs (G; cl_cn k1') T2 T2' f2
      -> ofs (G; cl_cn k1) T2
      -> trsubs G (sg_piapp T1 T2) (sg_piapp T1' T2')
         (tm_plam
            (kd_pi k1 k2)
            (tm_lam
               (cn_all
                  (shift1 k1) 
                  (cn_arrow
                     (subst (dot (var 0) (sh 2)) t1)
                     (subst (dot (cn_app (var 1) (var 0)) (dot (var 0) (sh 2))) t2)))
               (tm_plam
                  (subst (sh 2) k1')
                  (tm_lam
                     (subst (under 1 (sh 2)) t1')
                     (tm_app 
                        (tm_papp (subst (dot (var 1) (sh 4)) f2) (cn_app (var 3) (var 1)))
                        (tm_app
                           (tm_papp (var 2) (var 1))
                           (tm_app
                              (tm_papp (subst (sh 4) f1) (var 1))
                              (var 0))))))))

| trsubs_sigma {G T1 T1' T2 T2' k1 k1' t1 k2 t2 f1 f2}
    : split T1 k1 t1
      -> fsts T1' k1'
      -> split T2 k2 t2
      -> trsubs G T1 T1' f1
      -> trsubs (G; cl_cn k1) T2 T2' f2
      -> ofs (G; cl_cn k1') T2'
      -> trsubs G (sg_sigma T1 T2) (sg_sigma T1' T2')
         (tm_plam
            (kd_sigma k1 k2)
            (tm_lam
               (cn_prod 
                  (subst (dot (cn_pi1 (var 0)) sh1) t1)
                  (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2))
               (tm_pair
                  (tm_app
                     (tm_papp (subst (sh 2) f1) (cn_pi1 (var 1)))
                     (tm_pi1 (var 0)))
                  (tm_app
                     (tm_papp
                        (subst (dot (cn_pi1 (var 1)) (sh 2)) f2)
                        (cn_pi2 (var 1)))
                     (tm_pi2 (var 0))))))
.


Inductive trctx : context -> context -> Prop :=
| trctx_nil
    : trctx nil nil

| trctx_cn {G G' k}
    : trctx G G'
      -> trctx (G; cl_cn k) (G'; cl_cn k)

| trctx_tm {G G' t}
    : trctx G G'
      -> trctx (G; cl_tm t) (G'; cl_tm t)

| trctx_md {G G' T k t}
    : trctx G G'
      -> split T k t
      -> trctx (G; cl_cn k; cl_md (shift1 T) (var 0)) (G'; cl_cn k; cl_tm t)
.


Inductive trp : context -> (* md *) term -> (* sg *) term -> (* cn *) term -> (* tm *) term -> Prop :=
| trp_var {G i T c}
    : index i G (cl_md T c)
      -> trp G (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T)) (subst (sh (S i)) c) (var i)

| trp_star {G}
    : trp G md_star sg_one cn_star tm_star

| trp_stat {G c k}
    : ofc G c k
      -> trp G (md_stat c) (sg_stat (single c k)) c tm_star

| trp_dyn {G e t e'}
    : trt G e t e'
      -> trp G (md_dyn e) (sg_dyn t) cn_star e'

| trp_lamgn {G T1 T2 k1 t1 M e}
    : ofs G T1
      -> split T1 k1 t1
      -> tri (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) e
      -> trp G (md_lamgn T1 M) (sg_pigen T1 T2)
         cn_star (tm_plam k1 (tm_lam t1 e))

| trp_lamap {G T1 T2 k1 t1 M c e}
    : ofs G T1
      -> split T1 k1 t1
      -> trp (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) (shift1 c) e
      -> trp G (md_lamap T1 M) (sg_piapp T1 T2)
         (cn_lam k1 c)
         (tm_plam k1 (tm_lam t1 e))

| trp_appap {G T1 T2 M1 M2 c1 c2 e1 e2}
    : trp G M1 (sg_piapp T1 T2) c1 e1
      -> trp G M2 T1 c2 e2
      -> trp G (md_appap M1 M2) (subst1 c2 T2)
         (cn_app c1 c2)
         (tm_app (tm_papp e1 c2) e2)

| trp_pair {G T1 T2 M1 M2 c1 c2 e1 e2}
    : trp G M1 T1 c1 e1
      -> trp G M2 T2 c2 e2
      -> trp G (md_pair M1 M2) (sg_sigma T1 (shift1 T2)) (cn_pair c1 c2) (tm_pair e1 e2)

| trp_pi1 {G T1 T2 M c e}
    : trp G M (sg_sigma T1 T2) c e
      -> trp G (md_pi1 M) T1 (cn_pi1 c) (tm_pi1 e)

| trp_pi2 {G T1 T2 M c e}
    : trp G M (sg_sigma T1 T2) c e
      -> trp G (md_pi2 M) (subst1 (cn_pi1 c) T2) (cn_pi2 c) (tm_pi2 e)

| trp_lett {G t T e M e1 c2 e2}
    : trt G e t e1
      -> trp (G; cl_tm t) M (shift1 T) (shift1 c2) e2
      -> trp G (md_lett e M) T c2 (tm_lett e1 e2)

| trp_subsume {G T T' M c e f}
    : trp G M T c e
      -> trsubs G T T' f
      -> trp G M T' c (tm_app (tm_papp f c) e)


with tri : context -> (* md *) term -> (* sg *) term -> (* tm *) term -> Prop :=
| tri_appgn {G T1 T2 M1 M2 c2 e1 e2}
    : tri G M1 (sg_pigen T1 T2) e1
      -> trp G M2 T1 c2 e2
      -> tri G (md_appgn M1 M2) (subst1 c2 T2)
         (tm_unpack e1
            (tm_app (tm_papp (var 0) (subst (sh 2) c2)) (subst (sh 2) e2)))

| tri_appap {G T1 T2 M1 M2 c2 e1 e2 k2 t2}
    : tri G M1 (sg_piapp T1 T2) e1
      -> trp G M2 T1 c2 e2
      -> split T2 k2 t2
      -> tri G (md_appap M1 M2) (subst1 c2 T2)
         (tm_unpack e1 
            (tm_pack
               (cn_app (var 1) (subst (sh 2) c2))
               (tm_app (tm_papp (var 0) (subst (sh 2) c2)) (subst (sh 2) e2))
               (subst (dot (subst (sh 2) c2) (sh 2)) k2)
               (subst (dot (var 0) (dot (subst (sh 3) c2) (sh 3))) t2)))

| tri_pair {G T1 T2 M1 M2 e1 e2 k1 k2 t1 t2}
    : tri G M1 T1 e1
      -> tri G M2 T2 e2
      -> split T1 k1 t1
      -> split T2 k2 t2
      -> tri G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))
         (tm_unpack e1
            (tm_unpack (subst (sh 2) e2)
               (tm_pack
                  (cn_pair (var 3) (var 1))
                  (tm_pair (var 2) (var 0))
                  (kd_sigma (subst (sh 4) k1) (subst (sh 5) k2))
                  (cn_prod
                     (subst (dot (cn_pi1 (var 0)) (sh 5)) t1)
                     (subst (dot (cn_pi2 (var 0)) (sh 5)) t2)))))

| tri_unpack {G k t T e M e1 e2}
    : trt G e (cn_exists k t) e1
      -> tri (G; cl_cn k; cl_tm t) M (subst (sh 2) T) e2
      -> ofs G T
      -> tri G (md_unpack e M T) T 
         (tm_unpack e1 e2)

| tri_lett {G t T e M e1 e2}
    : trt G e t e1
      -> tri (G; cl_tm t) M (shift1 T) e2
      -> tri G (md_lett e M) T 
         (tm_lett e1 e2)

| tri_letm {G T1 T2 k1 M1 M2 e1 e2}
    : tri G M1 T1 e1
      -> fsts T1 k1
      -> tri (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M2 (subst (sh 2) T2) e2
      -> ofs G T2
      -> tri G (md_letm M1 M2 T2) T2
         (tm_unpack e1 e2)

| tri_seal {G T M e}
    : tri G M T e
      -> tri G (md_seal M T) T e

| tri_forget {G T M c e k t}
    : trp G M T c e
      -> split T k t
      -> tri G M T
         (tm_pack c e k t)

| tri_subsume {G T T' M e f k' t'}
    : tri G M T e
      -> trsubs G T T' f
      -> split T' k' t'
      -> tri G M T'
         (tm_unpack e 
            (tm_pack 
               (var 1)
               (tm_app (tm_papp (subst (sh 2) f) (var 1)) (var 0))
               (subst (sh 2) k')
               (subst (dot (var 0) (sh 3)) t')))


with trt : context -> (* tm *) term -> (* cn *) term -> (* tm *) term -> Prop :=
| trt_var {G i t}
    : index i G (cl_tm t)
      -> trt G (var i) (subst (sh (S i)) t) (var i)

| trt_star {G}
    : trt G tm_star cn_unit tm_star

| trt_lam {G t1 t2 e e'}
    : ofc G t1 kd_type
      -> trt (G; cl_tm t1) e (shift1 t2) e'
      -> trt G (tm_lam t1 e) (cn_arrow t1 t2) (tm_lam t1 e')

| trt_app {G t1 t2 e1 e1' e2 e2'}
    : trt G e1 (cn_arrow t1 t2) e1'
      -> trt G e2 t1 e2'
      -> trt G (tm_app e1 e2) t2 (tm_app e1' e2')

| trt_pair {G t1 t2 e1 e1' e2 e2'}
    : trt G e1 t1 e1'
      -> trt G e2 t2 e2'
      -> trt G (tm_pair e1 e2) (cn_prod t1 t2) (tm_pair e1' e2')

| trt_pi1 {G t1 t2 e e'}
    : trt G e (cn_prod t1 t2) e'
      -> trt G (tm_pi1 e) t1 (tm_pi1 e')

| trt_pi2 {G t1 t2 e e'}
    : trt G e (cn_prod t1 t2) e'
      -> trt G (tm_pi2 e) t2 (tm_pi2 e')

| trt_plam {G k t e e'}
    : ofk G k
      -> trt (G; cl_cn k) e t e'
      -> svalue e
      -> trt G (tm_plam k e) (cn_all k t) (tm_plam k e')

| trt_papp {G k t e e' c}
    : trt G e (cn_all k t) e'
      -> ofc G c k
      -> trt G (tm_papp e c) (subst1 c t) (tm_papp e' c)

| trt_pack {G k t c e e'}
    : ofc G c k
      -> trt G e (subst1 c t) e'
      -> ofc (G; cl_cn k) t kd_type
      -> trt G (tm_pack c e k t) (cn_exists k t) (tm_pack c e' k t)

| trt_unpack {G k t t' e1 e1' e2 e2'}
    : trt G e1 (cn_exists k t) e1'
      -> trt (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') e2'
      -> ofc G t' kd_type
      -> trt G (tm_unpack e1 e2) t' (tm_unpack e1' e2')

| trt_fix {G t e e'}
    : trt G e (cn_arrow (cn_arrow cn_unit t) t) e'
      -> trt G (tm_fix t e) t (tm_fix t e')

| trt_lett {G t1 t2 e1 e1' e2 e2'}
    : trt G e1 t1 e1'
      -> trt (G; cl_tm t1) e2 (shift1 t2) e2'
      -> trt G (tm_lett e1 e2) t2 (tm_lett e1' e2')

| trt_equiv {G t t' e e'}
    : trt G e t e'
      -> eqc G t t' kd_type
      -> trt G e t' e'

| trt_letm {G T k t M e e1 e2}
    : tri G M T e1
      -> fsts T k
      -> trt (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t) e2
      -> ofc G t kd_type
      -> trt G (tm_letm M e) t (tm_unpack e1 e2)

| trt_ext {G t M e}
    : tri G M (sg_dyn t) e
      -> trt G (tm_ext M) t 
         (tm_unpack e (var 0))
.


Scheme trp_mut_ind := Minimality for trp Sort Prop
with   tri_mut_ind := Minimality for tri Sort Prop
with   trt_mut_ind := Minimality for trt Sort Prop.
Combined Scheme trans_ind from trp_mut_ind, tri_mut_ind, trt_mut_ind.
