Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import ProveSubof.
Require Import Determinism.
Require Import Formation.
Require Import Regularity.
Require Import Singleton.
Require Import Principal.
Require Import Inversion.
Require Import Preservation.
Require Import Congruence.
Require Import Contextual.
Require Import Equivalence.
Require Import Reductive.
Require Import Translation.
Require Import Split.
Require Import Inverse.
Require Breduce.


(* Equality modulo erasure *)
Inductive teme : (* tm *) term -> (* tm *) term -> Prop :=
| teme_refl {e}
    : teme e e

| teme_lam {t t' e e'}
    : teme e e'
      -> teme (tm_lam t e) (tm_lam t' e')

| teme_app {e1 e1' e2 e2'}
    : teme e1 e1'
      -> teme e2 e2'
      -> teme (tm_app e1 e2) (tm_app e1' e2')

| teme_pair {e1 e1' e2 e2'}
    : teme e1 e1'
      -> teme e2 e2'
      -> teme (tm_pair e1 e2) (tm_pair e1' e2')

| teme_pi1 {e e'}
    : teme e e'
      -> teme (tm_pi1 e) (tm_pi1 e')

| teme_pi2 {e e'}
    : teme e e'
      -> teme (tm_pi2 e) (tm_pi2 e')

| teme_plam {k k' e e'}
    : teme e e'
      -> teme (tm_plam k e) (tm_plam k' e')

| teme_papp {e e' c c'}
    : teme e e'
      -> teme (tm_papp e c) (tm_papp e' c')

| teme_pack {c c' e e' k k' t t'}
    : teme e e'
      -> teme (tm_pack c e k t) (tm_pack c' e' k' t')

| teme_unpack {e1 e1' e2 e2'}
    : teme e1 e1'
      -> teme e2 e2'
      -> teme (tm_unpack e1 e2) (tm_unpack e1' e2')

| teme_lett {e1 e1' e2 e2'}
    : teme e1 e1'
      -> teme e2 e2'
      -> teme (tm_lett e1 e2) (tm_lett e1' e2')

| teme_letm {M M' e e'}
    : meme M M'
      -> teme e e'
      -> teme (tm_letm M e) (tm_letm M' e')

| teme_ext {M M'}
    : meme M M'
      -> teme (tm_ext M) (tm_ext M')

| teme_symm {e e'}
    : teme e e'
      -> teme e' e

| teme_trans {e1 e2 e3}
    : teme e1 e2
      -> teme e2 e3
      -> teme e1 e3

with meme : (* md *) term -> (* md *) term -> Prop :=
| meme_refl {M}
    : meme M M

| meme_stat {c c'}
    : meme (md_stat c) (md_stat c')

| meme_dyn {e e'}
    : teme e e'
      -> meme (md_dyn e) (md_dyn e')

| meme_lamgn {T T' M M'}
    : meme M M'
      -> meme (md_lamgn T M) (md_lamgn T' M')

| meme_appgn {M1 M1' M2 M2'}
    : meme M1 M1'
      -> meme M2 M2'
      -> meme (md_appgn M1 M2) (md_appgn M1' M2')

| meme_lamap {T T' M M'}
    : meme M M'
      -> meme (md_lamap T M) (md_lamap T' M')

| meme_appap {M1 M1' M2 M2'}
    : meme M1 M1'
      -> meme M2 M2'
      -> meme (md_appap M1 M2) (md_appap M1' M2')

| meme_pair {M1 M1' M2 M2'}
    : meme M1 M1'
      -> meme M2 M2'
      -> meme (md_pair M1 M2) (md_pair M1' M2')

| meme_pi1 {M M'}
    : meme M M'
      -> meme (md_pi1 M) (md_pi1 M')

| meme_pi2 {M M'}
    : meme M M'
      -> meme (md_pi2 M) (md_pi2 M')

| meme_unpack {e e' M M' T T'}
    : teme e e'
      -> meme M M'
      -> meme (md_unpack e M T) (md_unpack e' M' T')

| meme_lett {e1 e1' M2 M2'}
    : teme e1 e1'
      -> meme M2 M2'
      -> meme (md_lett e1 M2) (md_lett e1' M2')

| meme_letm {M1 M1' M2 M2' T T'}
    : meme M1 M1'
      -> meme M2 M2'
      -> meme (md_letm M1 M2 T) (md_letm M1' M2' T')

| meme_seal {M M' T T'}
    : meme M M'
      -> meme (md_seal M T) (md_seal M' T')

| meme_symm {M M'}
    : meme M M'
      -> meme M' M

| meme_trans {M1 M2 M3}
    : meme M1 M2
      -> meme M2 M3
      -> meme M1 M3
.


Scheme teme_mut_ind := Minimality for teme Sort Prop
with   meme_mut_ind := Minimality for meme Sort Prop.
Combined Scheme eme_ind from teme_mut_ind, meme_mut_ind.


Lemma eme_reduce :
  (forall G e e', teme e e' -> Breduce.treduce G e e' /\ Breduce.treduce G e' e)
  /\
  (forall G M M', meme M M' -> Breduce.mreduce G M M' /\ Breduce.mreduce G M' M).
Proof.
exploit
  (eme_ind
     (fun e e' => forall G, Breduce.treduce G e e' /\ Breduce.treduce G e' e)
     (fun M M' => forall G, Breduce.mreduce G M M' /\ Breduce.mreduce G M' M))
  as Hind;
  try (intros;
       repeat
         (match goal with
          | H : forall _, _ /\ _ |- _ => so (fun G => H G andel); so (fun G => H G ander); clear H
          end);
       split;
       eauto using Breduce.treduce, Breduce.mreduce;
       done);
  [].
destruct Hind; split; intros; eauto; done.
Qed.


Lemma teme_ceqt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> teme e e'
    -> ceqt G e e' t.
Proof.
intros G e e' t Hof Hof' Heme.
so (eme_reduce andel G _ _ Heme) as (Hred & Hred').
split; eapply treduce_capproxt; eauto.
Qed.


Lemma meme_ceqm_impure :
  forall G M M' T,
    ofm G impure M T
    -> ofm G impure M' T
    -> meme M M'
    -> ceqm G impure M M' T.
Proof.
intros G M M' T Hof Hof' Heme.
so (eme_reduce ander G _ _ Heme) as (Hred & Hred').
split.
  eapply mreduce_capproxm_impure; eauto; done.

  eapply mreduce_capproxm_impure; eauto; done.
Qed.


Lemma meme_ceqm_pure :
  forall G M M' T k c c',
    ofm G pure M T
    -> ofm G pure M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> meme M M'
    -> ceqm G pure M M' T.
Proof.
intros G M M' T k c c' Hof Hof' Hfst HfstM HfstM' Heqc Heme.
so (eme_reduce ander G _ _ Heme) as (Hred & Hred').
split.
  eapply mreduce_capproxm_pure; eauto; done.

  eapply mreduce_capproxm_pure; eauto using eqc_symm; done.
Qed.


(* signature with same structure *)
Inductive sss : (* sg *) term -> (* sg *) term -> Prop :=
| sss_one
    : sss sg_one sg_one

| sss_stat {k k'}
    : sss (sg_stat k) (sg_stat k')

| sss_dyn {t t'}
    : sss (sg_dyn t) (sg_dyn t')

| sss_pigen {T1 T1' T2 T2'}
    : sss T1 T1'
      -> sss T2 T2'
      -> sss (sg_pigen T1 T2) (sg_pigen T1' T2')

| sss_piapp {T1 T1' T2 T2'}
    : sss T1 T1'
      -> sss T2 T2'
      -> sss (sg_piapp T1 T2) (sg_piapp T1' T2')

| sss_sigma {T1 T1' T2 T2'}
    : sss T1 T1'
      -> sss T2 T2'
      -> sss (sg_sigma T1 T2) (sg_sigma T1' T2')
.


Lemma subst_sss :
  forall s T,
    hsg T
    -> sss T (subst s T).
Proof.
intros s T H.
revert s.
induct H; eauto using sss.
Qed.


Lemma sss_refl :
  forall T, hsg T -> sss T T.
Proof.
intros T H.
induct H; eauto using sss.
Qed.


Lemma sss_symm : forall T T', sss T T' -> sss T' T.
Proof.
intros T T' H.
induct H; eauto using sss.
Qed.


Lemma sss_trans :
  forall T1 T2 T3,
    sss T1 T2
    -> sss T2 T3
    -> sss T1 T3.
Proof.
intros T1 T2 T3 H12 H23.
revert T3 H23.
induct H12;
  try (intros;
       repeat
         (match goal with
          | H : sss ?T _ |- _ => isapp T; invert H; clear H
          end);
       intros;
       subst;
       eauto using sss;
       done).
Qed.


Lemma eqs_sss :
  forall G T T', eqs G T T' -> sss T T'.
Proof.
intros G T T' H.
induct H; eauto using sss, ofs_hsg, sss_refl, sss_symm, sss_trans.
Qed.


(* snd respects eqs *)

Lemma sss_snd_join :
  (forall T T' M M' e e',
     sss T T'
     -> meme M M'
     -> snd T M e
     -> snd T' M' e'
     -> teme e e')
  /\
  (forall T T' c c' e e' M M',
     sss T T'
     -> teme e e'
     -> join T c e M
     -> join T' c' e' M'
     -> meme M M').
Proof.
cut (forall T T',
       sss T T'
       -> (forall s s' M M' e e',
             meme M M'
             -> snd (subst s T) M e
             -> snd (subst s' T') M' e'
             -> teme e e')
          /\
          (forall s s' c c' e e' M M',
             teme e e'
             -> join (subst s T) c e M
             -> join (subst s' T') c' e' M'
             -> meme M M')).
  intros Hind.
  split.
    intros T T' M M' e e' Hsss Heme Hsnd Hsnd'.
    so (Hind _ _ Hsss andel id id) as H; simp_sub in H.
    eapply H; eauto; done.

    intros T T' c c' e e' M M' Hsss Heme Hjoin Hjoin'.
    so (Hind _ _ Hsss ander id id) as H; simp_sub in H.
    eapply H; eauto; done.
intros T T' Hsss.
#3 induct Hsss;
try (intros;
     split;
     intros;
     repeat
       (match goal with
        | H : snd _ _ _ |- _ => invert H; clear H
        | H : join _ _ _ _ |- _ => invert H; clear H
        end);
     intros; 
     subst;
     eauto 2 using teme, meme;
     done).

(* pigen *)
intros T1 T1' T2 T2' _ IH1 _ IH2.
split.
  intros s s' M M' e e' HemeM Hsnd Hsnd'.
  simp_sub in Hsnd; simp_sub in Hsnd'.
  invert Hsnd; clear Hsnd; [].
  intros k1 t1 k2 t2 N f _ _ Hjoin1 Hsnd2 <-.
  invert Hsnd'; clear Hsnd'; [].
  intros k1' t1' k2' t2' N' f' _ _ Hjoin1' Hsnd2' <-.
  simp_sub in Hjoin1; simp_sub in Hjoin1'; simp_sub in Hsnd2; simp_sub in Hsnd2'.
  so (IH1 ander _#8 teme_refl Hjoin1 Hjoin1').
  so (IH2 andel _#6 meme_refl Hsnd2 Hsnd2').
  apply teme_letm; auto; [].
  apply teme_plam; [].
  eauto using teme, meme; done.

  intros s s' c c' e e' M M' Hemee Hjoin Hjoin'.
  simp_sub in Hjoin; simp_sub in Hjoin'.
  invert Hjoin; clear Hjoin; [].
  intros f N Hsnd1 Hjoin2 <-.
  invert Hjoin'; clear Hjoin'; [].
  intros f' N' Hsnd1' Hjoin2' <-.
  simp_sub in Hsnd1; simp_sub in Hsnd1'; simp_sub in Hjoin2; simp_sub in Hjoin2'.
  so (IH1 andel _#6 meme_refl Hsnd1 Hsnd1').
  so (IH2 ander _#8 teme_refl Hjoin2 Hjoin2').
  eauto using teme, meme; done.

(* piapp *)
intros T1 T1' T2 T2' _ IH1 _ IH2.
split.
  intros s s' M M' e e' HemeM Hsnd Hsnd'.
  simp_sub in Hsnd; simp_sub in Hsnd'.
  invert Hsnd; clear Hsnd; [].
  intros k1 t1 N f _ Hjoin1 Hsnd2 <-.
  invert Hsnd'; clear Hsnd'; [].
  intros k1' t1' N' f' _ Hjoin1' Hsnd2' <-.
  simp_sub in Hjoin1; simp_sub in Hjoin1'; simp_sub in Hsnd2; simp_sub in Hsnd2'.
  so (IH1 ander _#8 teme_refl Hjoin1 Hjoin1').
  #2 sor (refine (IH2 andel _#6 _ Hsnd2 Hsnd2')).
    eauto using meme; done.
  apply teme_letm; auto; [].
  apply teme_plam; [].
  eauto using teme, meme; done.

  intros s s' c c' e e' M M' Hemee Hjoin Hjoin'.
  simp_sub in Hjoin; simp_sub in Hjoin'.
  invert Hjoin; clear Hjoin; [].
  intros f N Hsnd1 Hjoin2 <-.
  invert Hjoin'; clear Hjoin'; [].
  intros f' N' Hsnd1' Hjoin2' <-.
  simp_sub in Hsnd1; simp_sub in Hsnd1'; simp_sub in Hjoin2; simp_sub in Hjoin2'.
  so (IH1 andel _#6 meme_refl Hsnd1 Hsnd1').
  #2 sor (refine (IH2 ander _#8 _ Hjoin2 Hjoin2')).
    eauto using teme; done.
  eauto using teme, meme; done.

(* sigma *)
intros T1 T1' T2 T2' _ IH1 _ IH2.
split.
  intros s s' M M' e e' HemeM Hsnd Hsnd'.
  simp_sub in Hsnd; simp_sub in Hsnd'.
  invert Hsnd; clear Hsnd; [].
  intros e1 e2 Hsnd1 Hsnd2 <-.
  invert Hsnd'; clear Hsnd'; [].
  intros e1' e2' Hsnd1' Hsnd2' <-.
  simp_sub in Hsnd1; simp_sub in Hsnd1'; simp_sub in Hsnd2; simp_sub in Hsnd2'.
  so (IH1 andel _#6 meme_refl Hsnd1 Hsnd1').
  so (IH2 andel _#6 meme_refl Hsnd2 Hsnd2').
  eauto using teme, meme; done.

  intros s s' c c' e e' M M' Hemee Hjoin Hjoin'.
  simp_sub in Hjoin; simp_sub in Hjoin'.
  invert Hjoin; clear Hjoin; [].
  intros N1 N2 Hjoin1 Hjoin2 <-.
  invert Hjoin'; clear Hjoin'; [].
  intros N1' N2' Hjoin1' Hjoin2' <-.
  simp_sub in Hjoin1; simp_sub in Hjoin1'; simp_sub in Hjoin2; simp_sub in Hjoin2'.
  so (IH1 ander _#8 teme_refl Hjoin1 Hjoin1').
  so (IH2 ander _#8 teme_refl Hjoin2 Hjoin2').
  eauto using teme, meme; done.
Qed.


Lemma sss_snd :
  forall T T' M M' e e',
    sss T T'
    -> meme M M'
    -> snd T M e
    -> snd T' M' e'
    -> teme e e'.
Proof.
exact (sss_snd_join andel).
Qed.


Lemma sss_join :
  forall T T' c c' e e' M M',
    sss T T'
    -> teme e e'
    -> join T c e M
    -> join T' c' e' M'
    -> meme M M'.
Proof.
exact (sss_snd_join ander).
Qed.


Lemma snd_eqs :
  forall G T T' k t M c e e',
    eqs G T T'
    -> ofm G pure M T
    -> split T k t
    -> fstm G M c
    -> snd T M e
    -> snd T' M e'
    -> ceqt G e e' (subst1 c t).
Proof.
intros G T T' k t M c e e' Heqs HofM Hsplit HfstM Hsnd Hsnd'.
so (sss_snd _#6 (eqs_sss _#3 Heqs) meme_refl Hsnd Hsnd') as Heme.
apply ceqt_with_wf; intro Hwf; [].
so (snd_formation _#7 Hwf HofM Hsnd Hsplit HfstM) as Hofe.
so (can_split _ (ofs_hsg _ _ (eqs_reg _#3 Hwf Heqs ander))) as (k' & t' & Hsplit').
so (snd_formation _#7 Hwf (ofm_subsume HofM (subs_refl Heqs)) Hsnd' Hsplit' HfstM) as Hofe'.
eapply teme_ceqt; eauto; [].
eapply oft_equiv; eauto; [].
so (split_eqs _#7 Hwf Heqs Hsplit Hsplit') as (Heqk & Heqt).
so (split_fsts _#3 Hsplit) as Hfst.
so (ofm_fst _#6 Hwf HofM HfstM Hfst) as Hofc.
apply eqc_symm; [].
eapply subst1_eqc_type; eauto; done.
Qed.


Lemma join_eqs :
  forall G T T' k t c e M M',
    eqs G T T'
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> join T c e M
    -> join T' c e M'
    -> ceqm G pure M M' T.
Proof.
intros G T T' k t c e M M' Heqs Hsplit Hofc Hofe Hjoin Hjoin'.
so (sss_join _#8 (eqs_sss _#3 Heqs) teme_refl Hjoin Hjoin') as Heme.
apply ceqm_with_wf; intro Hwf; [].
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
so (can_split _ (ofs_hsg _ _ HofT')) as (k' & t' & Hsplit').
so (split_eqs _#7 Hwf Heqs Hsplit Hsplit') as (Heqk & Heqt).
so (join_formation _#7 Hwf HofT Hjoin Hsplit Hofc Hofe) as HofM.
assert (ofc G c k') as Hofc'.
  eapply ofc_subsume; eauto with static; done.
assert (oft G e (subst1 c t')) as Hofe'.
  eapply oft_equiv; eauto; [].
  eapply subst_eqc_type; eauto; prove_subof; auto; done.
so (join_formation _#7 Hwf HofT' Hjoin' Hsplit' Hofc' Hofe') as HofM'.
so (join_fst_o _#7 Hwf HofT Hjoin Hsplit Hofc Hofe) as (b & HfstM & Heqb).
so (join_fst_o _#7 Hwf HofT' Hjoin' Hsplit' Hofc' Hofe') as (b' & HfstM' & Heqb').
#3 eapply meme_ceqm_pure; eauto.
  eapply ofm_subsume; eauto with static; done.

  eapply split_fsts; eauto; done.

  #2 apply (@eqc_trans _ _ c).
    apply eqc_symm; auto; done.
  eapply eqc_subsume; eauto with static; done.
Qed.


(* snd respects selfification *)

Lemma singles_sss :
  forall c T,
    hsg T
    -> sss T (singles c T).
Proof.
intros c T H.
rewrite <- (subst_id T) at 2; [].
generalize id.
revert c.
induct H; try (intros; simpl; simp_sub; eauto using sss, sss_refl, subst_sss; done).
(* sigma *)
intros T1 T2 _ IH1 H2 IH2 c s.
simp_sub.
simpl.
rewrite -> subst_singles; auto using subst_hsg; [].
simp_sub.
apply sss_sigma; auto; done.
Qed.  


Lemma split_singles' :
  forall T c k k' t t',
    split T k t
    -> split (singles' c T) k' t'
    -> t = t'.
Proof.
intros T c k k' t t' Hsplit Hsplit'.
revert c k' t' Hsplit'.
induct Hsplit.

(* one *)
intros c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; [].
intros <- <-.
reflexivity.

(* stat *)
intros k c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; [].
intros <- <-.
reflexivity.

(* dyn *)
intros t c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; [].
intros <- <-.
reflexivity.

(* pigen *)
intros T1 T2 k1 k2 t1 t2 Hsplit1 _ Hsplit2 _ c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fun _#5 Hsplit1 Hsplit1') as (<- & <-).
so (split_fun _#5 Hsplit2 Hsplit2') as (<- & <-).
reflexivity.

(* piapp *)
intros T1 T2 k1 k2 t1 t2 Hsplit1 _ _ IH2 c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fun _#5 Hsplit1 Hsplit1') as (<- & <-).
clear Hsplit1'.
f_equal; [].
f_equal; [].
f_equal; eauto; done.

(* sigma *)
intros T1 T2 k1 k2 t1 t2 _ IH1 _ IH2 c k' t' Hsplit'.
simpl in Hsplit'.
invert Hsplit'; clear Hsplit'; [].
intros k1' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
#2 f_equal.
  f_equal; eauto; done.

  f_equal; eauto; done.
Qed.


Lemma split_singles :
  forall G T c k k' t t',
    wf G
    -> ofs G T
    -> ofc G c k
    -> split T k t
    -> split (singles c T) k' t'
    -> eqc (G; cl_cn k') t t' kd_type.
Proof.
intros G T c k k' t t' Hwf HofT Hofc Hsplit Hsplit'.
so (singles_singles' _#4 Hwf HofT (split_fsts _ _ _ Hsplit) Hofc) as Heqs.
so (can_split _ (eqs_hsg _#3 Heqs ander)) as (k'' & t'' & Hsplit'').
so (split_singles' _#6 Hsplit Hsplit''); subst t''.
so (split_eqs _#7 Hwf Heqs Hsplit' Hsplit'') as (_ & H).
exact (eqc_symm H).
Qed.


Lemma snd_self :
  forall G T k t M c e e',
    ofs G T
    -> split T k t
    -> ofm G pure M T
    -> fstm G M c
    -> snd T M e
    -> snd (singles c T) M e'
    -> ceqt G e e' (subst1 c t).
Proof.
intros G T k t M c e e' HofT Hsplit HofM HfstM Hsnd Hsnd'.
apply ceqt_with_wf; intro Hwf; [].
#3 apply teme_ceqt.
  eapply snd_formation; eauto; done.

  so (can_split _ (singles_hsg c _ (ofs_hsg _ _ HofT))) as (k' & t' & Hsplit').
  #2 eapply oft_equiv.
    eapply (snd_formation _ (singles c T)); eauto; [].
    eapply singles_ofm; eauto; done.

    so (split_fsts _#3 Hsplit) as Hfst.
    so (ofm_fst _#6 Hwf HofM HfstM Hfst) as Hofc.
    so (split_singles _#7 Hwf HofT Hofc Hsplit Hsplit') as Heqc.
    apply eqc_symm; [].
    refine (subst1_eqc_type _#5 _ Heqc); [].
    so (singles_fst c _ _ Hfst) as Hfst'.
    so (split_fsts_eq _#4 Hsplit' Hfst') as Heq.
    subst k'.
    eapply single_ofc; eauto; done.

  exact (sss_snd _#6 (singles_sss _ _ (ofs_hsg _ _ HofT)) meme_refl Hsnd Hsnd').
Qed.


Lemma join_self :
  forall G T k t c e M M',
    ofs G T
    -> split T k t
    -> ofc G c k
    -> oft G e (subst1 c t)
    -> join T c e M
    -> join (singles c T) c e M'
    -> ceqm G pure M M' T.
Proof.
intros G T k t c e M Ms HofT Hsplit Hofc Hofe Hjoin Hjoins.
apply ceqm_with_wf; intro Hwf; [].
so (split_fsts _#3 Hsplit) as Hfst.
so (join_formation _#7 Hwf HofT Hjoin Hsplit Hofc Hofe) as HofM.
so (singles_fst c _ _ Hfst) as Hfsts.
so (singles_ofs _#4 Hwf HofT Hfst Hofc) as HofTs.
so (can_split _ (ofs_hsg _ _ HofTs)) as (ks & ts & Hsplits).
so (split_fsts_eq _#4 Hsplits Hfsts); subst ks.
so (split_singles _#7 Hwf HofT Hofc Hsplit Hsplits) as Heqt.
so (single_ofc _#3 Hwf Hofc) as Hofcs.
assert (oft G e (subst1 c ts)) as Hofes.
  eapply oft_equiv; eauto; [].
  eapply subst1_eqc_type; eauto; done.
so (join_formation _#7 Hwf HofTs Hjoins Hsplits Hofcs Hofes) as HofMs.
so (singles_subs _#4 Hwf HofT Hfst Hofc) as Hsubs.
so (join_fst_o _#7 Hwf HofT Hjoin Hsplit Hofc Hofe) as (c' & HfstM & Heqc).
so (join_fst_o _#7 Hwf HofTs Hjoins Hsplits Hofcs Hofes) as (cs & HfstMs & Heqcs).
#3 eapply meme_ceqm_pure; eauto.
  eapply ofm_subsume; eauto; done.

  #2 eapply eqc_trans.
    eapply eqc_symm; eauto; done.

    eapply eqc_subsume; eauto; [].
    eapply single_subk; eauto; done.

  exact (sss_join _#8 (singles_sss _ _ (ofs_hsg _ _ HofT)) teme_refl Hjoin Hjoins).
Qed.


(* Correctness of trsubs *)

Lemma trsubs_subs :
  forall G T T' f,
    trsubs G T T' f
    -> subs G T T'.
Proof.
intros G T T' f H.
induct H; eauto using subs, split_fsts.
Qed.


Lemma subs_trsubs :
  forall G T T',
    subs G T T'
    -> exists f, trsubs G T T' f.
Proof.
intros G T T' H.
induct H.

(* refl *)
intros G T T' Heqs.
so (can_split _ (eqs_hsg _#3 Heqs andel)) as (k & t & Hsplit).
eexists; eauto using trsubs; done.

(* trans *)
intros G T1 T2 T3 Hsubs (f1 & IH1) _ (f2 & IH2).
so (can_split _ (subs_hsg _#3 Hsubs andel)) as (k & t & Hsplit).
eexists; eauto using trsubs; done.

(* stat *)
intros G k k' H.
eexists; eauto using trsubs; done.

(* pigen *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 (f1 & Htrsub1) Hfst1' Hsubs2 (f2 & Htrsub2) Hfst1 HofT2.
so (subs_hsg _#3 Hsubs1) as (Hhsg1' & Hhsg1).
so (subs_hsg _#3 Hsubs2) as (Hhsg2 & Hhsg2').
so (can_split _ Hhsg1) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ Hhsg1') as (k1a' & t1' & Hsplit1').
so (split_fsts_eq _#4 Hsplit1' Hfst1'); subst k1a'.
so (can_split _ Hhsg2) as (k2 & t2 & Hsplit2).
so (can_split _ Hhsg2') as (k2' & t2' & Hsplit2').
eexists; eauto using trsubs; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 (f1 & Htrsub1) Hfst1' Hsubs2 (f2 & Htrsub2) Hfst1 HofT2.
so (subs_hsg _#3 Hsubs1) as (Hhsg1' & Hhsg1).
so (subs_hsg _#3 Hsubs2) as (Hhsg2 & Hhsg2').
so (can_split _ Hhsg1) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ Hhsg1') as (k1a' & t1' & Hsplit1').
so (split_fsts_eq _#4 Hsplit1' Hfst1'); subst k1a'.
so (can_split _ Hhsg2) as (k2 & t2 & Hsplit2).
eexists; eauto using trsubs; done.

(* sigma *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 (f1 & Htrsub1) Hfst1 Hsubs2 (f2 & Htrsub2) Hfst1' HofT2'.
so (subs_hsg _#3 Hsubs1) as (Hhsg1 & Hhsg1').
so (subs_hsg _#3 Hsubs2) as (Hhsg2 & Hhsg2').
so (can_split _ Hhsg1) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ Hhsg1') as (k1a' & t1' & Hsplit1').
so (split_fsts_eq _#4 Hsplit1' Hfst1'); subst k1a'.
so (can_split _ Hhsg2) as (k2 & t2 & Hsplit2).
eexists; eauto using trsubs; done.
Qed.


Lemma trsubs_formation :
  forall G T T' f k k' t t',
    wf G
    -> trsubs G T T' f
    -> split T k t
    -> split T' k' t'
    -> oft G f (cn_all k (cn_arrow t t')).
Proof.
intros G T T' f k k' t t' Hwf Htr Hsplit Hsplit'.
revert k k' t t' Hwf Hsplit Hsplit'.
induct Htr.

(* refl *)
intros G T T' ka ta Hsplita Heqs k k' t t' Hwf Hsplit Hsplit'.
so (split_fun _#5 Hsplita Hsplit) as (-> & ->); clear Hsplita.
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (split_eqs _#7 Hwf Heqs Hsplit Hsplit') as (_ & Heqt).
apply oft_plam; auto with static; [].
apply oft_lam; auto; [].
#2 eapply oft_equiv.
  ofvar; done.

  eapply subst_eqc_type; eauto using weaken1; done.

(* trans *)
intros G T1 T2 T3 k1a t1a f1 f2 Hsplit1a Htr1 IH1 Htr2 IH2 k1 k3 t1 t3 Hwf Hsplit1 Hsplit3.
so (split_fun _#5 Hsplit1a Hsplit1) as (-> & ->); clear Hsplit1a.
so (can_split _ (subs_hsg _#3 (trsubs_subs _#4 Htr1) ander)) as (k2 & t2 & Hsplit2).
so (IH1 _#4 Hwf Hsplit1 Hsplit2) as Hoff1.
so (IH2 _#4 Hwf Hsplit2 Hsplit3) as Hoff2.
so (split_formation _#4 (subs_reg _#3 Hwf (trsubs_subs _#4 Htr1) andel) Hsplit1) as (Hofk1 & Hoft1).
apply oft_plam; auto with static; [].
apply oft_lam; auto; [].
#2 eapply oft_app.
  relquest.
    #2 eapply oft_papp.
      relquest.
        eapply subst_oft; eauto using weaken2; done.

        simp_sub; reflexivity.

      #2 eapply ofc_subsume.
        ofvar; done.

        eapply subst_subk; eauto using weaken2; [].
        clear Htr2.
        eapply subs_fst; eauto using trsubs_subs, split_fsts; done.

    simp_sub.
    rewrite -> !subst_dot1_sh; [].
    reflexivity.

  #2 eapply oft_app.
    relquest.
      #2 eapply oft_papp.
        relquest.
          eapply subst_oft; eauto using weaken2; done.

          simp_sub; reflexivity.

        ofvar; done.

      simp_sub.
      rewrite -> !subst_dot1_sh; [].
      reflexivity.
      
    ofvar; done.

(* stat *)
intros G k k' Hsubk ka ka' t t' Hwf Hsplit Hsplit'.
invert Hsplit; [].
intros <- <-.
invert Hsplit'; [].
intros <- <-.
so (subk_reg _#3 Hwf Hsubk) as (Hofk & _).
apply oft_plam; auto using svalue; [].
apply oft_lam; auto with static; [].
ofvar; done.

(* pigen *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 k2' t2 t2' f1 f2 Hsplit1 Hsplit1' Hsplit2 Hsplit2' Htr1 IH1 Htr2 IH2 HofT2 k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-); clear Hsplit1a.
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-); clear Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2a' t1a' t2a' Hsplit1a' Hsplit2a' <- <-.
so (split_fun _#5 Hsplit1' Hsplit1a') as (<- & <-); clear Hsplit1a'.
so (split_fun _#5 Hsplit2' Hsplit2a') as (<- & <-); clear Hsplit2a'.
so (trsubs_subs _#4 Htr1) as Hsubs1; clear Htr1.
so (trsubs_subs _#4 Htr2) as Hsubs2; clear Htr2.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_formation _#4 (subs_reg _#3 (wf_cn Hwf Hofk1') Hsubs2 ander) Hsplit2') as (Hofk2' & Hoft2').
so (subs_fst _#5 Hwf Hsubs1 (split_fsts _#3 Hsplit1') (split_fsts _#3 Hsplit1)) as Hsubk1.
so (subs_fst _#5 (wf_cn Hwf Hofk1') Hsubs2 (split_fsts _#3 Hsplit2) (split_fsts _#3 Hsplit2')) as Hsubk2.
apply oft_plam; auto with static; [].
simp_sub.
#2 apply oft_lam.
  #2 apply ofc_all.
    eapply subst_ofk; eauto using weaken1; done.

    #2 apply ofc_arrow.
      eapply subst_ofc_type; eauto; prove_subof; done.

      #2 apply ofc_exists.
        eapply subst_ofk; eauto; prove_subof; done.

        eapply subst_ofc_type; eauto; prove_subof; done.
simp_sub.
#2 apply oft_plam; auto using svalue.
  eapply subst_ofk; eauto using weaken2; done.
#2 apply oft_lam.
  eapply subst_ofc_type; eauto; prove_subof; done.
#3 eapply oft_unpack.
  #2 eapply oft_app.
    relquest.
      #2 eapply oft_papp.
        ofvar; done.

        #2 eapply ofc_subsume.
          ofvar; done.

          eapply subst_subk; eauto; [].
          prove_subof; done.

      simp_sub; reflexivity.

    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          relquest.
            eapply subst_oft; eauto; prove_subof; done.

            simp_sub; reflexivity.

          ofvar; done.

        simp_sub; reflexivity.

      ofvar; done.

  simp_sub.
  #3 apply oft_pack.
    #2 eapply ofc_subsume.
      ofvar; done.

      eapply subst_subk; eauto; prove_subof; done.

    simp_sub.
    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          relquest.
            eapply subst_oft; eauto using wf; prove_subof; done.

            simp_sub; reflexivity.

          ofvar; done.

        simp_sub; reflexivity.

      ofvar; done.

    eapply subst_ofc_type; eauto; prove_subof; done.

  simp_sub.
  #2 apply ofc_exists.
    eapply subst_ofk; eauto; prove_subof; done.

    eapply subst_ofc_type; eauto; prove_subof; done.

(* piapp *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 t2 f1 f2 Hsplit1 Hsplit1' Hsplit2 Htrsub1 IH1 Htrsub2 IH2 HofT2 k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-); clear Hsplit1a.
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-); clear Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2' t1a' t2' Hsplit1a' Hsplit2' <- <-.
so (split_fun _#5 Hsplit1' Hsplit1a') as (<- & <-); clear Hsplit1a'.
so (trsubs_subs _#4 Htrsub1) as Hsubs1; clear Htrsub1.
so (trsubs_subs _#4 Htrsub2) as Hsubs2; clear Htrsub2.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_formation _#4 (subs_reg _#3 (wf_cn Hwf Hofk1') Hsubs2 ander) Hsplit2') as (Hofk2' & Hoft2').
so (subs_fst _#5 Hwf Hsubs1 (split_fsts _#3 Hsplit1') (split_fsts _#3 Hsplit1)) as Hsubk1.
so (subs_fst _#5 (wf_cn Hwf Hofk1') Hsubs2 (split_fsts _#3 Hsplit2) (split_fsts _#3 Hsplit2')) as Hsubk2.
so (IH1 _#4 Hwf Hsplit1' Hsplit1) as Hoff1.
so (IH2 _#4 (wf_cn Hwf Hofk1') Hsplit2 Hsplit2') as Hoff2.
clear IH1 IH2.
apply oft_plam; auto using svalue with static; [].
#2 apply oft_lam.
  #2 apply ofc_all.
    apply weaken1_ofk; auto; done.

    #2 apply ofc_arrow.
      eapply subst_ofc_type; eauto; [].
      rewrite -> !under_succ; simp_sub; [].
      prove_subof; done.

      eapply subst_ofc_type; eauto; [].
      prove_subof; [].
      relquest.
        #2 eapply ofc_app.
          relquest.
            ofvar; done.

            simp_sub; reflexivity.

          ofvar; done.

        simp_sub; reflexivity.

  simp_sub.
  #2 apply oft_plam; auto using svalue.
    apply weaken2_ofk; eauto; done.

    #2 apply oft_lam.
      eapply subst_ofc_type; eauto; prove_subof; done.

      simp_sub.
      #2 eapply oft_app.
        relquest.
          #2 eapply oft_papp.
            relquest.
              eapply subst_oft; eauto; prove_subof; done.

              simp_sub; reflexivity.

            relquest.
              #2 eapply ofc_app.
                ofvar; done.

                #2 eapply ofc_subsume.
                  ofvar; done.

                  eapply subst_subk; eauto; prove_subof; done.

              simp_sub; reflexivity.

          simp_sub; reflexivity.

        #2 eapply oft_app.
          relquest.
            #2 eapply oft_papp.
              ofvar; done.

              #2 eapply ofc_subsume.
                ofvar; done.

                eapply subst_subk; eauto; prove_subof; done.

            simp_sub; reflexivity.

          #2 eapply oft_app.
            relquest.
              #2 eapply oft_papp.
                relquest.
                  eapply subst_oft; eauto; prove_subof; done.

                  simp_sub; reflexivity.

                ofvar; done.

              simp_sub; reflexivity.

            ofvar; done.

(* sigma *)
intros G T1 T1' T2 T2' k1 k1' t1 k2 t2 f1 f2 Hsplit1 Hfst1' Hsplit2 Htrsub1 IH1 Htrsub2 IH2 HofT2' k k' t t' Hwf Hsplit Hsplit'.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-); clear Hsplit1a.
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-); clear Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fsts_eq _#4 Hsplit1' Hfst1'); subst k1a'.
so (trsubs_subs _#4 Htrsub1) as Hsubs1; clear Htrsub1.
so (trsubs_subs _#4 Htrsub2) as Hsubs2; clear Htrsub2.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1 & HofT1').
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (split_formation _#4 HofT2' Hsplit2') as (Hofk2' & Hoft2').
so (split_formation _#4 (subs_reg _#3 (wf_cn Hwf Hofk1) Hsubs2 andel) Hsplit2) as (Hofk2 & Hoft2).
so (subs_fst _#5 Hwf Hsubs1 (split_fsts _#3 Hsplit1) (split_fsts _#3 Hsplit1')) as Hsubk1.
so (subs_fst _#5 (wf_cn Hwf Hofk1) Hsubs2 (split_fsts _#3 Hsplit2) (split_fsts _#3 Hsplit2')) as Hsubk2.
so (IH1 _#4 Hwf Hsplit1 Hsplit1') as Hoff1.
so (IH2 _#4 (wf_cn Hwf Hofk1) Hsplit2 Hsplit2') as Hoff2.
clear IH1 IH2.
apply oft_plam; auto with static; [].
simp_sub.
#2 apply oft_lam.
  #2 apply ofc_prod.
    eapply subst_ofc_type; eauto; prove_subof; [].
    eapply ofc_pi1; ofvar; done.

    #2 eapply subst_ofc_type; eauto; prove_subof.
      eapply ofc_pi1; ofvar; done.

      relquest.
        eapply ofc_pi2; ofvar; done.

        simp_sub; reflexivity.

  simp_sub.
  #2 apply oft_pair.
    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          relquest.
            eapply subst_oft; eauto; prove_subof; done.

            simp_sub; reflexivity.

          eapply ofc_pi1; ofvar; done.

        simp_sub; reflexivity.

      eapply oft_pi1; ofvar; done.

    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          relquest.
            eapply subst_oft; eauto; prove_subof; [].
            eapply ofc_pi1; ofvar; done.

            simp_sub; reflexivity.

          relquest.
            eapply ofc_pi2; ofvar; done.

            simp_sub; reflexivity.

        simp_sub; reflexivity.

      eapply oft_pi2; ofvar; done.
Qed.


(* With more work, we could get by with fcontext G. *)
Lemma trsubs_statically_closed :
  forall G T1 T2 f G' s s',
    wf G
    -> trsubs G T1 T2 f
    -> subrel sid G' s s' G
    -> subst s f = subst s' f.
Proof.
intros G T1 T2 f G' s s' Hwf Htrsubs Hss.
revert G' s s' Hwf Hss.
induct Htrsubs.

(* refl *)
intros G T T' k t Hsplit Heqs G' s s' Hwf Hss.
simp_sub.
so (split_formation _#4 (eqs_reg _#3 Hwf Heqs andel) Hsplit) as (Hofk & Hoft).
#2 f_equal.
  eapply sid_ofk; eauto; done.

  f_equal; [].
  eapply sid_ofc; eauto; [].
  apply subrel_sid_bind_cn; eauto; done.

(* trans *)
intros G T1 T2 T3 k1 t1 f1 f2 Hsplit1 Htrsub1 IH1 _ IH2 G' s s' Hwf Hss.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsub1)) as (HofT1 & _).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
simp_sub.
#2 f_equal.
  eapply sid_ofk; eauto; done.

  #2 f_equal.
    eapply sid_ofc; eauto; [].
    apply subrel_sid_bind_cn; eauto; done.

    #2 f_equal.
      f_equal; [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply IH2; eauto; done.

      f_equal; [].
      f_equal; [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply IH1; eauto; done.

(* stat *)
intros G k k' Hsubk G' s s' Hwf Hss.
simp_sub.
f_equal; [].
so (subk_reg _#3 Hwf Hsubk andel) as Hofk.
eapply sid_ofk; eauto; done.

(* pigen *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 k2' t2 t2' f1 f2 Hsplit1 Hsplit1' Hsplit2 Hsplit2' Htrsubs1 IH1 Htrsubs2 IH2 HofT2 G' s s' Hwf Hss.
reduce_subst; simp_sub.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsubs1)) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (subs_reg _#3 (wf_cn Hwf Hofk1') (trsubs_subs _#4 Htrsubs2)) as (_ & HofT2').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_formation _#4 HofT2' Hsplit2') as (Hofk2' & Hoft2').
f_equal; [].
#2 f_equal.
  #2 f_equal.
    rewrite -> !subst_compose; [].
    f_equal; [].
    eapply sid_ofk; eauto; done.

    #2 f_equal.
      rewrite -> (dot0_compose_split s (sh 2)); [].
      rewrite -> (dot0_compose_split s' (sh 2)); [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofc; eauto; [].
      eapply subrel_sid_bind_cn; eauto; done.

      #2 f_equal.
        rewrite -> (dot0_compose_split s (sh 2)); [].
        rewrite -> (dot0_compose_split s' (sh 2)); [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply sid_ofk; eauto; [].
        eapply subrel_sid_bind_cn; eauto; done.

        rewrite -> (dot01_compose_split s (sh 3)); [].
        rewrite -> (dot01_compose_split s' (sh 3)); [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply sid_ofc; eauto; [].
        eapply subrel_sid_bind_cn_cn; eauto; done.

  #2 f_equal.
    rewrite -> !subst_compose; [].
    f_equal; [].
    eapply sid_ofk; eauto; done.

    #2 f_equal.
      simp_sub.
      rewrite -> (dot0_compose_split s (sh 3)); [].
      rewrite -> (dot0_compose_split s' (sh 3)); [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofc; eauto; [].
      eapply subrel_sid_bind_cn; eauto; done.

      #2 f_equal.
        f_equal; [].
        f_equal; [].
        f_equal; [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply IH1; eauto; done.

        #3 f_equal.
          f_equal; [].
          f_equal; [].
          replsub (subst (dot (var 3) (compose s (sh 6))) f2) with (subst (sh 3) (subst (dot (var 0) (compose s (sh 3))) f2)).
          replsub (subst (dot (var 3) (compose s' (sh 6))) f2) with (subst (sh 3) (subst (dot (var 0) (compose s' (sh 3))) f2)).
          f_equal; [].
          rewrite -> (dot0_compose_split s (sh 3)); [].
          rewrite -> (dot0_compose_split s' (sh 3)); [].
          rewrite -> !subst_compose; [].
          f_equal; [].
          eapply IH2; eauto using wf; [].
          apply subrel_sid_bind_cn; eauto; done.
          
          replsub (subst (dot (var 3) (compose s (sh 6))) k2') with (subst (sh 3) (subst (dot (var 0) (compose s (sh 3))) k2')).
          replsub (subst (dot (var 3) (compose s' (sh 6))) k2') with (subst (sh 3) (subst (dot (var 0) (compose s' (sh 3))) k2')).
          f_equal; [].
          rewrite -> (dot0_compose_split s (sh 3)); [].
          rewrite -> (dot0_compose_split s' (sh 3)); [].
          rewrite -> !subst_compose; [].
          f_equal; [].
          eapply sid_ofk; eauto; [].
          apply subrel_sid_bind_cn; eauto; done.
          
          replsub (subst (dot (var 0) (dot (var 4) (compose s (sh 7)))) t2') with (subst (dot (var 0) (sh 4)) (subst (dot (var 0) (dot (var 1) (compose s (sh 4)))) t2')).
          replsub (subst (dot (var 0) (dot (var 4) (compose s' (sh 7)))) t2') with (subst (dot (var 0) (sh 4)) (subst (dot (var 0) (dot (var 1) (compose s' (sh 4)))) t2')).
          f_equal; [].
          rewrite -> (dot01_compose_split s (sh 4)); [].
          rewrite -> (dot01_compose_split s' (sh 4)); [].
          rewrite -> !subst_compose; [].
          f_equal; [].
          eapply sid_ofc; eauto; [].
          apply subrel_sid_bind_cn_cn; eauto; done.

(* piapp *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 t2 f1 f2 Hsplit1 Hsplit1' Hsplit2 Htrsubs1 IH1 Htrsubs2 IH2 HofT2 G' s s' Hwf Hss.
simp_sub.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsubs1)) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
#2 f_equal.
  #2 f_equal.
    eapply sid_ofk; eauto; done.

    eapply sid_ofk; eauto using subrel_sid_bind_cn; done.

  #2 f_equal.
    #2 f_equal.
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofk; eauto; done.

      #2 f_equal.
        rewrite -> (dot0_compose_split s (sh 2)); [].
        rewrite -> (dot0_compose_split s' (sh 2)); [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply sid_ofc; eauto using subrel_sid_bind_cn; done.

        rewrite -> split_dot_compose; [].
        symmetry.
        rewrite -> split_dot_compose; [].
        symmetry.
        f_equal; [].
        simp_sub.
        rewrite -> (dot01_compose_split s (sh 3)); [].
        rewrite -> (dot01_compose_split s' (sh 3)); [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply sid_ofc; eauto using subrel_sid_bind_cn_cn; done.

    #2 f_equal.
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofk; eauto; done.

      #2 f_equal.
        simp_sub.
        rewrite -> (dot0_compose_split s (sh 3)); [].
        rewrite -> (dot0_compose_split s' (sh 3)); [].
        rewrite -> !subst_compose; [].
        f_equal; [].
        eapply sid_ofc; eauto using subrel_sid_bind_cn; done.

        #2 f_equal.
          f_equal; [].
          replsub (subst (dot (var 1) (compose s (sh 4))) f2) with (subst sh1 (subst (dot (var 0) (compose s (sh 3))) f2)).
          replsub (subst (dot (var 1) (compose s' (sh 4))) f2) with (subst sh1 (subst (dot (var 0) (compose s' (sh 3))) f2)).
          f_equal; [].
          rewrite -> (dot0_compose_split s (sh 3)); [].
          rewrite -> (dot0_compose_split s' (sh 3)); [].
          rewrite -> !subst_compose; [].
          f_equal; [].
          eapply IH2; eauto using subrel_sid_bind_cn, wf; done.

          f_equal; [].
          f_equal; [].
          f_equal; [].
          rewrite -> !subst_compose; [].
          f_equal; [].
          eapply IH1; eauto; done.

(* sigma *)
intros G T1 T1' T2 T2' k1 k1' t1 k2 t2 f1 f2 Hsplit1 Hfst1' Hsplit2 Htrsubs1 IH1 Htrsubs2 IH2 _ G' s s' Hwf Hss.
simp_sub.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsubs1)) as (HofT1 & HofT1').
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (ofs_fst _#3 HofT1' Hfst1') as Hofk1'.
so (subs_reg _#3 (wf_cn Hwf Hofk1) (trsubs_subs _#4 Htrsubs2)) as (HofT2 & _).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
#2 f_equal.
  #2 f_equal.
    eapply sid_ofk; eauto; done.

    eapply sid_ofk; eauto using subrel_sid_bind_cn; done.

  #2 f_equal.
    #2 f_equal.
      rewrite -> split_dot_compose; [].
      symmetry.
      rewrite -> split_dot_compose; [].
      symmetry.
      f_equal; [].
      simp_sub.
      rewrite -> (dot0_compose_split s (sh 2)); [].
      rewrite -> (dot0_compose_split s' (sh 2)); [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofc; eauto using subrel_sid_bind_cn; done.

      rewrite -> split_dot_dot_compose; [].
      symmetry.
      rewrite -> split_dot_dot_compose; [].
      symmetry.
      f_equal; [].
      simp_sub.
      rewrite -> (dot01_compose_split s (sh 3)); [].
      rewrite -> (dot01_compose_split s' (sh 3)); [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply sid_ofc; eauto using subrel_sid_bind_cn_cn; done.

    #2 f_equal.
      f_equal; [].
      f_equal; [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply IH1; eauto; done.

      f_equal; [].
      f_equal; [].
      rewrite -> split_dot_compose; [].
      symmetry.
      rewrite -> split_dot_compose; [].
      symmetry.
      f_equal; [].
      simp_sub.
      rewrite -> (dot0_compose_split s (sh 3)); [].
      rewrite -> (dot0_compose_split s' (sh 3)); [].
      rewrite -> !subst_compose; [].
      f_equal; [].
      eapply IH2; eauto using wf, subrel_sid_bind_cn; done.
Qed.


Lemma snd_join_coercion_formation :
  forall G T T' k k' t t' M e,
    wf G
    -> subs G T T'
    -> split T k t
    -> split T' k' t'
    -> join (subst (sh 2) T) (var 1) (var 0) M
    -> snd (subst (sh 2) T') M e
    -> oft G (tm_plam k (tm_lam t e)) (cn_all k (cn_arrow t t')).
Proof.
intros G T T' k k' t t' M e Hwf Hsubs Hsplit Hsplit' Hjoin Hsnd.
so (subs_reg _#3 Hwf Hsubs) as (Hofs & Hofs').
so (split_formation _#4 Hofs Hsplit) as (Hofk & Hoft).
apply oft_plam; auto using svalue; [].
apply oft_lam; auto; [].
#4 sor (refine (join_formation_and_fst_o (G; cl_cn k; cl_tm t) _#6 _ (weaken2_ofs _#4 Hofs) Hjoin (subst_split _#4 Hsplit) _ _)) as (HofM & (c & HfstM & Heqc)).
  auto using wf; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
#2 eapply oft_equiv.
  #2 eapply (snd_formation _ (subst (sh 2) T')); eauto using wf.
    eapply ofm_subsume; eauto; [].
    apply weaken2_subs; auto; done.

    apply subst_split; exact Hsplit'.

  simp_sub.
  simp_sub.
  rewrite <- subst_dot1_sh; [].
  so (split_formation _#4 Hofs' Hsplit') as (_ & Hoft').
  eapply seq_ofc_type; eauto using wf; [].
  prove_subrel; [].
  apply eqc_symm; [].
  eapply eqc_subsume; eauto; [].
  apply weaken2_subk; [].
  eapply subs_fst; eauto using split_fsts; done.
Qed.


Local Ltac prove_wf :=
  (match goal with
   | |- wf (_; cl_cn _) => apply wf_cn; [prove_wf |]
   | |- wf (_; cl_tm _) => apply wf_tm; [prove_wf |]
   | |- wf (_; cl_md _ _) => try (apply wf_md; [prove_wf | |])
   | |- wf _ => idtac
   end).


Lemma trsubs_correct :
  forall G T T' k k' t t' f M e,
    trsubs G T T' f
    -> split T k t
    -> split T' k' t'
    -> join (subst (sh 2) T) (var 1) (var 0) M
    -> snd (subst (sh 2) T') M e
    -> ceqt G f (tm_plam k (tm_lam t e)) (cn_all k (cn_arrow t t')).
Proof.
intros G T T' k k' t t' f M e Htr Hsplit Hsplit' Hjoin Hsnd.
revert k k' t t' M e Hsplit Hsplit' Hjoin Hsnd.
induct Htr.

(* refl *)
intros G T T' ka ta Hsplita Heqs k k' t t' M e Hsplit Hsplit' Hjoin Hsnd.
so (split_fun _#5 Hsplit Hsplita) as (<- & <-); clear Hsplita.
apply ceqt_with_wf; [].
intro Hwf.
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (can_snd _ M (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (e' & Hsnd').
assert (ceqt (G; cl_cn k; cl_tm t) e' e (shift1 t)) as Hceqe.
  #3 sor (refine (join_formation _#7 Hwf' (weaken2_ofs _#4 HofT) Hjoin (subst_split _#4  Hsplit) _ _)) as HofM.
    ofvar; done.
  
    simp_sub.
    simp_sub.
    rewrite -> subst_dot1_sh; [].
    ofvar; done.
  #3 sor (refine (join_fst_o _#7 Hwf' (weaken2_ofs _#4 HofT) Hjoin (subst_split _#4 Hsplit) _ _)) as (c & HfstM & Heqc).
    ofvar; done.
  
    simp_sub.
    simp_sub.
    rewrite -> subst_dot1_sh; [].
    ofvar; done.
  so (snd_eqs _#9 (subst_eqs _#5 (weaken2 _#4) Heqs) HofM (subst_split _#4 Hsplit) HfstM Hsnd' Hsnd) as Hceqe.
  simp_sub in Hceqe.
  simp_sub in Hceqe.
  apply ceqt_comp; eapply tcc_equiv; eauto; [].
  rewrite <- subst_dot1_sh; [].
  eapply seq_ofc_type; eauto; [].
  prove_subrel; [].
  apply eqc_symm; auto; done.
apply ceqt_plam; auto with static; [].
apply ceqt_comp; apply tcc_lam; auto with static; [].
#2 apply ceqt_comp; apply (@tcc_equiv _#7 (shift1 t)).
  eapply ceqt_trans; eauto; [].
  repl (shift1 t) with (subst1 (var 1) (subst (under 1 (sh 2)) t)).
    simp_sub.
    simp_sub.
    rewrite -> subst_dot1_sh; auto; done.
  #6 eapply snd_join_inverse.
    apply weaken2_ofs; [].
    exact HofT.
  
    apply subst_split; eauto; done.

    ofvar; done.

    simp_sub.
    simp_sub.
    rewrite -> subst_dot1_sh; [].
    ofvar; done.

    exact Hjoin.

    exact Hsnd'.

  apply weaken1_eqc_type; [].
  so (split_eqs _#7 Hwf Heqs Hsplit Hsplit') as (_ & Heqt).
  exact Heqt.

(* trans *)
intros G T1 T2 T3 k1a t1a f1 f2 Hsplit1a Htrsubs1 IH1 Htrsubs2 IH2 k1 k3 t1 t3 M1 e3 Hsplit1 Hsplit3 Hjoin1 Hsnd3.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-); clear Hsplit1a.
apply ceqt_with_wf; intro Hwf; [].
so (trsubs_subs _#4 Htrsubs1) as Hsubs1.
so (trsubs_subs _#4 Htrsubs2) as Hsubs2.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1 & HofT2).
so (subs_hsg _#3 Hsubs2) as (Hhsg2 & Hhsg3).
so (can_split _ Hhsg2) as (k2 & t2 & Hsplit2).
so (can_snd _ M1 (subst_hsg (sh 2) _ Hhsg2)) as (e2 & Hsnd2).
so (IH1 _#6 Hsplit1 Hsplit2 Hjoin1 Hsnd2) as Hceq12.
clear IH1.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg2)) as (M2 & Hjoin2).
so (can_snd _ M2 (subst_hsg (sh 2) _ Hhsg3)) as (e3' & Hsnd3').
so (IH2 _#6 Hsplit2 Hsplit3 Hjoin2 Hsnd3') as Hceq23.
clear IH2.
clear Hhsg2 Hhsg3.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (wf_tm (wf_cn Hwf Hofk1) Hoft1) as Hwf1.
so (wf_tm (wf_cn Hwf Hofk2) Hoft2) as Hwf2.
so (subs_fst _#5 Hwf Hsubs1 (split_fsts _#3 Hsplit1) (split_fsts _#3 Hsplit2)) as Hsubk1.
sor (refine (join_formation _#7 Hwf1 (weaken2_ofs _#4 HofT1) Hjoin1 (subst_split _#4 Hsplit1) _ _)) as HofM1.
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
sor (refine (join_fst_o _#7 Hwf1 (subst_ofs _#4 (weaken2 _#4) HofT1) Hjoin1 (subst_split _#4 Hsplit1) _ _)) as (c & HfstM1 & Heqc).
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
so (snd_formation _#7 Hwf1 (ofm_subsume HofM1 (subst_subs _#5 (weaken2 _#4) Hsubs1)) Hsnd2 (subst_split _#4 Hsplit2) HfstM1) as Hofe2; simp_sub in Hofe2; simp_sub in Hofe2.
#3 sor (refine (join_formation _#7 Hwf2 (weaken2_ofs _#4 HofT2) Hjoin2 (subst_split _#4 Hsplit2) _ _)) as HofM2.
  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
#2 apply reqt_ceqt.
  eapply trsubs_formation; eauto using trsubs_trans; done.
clear Htrsubs1 Htrsubs2.
apply reqt_plam; [| exact svalue_lam].
apply reqt_lam; [].
#2 eapply reqt_trans.
  #2 apply reqt_app.
    #2 eapply reqt_trans.
      apply reqt_papp; [].
      eapply ceqt_reqt; [].
      eapply subst_ceqt; eauto using weaken2; done.
    simp_sub.
    apply reqt_papp_beta; done.

    #2 eapply reqt_trans.
      apply reqt_app; [| apply reqt_refl; done].
      #2 eapply reqt_trans.
        apply reqt_papp; [].
        eapply ceqt_reqt; [].
        eapply subst_ceqt; eauto using weaken2; done.
      simp_sub.
      apply reqt_papp_beta; done.
    simp_sub.
    apply reqt_app_beta; eauto using tvalue_thalt with dynamic sequence; done.
simp_sub.
rewrite -> subst_dot01_sh; [].
rewrite -> subst_dot1_sh; [].
#2 eapply reqt_trans.
  apply reqt_app_beta_instances; [].
  intros s Hs.
  so (subst_snd _#3 s Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
  #2 refine (snd_halt_iff _#4 _ Hsnd2x andel _).
    #2 eapply ofm_subsume.
      #2 eapply subst_ofm.
        exact (subof_aof_dof _#3 Hs).
      #2 eapply (join_formation _ (subst (sh 2) T1)); eauto using subst_split, weaken2_ofs.
        ofvar; done.

        simp_sub.
        simp_sub.
        rewrite -> subst_dot1_sh; [].
        ofvar; done.

      simp_sub.
      eapply subst_subs; eauto using subof_aof_sof, subof_sof_trans, weaken2; done.
  so (subst_join _#4 s Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
  #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit1) _ Hjoin1x andel _).
    simp_sub.
    assert (oft (G; cl_cn k1; cl_tm t1) (var 0) (shift1 t1)) as Hvar by ofvar.
    rewrite <- subst_dot1_sh in Hvar; [].
    so (subst_oft _#5 (subof_aof_dof _#3 Hs) Hvar) as H.
    simp_sub in H.
    exact H.

    assert (tvalue (G; cl_cn k1; cl_tm t1) (var 0)) as Hvar by ofvar.
    so (subst_tvalue _#4 (subof_aof_oof _#3 Hs) Hvar) as H.
    simp_sub in H.
    apply tvalue_thalt; exact H.
simp_sub.
so (subst_snd _#3 (dot e2 (dot (var 1) (sh 2))) Hsnd3') as Hsnd3x'; simp_sub in Hsnd3x'.
#2 sor (refine (snd_ceq _#9 Hwf1 (subst_split _#4 Hsplit3) _ HfstM1 Hsnd3 Hsnd3x')) as Hceqt.
  #2 apply ceqm_comp; apply (mcc_subsume _#6 _ _ (subst (sh 2) T2)).
    #2 eapply (ceqm_trans _ _ _ (subst (dot e2 (dot c (sh 2))) M2)).
      so (subst_join _#4 (dot e2 (dot c (sh 2))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
      refine (join_snd_inverse _#6 _ HfstM1 Hsnd2 Hjoin2x); [].
      #2 apply (@ofm_subsume _ _ (subst (sh 2) T1)).
        #2 eapply join_formation; eauto using weaken2_ofs, subst_split.
          ofvar; done.
    
          simp_sub.
          simp_sub.
          rewrite -> subst_dot1_sh; [].
          ofvar; done.
          
        apply weaken2_subs; eauto; done.

      replsub (subst (sh 2) T2) with (subst (dot e2 (dot c (sh 2))) (subst (sh 2) T2)).
      eapply funct_ceq_md; eauto; [].
      #3 prove_subrel.
        exact (eqc_reg _#4 Hwf1 (eqc_subsume Heqc (subst_subk _#5 (weaken2 _#4) Hsubk1)) anderl).

        exact (eqc_symm (eqc_subsume Heqc (subst_subk _#5 (weaken2 _#4) Hsubk1))).

        apply ceqt_refl; exact Hofe2.

    eapply subst_subs; eauto using weaken2; done.
eapply ceqt_reqt; apply ceqt_symm; exact Hceqt.

(* stat *)
intros G k k' Hsubk ka ka' t t' M e Hsplit Hsplit' Hjoin Hsnd.
invert Hsplit; clear Hsplit; [].
intros <- <-.
invert Hsplit'; clear Hsplit'; [].
intros <- <-.
invert Hjoin; [].
intros <-.
invert Hsnd; [].
intros <-.
apply ceqt_with_wf; intro Hwf; [].
#2 apply reqt_ceqt.
  eapply trsubs_formation; eauto using trsubs_stat, split.
apply reqt_plam; auto using svalue; [].
apply reqt_lam; [].
simp_sub.
eapply ceqt_reqt; [].
#3 apply ceqt_unit_ext_open.
  ofvar; done.

  #4 eapply oft_letm.
    #2 eapply (@ofm_lett _#3 (sg_stat (subst (sh 2) k))).
      ofvar; done.

      apply ofm_forget; [].
      relquest.
        eapply ofm_stat; [].
        ofvar; done.

        replsub (sg_stat (subst (sh 3) k)) with (shift1 (sg_stat (subst (sh 2) k))).
        reflexivity.

    apply fsts_stat; done.

    simp_sub.
    apply oft_star; done.

    apply ofc_unit; done.

  intros s Hs.
  simp_sub.
  repl (project s 0) with tm_star.
    invert (subof_index _ aof_behaved _#5 iscl_tm Hs index_0); [].
    intros Hof Hval.
    invert (Progress.canon_tm _ _ Hof Consistent.similar_unit Hval); auto; done.
  split.
    intros _.
    exists tm_star; split; auto with dynamic; [].
    #2 eapply star_step.
      eapply tstep_letm1; [].
      apply mstep_lett2; auto with dynamic; done.
    simp_sub.
    #2 eapply star_step.
      eapply tstep_letm2; auto with static dynamic; done.
    simp_sub.
    apply star_refl; done.

    intros _.
    apply tvalue_thalt; auto with dynamic; done.

(* pigen *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 k2' t2 t2' f1 f2 Hsplit1 Hsplit1' Hsplit2 Hsplit2' Htrsub1 IH1 Htrsub2 IH2 HofT2 k k' t t' M e Hsplit Hsplit' Hjoin Hsnd.
apply ceqt_with_wf; intro Hwf; [].
apply ceqt_symm; [].
#2 apply reqt_ceqt.
  eapply snd_join_coercion_formation; eauto; [].
  eapply subs_pigen; eauto using trsubs_subs, split_fsts; done.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-).
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-).
clear Hsplit1a Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2a' t1a' t2a' Hsplit1a' Hsplit2a' <- <-.
so (split_fun _#5 Hsplit1' Hsplit1a') as (<- & <-).
so (split_fun _#5 Hsplit2' Hsplit2a') as (<- & <-).
clear Hsplit1a' Hsplit2a'.
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub1)) as (Hhsg1' & Hhsg1).
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub2)) as (Hhsg2 & Hhsg2').
so (can_snd _ (var 0) (subst_hsg sh1 _ Hhsg1)) as (f & Hsnd1); simp_sub in Hsnd1.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg2)) as (N & Hjoin2).
so (can_snd _ (var 0) (subst_hsg sh1 _ Hhsg2')) as (f' & Hsnd2); simp_sub in Hsnd2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg1')) as (N' & Hjoin1).
clear Hhsg1 Hhsg1' Hhsg2 Hhsg2'.
simp_sub in Hjoin; simp_sub in Hsnd.
invert Hjoin; clear Hjoin; [].
intros fa Na Hsnd1a Hjoin2a <-.
simp_sub in Hsnd1a.
so (subst_snd _#3 (dot (var 0) (sh 5)) Hsnd1) as H; simp_sub in H.
so (snd_fun _#4 Hsnd1a H); subst fa; clear Hsnd1a H.
simp_sub in Hjoin2a.
so (subst_join _#4 (dot (var 0) (dot (var 1) (dot (var 3) (sh 7)))) Hjoin2) as H; simp_sub in H.
so (join_fun _#5 Hjoin2a H); subst Na; clear Hjoin2a H.
invert Hsnd; clear Hsnd; [].
intros k1a' t1a' k2a' t2a' Na' fa' Hsplit1a' Hsplit2a' Hjoin1a Hsnd2a <-.
so (subst_split _#3 (sh 2) Hsplit1') as H; simp_sub in H.
so (split_fun _#5 H Hsplit1a') as (<- & <-); clear H Hsplit1a'.
so (subst_split _#3 (under 1 (sh 2)) Hsplit2') as H; simp_sub in H.
so (split_fun _#5 H Hsplit2a') as (<- & <-); clear H Hsplit2a'.
simp_sub in Hjoin1a.
so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 6))) Hjoin1) as H; simp_sub in H.
so (join_fun _#5 H Hjoin1a); subst Na'; clear Hjoin1a H.
simp_sub in Hsnd2a.
so (subst_snd _#3 (dot (var 0) (dot (var 3) (sh 8))) Hsnd2) as H; simp_sub in H.
so (snd_fun _#4 H Hsnd2a); subst fa'; clear Hsnd2a H.
simp_sub.
apply reqt_plam; [| auto using svalue; done].
apply reqt_lam; [].
#2 eapply reqt_trans.
  apply reqt_letm; [].
  #2 eapply reqm_trans.
    apply reqm_lett_beta; eauto using tvalue_thalt with dynamic sequence; done.
  simp_sub.
  apply reqm_refl; done.
#2 eapply reqt_trans.
  apply reqt_letm_beta; eauto using mvalue_mhalt with static dynamic; done.
simp_sub.
apply reqt_plam; [| auto using svalue; done].
apply reqt_lam; [].
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsub1)) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (subs_reg _#3 (wf_cn Hwf Hofk1') (trsubs_subs _#4 Htrsub2)) as (_ & HofT2').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_formation _#4 HofT2' Hsplit2') as (Hofk2' & Hoft2').
so (subs_fst _#5 Hwf (trsubs_subs _#4 Htrsub1) (split_fsts _#3 Hsplit1') (split_fsts _#3 Hsplit1)) as Hsubk1.
#6 sor (refine (join_fst_o (G; cl_cn k1'; cl_tm t1') _#6 _ _ Hjoin1 _ _)) as (c & HfstN' & Heqc).
  eauto using wf; done.

  apply weaken2_ofs; auto; done.

  apply subst_split; eauto; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
#6 sor (refine (join_fst_o (G; cl_cn k1; cl_cn k2; cl_tm t2) _#6 _ _ Hjoin2 _ _)) as (d & HfstN & Heqd).
  eauto using wf; done.

  apply weaken2_ofs; auto; done.

  apply subst_split; eauto; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
#2 eapply reqt_trans.
  apply reqt_letm; [].
  #4 eapply (reqm_appgn_beta_instances_fstm_eqc _#6 (var 1)).
    intros s Hs.
    so (subst_join _#4 (compose (dot (var 0) (dot (var 1) (sh 4))) s) Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
    simp_sub.
    #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit1') _ Hjoin1x andel _).
      simp_sub.
      replsub (project s 0) with (subst s (var 0)).
      relquest.
        eapply subst_oft; eauto using subof_aof_dof; [].
        ofvar; done.

        simp_sub.
        reflexivity.

      apply tvalue_thalt; [].
      replsub (project s 0) with (subst s (var 0)).
      eapply subst_tvalue; eauto using subof_aof_oof; [].
      ofvar; done.

    eapply subst_fstm; eauto; [].
    prove_subof; done.

    eapply subst_fsts; eauto using split_fsts; done.

    apply eqc_symm; [].
    #2 eapply eqc_subsume.
      assert (var 1 = subst (dot (var 0) (dot (var 1) (sh 4))) (var 1)) as H by (simp_sub; reflexivity).
      rewrite -> H at 2; clear H; [].
      eapply subst_eqc; eauto; prove_subof; done.

      simp_sub.
      eapply subst_subk; eauto; prove_subof; done.
simp_sub.
clear c HfstN' Heqc.
#2 eapply reqt_trans.
  apply reqt_commute_letm_unpack; done.
#2 eapply reqt_unpack_bind.
  apply ceqt_symm; [].
  #2 apply ceqt_comp; eapply tcc_app.
    relquest.
      #2 apply ceqt_comp; eapply tcc_papp.
        relquest.
          apply ceqt_refl; [].
          ofvar; done.

          simp_sub.
          reflexivity.

        apply eqc_refl; [].
        #2 eapply ofc_subsume.
          ofvar; done.

          eapply subst_subk; eauto; prove_subof; done.

      simp_sub; reflexivity.

    so (subst_snd _#3 (dot N' (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
    so (IH1  _#6 Hsplit1' Hsplit1 Hjoin1 Hsnd1x) as Hceq; clear Hsnd1x IH1.
    #2 eapply ceqt_trans_formation.
      #2 apply ceqt_comp; eapply tcc_app.
        relquest.
          #2 apply ceqt_comp; eapply tcc_papp.
            relquest.
              eapply subst_ceqt; eauto; [].
              prove_subof; done.

              simp_sub; reflexivity.

            apply eqc_refl; [].
            ofvar; done.

          simp_sub; reflexivity.

        apply ceqt_refl; [].
        ofvar; done.
    intro Hof.
    apply reqt_ceqt; [exact Hof | clear Hof].
    simp_sub.
    #2 eapply reqt_trans.
      #2 apply reqt_app.
        apply reqt_papp_beta; done.
        
        apply reqt_refl; done.
    simp_sub.
    #2 eapply reqt_trans.
      apply reqt_app_beta; [].
      apply tvalue_thalt; [].
      ofvar; done.
    simp_sub.
    apply reqt_refl; done.

  clear IH1.
  simp_sub.
  assert (wf (G; cl_cn kd_one;
              cl_tm
                (cn_all (shift1 k1)
                   (cn_arrow (subst (dot (var 0) (sh 2)) t1)
                      (cn_exists (subst (dot (var 0) (sh 2)) k2)
                         (subst (dot (var 0) (dot (var 1) (sh 3))) t2))));
              cl_cn (subst (sh 2) k1'); cl_tm (subst (dot (var 0) (sh 3)) t1');
              cl_cn (subst (dot (var 1) (sh 4)) k2);
              cl_tm (subst (dot (var 0) (dot (var 2) (sh 5))) t2))) as Hwf'.
    #5 prove_wf; auto with static.
      #2 apply ofc_all.
        apply weaken1_ofk; auto; done.

        #2 apply ofc_arrow.
          eapply subst_ofc_type; eauto; prove_subof; done.

          #2 apply ofc_exists.
            eapply subst_ofk; eauto; prove_subof; done.

            eapply subst_ofc_type; eauto; prove_subof; done.

      eapply subst_ofk; eauto; prove_subof; done.

      eapply subst_ofc_type; eauto; prove_subof; done.

      eapply subst_ofk; eauto; prove_subof; [].
      #2 eapply ofc_subsume.
        ofvar; done.

        eapply subst_subk; eauto; prove_subof; done.

      eapply subst_ofc_type; eauto; prove_subof; [].
      #2 eapply ofc_subsume.
        ofvar; done.

        eapply subst_subk; eauto; prove_subof; done.
  #2 eapply reqt_trans.
    #2 apply reqt_letm_beta_instances.
      clear Hsplit2' Htrsub2 IH2 HofT2 HofT1' HofT1.
      intros s Hs.
      so (subst_join _#4 (compose (dot (var 0) (dot (var 1) (dot (var 3) (sh 6)))) s) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
      simp_sub.
      #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit2) _ Hjoin2x andel _); clear Hjoin2x.
        simp_sub.
        replsub (project s 0) with (subst s (var 0)).
        relquest.
          eapply subst_oft; eauto using subof_aof_dof; [].
          ofvar; done.

          simp_sub; reflexivity.

        apply tvalue_thalt; [].
        replsub (project s 0) with (subst s (var 0)).
        eapply subst_tvalue; eauto using subof_aof_oof with dynamic sequence; done.

      eapply subst_fstm; eauto; prove_subof; done.
  simp_sub.
  eapply ceqt_reqt; [].
  apply ceqt_symm; [].
  #4 apply ceqt_comp; apply tcc_pack.
    #2 eapply eqc_subsume.
      assert (var 1 = subst (dot (var 0) (dot (var 1) (dot (var 3) (sh 6)))) (var 1)) as H by (simp_sub; reflexivity).
      rewrite -> H at 3; clear H; [].
      eapply seq_eqc; [| | | exact Heqd]; eauto using wf; [].
      #2 prove_subrel.
        apply eqc_refl; [].
        #2 eapply ofc_subsume.
          ofvar; done.

          eapply subst_subk; eauto; prove_subof; done.

        apply eqc_refl; [].
        ofvar; done.

      simp_sub.
      so (subs_fst _#5 (wf_cn Hwf Hofk1') (trsubs_subs _#4 Htrsub2) (split_fsts _#3 Hsplit2) (split_fsts _#3 Hsplit2')) as Hsubk2.
      eapply subst_subk; eauto; prove_subof; done.

    apply eqk_refl; [].
    eapply subst_ofk; eauto; prove_subof; done.
    
    so (subst_snd _#3 (dot N (sh 2)) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
    so (IH2 _#6 Hsplit2 Hsplit2' Hjoin2 Hsnd2x) as Hceq; clear Hsnd2x IH2.
    #2 eapply ceqt_trans_formation.
      #2 apply ceqt_comp; eapply tcc_app.
        relquest.
          #2 apply ceqt_comp; eapply tcc_papp.
            relquest.
              eapply subst_ceqt; [| | exact Hceq]; eauto using wf; prove_subof; done.

              simp_sub; reflexivity.

            apply eqc_refl; ofvar; done.

          simp_sub; reflexivity.

        apply ceqt_refl; ofvar; done.
    intro Hof; apply reqt_ceqt; [exact Hof | clear Hof].
    simp_sub.
    #2 eapply reqt_trans.
      apply reqt_app; [| apply reqt_refl; done].
      apply reqt_papp_beta; done.
    simp_sub.
    #2 eapply reqt_trans.
      apply reqt_app_beta; [].
      apply tvalue_thalt; ofvar; done.
    simp_sub.
    apply reqt_refl; done.

    apply eqc_refl; [].
    eapply subst_ofc_type; eauto; prove_subof; done.

(* piapp *)
intros G T1 T1' T2 T2' k1 k1' t1 t1' k2 t2 f1 f2 Hsplit1 Hsplit1' Hsplit2 Htrsub1 IH1 Htrsub2 IH2 HofT2 k k' t t' M e Hsplit Hsplit' Hjoin Hsnd.
apply ceqt_with_wf; intro Hwf; [].
apply ceqt_symm; [].
#2 apply reqt_ceqt.
  eapply snd_join_coercion_formation; eauto; [].
  eapply subs_piapp; eauto using trsubs_subs, split_fsts; done.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-).
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-).
clear Hsplit1a Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2' t1a' t2' Hsplit1a' Hsplit2' <- <-.
so (split_fun _#5 Hsplit1' Hsplit1a') as (<- & <-).
clear Hsplit1a'.
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub1)) as (Hhsg1' & Hhsg1).
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub2)) as (Hhsg2 & Hhsg2').
so (can_snd _ (var 0) (subst_hsg sh1 _ Hhsg1)) as (f & Hsnd1); simp_sub in Hsnd1.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg2)) as (N & Hjoin2).
so (can_snd _ (var 0) (subst_hsg sh1 _ Hhsg2')) as (f' & Hsnd2); simp_sub in Hsnd2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg1')) as (N' & Hjoin1).
clear Hhsg1 Hhsg1' Hhsg2 Hhsg2'.
so (split_fsts _#3 Hsplit1) as Hfst1.
so (split_fsts _#3 Hsplit1') as Hfst1'.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsub1)) as (HofT1' & HofT1).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (subs_fst _#5 Hwf (trsubs_subs _#4 Htrsub1) Hfst1' Hfst1) as Hsubk1.
so (subst_snd _#3 (dot N' (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
so (IH1 _#6 Hsplit1' Hsplit1 Hjoin1 Hsnd1x) as Hceq1.
clear Hsnd1x IH1.
so (subst_snd _#3 (dot N (sh 2)) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
so (IH2 _#6 Hsplit2 Hsplit2' Hjoin2 Hsnd2x) as Hceq2.
clear Hsnd2x IH2.
simp_sub in Hjoin; simp_sub in Hsnd.
invert Hjoin; clear Hjoin; [].
intros fa Na Hsnd1a Hjoin2a <-.
simp_sub in Hsnd1a.
so (subst_snd _#3 (dot (var 0) (sh 5)) Hsnd1) as H; simp_sub in H.
so (snd_fun _#4 Hsnd1a H); subst fa; clear Hsnd1a H.
simp_sub in Hjoin2a.
so (subst_join _#4 (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (var 0) (sh 5)) f)) (dot (cn_app (var 4) (var 1)) (dot (var 1) (sh 5)))) Hjoin2) as H; simp_sub in H.
so (join_fun _#5 Hjoin2a H); subst Na; clear Hjoin2a H.
invert Hsnd; clear Hsnd; [].
intros k1a' t1a' Na' fa' Hsplit1a' Hjoin1a Hsnd2a <-.
so (subst_split _#3 (sh 2) Hsplit1') as H; simp_sub in H.
so (split_fun _#5 H Hsplit1a') as (<- & <-); clear H Hsplit1a'.
simp_sub in Hjoin1a.
so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 6))) Hjoin1) as H; simp_sub in H.
so (join_fun _#5 H Hjoin1a); subst Na'; clear Hjoin1a H.
simp_sub in Hsnd2a.
so (subst_snd _#3 (dot (md_appap (var 2) (subst (dot (var 0) (dot (var 1) (sh 6))) N')) (dot (var 1) (sh 6))) Hsnd2) as H; simp_sub in H.
so (snd_fun _#4 H Hsnd2a); subst fa'; clear Hsnd2a H.
simp_sub.
apply reqt_plam; [| apply svalue_lam; done].
apply reqt_lam; [].
#2 eapply reqt_trans.
  apply reqt_letm; [].
  apply reqm_lett_beta; [].
  apply tvalue_thalt; ofvar; done.
simp_sub.
sor (refine (join_formation_and_fst_o (G; cl_cn k1'; cl_tm t1') _#6 _ _ Hjoin1 (subst_split _#4 Hsplit1') _ _)) as (HofN' & c' & HfstN' & Heqc').
  prove_wf; auto; done.

  apply weaken2_ofs; eauto; done.

  ofvar; done.

  simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
  ofvar; done.
sor (refine (join_formation_and_fst_o (G; cl_cn k1; cl_cn k2; cl_tm t2) _#6 _ _ Hjoin2 (subst_split _#4 Hsplit2) _ _)) as (HofN & c & HfstN & Heqc).
  prove_wf; auto; done.

  apply weaken2_ofs; eauto; done.
  
  ofvar; done.

  simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
  ofvar; done.
so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
#4 sor (refine (snd_formation (G; cl_cn k1; cl_md (shift1 T1) (var 0)) _#6 _ _ Hsnd1x (subst_split _#4 Hsplit1) _)) as Hoff.
  apply wf_md; auto; done.

  ofvar; done.

  ofvar; done.
clear Hsnd1x.
simp_sub in Hoff.
simp_sub in Hoff.
rewrite -> subst_dot1_sh in Hoff; [].
sor (refine (eqc_reg _#4 _ Heqc)) as (_ & Hofc & _).
  prove_wf; auto; done.
#2 eapply reqt_trans.
  #2 apply reqt_letm_beta.
    apply mvalue_mhalt; auto with dynamic; done.

    #2 eapply (@fstm_lamap _#4 (subst (dot nonsense (dot (cn_app (var 2) (var 0)) (dot (var 0) (sh 3)))) c)).
      eapply subst_fsts; eauto; done.

      relquest.
        eapply subst_fstm; eauto; [].
        prove_subof; done.

        simp_sub.
        rewrite -> (subord_tm_cn _#4 Hofc) at 1; [].
        simp_sub; reflexivity.
simp_sub.
#2 eapply reqt_trans.
  apply reqt_plam; [| eapply svalue_lam; done].
  apply reqt_lam; [].
  eapply ceqt_reqt; [].
  apply ceqt_with_wf; intro Hwf'; [].
  so (subst_snd _#3 (dot (md_appap (md_lamap (subst (sh 4) T1) (subst (dot (tm_app (tm_papp (var 4) (var 1)) (subst (dot (var 0) (sh 6)) f)) (dot (cn_app (var 5) (var 1)) (dot (var 1) (sh 6)))) N)) (subst (dot (var 0) (dot (var 1) (sh 4))) N')) (dot (var 1) (sh 4))) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
  #3 refine (snd_ceq_2 _#9 Hwf' (subst_split _#4 Hsplit2') _ _ Hsnd2x _); clear Hsnd2x.
    #2 apply reqm_ceqm.
      clear Hceq1 Hceq2.
      #2 eapply ofm_subsume.
        #3 eapply ofm_appap.
          #3 eapply ofm_lamap.
            eapply subst_ofs; eauto; prove_subof; done.

            eapply subst_fsts; eauto; done.

            relquest.
              eapply subst_ofm; eauto; [].
              #2 prove_subof.
                relquest.
                  #2 eapply ofc_app.
                    relquest.
                      eapply ofc_var; auto 6 with sequence; done.

                      simp_sub; reflexivity.

                    ofvar; done.

                  simp_sub; reflexivity.

                #2 eapply oft_app.
                  relquest.
                    #2 eapply oft_papp.
                      ofvar; done.

                      ofvar; done.

                    simp_sub; reflexivity.

                  replsub (subst (dot (var 0) (sh 6)) f) with (subst (dot (var 0) (dot (var 1) (sh 6))) (subst (dot (var 0) (sh 2)) f)).
                  relquest.
                    eapply subst_oft; eauto; [].
                    prove_subof; done.

                    simp_sub; reflexivity.

              simp_sub.
              replsub (subst (dot (var 1) (sh 6)) T2) with (shift1 (subst (dot (var 0) (sh 5)) T2)).
              reflexivity.

          #2 eapply ofm_subsume.
            eapply subst_ofm; eauto; [].
            prove_subof; done.

            simp_sub.
            eapply subst_subs; eauto using trsubs_subs; prove_subof; done.

          eapply subst_fstm; eauto; prove_subof; done.

        simp_sub.
        eapply (seq_subs _ _ _ (G; cl_cn k1')); eauto using trsubs_subs, wf; [].
        prove_subrel; [].
        apply eqc_symm; [].
        assert (var 1 = subst (dot (var 0) (dot (var 1) (sh 4))) (var 1)) as H by (simp_sub; reflexivity); rewrite -> H at 2; clear H.
        relquest.
          eapply subst_eqc; eauto; prove_subof; done.

          simp_sub; reflexivity.
      clear Hwf'.
      #4 eapply reqm_appap_beta_instances_fstm_eqc.
        intros s Hs.
        simp_sub.
        so (subst_join _#4 (dot (project s 0) (dot (project s 1) (compose (sh 4) s))) Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
        #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit1') _ Hjoin1x andel _).
          simp_sub.
          replsub (project s 0) with (subst s (var 0)).
          relquest.
            eapply subst_oft; eauto using subof_aof_dof; [].
            ofvar; done.

            simp_sub; reflexivity.

          replsub (project s 0) with (subst s (var 0)).
          eapply subst_thalt; eauto using subof_aof_oof; [].
          apply tvalue_thalt; ofvar; done.

        eapply subst_fstm; eauto; [].
        prove_subof; done.

        eapply subst_fsts; eauto; done.

        eapply eqc_symm; [].
        #2 eapply eqc_subsume.
          eapply subst_eqc; eauto; prove_subof; done.

          simp_sub.
          eapply subst_subk; eauto; prove_subof; done.

    simp_sub.
    eapply subst_fstm; eauto; [].
    prove_subof; done.

    simp_sub.
    so (subst_snd _#3 (dot (subst (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (sh 4)) f)) (dot (cn_app (var 3) (var 1)) (dot (var 1) (sh 4)))) N) (dot (var 1) (sh 4))) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
    exact Hsnd2x.
eapply ceqt_reqt; [].
apply ceqt_symm; [].
#2 apply reqt_ceqt.
  so (trsubs_formation _#8 Hwf (trsubs_piapp Hsplit1 Hsplit1' Hsplit2 Htrsub1 Htrsub2 HofT2) (split_piapp Hsplit1 Hsplit2) (split_piapp Hsplit1' Hsplit2')) as H1.
  so (Inversion.plam_oft_inversion_all _#5 Hwf H1) as (_ & H2 & _); clear H1.
  so (Inversion.lam_oft_inversion_arrow _#5 (wf_cn Hwf (ofk_pi Hofk1 Hofk2)) H2) as (_ & H3); clear H2.
  exact H3.
apply reqt_plam; [| apply svalue_lam; done].
apply reqt_lam; [].
#2 eapply reqt_trans.
  #2 apply reqt_app.
    #2 eapply reqt_trans.
      apply reqt_papp; [].
      eapply ceqt_reqt; [].
      eapply subst_ceqt; [| | exact Hceq2]; eauto using wf; [].
      prove_subof; done.
    simp_sub.
    apply reqt_papp_beta; done.

    apply reqt_app; [apply reqt_refl; done |].
    #2 eapply reqt_trans.
      apply reqt_app; [| apply reqt_refl; done].
      #2 eapply reqt_trans.
        apply reqt_papp; [].
        eapply ceqt_reqt; [].
        eapply subst_ceqt; eauto; prove_subof; done.
      simp_sub.
      apply reqt_papp_beta; done.
    simp_sub.
    apply reqt_app_beta; [].
    apply tvalue_thalt; ofvar; done.
simp_sub.
clear Hceq1 Hceq2 c HfstN Heqc Hofc.
apply reqt_with_wf; intro Hwf'; [].
#2 eapply reqt_trans.
  apply reqt_app_beta_instances_conditional; [].
  intros s Hs Hhalt.
  simp_sub in Hhalt.
  simp_sub.
  assert (oft (G; cl_cn (kd_pi k1 k2); cl_tm (cn_all (shift1 k1) (cn_arrow (subst (dot (var 0) (sh 2)) t1) (subst (dot (cn_app (var 1) (var 0)) (dot (var 0) (sh 2))) t2))); cl_cn (subst (sh 2) k1'); cl_tm (subst (dot (var 0) (sh 3)) t1')) (tm_app (tm_papp (var 2) (var 1)) (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (sh 4)) f)) (subst (dot (cn_app (var 3) (var 1)) (dot (var 1) (sh 4))) t2)) as Hofmess.
    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          ofvar; done.

          #2 eapply ofc_subsume.
            ofvar; done.

            eapply subst_subk; eauto; prove_subof; done.

        simp_sub; reflexivity.

      replsub (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (sh 4)) f) with (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (dot (subst (dot (var 0) (dot (var 1) (sh 4))) c') (sh 4))) (subst (dot (var 0) (sh 2)) f)).
      #2 eapply oft_equiv.
        eapply subst_oft; eauto; [].
        #3 prove_subof.
          #2 eapply ofc_subsume.
            so (eqc_reg _#4 (wf_tm (wf_cn Hwf Hofk1') Hoft1') Heqc') as (_ & Hofc' & _).
            eapply subst_ofc; eauto; prove_subof; done.
  
            simp_sub.
            eapply subst_subk; eauto; prove_subof; done.
  
          #2 eapply ofm_subsume.
            eapply subst_ofm; eauto; prove_subof; done.
  
            simp_sub.
            eapply subst_subs; eauto using trsubs_subs; prove_subof; done.
  
          eapply subst_fstm; eauto; prove_subof; done.
          
        simp_sub.
        apply eqc_symm; [].
        eapply seq_ofc_type; eauto; [].
        prove_subrel; [].
        assert (var 1 = subst (dot (var 0) (dot (var 1) (sh 4))) (var 1)) as H by (simp_sub; reflexivity); rewrite -> H at 2; clear H; [].
        #2 eapply eqc_subsume.
          eapply subst_eqc; eauto; prove_subof; done.

          simp_sub.
          eapply subst_subk; eauto; prove_subof; done.
  so (subst_snd _#3 (dot (subst (dot (tm_app (tm_papp (project s 2) (project s 1)) (subst (dot (subst (dot (project s 0) (dot (project s 1) (compose (sh 4) s))) N') (compose (sh 4) s)) f)) (dot (cn_app (project s 3) (project s 1)) (dot (project s 1) (compose (sh 4) s)))) N) (dot (project s 1) (compose (sh 4) s))) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
  #2 sor (refine (snd_halt_iff impure _#3 _ Hsnd2x ander Hhalt)) as Hhalt'; clear Hsnd2x Hhalt.
    simp_sub.
    apply ofm_forget; [].
    replsub (subst (dot (tm_app (tm_papp (project s 2) (project s 1)) (subst (dot (subst (dot (project s 0) (dot (project s 1) (compose (sh 4) s))) N') (compose (sh 4) s)) f)) (dot (cn_app (project s 3) (project s 1)) (dot (project s 1) (compose (sh 4) s)))) N) with (subst s (subst (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (sh 4)) f)) (dot (cn_app (var 3) (var 1)) (dot (var 1) (sh 4)))) N)).
    replsub (subst (dot (project s 1) (compose (sh 4) s)) T2') with (subst s (subst (dot (var 1) (sh 4)) T2')).
    eapply subst_ofm; eauto using subof_aof_dof; [].
    #2 eapply ofm_subsume.
      eapply subst_ofm; eauto; [].
      #3 prove_subof.
        #2 eapply ofc_subsume.
          ofvar; done.
  
          eapply subst_subk; eauto; prove_subof; done.
  
        relquest.
          #2 eapply ofc_app.
            ofvar; done.
  
            #2 eapply ofc_subsume.
              ofvar; done.
  
              eapply subst_subk; eauto; prove_subof; done.
  
          simp_sub; reflexivity.
  
        exact Hofmess.

      simp_sub.
      eapply subst_subs; eauto using trsubs_subs; prove_subof; done.
  so (subst_join _#4 (dot (tm_app (tm_papp (project s 2) (project s 1)) (subst (dot (subst (dot (project s 0) (dot (project s 1) (compose (sh 4) s))) N') (compose (sh 4) s)) f)) (dot (cn_app (project s 3) (project s 1)) (dot (project s 1) (compose (sh 4) s)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
  #2 sor (refine (join_halt_iff _#6 (subst_split _#4 Hsplit2) _ Hjoin2x ander Hhalt')) as Hhalt''; clear Hjoin2x Hhalt'.
    simp_sub.
    replsub (tm_app (tm_papp (project s 2) (project s 1)) (subst (dot (subst (dot (project s 0) (dot (project s 1) (compose (sh 4) s))) N') (compose (sh 4) s)) f)) with (subst s (tm_app (tm_papp (var 2) (var 1)) (subst (dot (subst (dot (var 0) (dot (var 1) (sh 4))) N') (sh 4)) f))).
    replsub (subst (dot (cn_app (project s 3) (project s 1)) (dot (project s 1) (compose (sh 4) s))) t2) with (subst s (subst (dot (cn_app (var 3) (var 1)) (dot (var 1) (sh 4))) t2)).
    eapply subst_oft; eauto using subof_aof_dof; done.
  exact Hhalt''.
simp_sub.
apply reqt_refl; done.

(* sigma *)
intros G T1 T1' T2 T2' k1 k1' t1 k2 t2 f1 f2 Hsplit1 Hfst1' Hsplit2 Htrsub1 IH1 Htrsub2 IH2 HofT2' k k' t t' M e Hsplit Hsplit' Hjoin Hsnd.
apply ceqt_with_wf; intro Hwf; [].
apply ceqt_symm; [].
#2 apply reqt_ceqt.
  eapply snd_join_coercion_formation; eauto; [].
  eapply subs_sigma; eauto using trsubs_subs, split_fsts; done.
invert Hsplit; clear Hsplit; [].
intros k1a k2a t1a t2a Hsplit1a Hsplit2a <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-).
so (split_fun _#5 Hsplit2 Hsplit2a) as (<- & <-).
clear Hsplit1a Hsplit2a.
invert Hsplit'; clear Hsplit'; [].
intros k1a' k2' t1' t2' Hsplit1' Hsplit2' <- <-.
so (split_fsts_eq _#4 Hsplit1' Hfst1'); subst k1a'.
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub1)) as (Hhsg1' & Hhsg1).
so (subs_hsg _#3 (trsubs_subs _#4 Htrsub2)) as (Hhsg2 & Hhsg2').
so (can_snd _ (var 0) (subst_hsg (sh 2) _ Hhsg1)) as (e1 & Hsnd1); simp_sub in Hsnd1.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg2)) as (N2 & Hjoin2).
so (can_snd _ (var 0) (subst_hsg (sh 2) _ Hhsg2')) as (e2 & Hsnd2); simp_sub in Hsnd2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ Hhsg1')) as (N1 & Hjoin1).
clear Hhsg1 Hhsg1' Hhsg2 Hhsg2'.
so (split_fsts _#3 Hsplit1) as Hfst1.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsub1)) as (HofT1 & HofT1').
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT1' Hsplit1') as (Hofk1' & Hoft1').
so (subs_reg _#3 (wf_cn Hwf Hofk1) (trsubs_subs _#4 Htrsub2)) as (HofT2 & _).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (subs_fst _#5 Hwf (trsubs_subs _#4 Htrsub1) Hfst1 Hfst1') as Hsubk1.
#4 sor (refine (join_fst_o _#7 (wf_tm (wf_cn Hwf Hofk1) Hoft1) _ Hjoin1 (subst_split _#4 Hsplit1) _ _)) as (c1 & HfstN1 & Heqc1).
  apply weaken2_ofs; auto; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
#4 sor (refine (join_fst_o _#7 (wf_tm (wf_cn (wf_cn Hwf Hofk1) Hofk2) Hoft2) _ Hjoin2 (subst_split _#4 Hsplit2) _ _)) as (c2 & HfstN2 & Heqc2).
  apply weaken2_ofs; auto; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
so (subst_snd _#3 (dot N1 (dot c1 (sh 2))) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
so (IH1 _#6 Hsplit1 Hsplit1' Hjoin1 Hsnd1x) as Hceq1.
clear Hsnd1x IH1.
so (subst_snd _#3 (dot N2 (dot c2 (sh 2))) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
so (IH2 _#6 Hsplit2 Hsplit2' Hjoin2 Hsnd2x) as Hceq2.
clear Hsnd2x IH2.
simp_sub in Hjoin; simp_sub in Hsnd.
invert Hjoin; clear Hjoin; [].
intros N1a N2a Hjoin1a Hjoin2a <-.
simp_sub in Hjoin1a.
so (subst_join _#4 (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 2)) (sh 3))) Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
so (join_fun _#5 Hjoin1a Hjoin1x); subst N1a; clear Hjoin1a Hjoin1x.
simp_sub in Hjoin2a.
so (subst_join _#4 (dot (tm_pi2 (var 0)) (dot (cn_pi2 (var 2)) (dot (cn_pi1 (var 2)) (sh 3)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
so (join_fun _#5 Hjoin2a Hjoin2x); subst N2a; clear Hjoin2a Hjoin2x.
invert Hsnd; clear Hsnd; [].
intros e1a e2a Hsnd1a Hsnd2a <-.
simp_sub in Hsnd1a.
so (subst_snd _#3 (dot (md_pi1 (var 0)) (dot (cn_pi1 (var 1)) (sh 4))) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
so (snd_fun _#4 Hsnd1a Hsnd1x); subst e1a; clear Hsnd1a Hsnd1x.
simp_sub in Hsnd2a.
so (subst_snd _#3 (dot (md_pi2 (var 0)) (dot (cn_pi2 (var 1)) (dot (cn_pi1 (var 1)) (sh 4)))) Hsnd2) as Hsnd2x; simp_sub in Hsnd2x.
so (snd_fun _#4 Hsnd2a Hsnd2x); subst e2a; clear Hsnd2a Hsnd2x.
apply reqt_plam; [| auto using svalue; done].
apply reqt_lam; [].
#2 eapply reqt_trans.
  apply reqt_letm; [].
  apply reqm_lett_beta; [].
  apply tvalue_thalt; ofvar; done.
simp_sub.
assert (forall s,
        subof aof nil s
          (G; cl_cn (kd_sigma k1 k2);
           cl_tm
             (cn_prod (subst (dot (cn_pi1 (var 0)) sh1) t1)
                (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2))) ->
        mhalt nil
          (subst s (subst (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 1)) (sh 2))) N1))) as HhaltN1.
  intros s Hs.
  simp_sub.
  so (subst_join _#4 (compose (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 1)) (sh 2))) s) Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
  #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit1) _ Hjoin1x andel _).
    simp_sub.
    eapply oft_pi1; [].
    relquest.
      rewrite <- subst_var; [].
      eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

      simp_sub; reflexivity.

    rewrite <- subst_var; [].
    #2 eapply thalt_pi1.
      relquest.
        eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

        simp_sub; reflexivity.

      apply tvalue_thalt; [].
      eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.
assert (forall s,
        subof aof nil s
          (G; cl_cn (kd_sigma k1 k2);
           cl_tm
             (cn_prod (subst (dot (cn_pi1 (var 0)) sh1) t1)
                (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2))) ->
        mhalt nil
          (subst s
             (subst
                (dot (tm_pi2 (var 0))
                   (dot (cn_pi2 (var 1)) (dot (cn_pi1 (var 1)) (sh 2)))) N2))) as HhaltN2.
  intros s Hs.
  simp_sub.
  so (subst_join _#4 (compose (dot (tm_pi2 (var 0)) (dot (cn_pi2 (var 1)) (dot (cn_pi1 (var 1)) (sh 2)))) s) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
  #2 refine (join_halt_iff _#6 (subst_split _#4 Hsplit2) _ Hjoin2x andel _).
    simp_sub.
    eapply oft_pi2; [].
    relquest.
      rewrite <- subst_var; [].
      eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

      simp_sub; reflexivity.

    rewrite <- subst_var; [].
    #2 eapply thalt_pi2.
      relquest.
        eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

        simp_sub; reflexivity.

      apply tvalue_thalt; [].
      eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.
#2 eapply reqt_trans.
  #2 eapply reqt_letm_beta_instances.
    intros s Hs.
    simp_sub.
    #2 apply mhalt_pair.
      so (HhaltN1 s Hs) as H; simp_sub in H; exact H.

      so (HhaltN2 s Hs) as H; simp_sub in H; exact H.

  #2 eapply fstm_pair.
    eapply subst_fstm; eauto; prove_subof; done.

    eapply subst_fstm; eauto; prove_subof; done.
simp_sub.
eapply ceqt_reqt; [].
apply ceqt_symm; [].
#2 eapply reqt_ceqt.
  so (trsubs_formation _#8 Hwf (trsubs_sigma Hsplit1 Hfst1' Hsplit2 Htrsub1 Htrsub2 HofT2') (split_sigma Hsplit1 Hsplit2) (split_sigma Hsplit1' Hsplit2')) as H1.
  so (Inversion.plam_oft_inversion_all _#5 Hwf H1) as (_ & H2 & _); clear H1.
  so (Inversion.lam_oft_inversion_arrow _#5 (wf_cn Hwf (ofk_sigma Hofk1 Hofk2)) H2) as (_ & H3); clear H2.
  exact H3.
apply reqt_with_wf; intro Hwf'; [].
sor (refine (join_formation _#7 (wf_tm (wf_cn Hwf Hofk1) Hoft1) _ Hjoin1 (subst_split _#4 Hsplit1) _ _)) as HofN1.
  apply weaken2_ofs; auto; done.

  ofvar; done.

  simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
  ofvar; done.
sor (refine (join_formation _#7 (wf_tm (wf_cn (wf_cn Hwf Hofk1) Hofk2) Hoft2) _ Hjoin2 (subst_split _#4 Hsplit2) _ _)) as HofN2.
  apply weaken2_ofs; auto; done.

  ofvar; done.

  simp_sub; simp_sub; rewrite -> subst_dot1_sh; [].
  ofvar; done.
assert (subof dof (G; cl_cn (kd_sigma k1 k2);
                   cl_tm (cn_prod (subst (dot (cn_pi1 (var 0)) sh1) t1)
                            (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2)))
          (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 1)) (sh 2)))
          (G; cl_cn k1; cl_tm t1)) as Hofsub1.
  #2 prove_subof.
    eapply ofc_pi1; ofvar; done.

    eapply oft_pi1; ofvar; done.
assert (subof dof
          (G; cl_cn (kd_sigma k1 k2);
           cl_tm
             (cn_prod (subst (dot (cn_pi1 (var 0)) sh1) t1)
                (subst (dot (cn_pi2 (var 0)) (dot (cn_pi1 (var 0)) sh1)) t2)))
          (dot (tm_pi2 (var 0))
             (dot (cn_pi2 (var 1)) (dot (cn_pi1 (var 1)) (sh 2))))
          (G; cl_cn k1; cl_cn k2; cl_tm t2)) as Hofsub2.
  #3 prove_subof.
    eapply ofc_pi1; ofvar; done.
    
    relquest.
      eapply ofc_pi2; ofvar; done.

      simp_sub; reflexivity.

    eapply oft_pi2; ofvar; done.
#2 apply reqt_pair.
  #2 eapply reqt_trans.
    apply reqt_app; [| apply reqt_refl; done].
    #2 eapply reqt_trans.
      apply reqt_papp; [].
      eapply ceqt_reqt; [].
      refine (subst_ceqt _#6 _ _ Hceq1); auto using wf; prove_subof; done.
    simp_sub.
    apply reqt_papp_beta; done.
  simp_sub.
  #2 eapply reqt_trans.
    apply reqt_app_beta_instances; [].
    intros s Hs.
    simp_sub.
    rewrite <- subst_var; [].
    #2 eapply thalt_pi1.
      relquest.
        eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

        simp_sub; reflexivity.

      apply tvalue_thalt; [].
      eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.
  simp_sub.
  eapply ceqt_reqt; [].
  #3 sor (refine (snd_formation _#7 (wf_md Hwf HofT1' Hfst1') _ Hsnd1 (subst_split _#4 Hsplit1') _)) as Hofe1.
    ofvar; done.

    ofvar; done.
  simp_sub in Hofe1.
  simp_sub in Hofe1.
  rewrite -> subst_dot1_sh in Hofe1; [].
  #2 refine (funct_ceq_tm _#6 Hwf' _ _ Hofe1).
    prove_wf; auto; done.

    #4 apply subrel_deq_ceq_md; auto.
      prove_subrel; done.

      apply ceqm_symm; [].
      #3 eapply ceqm_pi1_beta_instances.
        #2 eapply ofm_subsume.
          eapply subst_ofm; eauto; done.

          simp_sub.
          eapply subst_subs; eauto using trsubs_subs; prove_subof; done.

        #2 eapply ofm_subsume.
          eapply subst_ofm; eauto; done.

          simp_sub.
          eapply subst_subs; eauto using trsubs_subs; prove_subof; [].
          eapply ofc_pi1; ofvar; done.

        exact HhaltN2.

      eapply subst_fstm; eauto using subof_dof_tof; done.

      apply fstm_pi1; [].
      #2 apply fstm_pair.
        eapply subst_fstm; eauto using subof_dof_tof; done.

        eapply subst_fstm; eauto using subof_dof_tof; done.

  #2 eapply reqt_trans.
    apply reqt_app; [| apply reqt_refl; done].
    #2 eapply reqt_trans.
      apply reqt_papp; [].
      eapply ceqt_reqt; [].
      refine (subst_ceqt _#6 _ _ Hceq2); auto using wf; [].
      prove_subof; [].
      eapply ofc_pi1; ofvar; done.
    simp_sub.
    apply reqt_papp_beta; done.
  simp_sub.
  #2 eapply reqt_trans.
    apply reqt_app_beta_instances; [].
    intros s Hs.
    simp_sub.
    rewrite <- subst_var; [].
    #2 eapply thalt_pi2.
      relquest.
        eapply subst_oft; eauto using subof_aof_dof; ofvar; done.

        simp_sub; reflexivity.

      apply tvalue_thalt; [].
      eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.
  simp_sub.
  eapply ceqt_reqt; [].
  so (split_fsts _#3 Hsplit2') as Hfst2'.
  #3 sor (refine (snd_formation _#7 (wf_md (wf_cn Hwf Hofk1') HofT2' Hfst2') _ Hsnd2 (subst_split _#4 Hsplit2') _)) as Hofe2.
    ofvar; done.

    ofvar; done.
  simp_sub in Hofe2.
  simp_sub in Hofe2.
  rewrite -> subst_dot1_sh in Hofe2; [].
  #2 refine (funct_ceq_tm _#6 Hwf' _ _ Hofe2).
    prove_wf; auto; done.
    
    #4 apply subrel_deq_ceq_md; auto.
      #2 apply subrel_deq_cn; auto.
        prove_subrel; done.

        #2 eapply eqc_subsume.
          eapply eqc_pi1; [].
          #2 eapply eqc_trans.
            apply Declarative.eqc_sigma_eta; auto; [].
            ofvar; done.
          #3 eapply eqc_pair.
            replsub (cn_pi1 (var 1)) with (subst (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 1)) (sh 2))) (var 1)).
            relquest.
              eapply subst_eqc; eauto; [].
              prove_subof; [].
              eapply ofc_pi1; ofvar; done.
  
              simp_sub; reflexivity.
  
            simp_sub.
            assert (cn_pi2 (var 1) = subst (dot (tm_pi2 (var 0)) (dot (cn_pi2 (var 1)) (dot (cn_pi1 (var 1)) (sh 2)))) (var 1)) as H by (simp_sub; reflexivity); rewrite -> H at 1; clear H.
            relquest.
              eapply subst_eqc; eauto; [].
              #2 prove_subof.
                eapply ofc_pi1; ofvar; done.
  
                relquest.
                  eapply ofc_pi2; ofvar; done.
  
                  simp_sub; reflexivity.
  
              simp_sub; reflexivity.
  
            eapply subst_ofk; eauto; prove_subof; done.
  
          eapply subst_subk; eauto; prove_subof; done.

      apply ceqm_symm; [].
      #3 eapply ceqm_pi2_beta_instances.
        eapply subst_ofm; eauto; done.

        #2 eapply ofm_subsume.
          eapply subst_ofm; eauto; done.

          simp_sub.
          eapply subst_subs; eauto using trsubs_subs; prove_subof; [].
          eapply ofc_pi1; ofvar; done.

        exact HhaltN1.

      eapply subst_fstm; eauto using subof_dof_tof; done.

      apply fstm_pi2; [].
      #2 apply fstm_pair.
        eapply subst_fstm; eauto using subof_dof_tof; done.

        eapply subst_fstm; eauto using subof_dof_tof; done.
Qed.


(* Expressions translate iff they are well-formed. *)

Lemma trp_fstm :
  forall G M T c e,
    trp G M T c e
    -> fstm G M c.
Proof.
intros G M T c e H.
induct H; eauto using split_fsts with static; [].

(* lett *)
intros G t T e M _ c2 _ _ _ HfstM.
apply fstm_lett; [].
assert (subof tof (G; cl_utm) (dot (var 0) sh1) (G; cl_tm t)) as Hs.
  prove_subof; done.
so (subst_fstm _#5 Hs HfstM) as H.
simp_sub in H.
exact H.
Qed.


Lemma trans_of :
  (forall G M T c e,
     wf G
     -> trp G M T c e
     -> ofm G pure M T)
  /\
  (forall G M T e,
     wf G
     -> tri G M T e
     -> ofm G impure M T)
  /\
  (forall G e t e',
     wf G
     -> trt G e t e'
     -> oft G e t).
Proof.
exploit
  (trans_ind
     (fun G M T c e => ofmp G pure M T)
     (fun G M T e => ofmp G impure M T)
     (fun G e t e' => oftp G e t))
  as Hind;
eauto using oftp, ofmp, trp_fstm, split_fsts, trsubs_subs with static; [].
destruct Hind as (H1 & H2 & H3).
do2 2 split; intros; eauto using ofmp_ofm, oftp_oft; done.
Qed.


Lemma trp_ofm :
  forall G M T c e,
    wf G
    -> trp G M T c e
    -> ofm G pure M T.
Proof.
exact (trans_of andel).
Qed.


Lemma tri_ofm :
  forall G M T e,
    wf G
    -> tri G M T e
    -> ofm G impure M T.
Proof.
exact (trans_of anderl).
Qed.


Lemma trt_oft :
  forall G e t e',
    wf G
    -> trt G e t e'
    -> oft G e t.
Proof.
exact (trans_of anderr).
Qed.


Lemma can_trans :
  (forall G M T,
     wf G
     -> ofm G pure M T
     -> exists c e, trp G M T c e)
  /\
  (forall G M T,
     wf G
     -> ofm G impure M T
     -> exists e, tri G M T e)
  /\
  (forall G e t,
     wf G
     -> oft G e t
     -> exists e', trt G e t e').
Proof.
#15 exploit
  (dynamic_principal_ind
     (fun G e t =>
        wf G -> exists e', trt G e t e')
     (fun G p M T =>
        wf G
        -> match p with
           | pure =>
               exists c e, trp G M T c e
           | impure =>
               exists e, tri G M T e
           end));
try
  (intros;
   repeat
     (match goal with
      | H : exists _,_ |- _ => destruct H

      | H : wf _ -> _ |- _ =>
          lapply H; [clear H; intro | eauto using wf; done]
      end);
   repeat eexists;
   eauto using wf, trp, tri, trt, can_split, ofs_hsg;
   done).

(* unpack *)
intros G k t t' e1 e2 _ IH1 _ IH2 Hoft' Hwf.
so (IH1 Hwf) as (e1' & Htre1).
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htre1))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (IH2 (wf_tm (wf_cn Hwf Hofk) Hoft)) as (e2' & Htre2).
eexists.
eauto using trt; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (e1' & Htre1).
so (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htre1)) as Hoft1.
so (IH2 (wf_tm Hwf Hoft1)) as (e2' & Htre2).
eexists.
eauto using trt; done.

(* letm *)
intros G T k t M e _ IH1 Hfst _ IH2 Hoft Hwf.
so (IH1 Hwf) as (e1 & HtrM).
so (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf HtrM)) as HofT.
so (IH2 (wf_md Hwf HofT Hfst)) as (e2 & Htre).
eexists; eauto using trt; done.

(* lamgn *)
intros G T1 T2 k1 M HofT1 Hfst1 _ IH Hwf.
so (IH (wf_md Hwf HofT1 Hfst1)) as (e & HtrM).
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
eauto using trp; done.

(* appgn *)
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 Hwf.
so (IH1 Hwf) as (e1 & HtrM1).
so (IH2 Hwf) as (c2a & e2 & HtrM2).
so (trp_fstm _#5 HtrM2) as HfstM2a.
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
eauto using tri; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst1 _ IH Hwf.
so (IH (wf_md Hwf HofT1 Hfst1)) as (c & e & HtrM).
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (trp_fstm _#5 HtrM) as Hfstm.
so (ofm_fst_o _#5 (wf_md Hwf HofT1 Hfst1) (trp_ofm _#5 (wf_md Hwf HofT1 Hfst1) HtrM) Hfstm) as (k2 & _ & Hofc).
so (subord_md_cn _#5 Hofc) as Heq.
replsub (subst (dot nonsense sh1) c) with (shift1 (subst1 nonsense c)) in Heq.
rewrite -> Heq in HtrM; [].
eauto using trp; done.

(* appap *)
intros G p T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 Hwf.
so (IH1 Hwf) as IH1'; clear IH1.
so (IH2 Hwf) as (c2a & e2 & HtrM2); clear IH2.
so (fstm_fun _#4 HfstM2 (trp_fstm _#5 HtrM2)); subst c2a.
destruct p.
  (* pure *)
  destruct IH1' as (c1 & e1 & HtrM1).
  eauto using trp; done.

  (* impure *)
  destruct IH1' as (e1 & HtrM1).
  invert (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf HtrM1)); [].
  intros k1 _ _ HofT2.
  so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
  eauto using tri; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as IH1'; clear IH1.
so (IH2 Hwf) as IH2'; clear IH2.
destruct p.
  destruct_all IH1'; destruct_all IH2'; eauto using trp; done.

  destruct IH1' as (e1 & HtrM1).
  destruct IH2' as (e2 & HtrM2).
  so (can_split _ (ofs_hsg _ _ (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf HtrM1)))) as (k1 & t1 & Hsplit1).
  so (can_split _ (ofs_hsg _ _ (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf HtrM2)))) as (k2 & t2 & Hsplit2).
  eauto using tri; done.

(* pi2 *)
intros G T1 T2 M c _ IH HfstM Hwf.
destruct (IH Hwf) as (ca & e & HtrM).
so (fstm_fun _#4 HfstM (trp_fstm _#5 HtrM)); subst ca.
eauto using trp; done.

(* unpack *)
intros G k t T e M _ IH1 _ IH2 HofT Hwf.
so (IH1 Hwf) as (e1' & Htre1).
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htre1))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (IH2 (wf_tm (wf_cn Hwf Hofk) Hoft)) as (e2 & Htre2).
eauto using tri; done.

(* lett *)
intros G p t T e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (e1' & Htre).
so (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htre)) as Hoft.
so (IH2 (wf_tm Hwf Hoft)) as IH2'; clear IH2.
destruct p.
  destruct IH2' as (c2 & e2 & HtrM).
  so (ofm_fst_o _#5 (wf_tm Hwf Hoft) (trp_ofm _#5 (wf_tm Hwf Hoft) HtrM) (trp_fstm _#5 HtrM)) as (k2 & _ & Hofc).
  so (subord_tm_cn _#4 Hofc) as Heq.
  replsub (subst (dot nonsense sh1) c2) with (shift1 (subst1 nonsense c2)) in Heq.
  rewrite -> Heq in HtrM; [].
  eauto using trp; done.

  destruct IH2' as (e2 & HtrM).
  eauto using tri; done.

(* letm *)
intros G T1 T2 k1 M1 M2 _ IH1 Hfst1 _ IH2 HofT2 Hwf.
so (IH1 Hwf) as (e1 & HtrM1).
so (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf HtrM1)) as HofT1.
so (IH2 (wf_md Hwf HofT1 Hfst1)) as (e2 & HtrM2).
eauto using tri; done.

(* forget *)
intros G T M _ IH Hwf.
so (IH Hwf) as (c & e & HtrM).
so (can_split _ (ofs_hsg _ _ (ofm_reg _#4 Hwf (trp_ofm _#5 Hwf HtrM)))) as (k & t & Hsplit).
eauto using tri; done.

(* subsume *)
intros G p T T' M _ IH1 Hsubs Hwf.
so (IH1 Hwf) as IH1'.
so (subs_trsubs _#3 Hsubs) as (f & Htrsub).
destruct p.
  destruct_all IH1'; eauto using trp; done.

  destruct IH1'.
  so (can_split _ (ofs_hsg _ _ (subs_reg _#3 Hwf Hsubs ander))) as (k & t & Hsplit).
  eauto using tri; done.

(* epilogue *)
destruct H as (Ht & Hm).
do2 2 split.
  intro G; intros.
  eapply (Hm G pure); eauto using ofm_ofmp; done.

  intro G; intros.
  eapply (Hm G impure); eauto using ofm_ofmp; done.

  eauto using oft_oftp; done.
Qed.


Lemma ofm_trp :
  forall G M T,
    wf G
    -> ofm G pure M T
    -> exists c e, trp G M T c e.
Proof.
exact (can_trans andel).
Qed.


Lemma ofm_tri :
  forall G M T,
    wf G
    -> ofm G impure M T
    -> exists e, tri G M T e.
Proof.
exact (can_trans anderl).
Qed.


Lemma oft_trt :
  forall G e t, 
    wf G
    -> oft G e t
    -> exists e', trt G e t e'.
Proof.
exact (can_trans anderr).
Qed.       


Lemma ofm_iff_trp :
  forall G M T,
    wf G
    -> (ofm G pure M T <-> exists c e, trp G M T c e).
Proof.
intros G M T Hwf.
split.
  auto using ofm_trp; done.

  intros (c & e & Htr).
  eauto using trp_ofm; done.
Qed.


Lemma ofm_iff_tri :
  forall G M T,
    wf G
    -> (ofm G impure M T <-> exists e, tri G M T e).
Proof.
intros G M T Hwf.
split.
  auto using ofm_tri; done.

  intros (e & Htr).
  eauto using tri_ofm; done.
Qed.


Lemma oft_iff_trt :
  forall G e t,
    wf G
    -> (oft G e t <-> exists e', trt G e t e').
Proof.
intros G e t Hwf.
split.
  auto using oft_trt; done.

  intros (e' & Htr).
  eauto using trt_oft; done.
Qed.


Lemma trt_reg :
  forall G e t e',
    wf G
    -> trt G e t e'
    -> ofc G t kd_type.
Proof.
intros G e t e' Hwf Htr.
exact (oft_reg _#3 Hwf (trt_oft _#4 Hwf Htr)).
Qed.


Lemma trp_reg :
  forall G M T c e,
    wf G
    -> trp G M T c e
    -> ofs G T.
Proof.
intros G M T c e Hwf Htr.
exact (ofm_reg _#4 Hwf (trp_ofm _#5 Hwf Htr)).
Qed.


Lemma tri_reg :
  forall G M T e,
    wf G
    -> tri G M T e
    -> ofs G T.
Proof.
intros G M T e Hwf Htr.
exact (ofm_reg _#4 Hwf (tri_ofm _#4 Hwf Htr)).
Qed.


Lemma trp_fst :
  forall G M T c e k,
    wf G
    -> trp G M T c e
    -> fsts T k
    -> ofc G c k.
Proof.
intros G M T c e k Hwf Htr Hfst.
exact (ofm_fst _#6 Hwf (trp_ofm _#5 Hwf Htr) (trp_fstm _#5 Htr) Hfst).
Qed.


(* Static correctness *)

Lemma trctx_sofeq :
  forall G G',
    trctx G G'
    -> subof sof G' (under (length G) id) G.
Proof.
intros G G' HtrG.
induct HtrG.

(* nil *)
apply subof_id; done.

(* cn *)
intros G G' k Htrs IH.
simpl length.
rewrite -> under_succ; [].
rewrite <- (subst_under_id (length G) (cl_cn k)) at 1; [].
apply subof_bind_sof_cn; auto; done.

(* tm *)
intros G G' t _ IH.
simpl length.
rewrite -> under_succ; [].
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using sof_wclo; done.

  apply sof_tm; done.

(* md *)
intros G G' T k t _ IH Hsplit.
simpl length.
rewrite -> !under_succ; [].
simp_sub.
#2 apply subof_dot.
  replsub (dot (var 1) (compose (under (length G) id) (sh 2))) with (compose (dot (var 0) (compose (under (length G) id) sh1)) sh1).
  apply subof_wclo_weaken1; auto using sof_wclo; [].
  rewrite <- (subst_under_id (length G) k) at 1; [].
  apply subof_bind_sof_cn; auto; done.

  simp_sub.
  apply sof_md; done.
Qed.


Lemma trctx_ofk :
  forall G G' k,
    trctx G G'
    -> ofk G k
    -> ofk G' k.
Proof.
intros G G' k HtrG Hofk.
so (subst_ofk _#4 (trctx_sofeq _ _ HtrG) Hofk) as Hofk'.
rewrite -> subst_under_id in Hofk'; auto.
Qed.


Lemma trctx_ofc :
  forall G G' c k,
    trctx G G'
    -> ofc G c k
    -> ofc G' c k.
Proof.
intros G G' c k HtrG Hofc.
so (subst_ofc _#5 (trctx_sofeq _ _ HtrG) Hofc) as Hofc'.
rewrite -> !subst_under_id in Hofc'; auto.
Qed.


Lemma trctx_subk :
  forall G G' k k',
    trctx G G'
    -> subk G k k'
    -> subk G' k k'.
Proof.
intros G G' k k' HtrG H.
so (subst_subk _#5 (trctx_sofeq _ _ HtrG) H) as H'.
simp_sub in H'.
exact H'.
Qed.


Lemma trctx_eqc :
  forall G G' c c' k,
    trctx G G'
    -> eqc G c c' k
    -> eqc G' c c' k.
Proof.
intros G G' c c' k HtrG Hofc.
so (subst_eqc _#6 (trctx_sofeq _ _ HtrG) Hofc) as Heqc'.
rewrite -> !subst_under_id in Heqc'; auto.
Qed.


Lemma trctx_formation :
  forall G G',
    trctx G G'
    -> wf G
    -> wf G'.
Proof.
intros G G' Htr Hwf.
revert Hwf.
induct Htr.

(* nil *)
auto; done.

(* cn *)
intros G G' k HtrG IH Hwfk.
invert Hwfk; [].
intros Hwf Hofk.
apply wf_cn; auto; [].
eapply trctx_ofk; eauto; done.

(* tm *)
intros G G' t HtrG IH Hwft.
invert Hwft; [].
intros Hwf Hoft.
apply wf_tm; auto; [].
eapply trctx_ofc; eauto; done.

(* md *)
intros G G' T k t HtrG IH Hsplit HwfT.
invert HwfT; [].
intros Ta Hwf HofT Hfst Heq.
f_apply (fun z => subst1 nonsense z) in Heq as Heq'.
simp_sub in Heq'.
subst Ta.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
#2 prove_wf; auto.
  eapply trctx_ofk; eauto; done.
  
  eapply trctx_ofc; eauto; [].
  apply trctx_cn; auto; done.
Qed.


Lemma trctx_index_tm :
  forall G G' i t,
    trctx G G'
    -> index i G (cl_tm t)
    -> index i G' (cl_tm t).
Proof.
intros G G' i t Htr Hindex.
revert i t Hindex.
induct Htr.

(* nil *)
intros i t Hindex.
invert Hindex; done.

(* cn *)
intros G G' k _ IH i t Hindex.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j Hindex' <-.
  apply index_S; [].
  apply IH; auto; done.

(* tm *)
intros G G' u _ IH i t Hindex.
#2 invert Hindex.
  intros <- Heq.
  injection Heq.
  intros ->.
  apply index_0; done.

  intros j Hindex' <-.
  apply index_S; [].
  apply IH; auto; done.

(* md *)
intros G G' T k u _ IH _ i t Hindex.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j Hindex' <-.
  #2 invert Hindex'.
    intros _ Heq'.
    discriminate Heq'.

    intros l Hindex'' <-.
    apply index_S; apply index_S; [].
    apply IH; auto; done.
Qed.


Lemma trctx_index_md :
  forall G G' i T c,
    trctx G G'
    -> index i G (cl_md T c)
    -> exists k t,
         split T k t
         /\ index i G' (cl_tm (subst1 c t)).
Proof.
intros G G' i T c Htr Hindex.
revert i T c Hindex.
induct Htr.

(* nil *)
intros i T c Hindex.
invert Hindex; done.

(* cn *)
intros G G' l _ IH i T c Hindex.
#2 invert Hindex.
  intros <- H.
  discriminate H.

  intros j Hindex' <-.
  so (IH _#3 Hindex') as (k & t & Hsplit & Hoft).
  exists k, t.
  split; auto using index_S; done.

(* tm *)
intros G G' u _ IH i T c Hindex.
#2 invert Hindex.
  intros <- H.
  discriminate H.

  intros j Hindex' <-.
  so (IH _#3 Hindex') as (k & t & Hsplit & Hoft).
  exists k, t.
  split; auto using index_S; done.

(* md *)
intros G G' T k t _ IH Hsplit i T' c Hindex.
#2 invert Hindex.
  intros <- Heq.
  injection Heq.
  intros <- <-.
  exists (shift1 k), (subst (dot (var 0) (sh 2)) t).
  split.
    eapply subst_split; eauto; done.

    simp_sub.
    exact index_0.

  intros j Hindex' <-.
  #2 invert Hindex'.
    intros <- H.
    discriminate H.

    intros j' Hindex'' <-.
    so (IH _#3 Hindex'') as (k' & t' & Hsplit' & Hoft).
    exists k', t'.
    split; auto using index_S; done.
Qed.


Lemma trctx_trsubs :
  forall G G' T1 T2 f,
    trctx G G'
    -> trsubs G T1 T2 f
    -> trsubs G' T1 T2 f.
Proof.
intros G G' T1 T2 f HtrG Htr.
revert G' HtrG.
assert (forall G G' T, trctx G G' -> ofs G T -> ofs G' T) as Hofs.
  clear_all.
  intros G G' T HtrG H.
  so (subst_ofs _#4 (trctx_sofeq _ _ HtrG) H) as H'.
  simp_sub in H'.
  exact H'.
assert (forall G G' T T', trctx G G' -> eqs G T T' -> eqs G' T T') as Heqs.
  clear_all.
  intros G G' T T' HtrG H.
  so (subst_eqs _#5 (trctx_sofeq _ _ HtrG) H) as H'.
  simp_sub in H'.
  exact H'.
induct Htr;
eauto 7 using trsubs, trctx, trctx_subk.
Qed.


Lemma trt_svalue :
  forall G e t e',
    trt G e t e'
    -> svalue e
    -> svalue e'.
Proof.
intros G e t e' Htr Hsvalue.
revert Hsvalue.
induct Htr;
intros; 
repeat
  (match goal with
   | H : svalue _ |- _ => invert H; clear H
   end);
intros;
subst;
auto using svalue.
Qed.


Lemma trans_formation :
  (forall G G' M T c e k t,
     wf G
     -> trctx G G'
     -> trp G M T c e
     -> split T k t
     -> oft G' e (subst1 c t))
  /\
  (forall G G' M T e k t,
     wf G
     -> trctx G G'
     -> tri G M T e
     -> split T k t
     -> oft G' e (cn_exists k t))
  /\
  (forall G G' e t e',
     wf G
     -> trctx G G'
     -> trt G e t e'
     -> oft G' e' t).
Proof.
#31 exploit
  (trans_ind
     (fun G M T c e =>
        forall G' k t,
          wf G
          -> trctx G G'
          -> split T k t
          -> oft G' e (subst1 c t))
     (fun G M T e =>
        forall G' k t,
          wf G
          -> trctx G G'
          -> split T k t
          -> oft G' e (cn_exists k t))
     (fun G e t e' =>
        forall G',
          wf G
          -> trctx G G'
          -> oft G' e' t))
  as Hind;
eauto with static.

(* var *)
intros G i T c Hindex G' ks ts Hwf HtrG Hsplits.
so (wf_index_md _#4 Hwf Hindex) as (Hofs & ->).
so (ofm_var Hindex) as Hofm.
so (fstm_var Hindex) as Hfstm.
so (singles_ofm _#4 Hwf Hofm Hfstm) as Hofms.
so (trctx_index_md _#5 HtrG Hindex) as (k & t & Hsplit & Hindex').
so (oft_var Hindex') as Hoft.
(*
so (trctx_index_md _#5 HtrG Hindex) as (k & t & Hsplit & Hoft).
*)
so (ofm_fst _#6 Hwf Hofm Hfstm (split_fsts _#3 (subst_split _#4 Hsplit))) as Hofc.
so (split_singles _#7 Hwf Hofs Hofc (subst_split _#4 Hsplit) Hsplits) as Heqc.
so (ofm_fst _#6 Hwf Hofms Hfstm (split_fsts _#3 Hsplits)) as Hofcs.
simp_sub.
simp_sub in Hoft.
eapply oft_equiv; eauto; [].
replsub (subst (dot (var (1 + i)) (sh (1 + i))) t) with (subst1 (var (1 + i)) (subst (under 1 (sh (S i))) t)).
simp_sub in Hofcs.
eapply subst1_eqc_type; eauto using trctx_ofc; [].
eapply trctx_eqc; eauto; [].
apply trctx_cn; auto; done.

(* star *)
intros G G' k t Hwf HtrG Hsplit.
invert Hsplit; [].
intros <- <-.
simp_sub; auto with static; done.

(* stat *)
intros G c k Hofc G' ka t Hwf HtrG Hsplits.
invert Hsplits; [].
intros <- <-.
simp_sub.
eauto using single_ofc, trctx_ofc with static; done.

(* dyn *)
intros G e t e' _ IH G' k ta Hwf HtrG Hsplit.
invert Hsplit; [].
intros <- <-.
simp_sub.
eauto with static; done.

(* lamgn *)
intros G T1 T2 k1 t1 M e HofT1 Hsplit1 HtrM IH G' k t Hwf HtrG Hsplit.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1a t2 Hsplit1a Hsplit2 <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-).
clear Hsplit1a.
simp_sub.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
apply oft_plam; eauto using trctx_ofk, svalue_lam; [].
apply oft_lam; eauto using trctx_ofc, trctx; [].
simp_sub.
so (IH _#3 (wf_md Hwf HofT1 (split_fsts _#3 Hsplit1)) (trctx_md HtrG Hsplit1) (subst_split _#4 Hsplit2)) as Hofe.
rewrite <- (subst_dot01_sh t2) in Hofe.
simp_sub in Hofe.
exact Hofe.

(* lamap *)
intros G T1 T2 k1 t1 M c e HofT1 Hsplit1 HtrM IH G' k t Hwf HtrG Hsplit.
invert Hsplit; clear Hsplit; [].
intros k1a k2 t1a t2 Hsplit1a Hsplit2 <- <-.
so (split_fun _#5 Hsplit1 Hsplit1a) as (<- & <-); clear Hsplit1a.
so (split_fsts _#3 Hsplit1) as Hfst1.
so (split_fsts _#3 Hsplit2) as Hfst2.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (wf_md Hwf HofT1 Hfst1) as Hwf'.
so (IH _#3 Hwf' (trctx_md HtrG Hsplit1) (subst_split _#4 Hsplit2)) as Hofe.
so (ofm_fst _#6 Hwf' (trp_ofm _#5 Hwf' HtrM) (trp_fstm _#5 HtrM) (subst_fsts _#3 Hfst2)) as Hofcw.
#2 sor (refine (subst_ofc (G; cl_cn k1) _ (dot nonsense id) _ _ _ Hofcw)) as Hofc.
  prove_subof; done.
simp_sub in Hofc; clear Hofcw.
simp_sub.
apply oft_plam; eauto using trctx_ofk, svalue_lam; [].
apply oft_lam; eauto using trctx_ofc, trctx; [].
simp_sub.
simp_sub in Hofe.
eapply oft_equiv; eauto; [].
so (ofm_reg _#4 Hwf' (trp_ofm _#5 Hwf' HtrM)) as HofT2w.
#2 sor (refine (subst_ofs (G; cl_cn k1) _ (dot nonsense id) _ _ HofT2w)) as HofT2.
  prove_subof; done.
simp_sub in HofT2.
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
eapply trctx_eqc; eauto using trctx; [].
rewrite <- (subst_dot01_sh t2) at 1; [].
simp_sub.
eapply seq_ofc_type; eauto using wf; [].
#2 prove_subrel.
  apply eqc_refl; ofvar; done.

  apply eqc_symm; [].
  relquest.
    #2 eapply rel_quest_1.
      #2 eapply eqc_beta.
        eapply subst_ofc; eauto; prove_subof; done.

        ofvar; done.

      simp_sub.
      apply subst_dot1_sh; auto; done.

    simp_sub; reflexivity.

(* appap *)
intros G T1 T2 M1 M2 c1 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' k t Hwf HtrG Hsplit.
invert (trp_reg _#5 Hwf HtrM1); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_fun _#5 Hsplit (subst_split _#4 Hsplit2)) as (-> & ->); clear Hsplit.
so (IH1 _#3 Hwf HtrG (split_piapp Hsplit1 Hsplit2)) as Hofe1.
so (IH2 _#3 Hwf HtrG Hsplit1) as Hofe2.
so (trp_fst _#6 Hwf HtrM2 Hfst1) as Hofc2.
simp_sub in Hofe1.
simp_sub.
#2 eapply oft_app.
  relquest.
    eapply oft_papp; eauto using trctx_ofc; done.

    simp_sub; reflexivity.

  exact Hofe2.

(* pair *)
intros G T1 T2 M1 M2 c1 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' k t Hwf HtrG Hsplit.
so (trp_reg _#5 Hwf HtrM1) as HofT1.
so (trp_reg _#5 Hwf HtrM2) as HofT2.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
invert Hsplit; clear Hsplit; [].
intros k1 k2a t1 t2a Hsplit1 Hsplit2a <- <-.
so (split_fun _#5 (subst_split _#4 Hsplit2) Hsplit2a) as (<- & <-); clear Hsplit2a.
so (IH1 _#3 Hwf HtrG Hsplit1) as Hofe1.
so (IH2 _#3 Hwf HtrG Hsplit2) as Hofe2.
so (trp_fst _#6 Hwf HtrM1 (split_fsts _#3 Hsplit1)) as Hofc1.
so (trp_fst _#6 Hwf HtrM2 (split_fsts _#3 Hsplit2)) as Hofc2.
simp_sub.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
#2 apply oft_pair.    
  eapply oft_equiv; eauto; [].
  eapply trctx_eqc; eauto; [].
  eapply funct_cn_type; eauto; [].
  apply eqc_symm; [].
  eapply eqc_beta1; eauto; done.

  eapply oft_equiv; eauto; [].
  eapply trctx_eqc; eauto; [].
  eapply funct_cn_type; eauto; [].
  apply eqc_symm; [].
  eapply eqc_beta2; eauto; done.

(* pi1 *)
intros G T1 T2 M c e HtrM IH G' k1 t1 Hwf HtrG Hsplit1.
invert (trp_reg _#5 Hwf HtrM); [].
intros k1a HofT1 Hfst1 HofT2.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (IH _#3 Hwf HtrG (split_sigma Hsplit1 Hsplit2)) as Hofe.
simp_sub in Hofe.
eapply oft_pi1; [].
exact Hofe.

(* pi2 *)
intros G T1 T2 M c e HtrM IH G' k t Hwf HtrG Hsplit.
invert (trp_reg _#5 Hwf HtrM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_fun _#5 Hsplit (subst_split _#4 Hsplit2)) as (-> & ->); clear Hsplit.
so (IH _#3 Hwf HtrG (split_sigma Hsplit1 Hsplit2)) as Hofe.
simp_sub in Hofe.
simp_sub.
eapply oft_pi2; eauto; done.

(* lett *)
intros G t T e M e1 c2 e2 Htre IH1 HtrM IH2 G' k u Hwf HtrG Hsplit.
so (trt_reg _#4 Hwf Htre) as Hoft.
so (IH1 _ Hwf HtrG) as Hofe1.
so (IH2 _#3 (wf_tm Hwf Hoft) (trctx_tm HtrG) (subst_split _#4 Hsplit)) as Hofe2.
simp_sub in Hofe2.
replsub (subst (dot (shift1 c2) sh1) u) with (shift1 (subst1 c2 u)) in Hofe2.
eapply oft_lett; eauto; done.

(* subsume *)
intros G T T' M c e f HtrM IH Htrsub G' k' t' Hwf HtrG Hsplit'.
so (trp_reg _#5 Hwf HtrM) as HofT.
so (can_split _ (ofs_hsg _ _ HofT)) as (k & t & Hsplit).
so (IH _#3 Hwf HtrG Hsplit) as Hofe.
so (trsubs_formation _#8 (trctx_formation _ _ HtrG Hwf) (trctx_trsubs _#5 HtrG Htrsub) Hsplit Hsplit') as Hoff.
so (trp_fst _#6 Hwf HtrM (split_fsts _#3 Hsplit)) as Hofc.
eapply oft_app; eauto; [].
relquest.
  eapply oft_papp; eauto using trctx_ofc; done.

  simp_sub; reflexivity.

(* appgn *)
intros G T1 T2 M1 M2 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' k t Hwf HtrG Hsplit.
invert (tri_reg _#4 Hwf HtrM1); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_fun _#5 Hsplit (subst_split _#4 Hsplit2)) as (-> & ->); clear Hsplit.
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (IH1 _#3 Hwf HtrG (split_pigen Hsplit1 Hsplit2)) as Hofe1.
so (IH2 _#3 Hwf HtrG Hsplit1) as Hofe2.
so (trp_fst _#6 Hwf HtrM2 Hfst1) as Hofc2.
#3 eapply oft_unpack.
  exact Hofe1.

  #2 eapply oft_app.
    relquest.
      #2 eapply oft_papp.
        ofvar; done.

        apply weaken2_ofc; [].
        eapply trctx_ofc; eauto; done.

      simp_sub; reflexivity.

    simp_sub.
    relquest.
      apply weaken2_oft; eauto; done.

      simp_sub; reflexivity.

  simp_sub.
  eapply trctx_ofc; eauto; [].
  #2 apply ofc_exists.
    eapply subst1_ofk; eauto; done.

    eapply subst_ofc_type; eauto; prove_subof; [].
    apply weaken1_ofc; auto; done.

(* appap *)
intros G T1 T2 M1 M2 c2 e1 e2 k2 t2 HtrM1 IH1 HtrM2 IH2 Hsplit2 G' k t Hwf HtrG Hsplit.
invert (tri_reg _#4 Hwf HtrM1); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (split_fun _#5 Hsplit (subst_split _#4 Hsplit2)) as (-> & ->); clear Hsplit.
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (IH1 _#3 Hwf HtrG (split_piapp Hsplit1 Hsplit2)) as Hofe1.
so (IH2 _#3 Hwf HtrG Hsplit1) as Hofe2.
so (trp_fst _#6 Hwf HtrM2 (split_fsts _#3 Hsplit1)) as Hofc2.
#3 eapply oft_unpack.
  exact Hofe1.

  simp_sub.
  #3 eapply oft_pack.
    relquest.
      #2 eapply ofc_app.
        ofvar; done.

        apply weaken2_ofc; eauto using trctx_ofc; done.

      simp_sub; reflexivity.

    simp_sub.
    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          ofvar; done.

          apply weaken2_ofc; eauto using trctx_ofc; done.

        simp_sub; reflexivity.

      relquest.
        eapply weaken2_oft; eauto; done.

        simp_sub; reflexivity.

    eapply trctx_ofc; eauto using trctx; [].
    eapply subst_ofc_type; eauto; [].
    prove_subof; [].
    eapply subst_ofc; eauto; prove_subof; done.

  eapply trctx_ofc; eauto; [].
  #2 apply ofc_exists.
    eapply subst1_ofk; eauto; done.

    simp_sub.
    eapply subst_ofc_type; eauto; prove_subof; [].
    apply weaken1_ofc; auto; done.

(* pair *)
intros G T1 T2 M1 M2 e1 e2 k1 k2 t1 t2 HtrM1 IH1 HtrM2 IH2 Hsplit1 Hsplit2 G' k t Hwf HtrG Hsplit.
so (split_fun _#5 Hsplit (split_sigma Hsplit1 (subst_split _#4 Hsplit2))) as (-> & ->); clear Hsplit.
so (tri_reg _#4 Hwf HtrM1) as HofT1.
so (tri_reg _#4 Hwf HtrM2) as HofT2.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (IH1 _#3 Hwf HtrG Hsplit1) as Hofe1.
so (IH2 _#3 Hwf HtrG Hsplit2) as Hofe2.
assert (ofc G
          (cn_exists (kd_sigma k1 (shift1 k2))
             (cn_prod (subst (dot (cn_pi1 (var 0)) sh1) t1)
                (subst (dot (cn_pi2 (var 0)) sh1) t2))) kd_type) as Hofex.
  #2 apply ofc_exists.
    apply ofk_sigma; auto; [].
    apply weaken1_ofk; auto; done.

    #2 apply ofc_prod.
      eapply subst_ofc_type; eauto; prove_subof; [].
      eapply ofc_pi1; ofvar; done.

      eapply subst_ofc_type; eauto; prove_subof; [].
      relquest.
        eapply ofc_pi2; ofvar; done.

        simp_sub; reflexivity.
#3 eapply oft_unpack.
  exact Hofe1.

  #3 eapply oft_unpack.
    relquest.
      apply weaken2_oft; eauto; done.

      simp_sub; reflexivity.

    simp_sub.
    #3 apply oft_pack.
      eapply trctx_ofc; eauto using trctx; [].
      #3 eapply ofc_pair.
        ofvar; done.

        ofvar; done.

        eapply subst_ofk; eauto; prove_subof; done.

      simp_sub.
      #2 apply oft_pair.
        #2 eapply oft_equiv.
          ofvar; done.

          eapply trctx_eqc; eauto using trctx; [].
          rewrite <- (subst_dot0_sh t1) at 2; [].
          rewrite <- subst_compose; [].
          rewrite -> compose_dot; [].
          simp_sub.
          #2 eapply seq_ofc_type; eauto.
            #2 prove_wf; eauto.
              apply weaken2_ofk; auto; done.

              eapply subst_ofc_type; eauto; prove_subof; done.

            prove_subrel; [].
            apply eqc_symm; [].
            eapply eqc_beta1; ofvar; done.

        #2 eapply oft_equiv.
          ofvar; done.

          eapply trctx_eqc; eauto using trctx; [].
          #2 eapply seq_ofc_type; eauto.
            #2 prove_wf; eauto.
              apply weaken2_ofk; auto; done.

              eapply subst_ofc_type; eauto; prove_subof; done.

            prove_subrel; [].
            apply eqc_symm; [].
            eapply eqc_beta2; ofvar; done.

      eapply trctx_ofc; eauto using trctx; [].
      #2 apply ofc_prod.
        eapply subst_ofc_type; eauto; prove_subof; [].
        eapply ofc_pi1; [].
        ofvar; done.

        eapply subst_ofc_type; eauto; prove_subof; [].
        relquest.
          eapply ofc_pi2; [].
          ofvar; done.

          simp_sub; reflexivity.

    eapply trctx_ofc; eauto using trctx; [].
    apply weaken2_ofc_type; [].
    simp_sub.
    exact Hofex.

  eapply trctx_ofc; eauto using trctx; [].  
  simp_sub.
  exact Hofex.

(* unpack *)
intros G k1 t1 T e M e1 e2 Htre IH1 HtrM IH2 HofT G' k2 t2 Hwf HtrG Hsplit.
so (ofc_ofcr _#3 (trt_reg _#4 Hwf Htre)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk1 Hoft1 _; clear l Hofcr.
so (wf_tm (wf_cn Hwf Hofk1) Hoft1) as Hwf'.
so (split_formation _#4 HofT Hsplit) as (Hofk2 & Hoft2).
so (IH1 _ Hwf HtrG) as Hofe1.
so (IH2 _ _ _ Hwf' (trctx_tm (trctx_cn HtrG)) (subst_split _#4 Hsplit)) as Hofe2.
#3 eapply oft_unpack.
  exact Hofe1.

  exact Hofe2.

  eapply trctx_ofc; eauto; [].
  eapply ofc_exists; eauto; done.

(* lett *)
intros G u T e M e1 e2 Htre IH1 HtrM IH2 G' k t Hwf HtrG Hsplit.
so (trt_reg _#4 Hwf Htre) as Hofu.
so (IH1 _ Hwf HtrG) as Hofe1.
so (IH2 _#3 (wf_tm Hwf Hofu) (trctx_tm HtrG) (subst_split _#4 Hsplit)) as Hofe2.
eapply oft_lett; eauto; done.

(* letm *)
intros G T1 T2 k1 M1 M2 e1 e2 HtrM1 IH1 Hfst1 HtrM2 IH2 HofT2 G' k2 t2 Hwf HtrG Hsplit2.
so (tri_reg _#4 Hwf HtrM1) as HofT1.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (IH1 _#3 Hwf HtrG Hsplit1) as Hofe1.
so (IH2 _#3 (wf_md Hwf HofT1 Hfst1) (trctx_md HtrG Hsplit1) (subst_split _#4 Hsplit2)) as Hofe2.
eapply oft_unpack; eauto; [].
eapply trctx_ofc; eauto; [].
apply ofc_exists; auto; done.

(* forget *)
intros G T M c e k t HtrM IH Hsplit G' ka ta Hwf HtrG Hsplita.
so (split_fun _#5 Hsplit Hsplita) as (<- & <-); clear Hsplita.
so (split_formation _#4 (trp_reg _#5 Hwf HtrM) Hsplit) as (Hofk & Hoft).
so (IH _#3 Hwf HtrG Hsplit) as Hofe.
so (trp_fst _#6 Hwf HtrM (split_fsts _#3 Hsplit)) as Hofc.
#3 apply oft_pack.
  eapply trctx_ofc; eauto; done.

  assumption.

  eapply trctx_ofc; eauto using trctx; done.

(* subsume *)
intros G T T' M e f k' t' HtrM IH Htrsub Hsplit' G' ka' ta' Hwf HtrG Hsplita'.
so (split_fun _#5 Hsplit' Hsplita') as (<- & <-); clear Hsplita'.
so (subs_reg _#3 Hwf (trsubs_subs _#4 Htrsub)) as (HofT & HofT').
so (can_split _ (ofs_hsg _ _ HofT)) as (k & t & Hsplit).
so (split_formation _#4 HofT' Hsplit') as (Hofk' & Hoft').
so (subs_fst _#5 Hwf (trsubs_subs _#4 Htrsub) (split_fsts _#3 Hsplit) (split_fsts _#3 Hsplit')) as Hsubk.
so (IH _#3 Hwf HtrG Hsplit) as Hofe.
so (trsubs_formation _#8 (trctx_formation _ _ HtrG Hwf) (trctx_trsubs _#5 HtrG Htrsub) Hsplit Hsplit') as Hoff.
#3 eapply oft_unpack.
  exact Hofe.

  simp_sub.
  #3 apply oft_pack.
    eapply trctx_ofc; eauto using trctx; [].
    #2 eapply ofc_subsume.
      ofvar; done.

      apply weaken2_subk; auto; done.

    simp_sub.
    rewrite -> subst_dot1_sh; [].
    #2 eapply oft_app.
      relquest.
        #2 eapply oft_papp.
          relquest.
            apply weaken2_oft; eauto; done.

            simp_sub; reflexivity.

          ofvar; done.

        simp_sub.
        rewrite -> !subst_dot1_sh; [].
        reflexivity.

      ofvar; done.

    eapply trctx_ofc; eauto using trctx; [].
    eapply subst_ofc_type; eauto; prove_subof; done.

  eapply trctx_ofc; eauto using trctx; [].
  apply ofc_exists; auto; done.

(* var *)
intros G i t Hindex G' Hwf HtrG.
so (trctx_index_tm _#4 HtrG Hindex) as Hindex'.
apply oft_var; auto; done.

(* lam *)
intros.
apply oft_lam; eauto using trctx_ofc, trctx, wf; done.

(* plam *)
intros.
apply oft_plam; eauto using trctx_ofk, trctx, wf, trt_svalue; done.

(* papp *)
intros.
eapply oft_papp; eauto using trctx_ofc, trctx, wf; done.

(* pack *)
intros.
apply oft_pack; eauto using trctx_ofc, trctx, wf; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' Htre1 IH1 _ IH2 Hoft' G' Hwf HtrG.
so (ofc_ofcr _#3 (trt_reg _#4 Hwf Htre1)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
eapply oft_unpack; eauto 6 using trctx_formation, trctx_ofc, trctx, wf; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' Htre1 IH1 _ IH2 G' Hwf HtrG.
so (trt_reg _#4 Hwf Htre1) as Hoft1.
eapply oft_lett; eauto using trctx, wf; done.

(* equiv *)
eauto using trctx_eqc with static; done.

(* letm *)
intros G T k u M e e1 e2 HtrM IH1 Hfst Htre IH2 Hofu G' Hwf HtrG.
so (tri_reg _#4 Hwf HtrM) as HofT.
so (can_split _ (ofs_hsg _ _ HofT)) as (ka & t & Hsplit).
so (split_fsts_eq _#4 Hsplit Hfst); subst ka.
so (IH1 _#3 Hwf HtrG Hsplit) as Hofe1.
so (IH2 _ (wf_md Hwf HofT Hfst) (trctx_md HtrG Hsplit)) as Hofe2.
eauto using trctx_ofc with static; done.

(* ext *)
intros G t M e HtrM IH G' Hwf HtrG.
invert (tri_reg _#4 Hwf HtrM); [].
intros Hoft.
so (IH _#3 Hwf HtrG split_dyn) as Hofe.
eapply oft_unpack; eauto using trctx_ofc; [].
ofvar; done.

(* epilogue *)
destruct Hind as (Hp & Hi & Ht).
do2 2 split; eauto; done.
Qed.


Lemma trp_formation :
  forall G G' M T c e k t,
    wf G
    -> trctx G G'
    -> trp G M T c e
    -> split T k t
    -> oft G' e (subst1 c t).
Proof.
exact (trans_formation andel).
Qed.


Lemma tri_formation :
  forall G G' M T e k t,
    wf G
    -> trctx G G'
    -> tri G M T e
    -> split T k t
    -> oft G' e (cn_exists k t).
Proof.
exact (trans_formation anderl).
Qed.


Lemma trt_formation :
  forall G G' e t e',
    wf G
    -> trctx G G'
    -> trt G e t e'
    -> oft G' e' t.
Proof.
exact (trans_formation anderr).
Qed.


(* Substitutions mediating source and target contexts *)

Inductive trs : context -> context -> sub -> Prop :=
| trs_nil
    : trs nil nil id

| trs_cn {G G' k s}
    : trs G G' s
      -> trs (G; cl_cn k) (G'; cl_cn k) (dot (var 0) (compose s sh1))

| trs_tm {G G' t s}
    : trs G G' s
      -> trs (G; cl_tm t) (G'; cl_tm t) (dot (var 0) (compose s sh1))

| trs_md {G G' T k t s e}
    : trs G G' s
      -> split T k t
      -> snd (subst (sh 2) T) (var 0) e
      -> trs (G; cl_cn k; cl_md (shift1 T) (var 0)) (G'; cl_cn k; cl_tm t)
         (dot e (dot (var 1) (compose s (sh 2))))
.


Lemma trs_sid :
  forall G G' s,
    trs G G' s
    -> subrel sid G' (under (length G) id) s G.
Proof.
intros G G' s Htrs.
induct Htrs.

(* nil *)
simpl.
apply subrel_id; done.

(* cn *)
intros G G' k s _ IH.
simpl length.
rewrite -> under_succ; [].
rewrite <- (subst_under_id (length G) (cl_cn k)) at 1; [].
eapply subrel_wclo_bind; auto using sid_behaved, sid_wclo, iscl; done.

(* tm *)
intros G G' t s _ IH.
simpl length.
rewrite -> under_succ; [].
rewrite <- (subst_under_id (length G) (cl_tm t)) at 1; [].
eapply subrel_wclo_bind; auto using sid_behaved, sid_wclo, iscl; done.

(* md *)
intros G G' T k t s M _ IH Hsplit Hjoin.
simpl length.
rewrite -> !under_succ; [].
#2 apply subrel_dot.
  replsub (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1).
  apply subrel_wclo_weaken1; auto using sid_wclo; [].
  rewrite <- (subst_under_id (length G) (cl_cn k)) at 1; [].
  eapply subrel_wclo_bind; auto using sid_behaved, sid_wclo, iscl; done.

  simp_sub.
  apply sid_md; done.
Qed.


Lemma trs_subst_ofk :
  forall G G' s k,
    trs G G' s
    -> ofk G k
    -> subst s k = k.
Proof.
intros G G' s k Htrs Hofk.
symmetry.
etransitivity.
  symmetry.
  apply (subst_under_id (length G)); done.

  eapply sid_ofk; eauto; [].
  apply trs_sid; eauto; done.
Qed.


Lemma trs_subst_ofc :
  forall G G' s c k,
    trs G G' s
    -> ofc G c k
    -> subst s c = c.
Proof.
intros G G' s c k Htrs Hofc.
symmetry.
etransitivity.
  symmetry.
  apply (subst_under_id (length G)); done.

  eapply sid_ofc; eauto; [].
  apply trs_sid; eauto; done.
Qed.


Lemma trs_subst_dot_ofc :
  forall G G' s c k k1 b,
    trs G G' s
    -> ofc (G; cl_cn k1) c k
    -> subst (dot b s) c = subst1 b c.
Proof.
intros G G' s c k k1 b Htrs Hofc.
rewrite <- (trs_subst_ofc _#5 (trs_cn Htrs) Hofc) at 2; [].
simp_sub.
reflexivity.
Qed.


Lemma trs_subst_dot2_ofc :
  forall G G' s c k k1 k2 b1 b2,
    trs G G' s
    -> ofc (G; cl_cn k1; cl_cn k2) c k
    -> subst (dot b2 (dot b1 s)) c = subst (dot b2 (dot b1 id)) c.
Proof.
intros G G' s c k k1 k2 b1 b2 Htrs Hofc.
rewrite <- (trs_subst_ofc _#5 (trs_cn (trs_cn Htrs)) Hofc) at 2; [].
simp_sub.
reflexivity.
Qed.


Lemma trs_formation :
  forall G G' s,
    wf G
    -> trs G G' s
    -> subof dof G s G'.
Proof.
intros G G' s Hwf Htrs.
revert Hwf.
induct Htrs.

(* nil *)
intros _.
apply subof_id; done.

(* cn *)
intros G G' k s Htrs IH Hwfk.
invert Hwfk; [].
intros Hwf Hofk.
rewrite <- (trs_subst_ofk _#4 Htrs Hofk) at 1; [].
apply subof_bind_dof_cn; auto; done.

(* tm *)
intros G G' t s Htrs IH Hwft.
invert Hwft; [].
intros Hwf Hoft.
rewrite <- (trs_subst_ofc _#5 Htrs Hoft) at 1; [].
apply subof_bind_dof_tm; auto; done.

(* md *)
intros G G' T k t s M Htrs IH Hsplit Hsnd HwfT.
invert HwfT; [].
intros Ta Hwf HofT Hfst Heq.
f_apply (fun z => subst1 nonsense z) in Heq as Heq'.
simp_sub in Heq'.
subst Ta.
so (ofs_fst _#3 HofT Hfst) as Hofk.
#2 apply subof_dot.
  replsub (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1).
  apply subof_wclo_weaken1; auto using dof_wclo; [].
  rewrite <- (trs_subst_ofk _#4 Htrs Hofk) at 1; [].
  apply subof_bind_dof_cn; auto; done.

  simp_sub.
  apply dof_tm; [].
  #3 sor (refine (snd_formation _#7 HwfT _ Hsnd (subst_split _#4 Hsplit) _)) as HofM.
    ofvar; done.

    ofvar; done.
  simp_sub in HofM.
  simp_sub in HofM.
  rewrite -> subst_dot1_sh in HofM; [].
  force_exact HofM; [].
  f_equal; [].
  replsub (subst (dot (var 1) (compose s (sh 2))) t) with (shift1 (subst (dot (var 0) (compose s sh1)) t)).
  f_equal; [].
  rewrite <- (subst_under_id (S (length G)) t) at 1; [].
  refine (sid_ofc _#6 _ (split_formation _#4 HofT Hsplit ander)); [].
  rewrite -> under_succ; [].
  apply subrel_sid_bind_cn; [].
  apply trs_sid; eauto; done.
Qed.


Lemma trs_trctx :
  forall G s G',
    trs G G' s
    -> trctx G G'.
Proof.
intros G s G' H.
induct H; eauto using trctx.
Qed.


Lemma trs_trp :
  forall G G' s M T c e k t,
    wf G
    -> trs G G' s
    -> trp G M T c e
    -> split T k t
    -> oft G (subst s e) (subst1 c t).
Proof.
intros G G' s M T c e k t Hwf Htrs Htr Hsplit.
so (trp_reg _#5 Hwf Htr) as HofT.
so (split_formation _#4 HofT Hsplit) as (_ & Hoft).
so (trp_formation _#8 Hwf (trs_trctx _#3 Htrs) Htr Hsplit) as Hofe.
so (subst_oft _#5 (trs_formation _#3 Hwf Htrs) Hofe) as Hofe'.
so (trp_fst _#6 Hwf Htr (split_fsts _#3 Hsplit)) as Hofc.
so (subst1_ofc_type _#4 Hofc Hoft) as Hoftx.
rewrite -> (trs_subst_ofc _#5 Htrs Hoftx) in Hofe'; [].
exact Hofe'.
Qed.


Lemma trs_tri :
  forall G G' s M T e k t,
    wf G
    -> trs G G' s
    -> tri G M T e
    -> split T k t
    -> oft G (subst s e) (cn_exists k t).
Proof.
intros G G' s M T e k t Hwf Htrs Htr Hsplit.
so (tri_reg _#4 Hwf Htr) as HofT.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (tri_formation _#7 Hwf (trs_trctx _#3 Htrs) Htr Hsplit) as Hofe.
so (subst_oft _#5 (trs_formation _#3 Hwf Htrs) Hofe) as Hofe'.
#2 erewrite -> (trs_subst_ofc _#3 (cn_exists k t) _ Htrs) in Hofe'.
  exact Hofe'.

  apply ofc_exists; auto; done.
Qed.


Lemma trs_trt :
  forall G G' s e t e',
    wf G
    -> trs G G' s
    -> trt G e t e'
    -> oft G (subst s e') t.
Proof.
intros G G' s e t e' Hwf Htrs Htr.
so (trt_formation _#5 Hwf (trs_trctx _#3 Htrs) Htr) as Hofe'.
so (subst_oft _#5 (trs_formation _#3 Hwf Htrs) Hofe') as Hofe''.
rewrite -> (trs_subst_ofc _#5 Htrs (trt_reg _#4 Hwf Htr)) in Hofe''; [].
exact Hofe''.
Qed.


(* Dynamic correctness *)

Lemma trs_index_tm :
  forall G G' s i t,
    trs G G' s
    -> index i G (cl_tm t)
    -> project s i = var i.
Proof.
intros G G' s i t Htrs Hindex.
revert i Hindex.
induct Htrs.

(* nil *)
intros i Hindex.
invert Hindex; done.

(* cn *)
intros G G' k s _ IH i Hindex.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j Hindex' <-.
  simp_sub.
  rewrite -> project_compose; [].
  rewrite -> (IH _ Hindex'); [].
  simp_sub; reflexivity.

(* tm *)
intros G G' u s _ IH i Hindex.  
#2 invert Hindex.
  intros <- Heq.
  simp_sub; reflexivity.

  intros j Hindex' <-.
  simp_sub.
  rewrite -> project_compose; [].
  rewrite -> (IH _ Hindex'); [].
  simp_sub; reflexivity.

(* md *)
intros G G' T k u s e _ IH _ _ i Hindex.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j Hindex' <-.
  #2 invert Hindex'.
    intros _ Heq.
    discriminate Heq.

    intros j' Hindex'' <-.
    simp_sub.
    rewrite -> project_compose; [].
    rewrite -> (IH _ Hindex''); [].
    simp_sub; reflexivity.
Qed.


Lemma trs_index_md :
  forall G G' s i T c,
    trs G G' s
    -> index i G (cl_md T c)
    -> exists e,
         snd (subst sh1 T) (var 0) e
         /\ project s i = subst (sh i) e.
Proof.
intros G G' s i T c Htrs Hindex.
revert i T c Hindex.
induct Htrs.

(* nil *)
intros i T c Hindex.
invert Hindex; done.

(* cn *)
intros G G' l s _ IH i T c Hindex.
#2 invert Hindex.
  intros <- H.
  discriminate H.

  intros j Hindex' <-.
  so (IH _#3 Hindex') as (e & Hsnd & Heq).
  exists e.
  split; auto; [].
  simp_sub.
  rewrite -> project_compose; [].
  f_apply shift1 in Heq as Heq'.
  simp_sub in Heq'.
  exact Heq'.

(* tm *)
intros G G' u s _ IH i T c Hindex.
#2 invert Hindex.
  intros <- H.
  discriminate H.

  intros j Hindex' <-.
  so (IH _#3 Hindex') as (e & Hsnd & Heq).
  exists e.
  split; auto; [].
  simp_sub.
  rewrite -> project_compose; [].
  f_apply shift1 in Heq as Heq'.
  simp_sub in Heq'.
  exact Heq'.

(* md *)
intros G G' T k t s e _ IH Hsplit Hsnd i T' c Hindex.
#2 invert Hindex.
  intros <- Heq.
  injection Heq.
  intros <- <-.
  exists e.
  split; simp_sub; auto; done.

  intros j Hindex' <-.
  #2 invert Hindex'.
    intros <- H.
    discriminate H.

    intros j' Hindex'' <-.
    so (IH _#3 Hindex'') as (e' & Hsnd' & Heq).
    exists e'.
    split; auto; [].
    simp_sub.
    rewrite -> project_compose; [].
    f_apply (fun m => subst (sh 2) m) in Heq as Heq'.
    simp_sub in Heq'.
    exact Heq'.
Qed.


Lemma trp_join_formation :
  forall G G' s M T c e M',
    wf G
    -> trs G G' s
    -> trp G M T c e
    -> join T c (subst s e) M'
    -> ofm G pure M' T.
Proof.
intros G G' s M T c e M' Hwf Htrs Htr Hjoin.
so (trp_reg _#5 Hwf Htr) as HofT.
so (can_split _ (ofs_hsg _ _ HofT)) as (k & t & Hsplit).
so (trp_fst _#6 Hwf Htr (split_fsts _#3 Hsplit)) as Hofc.
so (trs_trp _#9 Hwf Htrs Htr Hsplit) as Hofe.
eapply join_formation; eauto; done.
Qed.


Lemma tri_join_formation :
  forall G G' s M T e M',
    wf G
    -> trs G G' s
    -> tri G M T e
    -> join (subst (sh 2) T) (var 1) (var 0) M'
    -> ofm G impure (md_unpack (subst s e) M' T) T.
Proof.
intros G G' s M T e M' Hwf Htrs Htr Hjoin.
so (tri_reg _#4 Hwf Htr) as HofT.
so (can_split _ (ofs_hsg _ _ HofT)) as (k & t & Hsplit).
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (trs_tri _#8 Hwf Htrs Htr Hsplit) as Hof.
eapply ofm_unpack; eauto; [].
apply ofm_forget; [].
#3 eapply join_formation; eauto using wf, subst_split.
  apply weaken2_ofs; auto; done.

  ofvar; done.

  simp_sub.
  simp_sub.
  rewrite -> subst_dot1_sh; [].
  ofvar; done.
Qed.


Theorem trans_correct :
  (forall G G' s M T c e M',
     trs G G' s
     -> trp G M T c e
     -> join T c (subst s e) M'
     -> ceqm G pure M M' T)
  /\
  (forall G G' s M T e M',
     trs G G' s
     -> tri G M T e
     -> join (subst (sh 2) T) (var 1) (var 0) M'
     -> ceqm G impure M (md_unpack (subst s e) M' T) T)
  /\
  (forall G G' s e t e',
     trs G G' s
     -> trt G e t e'
     -> ceqt G e (subst s e') t).
Proof.
exploit
  (trans_ind
     (fun G M T c e =>
        forall G' s M',
          trs G G' s
          -> join T c (subst s e) M'
          -> ceqm G pure M M' T)
     (fun G M T e =>
        forall G' s M',
          trs G G' s
          -> join (subst (sh 2) T) (var 1) (var 0) M'
          -> ceqm G impure M (md_unpack (subst s e) M' T) T)
     (fun G e t e' =>
        forall G' s,
          trs G G' s
          -> ceqt G e (subst s e') t)) as Hind.

(* var *)
intros G i T c Hindex G' s Ms Htrs Hjoins.
apply ceqm_with_wf; intro Hwf; [].
so (ofm_var Hindex) as Hofm.
so (fstm_var Hindex) as Hfstm.
simp_sub in Hfstm.
so (wf_index_md _#4 Hwf Hindex) as (HofT & ->).
so (trs_index_md _#6 Htrs Hindex) as (e & Hsndg & Heq).
so (can_join _ (var (S i)) (subst (sh i) e) (ofs_hsg _ _ HofT)) as (M & Hjoin).
so (subst_snd _#3 (sh i) Hsndg) as Hsnd; simp_sub in Hsnd.
so (join_snd_inverse _#6 Hofm Hfstm Hsnd Hjoin) as Hceqm.
apply ceqm_self_respecting; auto; [].
eapply ceqm_trans; eauto; [].
simp_sub in Hjoins.
rewrite -> Heq in Hjoins; [].
simp_sub in Hjoin.
so (can_split _ (ofs_hsg _ _ HofT)) as (k & t & Hsplit).
#2 eapply join_self; eauto.
  so (ofm_fst _#6 Hwf Hofm Hfstm (split_fsts _#3 Hsplit)) as H.
  simp_sub in H.
  exact H.

  eapply snd_formation; eauto; done.
  
(* star *)
intros G G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
apply ceqm_symm; [].
#2 apply reqm_ceqm.
  eapply trp_join_formation; eauto using trp_star; done.
invert Hjoin; clear Hjoin; [].
intros <-.
apply reqm_lett_beta; auto; [].
simp_sub.
apply tvalue_thalt; exact tvalue_star.

(* stat *)
intros G c k H G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
apply ceqm_symm; [].
#2 apply reqm_ceqm.
  eapply trp_join_formation; eauto using trp_stat; done.
invert Hjoin; clear Hjoin; [].
intros <-.
#2 eapply reqm_trans.
  apply reqm_lett_beta; [].
  simp_sub.
  apply tvalue_thalt; exact tvalue_star.
simp_sub.
apply reqm_refl; done.

(* dyn *)
intros G e t e' Htre IH G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert Hjoin; clear Hjoin; [].
intros <-.
apply ceqm_comp; apply mcc_dyn; [].
eapply IH; eauto; done.

(* lamgn *)
intros G T1 T2 k1 t1 M e HofT1 Hsplit1 HtrM IH G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert (trp_reg _#5 Hwf (trp_lamgn HofT1 Hsplit1 HtrM)); [].
intros k1a _ Hfst1 HofT2.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f & Hsnd1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N & Hjoin2).
#3 eexploit IH as Hceqm.
  so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
  exact (trs_md Htrs Hsplit1 Hsnd1x).

  simp_sub.
  so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 3))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
  exact Hjoin2x.
clear IH.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_lamgn; eauto using eqs_refl; done.
clear Hceqm.
apply ceqm_symm; [].
#2 apply reqm_ceqm.
  eapply trp_join_formation; eauto using trp_lamgn; done.
invert Hjoin; clear Hjoin; [].
intros fa Na Hsnd1a Hjoin2a <-.
so (subst_snd _#3 (dot (var 0) (sh 3)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
so (snd_fun _#4 Hsnd1a Hsnd1x); subst fa; clear Hsnd1a Hsnd1x.
so (subst_join _#4 (dot (var 0) (dot (var 1) (dot (var 3) (sh 5)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
so (join_fun _#5 Hjoin2a Hjoin2x); subst Na; clear Hjoin2a Hjoin2x.
simp_sub.
#2 eapply reqm_trans.
  apply reqm_lett_beta; [].
  apply tvalue_thalt; exact tvalue_plam.
simp_sub.
eapply reqm_lamgn; eauto; [].
rewrite -> subst_dot1_sh; [].
rewrite -> (subst_eqsub _ _ N (eqsub_dot _#3 (eqsub_dot _#3 (eqsub_expand_sh 3)))); [].
apply reqm_unpack; [].
#2 eapply reqt_trans.
  apply reqt_app; [| apply reqt_refl; done].
  apply reqt_papp_beta; done.
simp_sub.
#2 eapply reqt_trans.
  apply reqt_app_beta_instances; [].
  intros s' Hs'.
  #4 apply (snd_halt_val_instances (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (subst (sh 2) T1) (var 0)).
    ofvar; done.

    ofvar; done.

    so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
    exact Hsnd1x.

    exact Hs'.
simp_sub.
apply reqt_refl; done.

(* lamap *)
intros G T1 T2 k1 t1 M c e HofT1 Hsplit1 HtrM IH G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert (trp_reg _#5 Hwf (trp_lamap HofT1 Hsplit1 HtrM)); [].
intros k1a _ Hfst1 HofT2.
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f & Hsnd1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N & Hjoin2).
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (trp_fst _#6 (wf_md Hwf HofT1 Hfst1) HtrM (subst_fsts _#3 Hfst2)) as Hofcw.
#2 sor (refine (subst_ofc (G; cl_cn k1) _ (dot nonsense id) _ _ _ Hofcw)) as Hofc.
  prove_subof; done.
simp_sub in Hofc; clear Hofcw.
#3 eexploit IH as Hceqm.
  so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
  exact (trs_md Htrs Hsplit1 Hsnd1x).

  so (subst_join _#4 (dot (subst
        (dot (subst (dot (var 0) (sh 2)) f) (dot (var 1) (compose s (sh 2))))
        e) (dot (shift1 c) sh1)) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
  exact Hjoin2x.
clear IH.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_lamap; eauto using eqs_refl; done.
clear Hceqm.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  #2 apply reqm_ceqm.
    eapply trp_join_formation; eauto using trp_lamap; done.
  invert Hjoin; clear Hjoin; [].
  intros fa Na Hsnd1a Hjoin2a <-.
  so (subst_snd _#3 (dot (var 0) (sh 3)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
  so (snd_fun _#4 Hsnd1a Hsnd1x); subst fa; clear Hsnd1a Hsnd1x.
  simp_sub in Hjoin2a.
  so (subst_join _#4
        (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (var 0) (sh 3)) f))
           (dot (cn_app (cn_lam (subst (sh 3) k1) (subst (dot (var 0) (sh 4)) c)) (var 1))
              (dot (var 1) (sh 3)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
  so (join_fun _#5 Hjoin2a Hjoin2x); subst Na; clear Hjoin2a Hjoin2x.
  simp_sub.
  #2 eapply reqm_trans.
    apply reqm_lett_beta; [].
    apply tvalue_thalt; exact tvalue_plam.
  simp_sub.
  apply reqm_refl; done.
clear Hjoin.
apply ceqm_comp; eapply mcc_lamap; eauto using eqs_refl; [].
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2a & t2 & Hsplit2).
so (split_fsts_eq _#4 Hsplit2 Hfst2); subst k2a.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (join_vars_formation _#5 (wf_cn Hwf Hofk1) HofT2 Hsplit2 Hjoin2) as HofN.
rewrite -> (subst_eqsub _ _ N (eqsub_dot _#3 (eqsub_dot _#3 (eqsub_expand_sh 1)))); [].
relquest.
  refine (funct_ceq_md _#7 _ _ _ HofN); eauto using wf; [].
  #2 apply subrel_deq_ceq_tm.
    #2 apply subrel_deq_cn; eauto using wf.
      #2 apply subrel_deq_cn; eauto using wf.
        prove_subrel; done.

        apply eqc_refl; [].
        ofvar; done.
        
      relquest.
        #2 eapply rel_quest_1.
          #2 eapply eqc_beta; eauto.
            eapply subst_ofc; eauto; prove_subof; done.

            ofvar; done.

          simp_sub.
          rewrite -> subst_dot1_sh; [].
          reflexivity.

        simp_sub; reflexivity.

    #2 apply reqt_ceqt.
      so (trs_formation _#3 Hwf Htrs) as Hs.
      repl (subst (compose s (sh 2)) k1) with (subst (sh 2) k1).
        rewrite <- (trs_subst_ofk _#4 Htrs Hofk1) at 1; [].
        simp_sub; reflexivity.
      repl (subst (dot (var 0) (compose s (sh 3))) t1) with (subst (dot (var 0) (sh 3)) t1).
        rewrite <- (trs_subst_ofc _#5 (trs_cn Htrs) Hoft1) at 1; [].
        simp_sub; reflexivity.
      #2 eapply oft_app.
        #2 eapply oft_equiv.
          #2 eapply oft_papp.
            #2 eapply oft_plam; eauto using svalue_lam.
              eapply subst_ofk; eauto; prove_subof; done.

              #2 eapply oft_lam.
                eapply subst_ofc_type; eauto; prove_subof; done.

                relquest.
                  so (trp_formation _#8 (wf_md Hwf HofT1 Hfst1) (trctx_md (trs_trctx _#3 Htrs) Hsplit1) HtrM (subst_split _#4 Hsplit2)) as Hofe.
                  simp_sub in Hofe.
                  relquest.
                    #3 eapply subst_oft; eauto; prove_subof.
                      eapply subof_dof_trans; eauto; prove_subof; done.
  
                      rewrite -> subst_compose; [].
                      rewrite -> (trs_subst_ofk _#4 Htrs Hofk1); [].
                      ofvar; done.
  
                      relquest.
                        ofvar; done.
  
                        rewrite <- (trs_subst_ofc _#5 (trs_cn Htrs) Hoft1) at 1; [].
                        simp_sub; reflexivity.

                    simp_sub.
                    replsub (subst (dot (subst (dot (var 1) (compose s (sh 4))) c) (dot (var 1) (compose s (sh 4)))) t2) with (shift1 (subst (dot (subst (dot (var 0) (compose s (sh 3))) c) (dot (var 0) (compose s (sh 3)))) t2)).
                    reflexivity.

                  repl (subst (dot (var 0) (compose s (sh 3))) c) with (subst (dot (var 0) (sh 3)) c).
                    rewrite <- (trs_subst_ofc _#5 (trs_cn Htrs) Hofc) at 1; [].
                    simp_sub; reflexivity.
                  repl (subst (dot (subst (dot (var 0) (sh 3)) c) (dot (var 0) (compose s (sh 3)))) t2) with (subst (dot (subst (dot (var 0) (sh 3)) c) (dot (var 0) (sh 3))) t2).
                    rewrite <- (trs_subst_ofc _#5 (trs_cn (trs_cn Htrs)) Hoft2) at 1; [].
                    simp_sub; reflexivity.
                  reflexivity.
                    
            ofvar; done.

          simp_sub.
          rewrite -> !subst_dot1_sh; [].          
          #2 eapply eqc_arrow.
            apply eqc_refl; [].
            apply weaken1_ofc_type; auto; done.

            eapply seq_ofc_type; eauto using wf; [].
            #2 prove_subrel.
              apply eqc_refl; [].
              ofvar; done.

              apply eqc_symm; [].
              relquest.
                #2 eapply rel_quest_1.
                  #2 eapply eqc_beta.
                    eapply subst_ofc; eauto; prove_subof; done.


                    ofvar; done.

                  simp_sub.
                  rewrite -> subst_dot1_sh; reflexivity.

                simp_sub; reflexivity.

        eapply snd_var_formation; eauto; done.
    #2 eapply reqt_trans.
      apply reqt_app; [| apply reqt_refl; done].
      apply reqt_papp_beta; done.
    simp_sub.
    #2 eapply reqt_trans.
      apply reqt_app_beta_instances; [].
      intros s' Hs'.
      #4 apply (snd_halt_val_instances (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (subst (sh 2) T1) (var 0)).
        ofvar; done.
    
        ofvar; done.
    
        so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as Hsnd1x; simp_sub in Hsnd1x.
        exact Hsnd1x.
    
        exact Hs'.
    simp_sub.
    apply reqt_refl; done.

  simp_sub.
  rewrite -> subst_dot1_sh; [].
  reflexivity.

(* appap *)
intros G T1 T2 M1 M2 c1 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert (trp_reg _#5 Hwf HtrM1); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f & Hsnd1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
#3 eexploit (@join_piapp T1 T2 c1 (subst s e1)) as Hjoinpi.
  so (subst_snd _#3 (dot (var 0) (sh 3)) Hsnd1) as H; simp_sub in H.
  exact H.

  simp_sub.
  so (subst_join _#4 (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (var 0) (sh 3)) f)) (dot (cn_app (subst (sh 3) c1) (var 1)) (dot (var 1) (sh 3)))) Hjoin2) as H; simp_sub in H.
  exact H.
so (IH1 _#3 Htrs Hjoinpi) as Hceq1.
clear Hjoinpi.
so (subst_join _#4 (dot (subst s e2) (dot c2 id)) Hjoin1) as Hjoin1x; simp_sub in Hjoin1x.
so (IH2 _#3 Htrs Hjoin1x) as Hceq2.
clear Hjoin1x IH1 IH2.
simp_sub in Hjoin.
so (subst_join _#4 (dot (tm_app (tm_papp (subst s e1) (subst s c2)) (subst s e2)) (dot (cn_app c1 c2) (dot c2 id))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
so (join_fun _#5 Hjoin2x Hjoin); subst M'.
clear Hjoin Hjoin2x.
so (trp_fstm _#5 HtrM1) as HfstM1.
so (trp_fstm _#5 HtrM2) as HfstM2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (d1 & HfstN1 & Heqd1).
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_fsts _#3 Hsplit2) as Hfst2.
so (ofm_fst _#6 Hwf (ceqm_reg _#5 Hwf Hceq1 andel) HfstM1 (fsts_piapp Hfst1 Hfst2)) as Hofc1.
so (ofm_fst _#6 Hwf (ceqm_reg _#5 Hwf Hceq2 andel) HfstM2 Hfst1) as Hofc2.
#2 eapply ceqm_trans_formation.
  exact (ceqm_comp _#5 (mcc_appap _#15 Hceq1 Hceq2 HfstM2)).
clear Hceq1 Hceq2.
intro Hof.
#2 eapply ceqm_trans.
  #2 apply keqm_ceqm.
    exact Hof.
  clear Hof.
  #2 eapply keqm_discharge.
    #2 eapply keqm_trans.
      apply keqm_appap; [| apply keqm_refl; done].
      apply keqm_lett_beta; done.
    simp_sub.
    #2 eapply keqm_trans.
      #2 eapply (keqm_appap_beta_fstm_eqc _#6 c2); eauto.
        eapply subst_fstm; eauto; prove_subof; done.
    
        simp_sub.
        apply eqc_symm; [].
        so (subst1_eqc _#6 Hofc2 Heqd1) as H; simp_sub in H.
        exact H.
    simp_sub.
    apply keqm_refl; done.

    intros s' Hs' Hhalt.
    unfold andp, truep; simpl.
    simp_sub in Hhalt.
    simp_sub.
    so (join_vars_halt_only_if _#5 Hjoin2 Hhalt) as Hhalt'.
    clear Hhalt.
    so (app_halt_inversion' _ _ Hhalt') as (Hhalt1 & Hhalt2).
    clear Hhalt'.
    #2 split; [split | split]; auto.
      exact (papp_halt_inversion' _ _ Hhalt1).

      clear Hhalt1.
      exact (snd_vars_halt_only_if _#4 Hsnd1 Hhalt2).
clear Hof.
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (join_vars_formation _#5 (wf_cn Hwf Hofk1) HofT2 Hsplit2 Hjoin2) as HofN2.
relquest.
  refine (funct_ceq_md _#7 _ _ _ HofN2); eauto using wf; [].
  #2 apply subrel_dot.
    #2 apply subrel_deq_cn; auto.
      #2 apply subrel_deq_cn; auto.
        apply subrel_id; done.

        simp_sub.
        apply eqc_refl; auto; done.

      apply eqc_refl; [].
      eapply ofc_app; eauto; done.

    simp_sub.
    apply deq_tm; [].
    rewrite -> (trs_subst_ofc _#5 Htrs Hofc2); [].
    #2 apply ceqt_comp; eapply tcc_app.
      apply ceqt_refl; [].
      relquest.
        #2 eapply oft_papp.
          so (trp_formation _#8 Hwf (trs_trctx _#3 Htrs) HtrM1 (split_piapp Hsplit1 Hsplit2)) as Hofe1.
          relquest.
            eapply subst_oft; eauto using trs_formation; done.

            simp_sub; reflexivity.

          rewrite -> (trs_subst_ofk _#4 Htrs Hofk1); [].
          exact Hofc2.
          
        simp_sub.
        rewrite -> (trs_subst_ofc _#5 Htrs Hofc1); [].
        rewrite -> (trs_subst_dot_ofc _#7 Htrs Hoft1); [].
        rewrite -> (trs_subst_dot2_ofc _#9 Htrs Hoft2); [].
        reflexivity.
        
      apply ceqt_symm; [].
      #3 refine (snd_join_inverse _#8 HofT1 Hsplit1 Hofc2 _ _ _).
        rewrite <- (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc2 Hoft1)); [].
        eapply subst_oft; eauto using trs_formation; [].
        eapply trp_formation; eauto using trs_trctx; done.

        so (subst_join _#4 (dot (subst s e2) (dot c2 id)) Hjoin1) as H; simp_sub in H.
        exact H.

        so (subst_snd _#3 (dot (subst (dot (subst s e2) (dot c2 id)) N1) id) Hsnd1) as H; simp_sub in H.
        exact H.

  simp_sub; reflexivity.

(* pair *)
intros G T1 T2 M1 M2 c1 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
so (trp_reg _#5 Hwf HtrM1) as HofT1.
so (trp_reg _#5 Hwf HtrM2) as HofT2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (subst_join _#4 (dot (subst s e1) (dot c1 id)) Hjoin1) as H; simp_sub in H.
so (IH1 _#3 Htrs H) as Hceq1; clear H IH1.
so (subst_join _#4 (dot (subst s e2) (dot c2 id)) Hjoin2) as H; simp_sub in H.
so (IH2 _#3 Htrs H) as Hceq2; clear H IH2.
invert Hjoin; clear Hjoin; [].
intros N1a N2a Hjoin1a Hjoin2a <-.
so (subst_join _#4 (dot (tm_pi1 (var 0)) (dot (cn_pi1 (shift1 (cn_pair c1 c2))) sh1)) Hjoin1) as H; simp_sub in H.
so (join_fun _#5 H Hjoin1a); subst N1a; clear H Hjoin1a.
simp_sub in Hjoin2a.
so (subst_join _#4 (dot (tm_pi2 (var 0)) (dot (cn_pi2 (shift1 (cn_pair c1 c2))) sh1)) Hjoin2) as H; simp_sub in H.
so (join_fun _#5 H Hjoin2a); subst N2a; clear H Hjoin2a.
simp_sub.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1 & t1 & Hsplit1).
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (trp_fst _#6 Hwf HtrM1 (split_fsts _#3 Hsplit1)) as Hofc1.
so (trp_fst _#6 Hwf HtrM2 (split_fsts _#3 Hsplit2)) as Hofc2.
so (trp_formation _#8  Hwf (trs_trctx _#3 Htrs) HtrM1 Hsplit1) as Hofe1.
so (trp_formation _#8  Hwf (trs_trctx _#3 Htrs) HtrM2 Hsplit2) as Hofe2.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_pair; eauto; done.
assert (ofm (G; cl_tm (cn_prod (subst s (subst1 c1 t1)) (subst s (subst1 c2 t2))))
          pure
          (md_lett (tm_pi1 (var 0))
             (md_lett (tm_pi2 (var 1))
                (md_pair
                   (subst
                      (dot (var 1)
                         (dot (subst (sh 3) (cn_pi1 (cn_pair c1 c2))) (sh 3))) N1)
                   (subst
                      (dot (var 0)
                         (dot (subst (sh 3) (cn_pi2 (cn_pair c1 c2))) (sh 3))) N2))))
          (shift1 (sg_sigma T1 (shift1 T2)))) as Hof.
  rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc1 Hoft1)); [].
  rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc2 Hoft2)); [].
  assert (wf
          (G; cl_tm (cn_prod (subst1 c1 t1) (subst1 c2 t2));
           cl_tm (subst (dot (shift1 c1) sh1) t1);
           cl_tm (subst (dot (subst (sh 2) c2) (sh 2)) t2))) as Hwf'.
    #3 prove_wf; auto.
      apply ofc_prod; eapply subst1_ofc_type; eauto; done.

      eapply subst_ofc_type; eauto; prove_subof; [].
      apply weaken1_ofc; auto; done.

      eapply subst_ofc_type; eauto; prove_subof; [].
      apply weaken2_ofc; auto; done.
  #2 eapply ofm_lett.
    eapply oft_pi1; [].
    ofvar; done.

    #2 eapply ofm_lett.
      eapply oft_pi2; [].
      ofvar; done.

      relquest.
        #2 eapply ofm_pair.
          so (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1) as HofN1.
          eapply subst_ofm; eauto; [].
          #2 prove_subof.
            eapply ofc_pi1; [].
            #3 eapply (@ofc_pair _ _ (subst (sh 4) k2)).
              eapply subst_ofc; eauto; prove_subof; done.

              relquest.
                eapply subst_ofc; eauto; prove_subof; done.

                simp_sub; reflexivity.

              eapply subst_ofk; eauto; prove_subof; done.

            #2 eapply oft_equiv.
              ofvar; done.

              eapply seq_ofc_type; eauto; [].
              prove_subrel; [].
              apply eqc_symm; [].
              #2 eapply eqc_beta1.
                eapply subst_ofc; eauto; prove_subof; done.

                eapply subst_ofc; eauto; prove_subof; done.

          so (join_vars_formation _#5 Hwf HofT2 Hsplit2 Hjoin2) as HofN2.
          eapply subst_ofm; eauto; [].
          #2 prove_subof.
            relquest.
              eapply ofc_pi2; [].
              #3 eapply (@ofc_pair _ _ (subst (sh 4) k2)).
                eapply subst_ofc; eauto; prove_subof; done.
  
                relquest.
                  eapply subst_ofc; eauto; prove_subof; done.
 
                  simp_sub; reflexivity.
  
                eapply subst_ofk; eauto; prove_subof; done.

              simp_sub; reflexivity.
  
            #2 eapply oft_equiv.
              ofvar; done.

              eapply seq_ofc_type; eauto; [].
              prove_subrel; [].
              apply eqc_symm; [].
              #2 eapply eqc_beta2.
                eapply subst_ofc; eauto; prove_subof; done.

                eapply subst_ofc; eauto; prove_subof; done.
  
        simp_sub; reflexivity.
#2 eapply (ceqm_trans _#3
             (md_lett (tm_pair (subst s e1) (subst s e2))
                (md_lett (tm_pi1 (var 0))
                   (md_lett (tm_pi2 (var 1))
                      (md_pair
                         (subst
                            (dot (var 1)
                               (dot (subst (sh 3) (cn_pi1 (cn_pair c1 c2))) (sh 3))) N1)
                         (subst
                            (dot (var 0)
                               (dot (subst (sh 3) (cn_pi2 (cn_pair c1 c2))) (sh 3))) N2)))))).
  apply ceqm_symm; [].
  #2 eapply ceqm_trans.
    #2 eapply keqm_ceqm.
      eapply ofm_lett; eauto; [].
      #2 apply oft_pair.
        relquest.
          eapply subst_oft; eauto using trs_formation; done.

          rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc1 Hoft1)); [].
          reflexivity.

        relquest.
          eapply subst_oft; eauto using trs_formation; done.

          rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc2 Hoft2)); [].
          reflexivity.
    clear Hof.
    #2 eapply keqm_discharge.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      #2 eapply keqm_trans.
        apply keqm_lett; [].
        apply keqt_pi1_beta; done.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      #2 eapply keqm_trans.
        apply keqm_lett; [].
        apply keqt_pi2_beta; done.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      apply keqm_refl; done.
  
      intros s' Hs' Hhalt.
      unfold andp, truep; simpl.
      simp_sub in Hhalt.
      so (mpair_halt_inversion' _ _ Hhalt) as (Hhalt1 & Hhalt2); clear Hhalt.
      so (join_vars_halt_only_if _#5 Hjoin1 Hhalt1) as Hhalt1'.
      so (join_vars_halt_only_if _#5 Hjoin2 Hhalt2) as Hhalt2'.
      clear Hhalt1 Hhalt2.
      simp_sub.
      repeat split; auto using thalt_pair; done.
  apply ceqm_symm; [].
  #2 apply ceqm_comp; apply mcc_pair.
    so (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1) as HofN1.
    relquest.
      refine (funct_ceq_md _#7 Hwf _ _ HofN1); eauto using wf; [].
      #2 apply subrel_deq_ceq_tm.
        #2 apply subrel_deq_cn; eauto.
          apply subrel_id; done.
  
          apply eqc_symm; [].
          eapply eqc_beta1; simp_sub; eauto; done.

        apply ceqt_refl; auto; [].
        relquest.
          eapply subst_oft; eauto using trs_formation; done.

          rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc1 Hoft1)); [].
          reflexivity.

      simp_sub; reflexivity.

    so (join_vars_formation _#5 Hwf HofT2 Hsplit2 Hjoin2) as HofN2.
    relquest.
      refine (funct_ceq_md _#7 Hwf _ _ HofN2); eauto using wf; [].
      #2 apply subrel_deq_ceq_tm.
        #2 apply subrel_deq_cn; eauto.
          apply subrel_id; done.
  
          apply eqc_symm; [].
          eapply eqc_beta2; simp_sub; eauto; done.

        apply ceqt_refl; auto; [].
        relquest.
          eapply subst_oft; eauto using trs_formation; done.

          rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc2 Hoft2)); [].
          reflexivity.

      simp_sub; reflexivity.

  #2 apply ceqm_comp; eapply mcc_lett.
    apply ceqt_refl; [].
    eapply oft_pair; eapply subst_oft; eauto using trs_formation; done.

    #2 apply keqm_ceqm.
      exact Hof.
    clear Hof.
    #2 eapply keqm_discharge.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      apply keqm_refl; done.

      intros s' Hs' _.
      unfold andp, truep; simpl.
      #2 repeat split; auto.
        simp_sub.
        replsub (project s' 0) with (subst s' (var 0)).
        #2 eapply thalt_pi1.
          relquest.
            eapply subst_oft; eauto using subof_aof_dof; ofvar; done.
            
            simp_sub; reflexivity.
  
          apply tvalue_thalt; [].
          eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.

        simp_sub.
        replsub (project s' 0) with (subst s' (var 0)).
        #2 eapply thalt_pi2.
          relquest.
            eapply subst_oft; eauto using subof_aof_dof; ofvar; done.
            
            simp_sub; reflexivity.
  
          apply tvalue_thalt; [].
          eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.

(* pi1 *)
intros G T1 T2 M c e HtrM IH G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert (trp_reg _#5 Hwf HtrM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (subst_join _#4 (dot (subst s (tm_pi1 e)) (dot (cn_pi1 c) id)) Hjoin1) as H; simp_sub in H.
so (join_fun _#5 H Hjoin); subst M'; clear H Hjoin.
#2 eexploit IH as Hceq; eauto.
  #2 eapply join_sigma.
    so (subst_join _#4 (dot (tm_pi1 (var 0)) (dot (cn_pi1 (shift1 c)) sh1)) Hjoin1) as H; simp_sub in H.
    exact H.

    so (subst_join _#4 (dot (tm_pi2 (var 0)) (dot (cn_pi2 (shift1 c)) (dot (cn_pi1 (shift1 c)) sh1))) Hjoin2) as H; simp_sub in H.
    exact H.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_pi1; eauto; done.
intro Hof.
eapply keqm_ceqm; [exact Hof | clear Hof].
#2 eapply keqm_discharge.
  #2 eapply keqm_trans.
    apply keqm_pi1; [].
    apply keqm_lett_beta; done.
  simp_sub.
  apply keqm_pi1_beta; done.

  intros s' Hs' Hhalt.
  unfold andp, truep; simpl.
  simp_sub in Hhalt.
  so (join_vars_halt_only_if _#5 Hjoin1 Hhalt) as Hhalt'.
  so (pi1_halt_inversion' _ Hhalt') as Hhalt''.
  split.
    simp_sub.
    exact Hhalt''.

    simp_sub.
    so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
    so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
    so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
    so (trs_trp _#9 Hwf Htrs HtrM (split_sigma Hsplit1 Hsplit2)) as Hofe.
    simp_sub in Hofe.
    so (subst_oft _#5 (subof_aof_dof _#3 Hs') Hofe) as Hofex; simp_sub in Hofex.
    #2 refine (join_vars_halt_if _#7 Hsplit2 _ Hjoin2 _).
      eapply oft_pi2; eauto; done.
  
      eapply thalt_pi2; eauto; done.

(* pi2 *)
intros G T1 T2 M c e HtrM IH G' s M' Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
invert (trp_reg _#5 Hwf HtrM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (subst_join _#4 (dot (subst s (tm_pi2 e)) (dot (cn_pi2 c) (dot (cn_pi1 c) id))) Hjoin2) as H; simp_sub in H.
simp_sub in Hjoin.
so (join_fun _#5 H Hjoin); subst M'; clear H Hjoin.
#2 eexploit IH as Hceq; eauto.
  #2 eapply join_sigma.
    so (subst_join _#4 (dot (tm_pi1 (var 0)) (dot (cn_pi1 (shift1 c)) sh1)) Hjoin1) as H; simp_sub in H.
    exact H.

    so (subst_join _#4 (dot (tm_pi2 (var 0)) (dot (cn_pi2 (shift1 c)) (dot (cn_pi1 (shift1 c)) sh1))) Hjoin2) as H; simp_sub in H.
    exact H.
clear IH.
so (trp_fstm _#5 HtrM) as HfstM.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_pi2; eauto; done.
intro Hof.
eapply keqm_ceqm; [exact Hof | clear Hof].
#2 eapply keqm_discharge.
  #2 eapply keqm_trans.
    apply keqm_pi2; [].
    apply keqm_lett_beta; done.
  simp_sub.
  apply keqm_pi2_beta; done.

  intros s' Hs' Hhalt.
  unfold andp, truep; simpl.
  simp_sub in Hhalt.
  so (join_vars_halt_only_if _#5 Hjoin2 Hhalt) as Hhalt'.
  so (pi2_halt_inversion' _ Hhalt') as Hhalt''.
  split.
    simp_sub.
    exact Hhalt''.

    simp_sub.
    so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
    so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
    so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
    so (trs_trp _#9 Hwf Htrs HtrM (split_sigma Hsplit1 Hsplit2)) as Hofe.
    simp_sub in Hofe.
    so (subst_oft _#5 (subof_aof_dof _#3 Hs') Hofe) as Hofex; simp_sub in Hofex.
    #2 refine (join_vars_halt_if _#7 Hsplit1 _ Hjoin1 _).
      eapply oft_pi1; eauto; done.
  
      eapply thalt_pi1; eauto; done.

(* lett *)
intros G t T e M e1 c2 e2 Htre IH1 HtrM IH2 G' s M' Htrs Hjoin'.
apply ceqm_with_wf; intro Hwf; [].
so (trt_reg _#4 Hwf Htre) as Hoft.
so (trp_reg _#5 Hwf (trp_lett Htre HtrM)) as HofT.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (N & Hjoin).
so (subst_join _#4 (dot (subst s (tm_lett e1 e2)) (dot c2 id)) Hjoin) as H; simp_sub in H.
simp_sub in Hjoin'.
so (join_fun _#5 Hjoin' H); subst M'; clear Hjoin' H.
so (IH1 _ _ Htrs) as Hceq1.
so (subst_join _#4 (dot (subst (dot (var 0) (compose s sh1)) e2) (dot (shift1 c2) sh1)) Hjoin) as H; simp_sub in H.
so (IH2 _#3 (trs_tm Htrs) H) as Hceq2.
clear IH1 IH2 H.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_lett; eauto; done.
intro Hofl.
#2 eapply (ceqm_trans_formation' _#3 
             (md_lett (tm_lett (subst s e1) (subst (dot (var 0) (compose s sh1)) e2))
                (subst (dot (var 0) (dot (shift1 c2) sh1)) N))).
  intro Hofr.
  #5 eapply keqm_middle_ceqm; eauto; clear Hofl Hofr.
    #2 eapply keqm_trans.
      apply keqm_lett_beta; done.
    simp_sub.
    apply keqm_refl; done.

    #2 eapply keqm_partial_discharge.
      #2 eapply keqm_trans.
        apply keqm_lett; [].
        apply keqt_lett_beta; done.
      simp_sub.
      #2 eapply keqm_trans.
        apply keqm_lett_beta; done.
      simp_sub.
      apply keqm_refl; done.

      intros s' Hs' H Hhalt.
      unfold andp, truep; simpl.
      #2 repeat split; auto.
        exact H.

        simp_sub in Hhalt.
        simp_sub.
        eapply join_vars_halt_only_if; eauto; done.

    intros Hof _.
    apply ceqm_refl; auto; done.

    intros s' Hs' Hhalt.
    unfold andp, truep; simpl.
    split; auto; [].
    simp_sub; simp_sub in Hhalt.
    so (mlett_halt_inversion' _ _ Hhalt) as Hhalt'.
    exact (lett_halt_inversion' _ _ Hhalt').

    intros s' Hs' Hhalt.
    unfold andp, truep; simpl.
    simp_sub; simp_sub in Hhalt.
    exact (mlett_halt_inversion' _ _ Hhalt).
clear Hofl.
#2 apply keqm_ceqm.
  so (can_split _ (ofs_hsg _ _ HofT)) as (k & t' & Hsplit).
  #2 eapply ofm_lett.
    #2 eapply (@oft_lett _ _ (subst s (subst1 c2 t'))).
      exact (ceqt_reg _#4 Hwf Hceq1 ander).

      so (trp_formation _#8 (wf_tm Hwf Hoft) (trctx_tm (trs_trctx _#3 Htrs)) HtrM (subst_split _#4 Hsplit)) as Hofe2.
      so (subst_oft _#5 (trs_formation _#3 (wf_tm Hwf Hoft) (trs_tm Htrs)) Hofe2) as Hofe2x.
      simp_sub in Hofe2x.
      simp_sub.
      exact Hofe2x.

    so (subst_ofc _#5 (unshift_tm _ _) (trp_fst _#6 (wf_tm Hwf Hoft) HtrM (subst_fsts _#3 (split_fsts _#3 Hsplit)))) as Hofc2; simp_sub in Hofc2.
    so (split_formation _#4 HofT Hsplit) as (_ & Hoft').
    rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc2 Hoft')); [].
    so (join_vars_formation _#5 Hwf HofT Hsplit Hjoin) as HofN.
    relquest.
      eapply subst_ofm; eauto; prove_subof; [].
      apply weaken1_ofc; auto; done.

      simp_sub; reflexivity.
#2 eapply keqm_discharge.
  #2 eapply keqm_trans.
    apply keqm_lett_beta; done.
  simp_sub.
  apply keqm_refl; done.

  intros s' Hs' Hhalt.
  unfold andp, truep; simpl.
  split; auto; [].
  simp_sub; simp_sub in Hhalt.
  eapply join_vars_halt_only_if; eauto; done.

(* subsume *)
intros G T1 T2 M c e f HtrM IH Htrsubs G' s M' Htrs HjoinM'.
apply ceqm_with_wf; intro Hwf; [].
so (trsubs_subs _#4 Htrsubs) as Hsubs.
so (subs_reg _#3 Hwf Hsubs) as (HofT1 & HofT2).
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1 & t1 & Hsplit1).
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
so (split_fsts _#3 Hsplit1) as Hfst1.
so (split_fsts _#3 Hsplit2) as Hfst2.
so (subs_fst _#5 Hwf Hsubs Hfst1 Hfst2) as Hsubk.
so (trp_fst _#6 Hwf HtrM Hfst1) as Hofc.
so (trp_formation _#8 Hwf (trs_trctx _#3 Htrs) HtrM Hsplit1) as Hofe.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (subst_join _#4 (dot (subst s e) (dot c id)) Hjoin1) as H; simp_sub in H.
so (IH _#3 Htrs H) as Hceqm.
clear H IH.
#2 eapply ceqm_trans; eauto.
  apply ceqm_comp; eapply mcc_subsume; eauto; done.
so (subst_join _#4 (dot (subst s (tm_app (tm_papp f c) e)) (dot c id)) Hjoin2) as H; simp_sub in H.
so (join_fun _#5 HjoinM' H); subst M'; clear HjoinM' H.
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT2))) as (d2 & Hsnd2).
so (subst_snd _#3 (dot N1 (sh 2)) Hsnd2) as H; simp_sub in H.
so (trsubs_correct _#10 Htrsubs Hsplit1 Hsplit2 Hjoin1 H) as Hceqt; clear H.
so (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1) as HofN1.
so (join_vars_formation _#5 Hwf HofT2 Hsplit2 Hjoin2) as HofN2.
so (snd_var_formation _#5 Hwf HofT2 Hsplit2 Hsnd2) as Hofd2.
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (b1 & HfstN1 & Heqb1).
assert (ofm G pure (subst (dot (subst s e) (dot c id)) N1) T1) as HofN1x.
  relquest.
    eapply subst_ofm; eauto; prove_subof; auto; [].
    relquest.
      eapply subst_oft; eauto using trs_formation; done.

      eapply trs_subst_ofc; eauto; [].
      eapply subst1_ofc; eauto; done.

    simp_sub; reflexivity.
assert (fstm G (subst (dot (subst s e) (dot c id)) N1) (subst1 c b1)) as HfstN1x.
  relquest.
    eapply subst_fstm; eauto; prove_subof; done.

    simp_sub; reflexivity.
so (join_snd_vars_inverse _#6 (ofm_subsume HofN1x Hsubs) HfstN1x Hsnd2 Hjoin2) as Hceq.
eapply ceqm_trans; eauto; [].
clear Hceq.
apply ceqm_symm; [].
relquest.
  #2 refine (funct_ceq_md _#7 Hwf _ _ HofN2).
    eauto using wf; done.
  #2 apply subrel_deq_ceq_tm.
    apply subrel_deq_cn; auto using subrel_id; [].
    #2 eapply eqc_subsume.
      #2 eapply rel_quest_2.
        eapply subst1_eqc; eauto; done.

        simp_sub; reflexivity.

      simp_sub; auto; done.

    rewrite -> (trs_subst_ofc _#5 Htrs Hofc); [].
    rewrite <- (trsubs_statically_closed _#7 Hwf Htrsubs (trs_sid _#3 Htrs)); [].
    simp_sub.
    #2 eapply keqt_ceqt.
      #2 eapply oft_app.
        relquest.
          #2 eapply oft_papp.
            exact (ceqt_reg _#4 Hwf Hceqt andel).

            exact Hofc.
            
          simp_sub; reflexivity.

        so (subst_oft _#5 (trs_formation _#3 Hwf Htrs) Hofe) as Hofex.
        rewrite -> (trs_subst_ofc _#5 Htrs (subst1_ofc _#5 Hofc Hoft1)) in Hofex; [].
        exact Hofex.
    #2 eapply keqt_discharge.
      #2 eapply keqt_trans.
        eapply keqt_app; [| apply keqt_refl; done].
        apply reqt_keqt; [].
        #2 eapply reqt_trans.
          eapply reqt_papp; [].
          eapply ceqt_reqt; eauto; done.
        apply reqt_papp_beta; done.
      simp_sub.
      #2 eapply keqt_trans.
        apply keqt_app_beta; done.
      simp_sub.
      apply keqt_refl; done.

      intros s' Hs' Hhalt.
      unfold andp, truep; simpl.
      repeat split; auto; [].
      simp_sub in Hhalt; simp_sub.
      so (snd_vars_halt_only_if _#4 Hsnd2 Hhalt) as Hhalt'.
      exact (join_vars_halt_only_if _#5 Hjoin1 Hhalt').

  simp_sub; reflexivity.

(* appgn *)
intros G T1 T2 M1 M2 c2 e1 e2 HtrM1 IH1 HtrM2 IH2 G' s M' Htrs HjoinM'.
apply ceqm_with_wf; intro Hwf; [].
invert (tri_reg _#4 Hwf HtrM1).
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (tri_formation _#7 Hwf (trs_trctx _ _ _ Htrs) HtrM1 (split_pigen Hsplit1 Hsplit2)) as Hofe1.
so (trs_trp _#9 Hwf Htrs HtrM2 Hsplit1) as Hofe2.
so (trp_fst _#6 Hwf HtrM2 Hfst1) as Hofc2.
so (trp_fstm _#5 HtrM2) as HfstM2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f1 & Hsnd1).
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (b1 & HfstN1 & Heqb1).
#3 eexploit (@join_pigen (subst (sh 2) T1) (subst (dot (var 0) (sh 3)) T2) (var 1) (var 0)) as Hjoinpi.
  simp_sub.
  so (subst_snd _#3 (dot (var 0) (sh 5)) Hsnd1) as H; simp_sub in H.
  exact H.

  simp_sub.
  so (subst_join _#4 (dot (var 0) (dot (var 1) (dot (var 3) (sh 7)))) Hjoin2) as H; simp_sub in H.
  exact H.
so (IH1 _#3 Htrs Hjoinpi) as Hceq1.
clear IH1 Hjoinpi.
so (subst_join _#4 (dot (subst s e2) (dot c2 id)) Hjoin1) as H; simp_sub in H.
so (IH2 _#3 Htrs H) as Hceq2.
clear IH2 H.
so (subst_join _#4 (dot (var 0) (dot (var 1) (dot (subst (sh 2) c2) (sh 2)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
simp_sub in HjoinM'.
so (join_fun _#5 HjoinM' Hjoin2x); subst M'.
clear HjoinM'.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_appgn; eauto; done.
clear Hceq1 Hceq2.
intro Hofl.
simp_sub in Hofl.
simp_sub.
#2 eapply ceqm_trans_formation.
  #2 apply reqm_ceqm.
    exact Hofl.
  clear Hofl.
  #2 eapply reqm_commute_appgn_unpack_fstm_eqc; eauto.
    eapply subst_fstm; eauto; prove_subof; done.

    simp_sub.
    eapply eqc_symm; [].
    relquest.
      eapply subst1_eqc; eauto; done.

      simp_sub; reflexivity.
clear Hofl.
intro Hofl'.
simp_sub.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  #2 apply reqm_ceqm.
    #2 sor (refine (tri_join_formation _#7 Hwf Htrs (tri_appgn HtrM1 HtrM2) _)) as Hofr.
      simp_sub.
      exact Hjoin2x.
    simp_sub in Hofr.
    simp_sub.
    exact Hofr.
  apply reqm_commute_unpack_unpack; done.
clear Hjoin2x.
apply ceqm_symm; [].
#2 apply reqm_ceqm.
  simp_sub in Hofl'.
  exact Hofl'.
clear Hofl'.
#2 eapply reqm_unpack_bind.
  apply ceqt_refl; [].
  relquest.
    eapply subst_oft; eauto using trs_formation; done.

    simp_sub; reflexivity.
apply keqm_reqm; [].
#2 eapply keqm_discharge.
  #2 eapply keqm_trans.
    eapply keqm_appgn; [| apply keqm_refl; done].
    apply reqm_keqm; [].
    apply reqm_lett_beta; [].
    apply tvalue_thalt; ofvar; done.
  simp_sub.
  #2 eapply keqm_trans.
    #2 eapply keqm_appgn_beta_fstm_eqc; eauto using subst_fsts.
      eapply subst_fstm; eauto; prove_subof; done.
  
      simp_sub.
      eapply eqc_symm; [].
      relquest.
        eapply subst_eqc; eauto; prove_subof; [].
        apply weaken2_ofc; auto; done.

        simp_sub; reflexivity.
  simp_sub.
  apply reqm_keqm; [].
  apply reqm_unpack; [].
  #2 apply reqt_app.
    rewrite -> (subst_compose c2); [].
    rewrite -> (trs_subst_ofc _#5 Htrs Hofc2); [].
    apply reqt_refl; done.

    eapply ceqt_reqt; [].
    apply ceqt_symm; [].
    so (snd_join_vars_inverse _#8 HofT1 Hsplit1 Hofc2 Hofe2 Hjoin1 Hsnd1) as Hceq.
    sor (refine (subst_ceqt _#6 (weaken2 _#4) Hwf Hceq)) as Hceqw; [].
    simp_sub in Hceqw.
    exact Hceqw.

  intros s' Hs' Hhalt.
  unfold andp, truep; simpl.
  simp_sub in Hhalt.
  so (munpack_halt_inversion' _#3 Hhalt) as Hhalt'; clear Hhalt.
  so (app_halt_inversion' _ _ Hhalt') as (_ & Hhalt''); clear Hhalt'.
  repeat split; auto; [].
  simp_sub.
  eapply (join_vars_halt_if T1); eauto; [].
  sor (refine (subst_oft _#5 (weaken2 _#4) Hofe2)) as H; [].
  so (subst_oft _#5 (subof_aof_dof _#3 Hs') H) as H'.
  clear H.
  simp_sub in H'.
  exact H'.

(* appap *)
intros G T1 T2 M1 M2 c2 e1 e2 k2 t2 HtrM1 IH1 HtrM2 IH2 Hsplit2 G' s M' Htrs HjoinM'.
apply ceqm_with_wf; intro Hwf; [].
invert (tri_reg _#4 Hwf HtrM1).
intros k1 HofT1 Hfst1 HofT2.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (trs_tri _#8 Hwf Htrs HtrM1 (split_piapp Hsplit1 Hsplit2)) as Hofe1.
so (trs_trp _#9 Hwf Htrs HtrM2 Hsplit1) as Hofe2.
so (trp_fst _#6 Hwf HtrM2 Hfst1) as Hofc2.
so (trp_fstm _#5 HtrM2) as HfstM2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f1 & Hsnd1).
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (b1 & HfstN1 & Heqb1).
#3 eexploit (@join_piapp (subst (sh 2) T1) (subst (dot (var 0) (sh 3)) T2) (var 1) (var 0)) as Hjoinpi.
  simp_sub.
  so (subst_snd _#3 (dot (var 0) (sh 5)) Hsnd1) as H; simp_sub in H.
  exact H.

  simp_sub.
  so (subst_join _#4 (dot (tm_app (tm_papp (var 2) (var 1)) (subst (dot (var 0) (sh 5)) f1)) (dot (cn_app (subst (sh 3) (var 1)) (var 1)) (dot (var 1) (sh 5)))) Hjoin2) as H; simp_sub in H.
  exact H.
so (IH1 _#3 Htrs Hjoinpi) as Hceq1.
clear IH1 Hjoinpi.
so (subst_join _#4 (dot (subst s e2) (dot c2 id)) Hjoin1) as H; simp_sub in H.
so (IH2 _#3 Htrs H) as Hceq2.
clear IH2 H.
so (subst_join _#4 (dot (var 0) (dot (var 1) (dot (subst (sh 2) c2) (sh 2)))) Hjoin2) as Hjoin2x; simp_sub in Hjoin2x.
simp_sub in HjoinM'.
so (join_fun _#5 HjoinM' Hjoin2x); subst M'.
clear HjoinM'.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_appap; eauto; done.
clear Hceq1 Hceq2.
intro Hofl.
simp_sub in Hofl.
simp_sub.
#2 eapply ceqm_trans_formation.
  apply reqm_ceqm; auto; [].
  clear Hofl.
  #2 eapply reqm_commute_appap_unpack_fstm_eqc; eauto.
    eapply subst_fstm; eauto; prove_subof; done.

    simp_sub.
    eapply eqc_symm; [].
    relquest.
      eapply subst1_eqc; eauto; done.

      simp_sub; reflexivity.
clear Hofl.
intro Hofl'.
apply ceqm_symm; [].
#2 eapply ceqm_trans_formation.
  #2 apply reqm_ceqm.
    #2 sor (refine (tri_join_formation _#7 Hwf Htrs (tri_appap HtrM1 HtrM2 Hsplit2) _)) as Hofr.
      simp_sub.
      exact Hjoin2x.
    simp_sub in Hofr.
    simp_sub.
    exact Hofr.
  apply reqm_commute_unpack_unpack; done.
clear Hjoin2x.
intro Hofr'.
apply ceqm_symm; [].
#3 apply ceqm_comp; eapply mcc_unpack.
  apply ceqt_refl; eauto; done.

  #7 eapply keqm_middle_ceqm.
    exact (unpack_ofm_inversion_given _#8 Hwf Hofl' Hofe1 andel).

    exact (unpack_ofm_inversion_given _#8 Hwf Hofr' Hofe1 andel).

    clear Hofl' Hofr'.
    #2 eapply keqm_trans.
      apply keqm_appap; [| apply keqm_refl; done].
      apply reqm_keqm; apply reqm_lett_beta; [].
      apply tvalue_thalt; ofvar; done.
    simp_sub.
    #2 eapply keqm_trans.
      #2 eapply keqm_appap_beta_fstm_eqc; eauto using subst_fsts.
        eapply subst_fstm; eauto; prove_subof; done.

        simp_sub.
        eapply eqc_symm; [].
        relquest.
          eapply subst_eqc; eauto; prove_subof; [].
          apply weaken2_ofc; eauto; done.

          simp_sub; reflexivity.
    simp_sub.
    apply keqm_refl; done.

    clear Hofl' Hofr'.
    #2 eapply keqm_discharge.
      #2 eapply keqm_trans.
        apply keqm_unpack_beta; done.
      simp_sub.
      rewrite -> subst_compose; [].
      rewrite -> (trs_subst_ofc _#5 Htrs Hofc2); [].
      apply keqm_refl; done.

      intros s' Hs' Hhalt.
      unfold andp, truep.
      split; auto; [].
      simp_sub.
      simp_sub in Hhalt.
      rewrite -> subst_compose; [].
      rewrite -> (trs_subst_ofc _#5 Htrs Hofc2); [].
      exact (join_vars_halt_only_if _#5 Hjoin2 Hhalt).

    clear Hofl' Hofr'.
    intros _ _.
    simp_sub.
    apply ceqm_symm; [].
    so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
    so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
    apply ceqm_with_wf; intro Hwf'; [].
    relquest.
      #3 eapply funct_ceq_md; auto.
        exact (wf_tm (wf_cn (wf_cn Hwf Hofk1) Hofk2) Hoft2).

        #2 apply subrel_deq_ceq_tm.
          #2 apply subrel_deq_cn; auto.
            #2 apply subrel_deq_cn; auto.
              prove_subrel; done.

              apply eqc_refl; [].
              apply weaken2_ofc; auto; done.

            apply eqc_refl; [].
            relquest.
              #2 eapply ofc_app.
                ofvar; done.

                apply weaken2_ofc; auto; done.

              simp_sub; reflexivity.

          #2 apply ceqt_comp; eapply tcc_app.
            apply ceqt_refl; [].
            relquest.
              #2 eapply oft_papp.
                ofvar; done.

                apply weaken2_ofc; auto; done.

              simp_sub; reflexivity.

            cut (ceqt G (subst s e2) (subst (dot (subst (dot (subst s e2) (dot c2 id)) N1) id) f1) (subst1 c2 t1)).
              intro Hceq.
              sor (refine (subst_ceqt _#6 (weaken2 _#4) Hwf Hceq)) as H; simp_sub in H; [].
              exact H.
            eapply snd_join_vars_inverse; eauto; done.

        eapply ofm_forget; [].
        apply join_vars_formation; eauto using wf; done.

      simp_sub; reflexivity.

    clear Hofl' Hofr'.
    intros s' Hs' Hhalt.
    unfold andp, truep; simpl.
    repeat split; auto; [].
    simp_sub in Hhalt.
    so (app_halt_inversion' _ _ (pack_halt_inversion' _#4 (munpack_halt_inversion' _#3 Hhalt)) ander) as Hhalt'.
    clear Hhalt.
    simp_sub.
    eapply join_vars_halt_if; eauto; [].
    rewrite -> (subst_compose e2); [].
    relquest.
      eapply subst_oft; eauto; [].
      eapply subof_dof_trans; eauto using subof_aof_dof; prove_subof; done.

      simp_sub; reflexivity.

    clear Hofl' Hofr'.
    intros s' Hs' _.
    unfold andp, truep.
    auto; done.

    clear Hofl' Hofr'.
    eapply eqs_refl; [].
    eapply subst1_ofs; eauto; done.

(* pair *)
intros G T1 T2 M1 M2 e1 e2 k1 k2 t1 t2 HtrM1 IH1 HtrM2 IH2 Hsplit1 Hsplit2 G' s M' Htrs HjoinM'.
apply ceqm_with_wf; intro Hwf; [].
so (tri_reg _#4 Hwf HtrM1) as HofT1.
so (tri_reg _#4 Hwf HtrM2) as HofT2.
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT2))) as (N2 & Hjoin2).
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (b1 & HfstN1 & Heqb1).
so (IH1 _#3 Htrs Hjoin1) as Hceq1.
so (IH2 _#3 Htrs Hjoin2) as Hceq2.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_pair; eauto; done.
clear IH1 IH2 Hceq1 Hceq2.
intro Hofl.
#3 sor (refine (@join_sigma (subst (sh 2) T1) (subst (sh 3) T2) (var 1) (var 0) _ _ _ _)) as Hjoinsig.
  simp_sub.
  so (subst_join _#4 (dot (tm_pi1 (var 0)) (dot (cn_pi1 (var 2)) (sh 3))) Hjoin1) as H; simp_sub in H.
  exact H.

  simp_sub.
  so (subst_join _#4 (dot (tm_pi2 (var 0)) (dot (cn_pi2 (var 2)) (sh 3))) Hjoin2) as H; simp_sub in H.
  exact H.
#2 sor (refine (tri_join_formation _#7 Hwf Htrs (tri_pair HtrM1 HtrM2 Hsplit1 Hsplit2) _)) as Hofr.
  simp_sub.
  exact Hjoinsig.
simp_sub in HjoinM'.
so (join_fun _#5 HjoinM' Hjoinsig); subst M'; clear HjoinM' Hjoinsig.
simp_sub.
#2 eapply ceqm_trans.
  eapply ceqm_commute_pair_unpack_both; eauto; done.
clear Hofl.
simp_sub.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  #2 apply reqm_ceqm.
    simp_sub in Hofr; exact Hofr.
  clear Hofr.
  #2 eapply reqm_trans.
    apply reqm_commute_unpack_unpack; done.
  #2 eapply reqm_unpack_bind.
    apply ceqt_refl; [].
    exact (trs_tri _#8 Hwf Htrs HtrM1 Hsplit1).
  #2 eapply reqm_trans.
    apply reqm_commute_unpack_unpack; done.
  #2 eapply reqm_unpack_bind.
    apply ceqt_refl; [].
    rewrite -> subst_compose; [].
    relquest.
      eapply subst_oft; eauto using trs_tri, weaken2; done.

      simp_sub; reflexivity.
  #2 eapply reqm_trans.
    apply reqm_unpack_beta; [].
    apply tvalue_thalt; apply tvalue_pair; eauto with dynamic sequence; done.
  simp_sub.
  #2 eapply reqm_trans.
    apply reqm_lett_beta; [].
    apply tvalue_thalt; apply tvalue_pair; eauto with dynamic sequence; done.
  simp_sub.
  apply reqm_refl; done.
clear Hofr.
apply ceqm_symm; [].
so (split_formation _#4 HofT1 Hsplit1) as (Hofk1 & Hoft1).
so (split_formation _#4 HofT2 Hsplit2) as (Hofk2 & Hoft2).
#3 apply ceqm_comp; eapply mcc_unpack.
  apply ceqt_refl; eauto using trs_tri; done.

  simp_sub.
  #3 apply ceqm_comp; eapply mcc_unpack.
    apply ceqt_refl; [].
    rewrite -> subst_compose; [].
    relquest.
      eapply subst_oft; eauto using trs_tri, weaken2; done.

      simp_sub; reflexivity.

    simp_sub.
    relquest.
      #2 apply ceqm_comp; apply mcc_pair.
        so (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1) as HofN1.
        apply ceqm_with_wf; intro Hwf'; [].
        #2 refine (funct_ceq_md _#7 Hwf' _ _ (ofm_forget HofN1)).
          auto using wf; done.
        #2 apply subrel_deq_ceq_tm.
          #3 apply subrel_deq_cn.
            #2 prove_wf; auto.
              apply weaken2_ofk; auto; done.

              eapply subst_ofc_type; eauto; prove_subof; done.

            prove_subrel; done.

            apply eqc_symm; [].
            eapply eqc_beta1; ofvar; done.
            
          apply ceqt_symm; [].
          #3 apply ceqt_pi1_beta.
            eapply oft_pi1; [].
            #2 eapply oft_pair.
              relquest.
                ofvar; done.

                rewrite <- (subst_dot0_sh t1) at 1; [].
                rewrite <- subst_compose; [].
                simp_sub; reflexivity.

              ofvar; done.

            apply tvalue_thalt; eauto with dynamic sequence; done.

            apply tvalue_thalt; eauto with dynamic sequence; done.
      
        so (join_vars_formation _#5 Hwf HofT2 Hsplit2 Hjoin2) as HofN2.
        apply ceqm_with_wf; intro Hwf'; [].
        #2 refine (funct_ceq_md _#7 Hwf' _ _ (ofm_forget HofN2)).
          auto using wf; done.
        #2 apply subrel_deq_ceq_tm.
          #3 apply subrel_deq_cn.
            #2 prove_wf; auto.
              apply weaken2_ofk; auto; done.

              eapply subst_ofc_type; eauto; prove_subof; done.

            prove_subrel; done.

            apply eqc_symm; [].
            eapply eqc_beta2; ofvar; done.
            
          apply ceqt_symm; [].
          #3 apply ceqt_pi2_beta.
            eapply oft_pi2; [].
            #2 eapply oft_pair.
              relquest.
                ofvar; done.

                rewrite <- (subst_dot0_sh t1) at 1; [].
                rewrite <- subst_compose; [].
                simp_sub; reflexivity.

              ofvar; done.

            apply tvalue_thalt; eauto with dynamic sequence; done.

            apply tvalue_thalt; eauto with dynamic sequence; done.

      simp_sub; reflexivity.

    apply eqs_refl; [].
    eapply ofs_sigma; eauto using subst_fsts, split_fsts; eapply subst_ofs; eauto; prove_subof; done.

  apply eqs_refl; [].
  eapply ofs_sigma; eauto using subst_fsts, split_fsts; eapply subst_ofs; eauto; prove_subof; done.

(* unpack *)
intros G k t T e M e1 e2 Htre IH1 HtrM IH2 HofT G' s N Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
so (ofc_exists_invert _#4 (trt_reg _#4 Hwf Htre)) as (Hofk & Hoft).
so (IH1 _ _ Htrs) as Hceq1.
#2 sor (refine (IH2 _#3 (trs_tm (trs_cn Htrs)) _)) as Hceq2.
  simp_sub.
  so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 4))) Hjoin) as H; simp_sub in H.
  exact H.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_unpack; eauto using eqs_refl; done.
simp_sub.
so (tri_join_formation _#7 Hwf Htrs (tri_unpack Htre HtrM HofT) Hjoin) as Hofr.
simp_sub in Hofr.
apply ceqm_symm; [].
apply ceqm_commute_unpack_unpack; auto; done.

(* lett *)
intros G t T e M e1 e2 Htre IH1 HtrM IH2 G' s N Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
so (IH1 _ _ Htrs) as Hceq1.
#2 sor (refine (IH2 _#3 (trs_tm Htrs) _)) as Hceq2.
  simp_sub.
  so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 3))) Hjoin) as H; simp_sub in H.
  exact H.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_lett; eauto; done.
simp_sub.
apply ceqm_symm; [].
apply ceqm_commute_unpack_lett; [].
so (tri_join_formation _#7 Hwf Htrs (tri_lett Htre HtrM) Hjoin) as Hofr.
simp_sub in Hofr.
exact Hofr.

(* letm *)
intros G T1 T2 k1 M1 M2 e1 e2 HtrM1 IH1 Hfst1 HtrM2 IH2 HofT2 G' s N2 Htrs Hjoin2.
apply ceqm_with_wf; intro Hwf; [].
so (tri_reg _#4 Hwf HtrM1) as HofT1.
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1a & t1 & Hsplit1).
so (split_fsts_eq _#4 Hsplit1 Hfst1); subst k1a.
so (can_split _ (ofs_hsg _ _ HofT2)) as (k2 & t2 & Hsplit2).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (c1 & HfstN1 & Heqc1).
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT1))) as (f1 & Hsnd1).
so (IH1 _#3 Htrs Hjoin1) as Hceq1.
#3 eexploit IH2 as Hceq2.
  eapply trs_md; eauto; [].
  so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd1) as H; simp_sub in H.
  exact H.

  simp_sub.
  so (subst_join _#4 (dot (var 0) (dot (var 1) (sh 4))) Hjoin2) as H; simp_sub in H.
  exact H.
#2 eapply ceqm_trans_formation.
  apply ceqm_comp; eapply mcc_letm; eauto using eqs_refl; done.
clear IH1 IH2 Hceq1 Hceq2.
intro Hofl.
simp_sub.
#2 eapply ceqm_trans_formation.
  apply ceqm_commute_letm_unpack; auto; done.
clear Hofl.
intro Hofl'; simp_sub in Hofl'.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  apply ceqm_commute_unpack_unpack; [].
  so (tri_join_formation _#7 Hwf Htrs (tri_letm HtrM1 Hfst1 HtrM2 HofT2) Hjoin2) as Hofr.
  simp_sub in Hofr.
  exact Hofr.
apply ceqm_symm; [].
simp_sub.
apply reqm_ceqm; auto; [].
#2 eapply reqm_unpack_bind.
  apply ceqt_refl; eauto using trs_tri; done.
#2 eapply reqm_trans.
  apply reqm_letm_beta_instances; eauto; [].
  intros s' Hs'.
  rewrite <- (subst_dot01_sh N1); [].
  simp_sub.
  #2 refine (join_vars_halt_if _#7 Hsplit1 _ Hjoin1 _).
    replsub (project s' 0) with (subst s' (var 0)).
    relquest.
      eapply subst_oft; eauto using subof_aof_dof; [].
      ofvar; done.

      rewrite <- (subst_dot1_sh t1) at 1; [].
      simp_sub; reflexivity.
      
    replsub (project s' 0) with (subst s' (var 0)).
    apply tvalue_thalt; eapply subst_tvalue; eauto using subof_aof_oof; ofvar; done.
simp_sub.
apply reqm_unpack; [].
eapply ceqt_reqt; [].
so (tri_formation _#7 (wf_md Hwf HofT1 Hfst1) (trctx_md (trs_trctx _#3 Htrs) Hsplit1) HtrM2 (subst_split _#4 Hsplit2)) as Hofe2.
simp_sub in Hofe2.
apply ceqt_with_wf; intro Hwf'.
#2 refine (funct_ceq_tm _#6 Hwf' _ _ Hofe2).
  refine (trctx_formation _ _ _ Hwf'); eauto using trctx, trs_trctx; done.
#2 apply subrel_deq_ceq_tm.
  #2 apply subrel_deq_cn; auto.
    apply subrel_deq_ceq_refl; [].
    eapply subof_dof_trans; eauto using weaken2, trs_formation; done.

    replsub (var 1) with (shift1 (var 0)).
    apply eqc_symm; [].
    relquest.
      apply weaken1_eqc; eauto; done.

      rewrite -> subst_compose; [].
      rewrite -> (trs_subst_ofk _#4 Htrs (ofs_fst _#3 HofT1 Hfst1)); [].
      simp_sub; reflexivity.

  apply ceqt_symm; [].
  #2 apply ceqt_comp; eapply tcc_equiv.
    #6 eapply (snd_join_inverse _#4 (var 1) (var 0)).
      apply weaken2_ofs; exact HofT1.

      eapply subst_split; eauto; done.

      ofvar; done.

      simp_sub.
      simp_sub.
      rewrite -> subst_dot1_sh; [].
      ofvar; done.

      exact Hjoin1.

      so (subst_snd _#3 (dot N1 (sh 2)) Hsnd1) as H; simp_sub in H.
      exact H.

    so (split_formation _#4 HofT1 Hsplit1) as (_ & Hoft1).
    simp_sub.
    simp_sub.
    replsub (subst (dot (shift1 c1) (compose s (sh 2))) t1) with (subst (dot (shift1 c1) (sh 2)) (subst (dot (var 0) (compose s sh1)) t1)).
    rewrite -> (trs_subst_ofc _#5 (trs_cn Htrs) Hoft1); [].
    eapply seq_ofc_type; eauto; prove_subrel; [].
    replsub (var 1) with (shift1 (var 0)).
    relquest.
      eapply subst_eqc; eauto; prove_subof; done.

      simp_sub; reflexivity.

(* seal *)
intros G T M e HtrM IH G' s N Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
so (IH _#3 Htrs Hjoin) as Hceq.
eapply ceqm_trans; eauto; [].
#2 apply mstep_ceqm.
  apply ofm_seal; [].
  exact (ceqm_reg _#5 Hwf Hceq andel).

  apply mstep_seal; done.

(* forget *)
intros G T M c e k t HtrM IH Hsplit G' s N Htrs Hjoin.
apply ceqm_with_wf; intro Hwf; [].
so (trp_fst _#6 Hwf HtrM (split_fsts _#3 Hsplit)) as Hofc.
sor (refine (IH _#3 Htrs _)) as Hceq.
  so (subst_join _#4 (dot (subst s e) (dot c id)) Hjoin) as Hjoinx; simp_sub in Hjoinx.
  exact Hjoinx.
#2 eapply ceqm_trans.
  apply ceqm_comp; apply mcc_forget; eauto; done.
simp_sub.
apply ceqm_symm; [].
#2 apply keqm_ceqm.
  so (tri_join_formation _#7 Hwf Htrs (tri_forget HtrM Hsplit) Hjoin) as H.
  simp_sub in H.
  exact H.
#2 eapply keqm_discharge.
  #2 eapply keqm_trans.
    apply keqm_unpack_beta; done.
  rewrite -> (trs_subst_ofc _#5 Htrs Hofc); [].
  apply keqm_refl; done.

  intros s' Hs' Hhalt.
  unfold andp, truep.
  split; auto; [].
  simp_sub.
  simp_sub in Hhalt.
  exact (join_vars_halt_only_if _#5 Hjoin Hhalt).

(* subsume *)
intros G T1 T2 M e f k2 t2 HtrM IH Htrsubs Hsplit2 G' s N2 Htrs Hjoin2.
apply ceqm_with_wf; intro Hwf; [].
so (trsubs_subs _#4 Htrsubs) as Hsubs.
so (subs_reg _#3 Hwf Hsubs) as (HofT1 & HofT2).
so (can_split _ (ofs_hsg _ _ HofT1)) as (k1 & t1 & Hsplit1).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT1))) as (N1 & Hjoin1).
so (join_vars_fst_o _#5 Hwf HofT1 Hsplit1 Hjoin1) as (c & HfstN1 & Heqc).
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT2))) as (f2 & Hsnd2).
so (IH _#3 Htrs Hjoin1) as Hceq.
#2 eapply ceqm_trans.
  apply ceqm_comp; eapply mcc_subsume; eauto; done.
clear IH Hceq.
simp_sub.
apply ceqm_symm; [].
#2 eapply ceqm_trans.
  #2 eapply reqm_ceqm.
    so (tri_join_formation _#7 Hwf Htrs (tri_subsume HtrM Htrsubs Hsplit2) Hjoin2) as H; simp_sub in H.
    exact H.
  #2 eapply reqm_trans.
    apply reqm_commute_unpack_unpack; done.
  #2 eapply reqm_unpack_bind.
    apply ceqt_refl; eapply trs_tri; eauto; done.
  #2 eapply reqm_trans.
    apply reqm_unpack; [].
    apply reqt_pack; [].
    #2 eapply reqt_trans.
      apply reqt_app; [| apply reqt_refl; done].
      #2 eapply reqt_trans.
        apply reqt_papp; [].
        eapply ceqt_reqt; [].
        #2 sor (refine (trsubs_correct _#10 Htrsubs Hsplit1 Hsplit2 Hjoin1 _)) as Hceq.
          so (subst_snd _#3 (dot N1 (sh 2)) Hsnd2) as H; simp_sub in H.
          exact H.
        rewrite -> subst_compose; [].
        rewrite <- (trsubs_statically_closed _#7 Hwf Htrsubs (trs_sid _#3 Htrs)); [].
        simp_sub.
        exact (subst_ceqt _#6 (weaken2 _#4) Hwf Hceq).
      simp_sub.
      apply reqt_papp_beta; done.
    simp_sub.
    #2 eapply reqt_trans.
      apply reqt_app_beta; [].
      apply tvalue_thalt; ofvar; done.
    simp_sub.
    apply reqt_refl; done.
  eapply keqm_reqm; [].
  #2 eapply keqm_discharge.
    apply keqm_unpack_beta; done.

    intros s' Hs' Hhalt.
    unfold andp, truep.
    simp_sub.
    simp_sub in Hhalt.
    exact (join_vars_halt_only_if _#5 Hjoin2 Hhalt).
#2 eapply ceqm_trans.
  #3 apply ceqm_comp; eapply mcc_unpack.
    apply ceqt_refl; eapply trs_tri; eauto; done.
  
    simp_sub.
    apply ceqm_symm; [].
    rewrite -> subst_dot01_sh; [].
    apply ceqm_comp; eapply mcc_forget; [].
    #6 eapply (join_snd_inverse_fstm_eqc _ N1).
      #2 eapply ofm_subsume.
        exact (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1).

        eapply subst_subs; eauto using weaken2; done.

      exact HfstN1.

      eapply subst_fsts; eauto using split_fsts, weaken2; done.

      eapply eqc_symm; [].
      #2 eapply eqc_subsume.
        eapply subst_eqc; eauto; prove_subof; done.

        simp_sub.
        refine (subst_subk _#5 (weaken2 _#4) _); [].
        eapply subs_fst; eauto using split_fsts; done.

      so (subst_snd _#3 (dot N1 (sh 2)) Hsnd2) as H; simp_sub in H.
      exact H.

      so (subst_join _#4 (dot (subst (dot N1 (sh 2)) f2) (dot (var 1) (sh 2))) Hjoin2) as H; simp_sub in H.
      exact H.

    apply eqs_refl; auto; done.
#3 eapply meme_ceqm_impure.
  eapply ofm_unpack; eauto using trs_tri; [].
  #2 eapply ofm_subsume.
    eapply ofm_forget; [].
    exact (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1).

    apply weaken2_subs; auto; done.

  eapply ofm_subsume; eauto; [].
  eapply ofm_unpack; eauto using trs_tri; [].
  apply ofm_forget; [].
  exact (join_vars_formation _#5 Hwf HofT1 Hsplit1 Hjoin1).

  apply meme_unpack; auto using teme_refl, meme_refl; done.
    
(* var *)
intros G i t Hindex G' s Htrs.
so (trs_index_tm _#5 Htrs Hindex) as Heq.
simp_sub.
rewrite -> Heq; [].
apply ceqt_refl; [].
apply oft_var; auto; done.

(* star *)
intros G G' s H.
simp_sub.
apply ceqt_refl; auto with static; done.

(* lam *)
intros G t1 t2 e e' Hoft1 _ IH G' s Htrs.
simp_sub.
#2 apply ceqt_comp; eapply tcc_lam.
  rewrite -> (trs_subst_ofc _#5 Htrs Hoft1); [].
  apply eqc_refl; auto; done.

  exact (IH _ _ (trs_tm Htrs)).

(* app *)
intros G t1 t2 e1 e1' e2 e2' _ IH1 _ IH2 G' s Htrs.
simp_sub.
apply ceqt_comp; eapply tcc_app; eauto; done.

(* pair *)
intros; simp_sub; apply ceqt_comp; eapply tcc_pair; eauto; done.

(* pi1 *)
intros; simp_sub; apply ceqt_comp; eapply tcc_pi1; eauto; done.

(* pi2 *)
intros; simp_sub; apply ceqt_comp; eapply tcc_pi2; eauto; done.

(* plam *)
intros G k t e e' Hofk Htr IH Hsval G' s Htrs.
simp_sub.
so (IH _ _ (trs_cn Htrs)) as Hceq.
so (trt_svalue _#4 Htr Hsval) as Hsval'.
rewrite -> (trs_subst_ofk _#4 Htrs Hofk); [].
apply ceqt_plam; auto using eqk_refl, subst_svalue; done.

(* papp *)
intros G k t e e' c _ IH Hofc G' s Htrs.
simp_sub.
rewrite -> (trs_subst_ofc _#5 Htrs Hofc); [].
apply ceqt_comp; eapply tcc_papp; eauto using eqc_refl; done.

(* pack *)
intros G k t c e e' Hofc _ IH Hoft G' s Htrs.
simp_sub.
apply ceqt_with_wf; intro Hwf; [].
so (ofc_reg _#3 Hwf Hofc) as Hofk.
rewrite -> (trs_subst_ofk _#4 Htrs Hofk); [].
rewrite -> (trs_subst_ofc _#5 Htrs Hofc); [].
rewrite -> (trs_subst_ofc _#5 (trs_cn Htrs) Hoft); [].
apply ceqt_comp; eapply tcc_pack; eauto with static; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' _ IH1 _ IH2 Hoft' G' s Htrs.
simp_sub.
apply ceqt_comp; eapply tcc_unpack; eauto; [].
so (IH2 _ _ (trs_tm (trs_cn Htrs))) as H.
simp_sub in H.
exact H.

(* fix *)
intros G t e e' Htr IH G' s Htrs.
simp_sub.
apply ceqt_with_wf; intro Hwf; [].
so (ofc_arrow_invert _#4 (trt_reg _#4 Hwf Htr)) as (_ & Hoft).
rewrite -> (trs_subst_ofc _#5 Htrs Hoft); [].
apply ceqt_comp; eapply tcc_fix; eauto using eqc_refl; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' _ IH1 _ IH2 G' s Htrs.
simp_sub.
apply ceqt_comp; eapply tcc_lett; eauto using trs; done.

(* equiv *)
intros G t t' e e' _ IH Heqt G' s Htrs.
apply ceqt_comp; eapply tcc_equiv; eauto; done.

(* letm *)
intros G T k u M e e1 e2 HtrM IH1 Hfst Htre IH2 Hofu G' s Htrs.
apply ceqt_with_wf; intro Hwf; [].
so (tri_reg _#4 Hwf HtrM) as HofT.
so (can_split _ (ofs_hsg _ _ HofT)) as (ka & t & Hsplit).
so (split_fsts_eq _#4 Hsplit Hfst); subst ka.
so (split_formation _#4 HofT Hsplit) as (Hofk & Hoft).
so (can_join _ (var 1) (var 0) (subst_hsg (sh 2) _ (ofs_hsg _ _ HofT))) as (N & Hjoin).
so (join_vars_fst_o _#5 Hwf HofT Hsplit Hjoin) as (c & HfstN & Heqc).
so (can_snd _ (var 0) (subst_hsg sh1 _ (ofs_hsg _ _ HofT))) as (f & Hsnd).
so (IH1 _#3 Htrs Hjoin) as Hceq1.
so (subst_snd _#3 (dot (var 0) (sh 2)) Hsnd) as H; simp_sub in H.
so (IH2 _ _ (trs_md Htrs Hsplit H)) as Hceq2; clear H.
#2 eapply ceqt_trans_formation.
  apply ceqt_comp; eapply tcc_letm; eauto; done.
clear IH1 IH2 Hceq1 Hceq2.
intro Hofl.
#2 apply reqt_ceqt.
  exact Hofl.
clear Hofl.
#2 eapply reqt_trans.
  apply reqt_commute_letm_unpack; done.
simp_sub.
#2 eapply reqt_unpack_bind.
  apply ceqt_refl; eauto using trs_tri; done.
#2 eapply reqt_trans.
  eapply reqt_letm_beta_instances; eauto; [].
  intros s' Hs'.
  rewrite <- (subst_dot01_sh N); [].
  simp_sub.
  #2 refine (join_vars_halt_if _#7 Hsplit _ Hjoin _).
    replsub (project s' 0) with (subst s' (var 0)).
    relquest.
      eapply subst_oft; eauto using subof_aof_dof; [].
      ofvar; done.

      (* Argh! *)
      simp_sub.
      rewrite <- (subst_dot0_sh t) at 1; [].
      rewrite <- subst_compose; [].
      rewrite <- compose_assoc; [].
      simp_sub.
      rewrite -> project_compose; [].
      simp_sub.
      rewrite <- compose_assoc; [].
      unfold sh1.
      rewrite -> compose_sh_sh; [].
      simp_sub.
      reflexivity.

    apply tvalue_thalt; [].
    replsub (project s' 0) with (subst s' (var 0)).
    eapply subst_tvalue; eauto using subof_aof_oof; [].
    ofvar; done.
simp_sub.
eapply ceqt_reqt; [].
so (trt_formation _#5 (wf_md Hwf HofT Hfst) (trctx_md (trs_trctx _#3 Htrs) Hsplit) Htre) as Hofe2.
apply ceqt_with_wf; intro Hwf'.
#2 refine (funct_ceq_tm _#6 Hwf' _ _ Hofe2).
  refine (trctx_formation _ _ _ Hwf'); eauto using trctx, trs_trctx; done.
#2 apply subrel_deq_ceq_tm.
  #2 apply subrel_deq_cn; auto.
    apply subrel_deq_ceq_refl; [].
    eapply subof_dof_trans; eauto using weaken2, trs_formation; done.

    replsub (var 1) with (shift1 (var 0)).
    apply eqc_symm; [].
    relquest.
      apply weaken1_eqc; eauto; done.

      rewrite -> subst_compose; [].
      rewrite -> (trs_subst_ofk _#4 Htrs (ofs_fst _#3 HofT Hfst)); [].
      simp_sub; reflexivity.

  apply ceqt_symm; [].
  #2 apply ceqt_comp; eapply tcc_equiv.
    #6 eapply (snd_join_inverse _#4 (var 1) (var 0)).
      apply weaken2_ofs; exact HofT.

      eapply subst_split; eauto; done.

      ofvar; done.

      simp_sub.
      simp_sub.
      rewrite -> subst_dot1_sh; [].
      ofvar; done.

      exact Hjoin.

      so (subst_snd _#3 (dot N (sh 2)) Hsnd) as H; simp_sub in H.
      exact H.

    simp_sub.
    simp_sub.
    replsub (subst (dot (shift1 c) (compose s (sh 2))) t) with (subst (dot (shift1 c) (sh 2)) (subst (dot (var 0) (compose s sh1)) t)).
    rewrite -> (trs_subst_ofc _#5 (trs_cn Htrs) Hoft); [].
    eapply seq_ofc_type; eauto; prove_subrel; [].
    replsub (var 1) with (shift1 (var 0)).
    relquest.
      eapply subst_eqc; eauto; prove_subof; done.

      simp_sub; reflexivity.

(* ext *)
intros G t M e Htr IH G' s Htrs.
apply ceqt_with_wf; intro Hwf; [].
simp_sub in IH.
so (IH _#3 Htrs join_dyn) as Hceq.
#2 eapply ceqt_trans_formation.
  apply ceqt_comp; eapply tcc_ext; eauto; done.
simp_sub.
intro Hofl.
#2 eapply ceqt_trans.
  apply ceqt_commute_ext_unpack; auto; done.
#3 apply ceqt_comp; eapply tcc_unpack.
  apply ceqt_refl; eauto using trs_tri, split; done.

  #2 apply tstep_ceqt.
    apply oft_ext; [].
    apply ofm_forget; [].
    apply ofm_dyn; [].
    ofvar; done.
  
    apply tstep_ext2; [].
    ofvar; done.

  exact (oft_reg _#3 Hwf Hofl).

(* epilogue *)
destruct_all Hind.
do2 2 split; eauto; done.
Qed.


Corollary trp_correct :
  forall G G' s M T c e M',
    trs G G' s
    -> trp G M T c e
    -> join T c (subst s e) M'
    -> ceqm G pure M M' T.
Proof.
exact (trans_correct andel).
Qed.


Corollary tri_correct :
  forall G G' s M T e M',
    trs G G' s
    -> tri G M T e
    -> join (subst (sh 2) T) (var 1) (var 0) M'
    -> ceqm G impure M (md_unpack (subst s e) M' T) T.
Proof.
exact (trans_correct anderl).
Qed.


Corollary trt_correct :
  forall G G' s e t e',
    trs G G' s
    -> trt G e t e'
    -> ceqt G e (subst s e') t.
Proof.
exact (trans_correct anderr).
Qed.


Corollary closed_term_translation :
  forall e t,
    oft nil e t
    -> exists e',
         trt nil e t e'
         /\ ceqt nil e e' t.
Proof.
intros e t Hof.
so (can_trans anderr _#3 wf_nil Hof) as (e' & Htr).
exists e'.
split; auto; [].
so (trt_correct _#6 trs_nil Htr) as Hceq.
simp_sub in Hceq.
exact Hceq.
Qed.
