
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Dynamic.
Require Import Determinism.
Require Import Regularity.
Require Import Singleton.
Require Import Declarative.
Require Import Breduce.
Require Import Congruence.
Require Import TCongruence.
Require Import BreduceCong.
Require Import TBreduceCong.
Require Import Contextual.
Require Import TContextual.
Require Import Closure.
Require Import Candidate.
Require Import Logical.
Require Import ModelRegular.
Require Import Fundamental.
Require Import Translation.
Require Import Split.
Require Import Correctness.
Require Import Denouement.


Inductive targetctx : context -> Prop :=
| targetctx_nil
    : targetctx nil

| targetctx_cn {G k}
    : targetctx G
      -> targetctx (G; cl_cn k)

| targetctx_tm {G t}
    : targetctx G
      -> targetctx (G; cl_tm t)
.


Lemma targetctx_tail :
  forall G A, targetctx (G; A) -> targetctx G.
Proof.
intros G A H.
invert H; intros; subst; auto; done.
Qed.


Lemma trctx_target :
  forall G G',
    trctx G G'
    -> targetctx G'.
Proof.
intros G G' H.
induct H; intros; eauto using targetctx.
Qed.


Lemma trsubs_target :
  forall G T T' f,
    trsubs G T T' f
    -> target f.
Proof.
intros G T T' f H.
#3 induct H; try (intros; eauto 12 using target, targetsub, subst_target, under_target; done).

intros.
apply target_plam;
apply target_lam;
apply target_plam;
apply target_lam;
apply target_unpack;
eauto 10 using target, subst_target, targetsub; done.

intros.
apply target_plam;
apply target_lam;
apply target_plam;
apply target_lam;
apply target_app; eauto 10 using target, subst_target, targetsub; done.

intros.
apply target_plam;
apply target_lam;
apply target_pair;
eauto 10 using target, subst_target, targetsub; done.
Qed.


Inductive wwf : context -> Prop :=
| wwf_nil
    : wwf nil

| wwf_cn {G k}
    : wwf G
      -> wwf (G; cl_cn k)

| wwf_tm {G t}
    : wwf G
      -> wwf (G; cl_tm t)

| wwf_md {G k T}
    : wwf G
      -> wwf (G; cl_cn k; cl_md T (var 0))

| wwf_utm {G}
    : wwf G
      -> wwf (G; cl_utm)
.


Lemma wf_wwf :
  forall G, wf G -> wwf G.
Proof.
intros G H.
induct H; eauto using wwf.
Qed.


Lemma wwf_tail :
  forall G A, wwf (G; A) -> wwf G.
Proof.
intros G A Hwf.
invert Hwf; try (intros; auto; done); [].
intros G' k T Hwf' <- <-.
apply wwf_cn; auto; done.
Qed.


Lemma trp_target_cn :
  forall G M T c e,
    wwf G
    -> trp G M T c e
    -> target c.
Proof.
intros G M T c e Hwwf Htr.
revert Hwwf.
#3 induct Htr;
try (intros; eauto using target, wwf, ofc_target; done).

(* var *)
intros G i T c Hindex Hwf.
assert (c = var 0) as Heqc.
  remember (cl_md T c) as A eqn:HeqA.
  revert T c HeqA Hwf.
  #2 induct Hindex.
    (* 0 *)
    intros ? G T c -> Hwf.
    invert Hwf; intros; subst; eauto; done.

    (* S *)
    intros i A G ? _ IH T c -> Hwf.
    eapply IH; eauto using wwf_tail; done.
subst c.
simp_sub.
apply target_var; done.

(* lam *)
intros G T1 T2 k1 t1 M c e _ _ _ IH Hwwf.
so (IH (wwf_md Hwwf)) as Htar.
so (subst_target (dot nonsense id) _ (targetsub_dot target_nonsense targetsub_sh) Htar) as Htar'.
simp_sub in Htar'.
eauto using target; done.

(* lett *)
intros G t T e M e1 c2 e2 _ _ IH Hwwf.
so (IH (wwf_tm Hwwf)) as Htar.
so (subst_target (dot nonsense id) _ (targetsub_dot target_nonsense targetsub_sh) Htar) as Htar'.
simp_sub in Htar'.
eauto using target; done.
Qed.


Lemma tr_target :
  (forall G M T c e,
     wf G
     -> trp G M T c e
     -> target e)
  /\
  (forall G M T e,
     wf G
     -> tri G M T e
     -> target e)
  /\
  (forall G e t e',
     wf G
     -> trt G e t e'
     -> target e').
Proof.
exploit
  (trans_ind
     (fun G M T c e => wwf G -> target e)
     (fun G M T e => wwf G -> target e)
     (fun G e t e' => wwf G -> target e')) as Hind;
  try (intros;
       repeat
         (match goal with
          | H : trp _ _ _ _ _, Hwf : wwf _ |- _ =>
              so (trp_target_cn _#5 Hwf H);
              revert H
          end);
       intros;
       eauto 11 using target, wwf, trsubs_target, subst_target, targetsub, ofc_target);
  [].

destruct_all Hind.
do2 2 split; eauto using wf_wwf; done.
Qed.


Lemma trp_target :
  forall G M T c e,
    wf G
    -> trp G M T c e
    -> target e.
Proof.
exact (tr_target andel).
Qed.


Lemma tri_target :
  forall G M T e,
    wf G
    -> tri G M T e
    -> target e.
Proof.
exact (tr_target anderl).
Qed.


Lemma trt_target :
  forall G e t e',
    wf G
    -> trt G e t e'
    -> target e'.
Proof.
exact (tr_target anderr).
Qed.


Lemma trs_target :
  forall G G' s,
    targetctx G
    -> trs G G' s
    -> targetsub s.
Proof.
intros G G' s Htar H.
revert Htar.
induct H.

(* nil *)
intros _.
apply targetsub_sh; done.

(* cn *)
intros G G' k s Htrs IH HtarGx.
invert HtarGx; [].
intro HtarG.
eauto using targetsub, target, compose_target; done.

(* tm *)
intros G G' k s Htrs IH HtarGx.
invert HtarGx; [].
intro HtarG.
eauto using targetsub, target, compose_target; done.

(* md *)
intros G G' T k t s e _ _ _ _ Htar.
invert Htar; done.
Qed.


(* Contextual equivalence implies target equivalence *)

Lemma capprox_tquasicongruence :
  tquasicongruence (ttreg capproxt) (tvreg capproxv).
Proof.
#3 apply reg_tquasicongruence.
  split.
    intros; apply capproxt_refl; auto; done.

    intros; apply capproxv_refl; auto; done.

  split.
    intros; eapply subst_capproxt; eauto; done.

    intros; eapply subst_capproxv; eauto; done.

  split.
    intros G e e' t _ Hcc.
    apply capproxt_comp_weak; [].
    induct Hcc; eauto using tcc; done.

    intros G e e' t _ Hcc.
    apply capproxv_comp_weak; [].
    induct Hcc; eauto using vcc; done.
Qed.


Lemma capproxt_implies_tcapproxt :
  forall G e e' t,
    target e
    -> target e'
    -> capproxt G e e' t
    -> tcapproxt G e e' t.
Proof.
intros G e e' t Htar Htar' Hap.
apply tcapprox_with_wf; intro Hwf.
#2 eapply tcapproxt_univ_quasi; eauto using capprox_tquasicongruence.
  clear_all.
  intros e e' t H Hhalt.
  so (ttreg_elim _#5 wf_nil H) as Hap.
  eapply capproxt_halt; eauto; done.

  so (capproxt_reg _#4 Hwf Hap) as (Hofe & Hofe').
  do2 4 split; auto; done.
Qed.


Lemma ceqt_implies_tceqt :
  forall G e e' t,
    target e
    -> target e'
    -> ceqt G e e' t
    -> tceqt G e e' t.
Proof.
intros G e e' t Htar Htar' Hceq.
destruct Hceq.
split; eauto using capproxt_implies_tcapproxt.
Qed.


(* Target equivalence implies contextual equivalence *)

Lemma doppleganger :
  forall G e t,
    wf G
    -> targetctx G
    -> oft G e t
    -> exists e',
         target e'
         /\ ceqt G e e' t.
Proof.
intros G e t Hwf HtarG Hof.
so (can_trs _ Hwf) as (G' & s & Htrs).
so (oft_trt _#3 Hwf Hof) as (e' & Htrt).
so (trs_target _#3 HtarG Htrs) as Htars.
so (trt_target _#4 Hwf Htrt) as Htar.
exists (subst s e').
split.
  apply subst_target; auto; done.
  
  eapply trt_correct; eauto; done.
Qed.


Lemma doppleganger_val :
  forall v t,
    oft nil v t
    -> tvalue nil v
    -> exists v',
         target v'
         /\ tvalue nil v'
         /\ ceqt nil v v' t.
Proof.
intros v t Hofv Hval.
so (doppleganger _#3 wf_nil targetctx_nil Hofv) as (e & Htar & Hceq).
so (ceqt_halt _#3 Hceq (tvalue_thalt _ _ Hval)) as (v' & Hsteps & Hval').
exists v'.
do2 2 split; eauto using tsteps_target; [].
eapply ceqt_tsteps_trans; eauto; done.
Qed.


Lemma doppleganger_sub :
  forall G s,
    wf G
    -> targetctx G
    -> subof aof nil s G
    -> exists s',
         targetsub s'
         /\ subof aof nil s' G
         /\ subrel (deq ceqt ceqm) nil s s' G.
Proof.
intros G s Hwf HtarG Hs.
remember nil as G' eqn:HeqG'.
revert Hwf HtarG HeqG'.
induct Hs.

(* dot *)
intros ? m s A G Hs IH Haof HwfGx HtarGx ->.
so (wf_tail _ _ HwfGx) as Hwf.
so (targetctx_tail _ _ HtarGx) as HtarG.
so (IH Hwf HtarG (eq_refl _)) as (s' & Htar' & Hs' & Hss).
so (subrel_mono _ _ (deq_seq ceqt ceqm) _#4 Hss) as Hss'.
#2 invert HtarGx.
  (* cn *)
  intros k _ <-.
  rename m into c.
  invert HwfGx; [].
  intros _ Hofk.
  simp_sub in Haof.
  invert Haof; [].
  intros Hofc.
  exists (dot c s').
  do2 2 split.
    apply targetsub_dot; auto; [].
    eapply ofc_target; eauto; done.

    apply subof_dot; auto; [].
    simp_sub.
    apply aof_cn; [].
    eapply ofc_subsume; eauto; [].
    apply subk_refl; [].
    eapply seq_ofk; eauto using wf_nil; done.

    apply subrel_dot; auto; [].
    simp_sub.
    apply deq_cn; auto with static; done.


  (* tm *)
  intros t _ <-.
  rename m into v.
  invert HwfGx; [].
  intros _ Hoft.
  simp_sub in Haof.
  invert Haof; [].
  intros Hofv Hval.
  so (doppleganger_val _ _ Hofv Hval) as (v' & Htar & Hval' & Hceq).
  so (ceqt_reg _#4 wf_nil Hceq) as (_ & Hofv').
  exists (dot v' s').
  do2 2 split.
    apply targetsub_dot; auto; done.

    apply subof_dot; auto; [].
    simp_sub.
    apply aof_tm; auto; [].
    eapply oft_equiv; eauto; [].
    eapply seq_ofc_type; eauto using wf_nil; done.

    apply subrel_dot; auto; [].
    simp_sub.
    apply deq_tm; auto; done.

(* sh *)
intros ? n G Htrunc Hwf HtarG ->.
exists (sh n).
do2 2 split; auto using targetsub, subof, subrel; done.
Qed.


(* st-closure respects target equivalence *)
Lemma tpst_tceqt :
  forall t1 t2 R v1 v2 v1' v2',
    tvrel t1 t2 R
    -> tpclosed t1 t2 R
    -> ceqt nil v1 v1' t1
    -> tceqt nil v2 v2' t2
    -> tvalue nil v1'
    -> tvalue nil v2'
    -> R v1 v2
    -> R v1' v2'.
Proof.
intros t1 t2 R v1 v2 v1' v2' _ Hclosed Hceq1 Hceq2 Hval1 Hval2 HR.
so (ceqt_reg _#4 wf_nil Hceq1 ander) as Hof1.
so (tceqt_reg _#4 wf_nil Hceq2) as (Hof2 & Hof2' & _).
apply Hclosed; [].
split; auto; [].
intros f1 f2 HRs.
so (HRs andel _ _ HR) as Hiff.
so (HRs ander) as (Hoff1 & Hoff2 & _ & _).
so (doppleganger _#3 wf_nil targetctx_nil Hoff2) as (g & Htarg & Hceqg).
so (ceqt_reg _#4 wf_nil Hceqg) as (_ & Hofg).
transitivity (thalt nil (tm_app f1 v1)).
  symmetry.
  eapply ceqt_halt_iff; [].
  apply ceqt_comp; eapply tcc_app; eauto; [].
  apply ceqt_refl; eauto; done.
transitivity (thalt nil (tm_app f2 v2)).
  assumption.
transitivity (thalt nil (tm_app g v2)).
  eapply ceqt_halt_iff; [].
  apply ceqt_comp; eapply tcc_app; eauto; [].
  apply ceqt_refl; eauto; done.
transitivity (thalt nil (tm_app g v2')).
  eapply tceqt_halt_iff; [].
  apply tceqt_comp; eapply ttcc_app; eauto; [].
  apply tceqt_refl; eauto; done.
eapply ceqt_halt_iff; [].
apply ceqt_symm; [].
apply ceqt_comp; eapply tcc_app; eauto; [].
apply ceqt_refl; eauto; done.
Qed.


Lemma tpclosed_tev_tceqt :
  forall t1 t2 R e1 e2 e1' e2',
    tvrel t1 t2 R
    -> tpclosed t1 t2 R
    -> ceqt nil e1 e1' t1
    -> tceqt nil e2 e2' t2
    -> tev t1 t2 R e1 e2
    -> tev t1 t2 R e1' e2'.
Proof.
intros t1 t2 R e1 e2 e1' e2' Hrel Hclosed Hceq1 Hceq2 HR.
so (ceqt_reg _#4 wf_nil Hceq1 ander) as Hofe1'.
so (tceqt_reg _#4 wf_nil Hceq2 ander) as (Hofe2' & _).
destruct HR as (Hhaltiff & Hall & Hofs).
#2 do2 2 split; auto.
  split.
    intros Hhalt.
    eapply tceqt_halt; eauto; [].
    apply Hhaltiff; [].
    eapply ceqt_halt; eauto using ceqt_symm; done.

    intros Hhalt.
    eapply ceqt_halt; eauto; [].
    apply Hhaltiff; [].
    eapply tceqt_halt; eauto using tceqt_symm; done.

  intros v1' v2' Hsteps1' Hsteps2' Hval1' Hval2'.
  so (ceqt_halt _#3 (ceqt_symm _#4 Hceq1) (eintro v1' (conj Hsteps1' Hval1'))) as (v1 & Hsteps1 & Hval1).
  so (tceqt_halt _#3 (tceqt_symm _#4 Hceq2) (eintro v2' (conj Hsteps2' Hval2'))) as (v2 & Hsteps2 & Hval2).
  so (Hall _ _ Hsteps1 Hsteps2 Hval1 Hval2) as HR.
  #2 eapply tpst_tceqt; eauto using tpst_idem, tpst_tvrel.
    eapply anti_tsteps_ceqt_trans; eauto; [].
    eapply ceqt_tsteps_trans; eauto; done.
    
    eapply anti_tsteps_tceqt_trans; eauto; [].
    eapply tceqt_tsteps_trans; eauto; done.
Qed.


Lemma tpclosed_tev_ceqt_tceqt_ceqt :
  forall t1 t2 R e1 e2 e1' e2' e2'' e2''',
    tvrel t1 t2 R
    -> tpclosed t1 t2 R
    -> ceqt nil e1 e1' t1
    -> ceqt nil e2 e2' t2
    -> tceqt nil e2' e2'' t2
    -> ceqt nil e2'' e2''' t2
    -> tev t1 t2 R e1 e2
    -> tev t1 t2 R e1' e2'''.
Proof.
intros t1 t2 R e1 e2 e1' e2' e2'' e2''' Hrel Hclosed Hceq1 Hceq2 Hceq2' Hceq2'' HR.
assert (tev t1 t2 R e1' e2') as HR'.
  eapply tpclosed_tev_ceqt; eauto; done.
clear HR.
assert (tev t1 t2 R e1' e2'') as HR''.
  eapply (tpclosed_tev_tceqt _ _ _ e1'); eauto; [].
  apply ceqt_refl; [].
  exact (ceqt_reg _#4 wf_nil Hceq1 ander).
clear HR'.
eapply (tpclosed_tev_ceqt _ _ _ e1'); eauto; [].
apply ceqt_refl; [].
exact (ceqt_reg _#4 wf_nil Hceq1 ander).
Qed.


(* Target equivalence implies logical equivalence *)
Lemma tceqt_implies_leqt :
  forall G e e' t,
    targetctx G
    -> tceqt G e e' t
    -> leqt G e e' t.
Proof.
intros G e e' t HtarG Hceqt.
intro Hwf.
so (tceqt_reg _#4 Hwf Hceqt) as (Hofe & Hofe' & Htar & Htar').
so (oft_reg _#3 Hwf Hofe) as Hoft.
so (leqt_refl _#3 Hofe) as Hleqt.
do2 2 split; auto; [].
intros E HintE.
so (right_subof_aof _ _ Hwf HintE) as Hright.
so (doppleganger_sub _ _ Hwf HtarG Hright) as (s & Htars & Hs & Hss).
#7 apply (tpclosed_tev_ceqt_tceqt_ceqt _ _ _ (subst (left E) e) (subst (right E) e) _ (subst s e) (subst s e') (subst (right E) e')).
  eapply intt_tvrel; eauto using fintctx_stat; done.

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  eapply subst_ceqt; eauto using left_subof_aof, ceqt_refl; done.

  eapply funct_ceq_tm; eauto using wf_nil; done.

  #2 apply tceqt_comp; eapply ttcc_equiv.
    eapply subst_tceqt; eauto; done.

    apply eqc_symm; [].
    eapply seq_ofc_type; eauto using wf_nil; [].
    eapply subrel_deq_seq; eauto; done.

  apply ceqt_symm; [].
  eapply funct_ceq_tm; eauto using wf_nil; done.

  exact (Hleqt Hwf anderr _ HintE).
Qed.


Theorem tceqt_implies_ceqt :
  forall G e e' t,
    targetctx G
    -> tceqt G e e' t
    -> ceqt G e e' t.
Proof.
intros G e e' t Htar Htceq.
apply leqt_implies_ceqt; [].
apply tceqt_implies_leqt; auto; done.
Qed.


(* Stronger full abstraction *)

Theorem trp_full_abstraction :
  forall G G' M M' T k t c c' e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> trp G M T c e
    -> trp G M' T c' e'
    -> (ceqm G pure M M' T <-> (eqc G' c c' k /\ tceqt G' e e' (subst1 c t))).
Proof.
intros G G' M M' T k t c c' e e' Hwf HtrG Hsplit Htrp Htrp'.
split.
  intro HceqM.
  split.
    so (trp_fstm _#5 Htrp) as HfstM.
    so (trp_fstm _#5 Htrp') as HfstM'.
    so (split_fsts _#3 Hsplit) as Hfst.
    so (ceqm_fst' _#7 Hwf HceqM Hfst HfstM HfstM') as Heqc.
    so (can_trsc _ Hwf) as (G'a & s & Htrsc).
    so (trctx_fun _#3 HtrG (trsc_trctx _#3 Htrsc)); subst G'a.
    so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & Hofk).
    so (trsc_subst_ofc _#5 Hwf Htrsc Hofc) as Heqc1.
    so (trsc_subst_ofc _#5 Hwf Htrsc Hofc') as Heqc3.
    so (subst_eqc _#6 (subof_aaof_sof _#3 (trsc_formation _#3 Hwf Htrsc)) Heqc) as Heqc2.
    eapply eqc_trans; eauto; [].
    apply (@eqc_trans _ _ (subst s c')); eauto using eqc_symm; [].
    eapply eqc_subsume; eauto; [].
    apply subk_refl; [].
    apply eqk_symm; [].
    eapply trsc_subst_ofk; eauto; done.

    apply ceqt_implies_tceqt; eauto using trp_target; [].
    eapply trp_abstraction_preservation; eauto; done.

  intros (Heqc & Hceqe).
  so (tceqt_implies_ceqt _#4 (trctx_target _ _ HtrG) Hceqe) as Hceqe'.
  eapply trp_abstraction_preservation_conv; eauto; done.
Qed.


Theorem tri_full_abstraction :
  forall G G' M M' T k t e e',
    wf G
    -> trctx G G'
    -> split T k t
    -> tri G M T e
    -> tri G M' T e'
    -> (ceqm G impure M M' T <-> tceqt G' e e' (cn_exists k t)).
Proof.
intros G G' M M' T k t e e' Hwf HtrG Hsplit Htri Htri'.
split.
  intro Hceqm.
  apply ceqt_implies_tceqt; eauto using tri_target; [].
  eapply tri_abstraction_preservation; eauto; done.

  intros Hceqe.
  so (tceqt_implies_ceqt _#4 (trctx_target _ _ HtrG) Hceqe) as Hceqe'.
  eapply tri_abstraction_preservation_conv; eauto; done.
Qed.


Theorem trt_full_abstraction :
  forall G G' e e' t f f',
    wf G
    -> trctx G G'
    -> trt G e t f
    -> trt G e' t f'
    -> (ceqt G e e' t <-> tceqt G' f f' t).
Proof.
intros G G' e e' t f f' Hwf HtrG Htre Htre'.
split.
  intro Hceq.
  apply ceqt_implies_tceqt; eauto using trt_target; [].
  eapply trt_abstraction_preservation; eauto; done.

  intro Hceqf.
  so (tceqt_implies_ceqt _#4 (trctx_target _ _ HtrG) Hceqf) as Hceqf'.
  eapply trt_abstraction_preservation_conv; eauto; done.
Qed.
